@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -34,31 +34,30 @@
 /*
  * uimΥХåؿ
  */
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 #ifndef DEBUG
 #define NDEBUG
 #endif
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_STRING_H
+#ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-#if HAVE_ASSERT_H
+#ifdef HAVE_ASSERT_H
 #include <assert.h>
 #endif
+#include <iconv.h>
+#include <uim/uim-util.h>
 #include "uim-fep.h"
 #include "str.h"
 #include "callbacks.h"
 
 static uim_context s_context;
-/* TRUEʤ饫֤ȿžʤ */
-static int s_cursor_no_reverse;
-/* ơ饤 */
-static int s_statusline_width;
-
+/* ơ饤κ */
+static int s_max_width;
 static char *s_commit_str;
 static char *s_statusline_str;
 static char *s_candidate_str;
@@ -66,6 +65,7 @@ static int s_candidate_col;
 static char *s_index_str;
 static struct preedit_tag *s_preedit;
 static int s_mode;
+static char *s_nokori_str;
 
 static void start_callbacks(void);
 static void end_callbacks(void);
@@ -77,8 +77,6 @@ static void clear_cb(void *ptr);
 static void pushback_cb(void *ptr, int attr, const char *str);
 static void update_cb(void *ptr);
 static void mode_update_cb(void *ptr, int mode);
-static void property_update_cb(void *ptr, const char *str);
-static void property_list_update_cb(void *ptr, const char *str);
 static struct preedit_tag *dup_preedit(struct preedit_tag *p);
 static void make_page_strs(void);
 static int numwidth(int n);
@@ -122,11 +120,13 @@ static struct candidate_tag s_candidate = {
 /*
  * 
  */
-void init_callbacks(uim_context context, int status_type, int cursor_no_reverse, int statusline_width)
+void init_callbacks(uim_context context)
 {
   s_context = context;
-  s_cursor_no_reverse = cursor_no_reverse;
-  s_statusline_width = statusline_width;
+  s_max_width = g_win->ws_col;
+  if (g_opt.statusline_width != UNDEFINED && g_opt.statusline_width <= s_max_width) {
+    s_max_width = g_opt.statusline_width;
+  }
   s_commit_str = strdup("");
   s_candidate_str = strdup("");
   s_statusline_str = strdup("");
@@ -136,17 +136,42 @@ void init_callbacks(uim_context context, int status_type, int cursor_no_reverse,
   s_preedit = create_preedit();
   uim_set_preedit_cb(s_context, clear_cb, pushback_cb, update_cb);
   uim_set_mode_cb(s_context, mode_update_cb);
-  uim_set_prop_label_update_cb(s_context, property_update_cb);
-  uim_set_prop_list_update_cb(s_context, property_list_update_cb);
-  if (status_type != NONE) {
+  if (g_opt.status_type != NONE) {
     uim_set_candidate_selector_cb(s_context, activate_cb, select_cb, shift_page_cb, deactivate_cb);
   }
+
+  if (g_opt.ddskk) {
+    iconv_t cd;
+    char *nokori_str = "Ĥ";
+    size_t inbytesleft = strlen("Ĥ");
+    size_t outbytesleft = 6;
+    const char *enc;
+
+    s_nokori_str = malloc(outbytesleft + 1);
+    strcpy(s_nokori_str, "Ĥ");
+    if (strcmp(enc = get_enc(), "EUC-JP") != 0) {
+      cd = (iconv_t)uim_iconv_open(enc, "EUC-JP");
+      if (cd == (iconv_t)-1) {
+        perror("error in iconv_open");
+        puts("-d option is not available");
+        done(EXIT_FAILURE);
+      }
+      if (iconv(cd, (ICONV_CONST char**)&nokori_str, &inbytesleft, &s_nokori_str, &outbytesleft) == (size_t)-1) {
+        perror("error in iconv");
+        puts("-d option is not available");
+        done(EXIT_FAILURE);
+      }
+      s_nokori_str[0] = '\0';
+      s_nokori_str -= (6 - outbytesleft);
+      iconv_close(cd);
+    }
+  }
 }
 
 int press_key(int key, int key_state)
 {
   int raw;
-#if DEBUG > 2
+#if defined DEBUG && DEBUG > 2
   if (32 <= key && key <= 127) {
     debug2(("press key = %c key_state = %d\n", key, key_state));
   } else {
@@ -281,6 +306,13 @@ int get_mode(void)
   return s_mode;
 }
 
+char *get_mode_str(void)
+{
+  char *mode_str = strdup(uim_get_mode_name(s_context, s_mode));
+  strhead(mode_str, s_max_width);
+  return mode_str;
+}
+
 /*
  * ɽȤ˸ƤФ롣
  * s_candidate.nr = nr()
@@ -385,7 +417,7 @@ static void pushback_cb(void *ptr, int attr, const char *str)
   /* ʸ̵ */
   if (width > 0) {
     /* ֤ʸȿžʤ */
-    if (s_cursor_no_reverse && cursor && attr & UPreeditAttr_Reverse && s_preedit->cursor != UNDEFINED) {
+    if (g_opt.cursor_no_reverse && cursor && attr & UPreeditAttr_Reverse && s_preedit->cursor != UNDEFINED) {
       int *rval = width2byte2(str, 1);
       int first_char_byte = rval[0];
       int first_char_width = rval[1];
@@ -434,15 +466,6 @@ static void mode_update_cb(void *ptr, int mode)
   s_mode = mode;
 }
 
-static void property_update_cb(void *ptr, const char *str)
-{
-  debug2(("prop(\"%s\")", str));
-}
-static void property_list_update_cb(void *ptr, const char *str)
-{
-  debug2(("proplist(\"%s\")", str));
-}
-
 /*
  * ץꥨǥåȤꡤݥ󥿤֤
  */
@@ -492,9 +515,9 @@ static struct preedit_tag *dup_preedit(struct preedit_tag *p)
 
 /*
  * s_candidate.page_strs = ڡʸ
- * ʸmax_widthۤȤϡϤ߽Ф򼡤Υڡ˰ܤ
+ * ʸs_max_widthۤȤϡϤ߽Ф򼡤Υڡ˰ܤ
  * 1Ĥ䤬ʤƤϤ߽ФȤϡܤʤ
- * max_widthüץǻꤵ줿
+ * s_max_widthüץǻꤵ줿
  * s_candidate.page2index = ڡκǽθindex
  * s_candidate.cand_col = ΰ
  * s_candidate.nr_pages = ڡ
@@ -510,12 +533,6 @@ static void make_page_strs(void)
 
   int index;
 
-  int max_width = g_win->ws_col;
-
-  if (s_statusline_width != UNDEFINED && s_statusline_width <= max_width) {
-    max_width = s_statusline_width;
-  }
-
   assert(s_candidate.nr != UNDEFINED);
   assert(s_candidate.limit != UNDEFINED);
   assert(s_candidate.cand_col != NULL);
@@ -523,19 +540,27 @@ static void make_page_strs(void)
   s_candidate.nr_pages = 0;
 
   for (index = 0; index < s_candidate.nr; index++) {
+    /* A:  S:  D:  F:  J:  K:  L:  [Ĥ 227] */
     int next = FALSE;
     /* "[10/20]"  */
-    int index_width = 3 + numwidth(index + 1) + numwidth(s_candidate.nr);
-
+    int index_width;
     uim_candidate cand = uim_get_candidate(s_context, index, index_in_page);
     const char *cand_str_label = uim_candidate_get_heading_label(cand);
-    const char *cand_str_cand = uim_candidate_get_cand_str(cand);
+    char *cand_str_cand = tab2space(uim_candidate_get_cand_str(cand));
     int cand_label_width = strwidth(cand_str_label);
-    int cand_width = cand_label_width + 1 + strwidth(cand_str_cand) + 1;
-    int cand_byte = strlen(cand_str_label) + 1 + strlen(cand_str_cand) + 1;
+    int cand_width = cand_label_width + strlen(":") + strwidth(cand_str_cand) + strlen(" ");
+    int cand_byte = strlen(cand_str_label) + strlen(":") + strlen(cand_str_cand) + strlen(" ");
     char *cand_str = malloc(cand_byte + 1);
+
+    if (g_opt.ddskk) {
+      index_width = strlen("[xxxx ]") + numwidth(s_candidate.nr - index - 1);
+    } else {
+      index_width = strlen("[/]") + numwidth(index + 1) + numwidth(s_candidate.nr);
+    }
+
     sprintf(cand_str, "%s:%s ", cand_str_label, cand_str_cand);
     uim_candidate_free(cand);
+    free(cand_str_cand);
 
     if (index_in_page == 0) {
       s_candidate.page2index = realloc(s_candidate.page2index, (s_candidate.nr_pages + 1) * sizeof(int));
@@ -543,28 +568,55 @@ static void make_page_strs(void)
       s_candidate.index_col = realloc(s_candidate.index_col, (s_candidate.nr_pages + 1) * sizeof(int));
     }
 
-    if (page_width + cand_width + index_width > max_width && index_in_page != 0) {
+    if (page_width + cand_width + index_width > s_max_width && index_in_page != 0) {
       /* Ϥ߽ФΤǼΥڡ˰ܤ */
       index--;
-      index_width = 3 + numwidth(index + 1) + numwidth(s_candidate.nr);
+      if (g_opt.ddskk) {
+        index_width = strlen("[xxxx ]") + numwidth(s_candidate.nr - index - 1);
+      } else {
+        index_width = strlen("[/]") + numwidth(index + 1) + numwidth(s_candidate.nr);
+      }
       next = TRUE;
     } else {
-      if (cand_width + index_width > max_width && index_in_page == 0) {
+
+      s_candidate.cand_col[index] = page_width + cand_label_width + strlen(":");
+
+      if (cand_width + index_width > s_max_width && index_in_page == 0) {
+        /* Ϥ߽Ф˰ܤʤ */
         assert(page_width == 0);
-        /* Ϥ߽Ф˰ܤʤ */      /* ڡ1 */
-        cand_width = max_width - index_width - 1;
-        cand_width = strhead(cand_str, cand_width);
-        if (cand_label_width > cand_width) {
-          cand_label_width = cand_width;
-        }
-        cand_width++;
-        cand_byte = strlen(cand_str);
-        cand_str[cand_byte++] = ' ';
-        cand_str[cand_byte] = '\0';
         next = TRUE;
+                                                  /* 1ʸ */
+        if (s_max_width >= cand_label_width + strlen(":") + 2 + strlen(" ") + index_width) {
+          /*  + ǥå */
+
+          cand_width = s_max_width - index_width - strlen(" ");
+          cand_width = strhead(cand_str, cand_width);
+          assert(cand_width > cand_label_width);
+          cand_width += strwidth(" ");
+          cand_byte = strlen(cand_str);
+          cand_str[cand_byte++] = ' ';
+          cand_str[cand_byte] = '\0';
+        } else {
+          /* ǥåϤʤ */
+
+          index_width = UNDEFINED;
+          if (cand_width > s_max_width) {
+            cand_width = s_max_width;
+          }
+          cand_width -= strlen(" ");
+          cand_width = strhead(cand_str, cand_width);
+          if (cand_width <= cand_label_width + strlen(":")) {
+            cand_width = 1;
+            strcpy(cand_str, " ");
+            s_candidate.cand_col[index] = UNDEFINED;
+          } else {
+            cand_byte = strlen(cand_str);
+            cand_str[cand_byte++] = ' ';
+            cand_str[cand_byte] = '\0';
+          }
+        }
       }
-                                                             /* ':' */
-      s_candidate.cand_col[index] = page_width + cand_label_width + 1;
+
       page_width += cand_width;
       page_byte += cand_byte;
       page_str = realloc(page_str, page_byte + 1);
@@ -577,32 +629,35 @@ static void make_page_strs(void)
     }
 
     if (next) {
-      int index_byte = index_width;
-      char *index_str = malloc(index_byte + 1);
-      int i;
-      sprintf(index_str, "[%d/%d]", index + 1, s_candidate.nr);
-      for (i = 0; i < numwidth(index + 1); i++) {
-        index_str[1 + i] = ' ';
+      if (index_width == UNDEFINED) {
+        s_candidate.index_col[s_candidate.nr_pages] = UNDEFINED;
+      } else {
+        int index_byte = index_width + 2/* utf-8 */;
+        char *index_str = malloc(index_byte + 1);
+        int i;
+        if (g_opt.ddskk) {
+          sprintf(index_str, "[%s %d]", s_nokori_str, s_candidate.nr - index - 1);
+        } else {
+          sprintf(index_str, "[%d/%d]", index + 1, s_candidate.nr);
+          for (i = 0; i < numwidth(index + 1); i++) {
+            index_str[1 + i] = ' ';
+          }
+          index_str[i] = '-';
+        }
+        assert(page_width + index_width <= s_max_width);
+        s_candidate.index_col[s_candidate.nr_pages] = page_width + strlen("[");
+        page_byte += index_byte;
+        page_str = realloc(page_str, page_byte + 1);
+        strcat(page_str, index_str);
+        free(index_str);
       }
-      index_str[i] = '-';
-      /* ۤüʤϤ߽Фʤ */
-      if (page_width + index_width > max_width) {
-        index_width = max_width - page_width;
-        index_width = strhead(index_str, index_width);
-        index_byte = strlen(index_str);
-      }
-      s_candidate.index_col[s_candidate.nr_pages] = page_width + 1;
-      page_byte += index_byte;
-      page_str = realloc(page_str, page_byte + 1);
-      strcat(page_str, index_str);
-      s_candidate.page_strs = realloc(s_candidate.page_strs, (s_candidate.nr_pages + 1)* sizeof(char *));
+      s_candidate.page_strs = realloc(s_candidate.page_strs, (s_candidate.nr_pages + 1) * sizeof(char *));
       s_candidate.page_strs[s_candidate.nr_pages] = strdup(page_str);
       s_candidate.nr_pages++; 
 
       page_byte = 0;
       page_width = 0;
       index_in_page = 0;
-      free(index_str);
       free(page_str);
       page_str = strdup("");
     }
@@ -682,26 +737,37 @@ static void get_candidate(void)
   int cand_width;
   /* "[10/20]" */
   int index_width;
-  int max_width = g_win->ws_col;
-  /* üθΥǥå */
-  int right_edge_cand_index = s_candidate.page + 1 == s_candidate.nr_pages ? s_candidate.nr : s_candidate.page2index[s_candidate.page + 1];
-  /* üθΥǥå */
-  int right_edge_cand_index_width = numwidth(right_edge_cand_index + 1);
-  /* ߤθΥǥå */
-  int cand_index_width = numwidth(s_candidate.index + 1);
-  int i;
-  s_index_str = malloc(right_edge_cand_index_width + 1);
-  for (i = 0; i < right_edge_cand_index_width - cand_index_width; i++) {
-    s_index_str[i] = ' ';
-  }
-  s_index_str[i] = '\0';
-  sprintf(s_index_str, "%s%d", s_index_str, s_candidate.index + 1);
 
-  if (s_statusline_width != UNDEFINED && s_statusline_width <= max_width) {
-    max_width = s_statusline_width;
+  if (s_candidate.index_col[s_candidate.page] == UNDEFINED) {
+    s_index_str = strdup("");
+    index_width = 0;
+  } else {
+    /* üθΥǥå */
+    int right_edge_cand_index = s_candidate.page + 1 == s_candidate.nr_pages ? s_candidate.nr - 1 : s_candidate.page2index[s_candidate.page + 1] - 1;
+    /* üθΥǥå */
+    int right_edge_cand_index_width = numwidth(right_edge_cand_index + 1);
+    /* ߤθΥǥå */
+    int cand_index_width = numwidth(s_candidate.index + 1);
+    int i;
+    s_index_str = malloc(right_edge_cand_index_width + 1);
+    for (i = 0; i < right_edge_cand_index_width - cand_index_width; i++) {
+      s_index_str[i] = ' ';
+    }
+    s_index_str[i] = '\0';
+    sprintf(s_index_str, "%s%d", s_index_str, s_candidate.index + 1);
+    if (g_opt.ddskk) {
+      index_width = strlen("[xxxx ]") + numwidth(s_candidate.nr - s_candidate.index - 1);
+    } else {
+      index_width = strlen("[/]") + numwidth(s_candidate.index + 1) + numwidth(s_candidate.nr);
+    }
   }
 
+
   s_candidate_col = s_candidate.cand_col[s_candidate.index];
+  if (s_candidate_col == UNDEFINED) {
+    s_candidate_str = strdup("");
+    return;
+  }
   cand = uim_get_candidate(s_context, s_candidate.index, 0);
   if (uim_candidate_get_cand_str(cand) == NULL) {
     s_candidate_str = strdup("");
@@ -709,18 +775,20 @@ static void get_candidate(void)
     uim_candidate_free(cand);
     return;
   }
-  s_candidate_str = strdup(uim_candidate_get_cand_str(cand));
+  s_candidate_str = tab2space(uim_candidate_get_cand_str(cand));
   cand_width = strwidth(s_candidate_str);
-  index_width = 3 + numwidth(s_candidate.index + 1) + numwidth(s_candidate.nr);
-  /* ۤüʤϤ߽Фʤ */
-  if (s_candidate_col + cand_width + index_width > max_width) {
-    strhead(s_candidate_str, max_width - s_candidate_col - index_width - 1);
+  if (s_candidate_col + cand_width + strlen(" ") + index_width > s_max_width) {
+    strhead(s_candidate_str, s_max_width - s_candidate_col - strlen(" ") - index_width);
   }
   uim_candidate_free(cand);
 }
 
 void callbacks_winch(void)
 {
+  s_max_width = g_win->ws_col;
+  if (g_opt.statusline_width != UNDEFINED && g_opt.statusline_width <= s_max_width) {
+    s_max_width = g_opt.statusline_width;
+  }
   if (s_candidate.nr != UNDEFINED) {
     if (s_candidate.page_strs != NULL) {
       int i;
