@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -34,19 +34,22 @@
 /*
  * ץꥨǥåȤ䥹ơ饤褹
  */
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
-#if HAVE_STRING_H
+#ifndef DEBUG
+#define NDEBUG
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_ASSERT_H
+#ifdef HAVE_ASSERT_H
 #include <assert.h>
 #endif
 
@@ -78,9 +81,6 @@ static int *s_prev_line2width = NULL;
 static int s_preedit_lines = 0;
 static int s_prev_preedit_lines = 0;
 
-static uim_context s_context;
-/* ơ饤μ */
-static int s_status_type;
 /* üޥΥե뵭һ */
 static int s_master;
 /* ⡼ɾʸ */
@@ -93,7 +93,6 @@ static char s_win_no[WINNOSIZE];
 static const char *s_path_getmode;
 /* üѴȤTRUE */
 static int s_winch = FALSE;
-static int s_on_the_spot = FALSE;
 
 static void init_backtick(void);
 static void update_backtick(void);
@@ -101,10 +100,9 @@ static void start_preedit(void);
 static void end_preedit(void);
 static void draw_statusline(int force, int restore, int visible, int draw_background);
 static void draw_preedit(struct preedit_tag *preedit, struct preedit_tag *prev_preedit);
-static void draw_subpreedit(struct preedit_tag *p, int start, int end);
 static int is_eq_region(void);
+static void draw_subpreedit(struct preedit_tag *p, int start, int end);
 static void draw_pseg(struct preedit_segment_tag *pseg, int start_width);
-static void change_attr(int attr);
 static int compare_preedit(struct preedit_tag *p1, struct preedit_tag *p2);
 static int compare_preedit_rev(struct preedit_tag *p1, struct preedit_tag *p2);
 static int min(int a, int b);
@@ -118,21 +116,17 @@ static struct point_tag width2point_col(int width);
 static int width2lineno_char(int width);
 static int width2lineno_col(int width);
 
-#if DEBUG > 2
+#if defined(DEBUG) && DEBUG > 2
 static void print_preedit(struct preedit_tag *p);
 #endif
 
-void init_draw(uim_context context, int on_the_spot, int status_type, int master, const char *path_getmode)
+void init_draw(int master, const char *path_getmode)
 {
-  s_context = context;
-  s_status_type = status_type;
   s_master = master;
   s_path_getmode = path_getmode;
   s_preedit = create_preedit();
-  s_on_the_spot = on_the_spot;
-  if (s_status_type == BACKTICK) {
+  if (g_opt.status_type == BACKTICK) {
     init_backtick();
-    draw_statusline_no_restore();
   }
 }
 
@@ -169,25 +163,10 @@ void draw(void)
 
   /* üѹ줿Ȥs_headѹ */
   if (s_winch && g_start_preedit) {
-    /* üѹΥ */
-    struct point_tag cursor = width2point_char(s_preedit->cursor);
-    /* ץꥨǥåƬа */
-    int diff_row = cursor.row - s_head.row;
-    int diff_col = cursor.col - s_head.col;
-    /* s_headѹ */
-    s_head = get_cursor_position();
-    s_head.row -= diff_row;
-    s_head.col -= diff_col;
-    /* s_head̤Фʤ褦˽ */
-    if (s_head.row < 0) {
-      s_head.row = 0;
-    } else if (s_head.row > g_win->ws_row - 1) {
-      s_head.row = g_win->ws_row - 1;
-    }
-    if (s_head.col < 0) {
-      s_head.col = 0;
-    } else if (s_head.col > g_win->ws_col - 1) {
-      s_head.col = g_win->ws_col - 1;
+    if (g_opt.no_report_cursor) {
+      s_preedit->cursor = 0;
+    } else {
+      s_head = get_cursor_position();
     }
   }
 
@@ -206,7 +185,7 @@ void draw(void)
   s_save_preedit.width = s_save_preedit.cursor = s_save_preedit.nr_psegs = 0;
 
   /* ץꥨǥåȤ̵Х᤹ */
-  draw_statusline(FALSE, !g_start_preedit, FALSE, FALSE);
+  draw_statusline(FALSE, !g_start_preedit || g_opt.no_report_cursor, FALSE, FALSE);
 
   /* ߥåȤ줿 */
   if (commit_str[0] != '\0') {
@@ -214,7 +193,17 @@ void draw(void)
     /* ץꥨǥåȤäɬפ뤫 */
     if (prev_preedit->width > 0) {
       put_cursor_invisible();
-      erase_preedit();
+      if (g_opt.no_report_cursor) {
+        put_cursor_left(prev_preedit->cursor);
+        if (g_opt.on_the_spot) {
+          put_delete(prev_preedit->width);
+        } else {
+          put_erase(prev_preedit->width);
+          put_cursor_left(prev_preedit->width);
+        }
+      } else {
+        erase_preedit();
+      }
       end_preedit();
     }
     write(s_master, commit_str, strlen(commit_str));
@@ -230,7 +219,7 @@ void draw(void)
         end_preedit();
       }
     }
-  } else {
+  } else { /* if (g_commit) */
     if (s_preedit->width > 0) {
       /* ߥåȤ줿Ȥ˥ץꥨǥåȤäΤǡ夫Ϥ
        * 뤿˥ץꥨǥåȤ¸  */
@@ -254,6 +243,11 @@ static void start_preedit(void)
 {
   if (!g_start_preedit) {
     debug2(("start_preedit()\n"));
+    g_start_preedit = TRUE;
+    if (g_opt.no_report_cursor) {
+      return;
+    }
+
     s_head = get_cursor_position();
     if (s_head.col == g_win->ws_col - 1) {
       s_head.col = 0;
@@ -267,7 +261,6 @@ static void start_preedit(void)
     s_line2width = malloc(sizeof(int));
     s_line2width[0] = s_head.col;
     s_preedit_lines = 1;
-    g_start_preedit = TRUE;
   }
 }
 
@@ -277,8 +270,13 @@ static void start_preedit(void)
  */
 static void end_preedit(void)
 {
+  debug2(("end_preedit()\n"));
   assert(g_start_preedit);
   g_start_preedit = FALSE;
+  if (g_opt.no_report_cursor) {
+    return;
+  }
+
   put_cursor_address_p(&s_head);
   s_preedit_lines = 0;
   if (s_line2width != NULL) {
@@ -289,7 +287,6 @@ static void end_preedit(void)
     free(s_prev_line2width);
     s_prev_line2width = NULL;
   }
-  debug2(("end_preedit()\n"));
 }
 
 /*
@@ -332,18 +329,14 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
   static char *candidate_str = NULL;
   static int candidate_col = UNDEFINED;
   static int mode = UNDEFINED;
-  static int mode_width = 0;
   static char *index_str = NULL;
   static int index_col = UNDEFINED;
 
-  const char *mode_str;
-
   char *prev_statusline_str;
   int prev_statusline_str_width;
   char *prev_candidate_str;
   int prev_candidate_col;
   int prev_mode;
-  int prev_mode_width;
   char *prev_index_str;
   int prev_index_col;
 
@@ -363,12 +356,10 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
   prev_candidate_str = candidate_str;
   prev_candidate_col = candidate_col;
   prev_mode = mode;
-  prev_mode_width = mode_width;
   prev_index_str = index_str;
   prev_index_col = index_col;
 
   statusline_str = get_statusline_str();
-  statusline_str_width = strwidth(statusline_str);
   candidate_str = get_candidate_str();
   candidate_col = get_candidate_col();
   mode = get_mode();
@@ -386,16 +377,10 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
 
   /* õ */
   if (statusline_str[0] == '\0' && prev_statusline_str[0] != '\0') {
-    if (s_status_type == LASTLINE) {
-      if (restore) {
-        put_save_cursor();
-      }
-      put_cursor_invisible();
-      put_goto_lastline(0);
-      put_clear_to_end_of_line(prev_statusline_str_width);
+    if (g_opt.status_type == LASTLINE) {
       /* äϥ⡼ɤ褹ɬפ */
       force = TRUE;
-    } else if (s_status_type == BACKTICK) {
+    } else if (g_opt.status_type == BACKTICK) {
       s_candbuf[0] = '\0';
     }
   } else {
@@ -404,65 +389,68 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
       /* ʤΤθϤʤ */
       prev_candidate_col = UNDEFINED;
       prev_index_col = UNDEFINED;
-      if (s_status_type == LASTLINE) {
+      if (g_opt.status_type == LASTLINE) {
         if (restore) {
           put_save_cursor();
         }
         put_cursor_invisible();
-        put_exit_standout_mode();
-        put_exit_underline_mode();
         put_goto_lastline(0);
         /* 䤬򤵤Ƥ뤫 */
         if (candidate_col != UNDEFINED) {
           int byte_cand = (width2byte(statusline_str, candidate_col))[0];
-          int byte_index = (width2byte(statusline_str, index_col))[0];
-          put_uim_str_len(statusline_str, byte_cand);
-          put_enter_standout_mode();
-          put_uim_str(candidate_str);
-          put_exit_standout_mode();
-          put_uim_str_len(statusline_str + byte_cand + strlen(candidate_str), byte_index - byte_cand - strlen(candidate_str));
-          assert(index_col != UNDEFINED);
-          put_uim_str(index_str);
-          put_uim_str(statusline_str + byte_index + strlen(index_str));
+          int byte_index;
+          put_uim_str_len(statusline_str, UPreeditAttr_None, byte_cand);
+          put_uim_str(candidate_str, UPreeditAttr_Reverse);
+          if (index_col == UNDEFINED) {
+            put_uim_str(statusline_str + byte_cand + strlen(candidate_str), UPreeditAttr_None);
+          } else {
+            if (g_opt.ddskk) {
+              put_uim_str(statusline_str + byte_cand + strlen(candidate_str), UPreeditAttr_None);
+            } else {
+              byte_index = (width2byte(statusline_str, index_col))[0];
+              put_uim_str_len(statusline_str + byte_cand + strlen(candidate_str),
+                  UPreeditAttr_None,
+                  byte_index - byte_cand - strlen(candidate_str));
+              put_uim_str(index_str, UPreeditAttr_None);
+              put_uim_str(statusline_str + byte_index + strlen(index_str), UPreeditAttr_None);
+            }
+          }
         } else {
-          put_uim_str(statusline_str);
+          put_uim_str(statusline_str, UPreeditAttr_None);
         }
+        statusline_str_width = strwidth(statusline_str);
         if (draw_background) {
           put_clear_to_end_of_line(g_win->ws_col - statusline_str_width);
         } else if (statusline_str_width < prev_statusline_str_width) {
           put_clear_to_end_of_line(prev_statusline_str_width - statusline_str_width);
         }
         goto end_candidate;
-      } else if (s_status_type == BACKTICK) {
+      } else if (g_opt.status_type == BACKTICK) {
         strncpy(s_candbuf, statusline_str, CANDSIZE - 1);
       }
     }
     if (prev_candidate_col != candidate_col) {
       /* θȿž᤹ */
       if (prev_candidate_col != UNDEFINED) {
-        if (s_status_type == LASTLINE) {
+        if (g_opt.status_type == LASTLINE) {
           if (restore) {
             put_save_cursor();
           }
           put_cursor_invisible();
-          put_exit_standout_mode();
-          put_exit_underline_mode();
           put_goto_lastline(prev_candidate_col);
-          put_uim_str(prev_candidate_str);
+          put_uim_str(prev_candidate_str, UPreeditAttr_None);
         }
       }
       /* 򤵤줿ȿž */
       if (candidate_col != UNDEFINED) {
-        if (s_status_type == LASTLINE) {
+        if (g_opt.status_type == LASTLINE) {
           if (restore) {
             put_save_cursor();
           }
           put_cursor_invisible();
-          put_exit_underline_mode();
-          put_enter_standout_mode();
           put_goto_lastline(candidate_col);
-          put_uim_str(candidate_str);
-        } else if (s_status_type == BACKTICK) {
+          put_uim_str(candidate_str, UPreeditAttr_Reverse);
+        } else if (g_opt.status_type == BACKTICK) {
           int byte;
           strncpy(s_candbuf, statusline_str, CANDSIZE - 1);
           byte = (width2byte(statusline_str, candidate_col))[0] + strlen(candidate_str);
@@ -476,15 +464,13 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
         }
       }
     }
-    if (index_col != UNDEFINED) {
-      if (s_status_type == LASTLINE) {
+    if (index_col != UNDEFINED && !g_opt.ddskk) {
+      if (g_opt.status_type == LASTLINE) {
         int i = 0;
         if (restore) {
           put_save_cursor();
         }
         put_cursor_invisible();
-        put_exit_underline_mode();
-        put_exit_standout_mode();
         /* index_strascii */
         if (prev_index_col != UNDEFINED) {
           for (i = 0; i < (int)strlen(index_str); i++) {
@@ -495,9 +481,9 @@ static void draw_statusline(int force, int restore, int visible, int draw_backgr
         }
         if (i < (int)strlen(index_str)) {
           put_goto_lastline(index_col + i);
-          put_uim_str(index_str + i);
+          put_uim_str(index_str + i, UPreeditAttr_None);
         }
-      } else if (s_status_type == BACKTICK) {
+      } else if (g_opt.status_type == BACKTICK) {
         memcpy(s_candbuf + (width2byte(statusline_str, index_col))[0], index_str, strlen(index_str));
       }
     }
@@ -506,7 +492,7 @@ end_candidate:
 
   if (force || mode != prev_mode) {
 
-    /* ߤΥ⡼ɤUIM_FEP_SETMODE˽񤭹 */
+    /* ߤΥ⡼ɤUIM_FEP_GETMODE˽񤭹 */
     if (s_path_getmode[0] != '\0') {
       FILE *fp = fopen(s_path_getmode, "wt");
       if (fp) {
@@ -515,27 +501,26 @@ end_candidate:
       }
     }
 
-    if (s_status_type != NONE && statusline_str[0] == '\0') {
-      mode_str = uim_get_mode_name(s_context, mode);
+    if (g_opt.status_type != NONE && statusline_str[0] == '\0') {
+      char *mode_str = get_mode_str();
       return_if_fail(mode_str != NULL);
-      if (s_status_type == LASTLINE) {
+      if (g_opt.status_type == LASTLINE) {
         if (restore) {
           put_save_cursor();
         }
         put_cursor_invisible();
         put_goto_lastline(0);
-        put_exit_standout_mode();
-        put_exit_underline_mode();
-        put_uim_str(mode_str);
-        mode_width = strwidth(mode_str);
+        put_uim_str(mode_str, UPreeditAttr_None);
+        statusline_str_width = strwidth(mode_str);
         if (draw_background) {
-          put_clear_to_end_of_line(g_win->ws_col - mode_width);
-        } else if (prev_mode_width > mode_width) {
-          put_clear_to_end_of_line(prev_mode_width - mode_width);
+          put_clear_to_end_of_line(g_win->ws_col - statusline_str_width);
+        } else if (statusline_str_width < prev_statusline_str_width) {
+          put_clear_to_end_of_line(prev_statusline_str_width - statusline_str_width);
         }
-      } else if (s_status_type == BACKTICK) {
+      } else if (g_opt.status_type == BACKTICK) {
         strncpy(s_modebuf, mode_str, MODESIZE - 1);
       }
+      free(mode_str);
     }
   }
   free(prev_candidate_str);
@@ -544,7 +529,7 @@ end_candidate:
   if (restore) {
     put_restore_cursor();
   }
-  if (s_status_type == BACKTICK) {
+  if (g_opt.status_type == BACKTICK) {
     update_backtick();
   }
   if (visible) {
@@ -562,22 +547,12 @@ void draw_statusline_restore(void)
   draw_statusline(FALSE, TRUE, TRUE, FALSE);
 }
 
-/*
- * ơ饤Υ⡼ɽmodeˤ
- * ϺǲԤ˰ư
- */
-void draw_statusline_no_restore(void)
-{
-  draw_statusline(FALSE, FALSE, FALSE, TRUE);
-}
-
 /*
  * ǲԤ褹
  * ϺǲԤ˰ư
  */
 void draw_statusline_force_no_restore(void)
 {
-  assert(s_status_type == LASTLINE);
   draw_statusline(TRUE, FALSE, FALSE, TRUE);
 }
 
@@ -587,7 +562,6 @@ void draw_statusline_force_no_restore(void)
  */
 void draw_statusline_force_restore(void)
 {
-  assert(s_status_type == LASTLINE);
   draw_statusline(TRUE, TRUE, TRUE, TRUE);
 }
 
@@ -597,7 +571,7 @@ void draw_statusline_force_restore(void)
  */
 void clear_lastline(void)
 {
-  assert(s_status_type == LASTLINE);
+  assert(g_opt.status_type == LASTLINE);
   put_goto_lastline(0);
   put_clear_to_end_of_line(g_win->ws_col);
 }
@@ -607,7 +581,7 @@ void clear_lastline(void)
  */
 void clear_backtick(void)
 {
-  assert(s_status_type == BACKTICK);
+  assert(g_opt.status_type == BACKTICK);
   sendline("");
 }
 
@@ -640,30 +614,61 @@ static void draw_preedit(struct preedit_tag *preedit, struct preedit_tag *prev_p
 
   /* preedit == prev_preeditΤȤϡΰư */
   if (eq_width == preedit->width && eq_width == prev_preedit->width && eq_width > 0) {
-    goto_char(preedit->cursor);
+    if (g_opt.no_report_cursor) {
+      put_move_cur(prev_preedit->cursor, preedit->cursor);
+    } else {
+      goto_char(preedit->cursor);
+    }
     return;
   }
 
-  set_line2width(preedit);
+  if (!g_opt.no_report_cursor) {
+    set_line2width(preedit);
+  }
 
   /* Ϥ֤˰ư */
-  goto_col(eq_width);
+  if (g_opt.no_report_cursor) {
+    put_move_cur(prev_preedit->cursor, eq_width);
+  } else {
+    goto_col(eq_width);
+  }
 
   /* ΰ褬ѤäƤʤΤѹʬ */
-  if (is_eq_region()) {
+  if ((g_opt.no_report_cursor && preedit->width == prev_preedit->width) || (!g_opt.no_report_cursor && is_eq_region())) {
     int eq_width_rev = compare_preedit_rev(preedit, prev_preedit);
     debug2(("eq_width_rev = %d\n", eq_width_rev));
     draw_subpreedit(preedit, eq_width, preedit->width - eq_width_rev);
-    goto_char(preedit->cursor);
+    if (g_opt.no_report_cursor) {
+      put_move_cur(preedit->width - eq_width_rev, preedit->cursor);
+    } else {
+      goto_char(preedit->cursor);
+    }
     return;
   }
 
+  if (g_opt.no_report_cursor && g_opt.on_the_spot && preedit->width > prev_preedit->width) {
+    put_insert(preedit->width - prev_preedit->width);
+  }
   draw_subpreedit(preedit, eq_width, preedit->width);
 
-  erase_prev_preedit();
+  if (g_opt.no_report_cursor) {
+    if (preedit->width > prev_preedit->width) {
+      put_cursor_left(preedit->width - preedit->cursor);
+    } else {
+      if (g_opt.on_the_spot) {
+        put_delete(prev_preedit->width - preedit->width);
+        put_cursor_left(preedit->width - preedit->cursor);
+      } else {
+        put_erase(prev_preedit->width - preedit->width);
+        put_cursor_left(prev_preedit->width - preedit->cursor);
+      }
+    }
+  } else {
+    erase_prev_preedit();
+    /* ΰ֤˰ư */
+    goto_char(preedit->cursor);
+  }
 
-  /* ΰ֤˰ư */
-  goto_char(preedit->cursor);
 }
 
 static int is_eq_region(void)
@@ -721,7 +726,11 @@ static void draw_subpreedit(struct preedit_tag *p, int start, int end)
     char *seg_str = p->pseg[i].str;
     int seg_w = strwidth(seg_str);
     if (w + seg_w <= width) {
-      draw_pseg(&(p->pseg[i]), start + w);
+      if (g_opt.no_report_cursor) {
+        put_uim_str(seg_str, p->pseg[i].attr);
+      } else {
+        draw_pseg(&(p->pseg[i]), start + w);
+      }
       w += seg_w;
       if (w == width) {
         break;
@@ -731,7 +740,11 @@ static void draw_subpreedit(struct preedit_tag *p, int start, int end)
       int byte = byte_width[0];
       int save_char = seg_str[byte];
       seg_str[byte] = '\0';
-      draw_pseg(&(p->pseg[i]), start + w);
+      if (g_opt.no_report_cursor) {
+        put_uim_str(seg_str, p->pseg[i].attr);
+      } else {
+        draw_pseg(&(p->pseg[i]), start + w);
+      }
       seg_str[byte] = save_char;
       w += byte_width[1];
       break;
@@ -753,20 +766,19 @@ static void draw_pseg(struct preedit_segment_tag *pseg, int start_width)
   int seg_w = strwidth(seg_str);
 
   assert(margin >= 0 && margin <= g_win->ws_col);
-  change_attr(pseg->attr);
 
   while (TRUE) {
     int *byte_width;
     int byte;
     int width;
 
-    if (s_on_the_spot && s_line2width[lineno] > s_prev_line2width[lineno]) {
+    if (g_opt.on_the_spot && s_line2width[lineno] > s_prev_line2width[lineno]) {
       int margin2 = margin - g_win->ws_col - s_prev_line2width[lineno];
       if (seg_w >= margin2) {
         byte_width = width2byte(seg_str, margin2);
         byte = byte_width[0];
         width = byte_width[1];
-        put_uim_str_len(seg_str, byte);
+        put_uim_str_len(seg_str, pseg->attr, byte);
         if (width < margin2) {
           put_cursor_address(s_head.row + lineno, s_prev_line2width[lineno]);
           put_insert(s_line2width[lineno] - s_prev_line2width[lineno]);
@@ -783,7 +795,7 @@ static void draw_pseg(struct preedit_segment_tag *pseg, int start_width)
 
     /* ޤ֤ɬפʤ */
     if (seg_w < margin) {
-      put_uim_str(seg_str);
+      put_uim_str(seg_str, pseg->attr);
       break;
     }
 
@@ -792,16 +804,15 @@ static void draw_pseg(struct preedit_segment_tag *pseg, int start_width)
     width = byte_width[1];
 
     /* ޤǽ */
-    put_uim_str_len(seg_str, byte);
+    put_uim_str_len(seg_str, pseg->attr, byte);
 
     /* üʸäɬפ뤫 */
     if (s_line2width[lineno] < s_prev_line2width[lineno]) {
-      if (s_on_the_spot) {
+      if (g_opt.on_the_spot) {
         put_delete(s_prev_line2width[lineno] - s_line2width[lineno]);
       } else {
         put_erase(s_prev_line2width[lineno] - s_line2width[lineno]);
       }
-      change_attr(pseg->attr);
       s_prev_line2width[lineno] = s_line2width[lineno];
     }
 
@@ -820,25 +831,6 @@ static void draw_pseg(struct preedit_segment_tag *pseg, int start_width)
   }
 }
 
-/*
- * °attrˤ
- */
-static void change_attr(int attr)
-{
-  if (!(attr & UPreeditAttr_Reverse)) {
-    put_exit_standout_mode();
-  }
-  if (!(attr & UPreeditAttr_UnderLine)) {
-    put_exit_underline_mode();
-  }
-  if (attr & UPreeditAttr_Reverse) {
-    put_enter_standout_mode();
-  }
-  if (attr & UPreeditAttr_UnderLine) {
-    put_enter_underline_mode();
-  }
-}
-
 /*
  * p1p2Ƭζʬʸ(°)֤
  */
@@ -898,10 +890,10 @@ static void erase_prev_preedit(void)
 {
   int lineno;
   for (lineno = 0; lineno < s_preedit_lines; lineno++) {
-    assert(!s_on_the_spot || s_line2width[lineno] <= s_prev_line2width[lineno]);
+    assert(!g_opt.on_the_spot || s_line2width[lineno] <= s_prev_line2width[lineno]);
     if (s_prev_line2width[lineno] > s_line2width[lineno]) {
       put_cursor_address(s_head.row + lineno, s_line2width[lineno]);
-      if (s_on_the_spot) {
+      if (g_opt.on_the_spot) {
         put_delete(s_prev_line2width[lineno] - s_line2width[lineno]);
       } else {
         put_erase(s_prev_line2width[lineno] - s_line2width[lineno]);
@@ -911,7 +903,7 @@ static void erase_prev_preedit(void)
   for (; lineno < s_prev_preedit_lines; lineno++) {
     if (s_prev_line2width[lineno] > 0) {
       put_cursor_address(s_head.row + lineno, 0);
-      if (s_on_the_spot) {
+      if (g_opt.on_the_spot) {
         put_delete(s_prev_line2width[lineno]);
       } else {
         put_erase(s_prev_line2width[lineno]);
@@ -1087,7 +1079,7 @@ void draw_winch(void)
   s_winch = TRUE;
 }
 
-#if DEBUG > 2
+#if defined(DEBUG) && DEBUG > 2
 static void print_preedit(struct preedit_tag *p)
 {
   int i;
