@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -34,9 +34,7 @@
 #ifndef DRAW_H
 #define DRAW_H
 
-#if HAVE_UIM_UIM_H
 #include <uim/uim.h>
-#endif
 
 #define FORCE -1
 #define CLEAR -2
@@ -46,10 +44,9 @@
 extern int g_start_preedit;
 extern int g_commit;
 
-void init_draw(uim_context context, int on_the_spot, int status_type, int master, const char *path_getmode);
+void init_draw(int master, const char *path_getmode);
 void draw(void);
 void draw_statusline_restore(void);
-void draw_statusline_no_restore(void);
 void draw_statusline_force_no_restore(void);
 void draw_statusline_force_restore(void);
 void clear_lastline(void);
