@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,32 +31,34 @@
 
 */
 
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 #ifndef DEBUG
 #define NDEBUG
 #endif
 #include <stdio.h>
-#if HAVE_CURSES_H
+#ifdef HAVE_CURSES_H
 #include <curses.h>
 #endif
-#if HAVE_TERM_H
+#ifdef HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#if HAVE_ASSERT_H
+#ifdef HAVE_ASSERT_H
 #include <assert.h>
 #endif
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_CTYPE_H
+#ifdef HAVE_CTYPE_H
 #include <ctype.h>
 #endif
-#if HAVE_STRING_H
+#ifdef HAVE_STRING_H
 #include <string.h>
 #endif
 
@@ -66,10 +68,9 @@
 #include "str.h"
 #include "read.h"
 
-/* ä */
-static int s_use_civis = FALSE;
-/* ơ饤μ */
-static int s_status_type;
+#define my_putp(str) tputs(str, 1, my_putchar);
+
+
 /* TRUE */
 static int s_init = FALSE;
 /* ߤΥ */
@@ -82,8 +83,6 @@ static int s_cursor_invisible = FALSE;
 static int s_save = FALSE;
 /* ¸ */
 static struct point_tag s_save_cursor;
-/* ץꥨǥåȤȥơ饤褹⡼ */
-static int s_uim_mode = FALSE;
 /* uim_modeˤʤȤ˽Ϥ륨ץ */
 static const char *s_enter_uim_mode;
 /* enter_underline_mode˴ޤޤ */
@@ -139,30 +138,48 @@ static struct attribute_tag s_attr = {
   FALSE      /* background */
 };
 
-static void fixtty(void);
+static void check_escseq(void);
 static char *escseq2n(const char *escseq);
 static void escseq2n2(const char *escseq, const char **first, const char **second);
-static const char *attr2escseq(const struct attribute_tag *attr);
-static void set_attr(const char *str);
-static void put_enter_uim_mode(void);
-static void put_exit_uim_mode(void);
 #ifndef HAVE_CFMAKERAW
 static int cfmakeraw(struct termios *termios_p);
 #endif
+static void change_background_attr(struct attribute_tag *from, struct attribute_tag to);
+static const char *attr2escseq(const struct attribute_tag *attr);
+static void set_attr(const char *str, int len);
+static int my_putchar(int c);
 
-#if DEBUG > 1
+#if defined(DEBUG) && DEBUG > 1
 static void print_attr(struct attribute_tag *attr);
 #endif
 
+void init_escseq(const struct attribute_tag *attr_uim)
+{
+  s_attr_uim = *attr_uim;
+
+  s_enter_underline_num = escseq2n(enter_underline_mode);
+  s_exit_underline_num = escseq2n(exit_underline_mode);
+  s_enter_standout_num = escseq2n(enter_standout_mode);
+  s_exit_standout_num = escseq2n(exit_standout_mode);
+  s_bold_num = escseq2n(enter_bold_mode);
+  s_blink_num = escseq2n(enter_blink_mode);
+  s_orig_pair_num = escseq2n(orig_pair);
+  escseq2n2(orig_pair, &s_orig_fore_num, &s_orig_back_num);
+  s_enter_uim_mode = attr2escseq(&s_attr_uim);
+  if (s_enter_uim_mode != NULL) {
+    s_enter_uim_mode = strdup(s_enter_uim_mode);
+  }
+
+  fixtty();
+  check_escseq();
+  s_init = TRUE;
+}
+
 /*
  * termcap/terminfoΥȥ꤬뤫ǧ
  */
-void init_escseq(int use_civis, int use_ins_del, int status_type, const struct attribute_tag *attr_uim)
+static void check_escseq(void)
 {
-  s_use_civis = use_civis;
-  s_status_type = status_type;
-  s_attr_uim = *attr_uim;
-
   if (enter_underline_mode == NULL) {
     printf("enter_underline_mode is not available\n");
     done(EXIT_FAILURE);
@@ -183,21 +200,53 @@ void init_escseq(int use_civis, int use_ins_del, int status_type, const struct a
     printf("exit_attribute_mode is not available\n");
     done(EXIT_FAILURE);
   }
-  if (clr_eol == NULL) {
-    printf("clr_eol is not available\n");
-    done(EXIT_FAILURE);
-  }
-  if (cursor_address == NULL) {
-    printf("cursor_address is not available\n");
-    done(EXIT_FAILURE);
-  }
-  if (s_status_type == LASTLINE) {
+
+  if (g_opt.status_type == LASTLINE) {
+    if (cursor_address == NULL) {
+      printf("cursor_address is not available\n");
+      done(EXIT_FAILURE);
+    }
+    if (clr_eol == NULL) {
+      printf("clr_eol is not available\n");
+      done(EXIT_FAILURE);
+    }
     if (change_scroll_region == NULL) {
       printf("change_scroll_region is not available\n");
       done(EXIT_FAILURE);
     }
+    if (g_opt.no_report_cursor) {
+      if (cursor_up == NULL) {
+        printf("cursor_up is not available.\n");
+        done(EXIT_FAILURE);
+      }
+    }
   }
-  if (use_ins_del) {
+
+  if (g_opt.no_report_cursor) {
+    if (save_cursor == NULL) {
+      printf("save_cursor is not available.\n");
+      done(EXIT_FAILURE);
+    }
+    if (restore_cursor == NULL) {
+      printf("restore_cursor is not available.\n");
+      done(EXIT_FAILURE);
+    }
+    if (cursor_left == NULL) {
+      printf("cursor_left is not available\n");
+      done(EXIT_FAILURE);
+    }
+    if (cursor_right == NULL) {
+      printf("cursor_right is not available\n");
+      done(EXIT_FAILURE);
+    }
+  } else {
+    if (cursor_address == NULL) {
+      printf("cursor_address is not available\n");
+      done(EXIT_FAILURE);
+    }
+  }
+
+  if (g_opt.on_the_spot) {
     if (parm_ich == NULL) {
       printf("parm_ich is not available\n");
       done(EXIT_FAILURE);
@@ -207,20 +256,6 @@ void init_escseq(int use_civis, int use_ins_del, int status_type, const struct a
       done(EXIT_FAILURE);
     }
   }
-  s_enter_underline_num = escseq2n(enter_underline_mode);
-  s_exit_underline_num = escseq2n(exit_underline_mode);
-  s_enter_standout_num = escseq2n(enter_standout_mode);
-  s_exit_standout_num = escseq2n(exit_standout_mode);
-  s_bold_num = escseq2n(enter_bold_mode);
-  s_blink_num = escseq2n(enter_blink_mode);
-  s_orig_pair_num = escseq2n(orig_pair);
-  escseq2n2(orig_pair, &s_orig_fore_num, &s_orig_back_num);
-  s_enter_uim_mode = attr2escseq(&s_attr_uim);
-  if (s_enter_uim_mode != NULL) {
-    s_enter_uim_mode = strdup(s_enter_uim_mode);
-  }
-  fixtty();
-  s_init = TRUE;
 }
 
 /*
@@ -240,7 +275,7 @@ static char *escseq2n(const char *escseq)
     return NULL;
   }
   n2 = n += 2;
-  while (isdigit(n2[0])) {
+  while (isdigit((unsigned char)n2[0])) {
     n2++;
   }
   if (n2[0] == 'm') {
@@ -270,7 +305,7 @@ static void escseq2n2(const char *escseq, const char **first, const char **secon
     return;
   }
   n2 = n += 2;
-  while (isdigit(n2[0])) {
+  while (isdigit((unsigned char)n2[0])) {
     n2++;
   }
   if (n2[0] == ';') {
@@ -281,7 +316,7 @@ static void escseq2n2(const char *escseq, const char **first, const char **secon
     return;
   }
   n = ++n2;
-  while (isdigit(n2[0])) {
+  while (isdigit((unsigned char)n2[0])) {
     n2++;
   }
   if (n2[0] == 'm') {
@@ -295,15 +330,15 @@ static void escseq2n2(const char *escseq, const char **first, const char **secon
 
 void quit_escseq(void)
 {
-  if (s_init == FALSE) {
+  if (!s_init) {
     return;
   }
   put_exit_attribute_mode();
-  if (s_status_type == LASTLINE) {
+  if (g_opt.status_type == LASTLINE) {
     put_save_cursor();
     put_change_scroll_region(0, g_win->ws_row);
     put_goto_lastline(0);
-    putp(clr_eol);
+    my_putp(clr_eol);
     put_restore_cursor();
   }
   put_restore_cursor();
@@ -313,53 +348,79 @@ void quit_escseq(void)
 /*
  * üraw⡼ɤˤ
  */
-static void fixtty(void)
+void fixtty(void)
 {
   struct point_tag start_cursor;
   struct point_tag cursor;
   struct point_tag cursor2;
   static struct termios tios;
 
-  tcgetattr(STDIN_FILENO, &tios);
+  tcgetattr(g_win_in, &tios);
   cfmakeraw(&tios);
   /* read_stdinޤǤɤޤʤФʤʤǾʸ */
   tios.c_cc[VMIN] = 0;
   /* read_stdinޤǤΥॢ 0.1ñ */
   tios.c_cc[VTIME] = 3;
-  tcsetattr(STDIN_FILENO, TCSAFLUSH, &tios);
-  /* ϰ֤¸ */
-  start_cursor = get_cursor_position();
-  if (start_cursor.row == UNDEFINED) {
-    printf("Report Cursor Position is not available\r\n");
-    done(EXIT_FAILURE);
+  tcsetattr(g_win_in, TCSANOW, &tios);
+
+  if (s_init) {
+    s_cursor.row = s_cursor.col = UNDEFINED;
+    /* λΰ֤ */
+    put_restore_cursor();
   }
+
+  if (!g_opt.no_report_cursor) {
+    /* ϰ֤¸ */
+    start_cursor = get_cursor_position();
+    if (start_cursor.row == UNDEFINED) {
+      g_opt.no_report_cursor = TRUE;
+    }
+  }
+
+  if (g_opt.no_report_cursor) {
+    if (g_opt.status_type == LASTLINE) {
+      put_cursor_invisible();
+      put_crlf();
+      my_putp(cursor_up);
+      put_save_cursor();
+      put_change_scroll_region(0, g_win->ws_row - 1);
+      put_restore_cursor();
+      put_cursor_normal();
+    }
+    draw_statusline_force_restore();
+    return;
+  }
+
   put_cursor_invisible();
   /* ǲԤ鳫ϤȤΤ˥ */
-  if (s_status_type == LASTLINE) {
-    putchar('\n');
+  if (g_opt.status_type == LASTLINE) {
+    write(g_win_out, "\n", strlen("\n"));
   }
-  /* ʰ֤˰ư */
-  put_cursor_address(1, 1);
-  s_cursor.row = s_cursor.col = UNDEFINED;
-  /* ֤ */
-  cursor = get_cursor_position();
-  s_cursor.row = s_cursor.col = UNDEFINED;
-  /* ֤˰ư */
-  put_cursor_address_p(&cursor);
-  s_cursor.row = s_cursor.col = UNDEFINED;
-  /* Ʊ֤뤫 */
-  cursor2 = get_cursor_position();
-  /* ʤä麹ʬĴ٤ */
-  s_cursor_diff.row = cursor2.row - cursor.row;
-  s_cursor_diff.col = cursor2.col - cursor.col;
-  start_cursor.row -= s_cursor_diff.row;
-  start_cursor.col -= s_cursor_diff.col;
-  if (s_status_type == LASTLINE) {
+
+  if (!s_init) {
+    s_cursor.row = s_cursor.col = UNDEFINED;
+    /* ʰ֤˰ư */
+    put_cursor_address(1, 1);
+    s_cursor.row = s_cursor.col = UNDEFINED;
+    /* ֤ */
+    cursor = get_cursor_position();
+    s_cursor.row = s_cursor.col = UNDEFINED;
+    /* ֤˰ư */
+    put_cursor_address_p(&cursor);
+    s_cursor.row = s_cursor.col = UNDEFINED;
+    /* Ʊ֤뤫 */
+    cursor2 = get_cursor_position();
+    /* ʤä麹ʬĴ٤ */
+    s_cursor_diff.row = cursor2.row - cursor.row;
+    s_cursor_diff.col = cursor2.col - cursor.col;
+    start_cursor.row -= s_cursor_diff.row;
+    start_cursor.col -= s_cursor_diff.col;
+  }
+
+  if (g_opt.status_type == LASTLINE) {
     put_change_scroll_region(0, g_win->ws_row - 1);
   }
-  if (s_status_type != BACKTICK) {
-    draw_statusline_no_restore();
-  }
+  draw_statusline_force_no_restore();
   /* ϰ֤ */
   put_cursor_address_p(&start_cursor);
   put_cursor_normal();
@@ -377,6 +438,33 @@ static int cfmakeraw(struct termios *termios_p)
 }
 #endif
 
+void put_move_cur(int from, int to)
+{
+  if (to < from) {
+    put_cursor_left(from - to);
+  } else if (from < to) {
+    put_cursor_right(to - from);
+  }
+}
+
+void put_cursor_left(int n)
+{
+  int i;
+  debug(("<left %d>", n));
+  for (i = 0; i < n; i++) {
+    my_putp(cursor_left);
+  }
+}
+
+void put_cursor_right(int n)
+{
+  int i;
+  debug(("<right %d>", n));
+  for (i = 0; i < n; i++) {
+    my_putp(cursor_right);
+  }
+}
+
 /*
  * ֤¸
  */
@@ -384,8 +472,13 @@ void put_save_cursor(void)
 {
   if (!s_save) {
     s_save = TRUE;
-    s_save_cursor = get_cursor_position();
     debug(("<put_save_cursor>"));
+    if (g_opt.no_report_cursor) {
+      my_putp(save_cursor);
+      s_save_cursor = s_cursor;
+    } else {
+      s_save_cursor = get_cursor_position();
+    }
   }
 }
 
@@ -396,8 +489,15 @@ void put_restore_cursor(void)
 {
   if (s_save) {
     s_save = FALSE;
-    put_cursor_address_p(&s_save_cursor);
     debug(("<put_restore_cursor>"));
+    if (g_opt.no_report_cursor) {
+      my_putp(restore_cursor);
+      /* DOSץץȤǤ1restore_cursorǤʤ */
+      my_putp(restore_cursor);
+      s_cursor = s_save_cursor;
+    } else {
+      put_cursor_address_p(&s_save_cursor);
+    }
   }
 }
 
@@ -408,96 +508,74 @@ void put_restore_cursor(void)
  */
 struct point_tag get_cursor_position(void)
 {
-  char ibuf[100];
-  char *ibuf_offset;
-  ssize_t len;
-  char *escseq = NULL;
+  char ibuf[BUFSIZ];
+  ssize_t len = 0;
+  ssize_t read_len = 0;
+  char *escseq = ibuf - 1;
   int loop_count = 0;
-  char *unget_buf = NULL;
-  char *unget_buf2 = NULL;
-  int buf_size = 0;
-  int offset = 0;
+
+  assert(!g_opt.no_report_cursor);
+
   if (s_cursor.row != UNDEFINED) {
     return s_cursor;
   }
-  printf("\033[6n");
 
-  do {
-    /* 10롼פƤä齪λ */
-    if (loop_count++ >= 10) {
-      break;
-    }
-    ibuf_offset = ibuf + offset;
-    len = read_stdin(ibuf_offset, sizeof(ibuf) - offset);
-    if (len <= 0) {
-      continue;
-    }
-    ibuf_offset[len] = '\0';
-    debug2(("get = %s\n", ibuf));
-    escseq = strstr(ibuf, "\033");
-    if (escseq != NULL) {
-      int n;
-      unget_buf2 = malloc(offset + len);
-      /* unget_buf2[0] == 'R' */
-      n = sscanf(escseq, "\033[%d;%d%s", &(s_cursor.row), &(s_cursor.col), unget_buf2);
-      if (n < 3) {
-        /* ESCϤ뤬, ץ󥹤ȤƤԽʬ */
-        offset += len;
-        s_cursor.row = s_cursor.col = UNDEFINED;
-      } else {
-        char *R = strchr(escseq, 'R');
-        /* Rɬ */
-        assert(R != NULL);
-        /* Rμʸ뤫 */
-        if (R[1] != '\0') {
-          strcpy(unget_buf2, R + 1);
-        } else {
-          free(unget_buf2);
-          unget_buf2 = NULL;
-        }
+  write(g_win_out, "\033[6n", strlen("\033[6n"));
+
+  while (TRUE) {
+    char *next_escseq;
+    len += (read_len = read_stdin(ibuf + len, sizeof(ibuf) - len));
+    if (read_len == 0) {
+      debug(("loop = %d\n", loop_count + 1));
+      if (++loop_count == 5) {
+        break;
       }
-    } else {
-      offset += len;
     }
-  } while (s_cursor.row == UNDEFINED);
+    ibuf[len] = '\0';
 
-  /* ץ󥹤ʸ󤬤뤫 */
-  if (escseq > ibuf) {
-    unget_buf = malloc(escseq - ibuf);
-    memcpy(unget_buf, ibuf, escseq - ibuf);
-    buf_size += (escseq - ibuf);
+    debug2(("get = \""));
+    debug_write2(ibuf, len);
+    debug2(("\"\n"));
+
+    if (escseq != ibuf - 1) {
+      escseq--;
+    }
+    while ((next_escseq = memchr(escseq + 1, ESCAPE_CODE, len - (escseq - ibuf) - 1)) != NULL) {
+      int n;
+      char R;
+      escseq = next_escseq;
+      n = sscanf(escseq, "\033[%d;%d%c", &(s_cursor.row), &(s_cursor.col), &R);
+      if (n == 3 && R == 'R') {
+        char *next_to_R = strchr(escseq, 'R') + 1;
+        /* ץ󥹤ʸ󤬤뤫 */
+        if (escseq > ibuf) {
+          unget_stdin(ibuf, escseq - ibuf);
+        }
+
+        /* ץ󥹤θʸ */
+        unget_stdin(next_to_R, len - (next_to_R - ibuf));
+
+        s_cursor.row--;
+        s_cursor.col--;
+        s_cursor.row -= s_cursor_diff.row;
+        s_cursor.col -= s_cursor_diff.col;
+
+        /* GNU screen ǤϤʤ뤳Ȥ */
+        if (s_cursor.col > g_win->ws_col - 1) {
+          put_crlf();
+        }
+        debug(("<get row = %d col = %d>", s_cursor.row, s_cursor.col));
+        return s_cursor;
+
+      } else {
+        /* ESCϤ뤬, ץ󥹤ȤƤԽʬ */
+        s_cursor.row = s_cursor.col = UNDEFINED;
+      }
+    }
   }
 
-  /* ץ󥹤θʸ󤬤뤫 */
-  if (unget_buf2 != NULL) {
-    int buf_size2 = strlen(unget_buf2);
-    unget_buf = realloc(unget_buf, buf_size + buf_size2);
-    memcpy(unget_buf + buf_size, unget_buf2, buf_size2);
-    buf_size += buf_size2;
-    free(unget_buf2);
-  }
-
-  if (buf_size > 0) {
-    unget_stdin(unget_buf, buf_size);
-    free(unget_buf);
-  }
-
-  if (s_cursor.row == UNDEFINED) {
-    /*  */
-    /* s_cursor.row = s_cursor.col = 0; */
-    return s_cursor;
-  }
-
-  s_cursor.row--;
-  s_cursor.col--;
-  s_cursor.row -= s_cursor_diff.row;
-  s_cursor.col -= s_cursor_diff.col;
-
-  /* screenǤϤʤ뤳Ȥ */
-  if (s_cursor.col > g_win->ws_col - 1) {
-    put_crlf();
-  }
-  debug(("<get row = %d col = %d>", s_cursor.row, s_cursor.col));
+  unget_stdin(ibuf, len);
+  /*  */
   return s_cursor;
 }
 
@@ -506,9 +584,9 @@ struct point_tag get_cursor_position(void)
  */
 void put_cursor_invisible(void)
 {
-  if (s_use_civis && !s_cursor_invisible && cursor_invisible != NULL && cursor_normal != NULL) {
+  if (g_opt.use_civis && !s_cursor_invisible && cursor_invisible != NULL && cursor_normal != NULL) {
     s_cursor_invisible = TRUE;
-    putp(cursor_invisible);
+    my_putp(cursor_invisible);
     debug(("<invis>"));
   }
 }
@@ -520,78 +598,11 @@ void put_cursor_normal(void)
 {
   if (s_cursor_invisible) {
     s_cursor_invisible = FALSE;
-    putp(cursor_normal);
+    my_putp(cursor_normal);
     debug(("<norm>"));
   }
 }
 
-/*
- * underline⡼ɤ򳫻Ϥ
- */
-void put_enter_underline_mode(void)
-{
-  put_enter_uim_mode();
-  if (!s_attr.underline) {
-    s_attr.underline = TRUE;
-    putp(enter_underline_mode);
-    debug(("<{_>"));
-  }
-}
-
-/*
- * underline⡼ɤˤʤäƤ顢underline⡼ɤstandout⡼ɤλ
- */
-void put_exit_underline_mode(void)
-{
-  if (s_attr.underline) {
-    if (!s_uim_mode) {
-      s_uim_mode = TRUE;
-      debug(("<{uim>"));
-      s_attr_pty = s_attr;
-    }
-    put_exit_attribute_mode();
-    if (s_enter_uim_mode != NULL) {
-      putp(s_enter_uim_mode);
-      debug(("%s", s_enter_uim_mode));
-    }
-    /* debug(("<_}>")); */
-  }
-}
-
-/*
- * standout⡼ɤ򳫻Ϥ
- */
-void put_enter_standout_mode(void)
-{
-  put_enter_uim_mode();
-  if (!s_attr.standout) {
-    s_attr.standout = TRUE;
-    putp(enter_standout_mode);
-    debug(("<{r>"));
-  }
-}
-
-/*
- * standout⡼ɤˤʤäƤ顢underline⡼ɤstandout⡼ɤλ
- */
-void put_exit_standout_mode(void)
-{
-  if (s_attr.standout) {
-    if (!s_uim_mode) {
-      s_uim_mode = TRUE;
-      debug(("<{uim>"));
-      s_attr_pty = s_attr;
-    }
-    put_exit_attribute_mode();
-    if (s_enter_uim_mode != NULL) {
-      putp(s_enter_uim_mode);
-      debug(("%s", s_enter_uim_mode));
-      debug(("<{uim>"));
-    }
-    /* debug(("<r}>")); */
-  }
-}
-
 /*
  * standout⡼ɡunderline⡼ɤλ
  * ʸطʿ򸵤᤹
@@ -599,22 +610,23 @@ void put_exit_standout_mode(void)
 void put_exit_attribute_mode(void)
 {
   s_attr = s_attr_none;
-  putp(exit_attribute_mode);
+  my_putp(exit_attribute_mode);
   debug(("<a}>"));
 }
 
 /*
- * °fromtoѹ
+ * °fromtoѹfrom <= to
  */
-static void change_attr(const struct attribute_tag *from, const struct attribute_tag *to)
+static void change_attr(struct attribute_tag *from, const struct attribute_tag *to)
 {
   const char *escseq;
-  if ((from->underline && !to->underline)
-      || (from->standout && !to->standout)
-      || (from->bold && !to->bold)
-      || (from->blink && !to->blink)
-      || (from->foreground != FALSE && to->foreground == FALSE)
-      || (from->background != FALSE && to->background == FALSE)
+
+  if (   (from->underline           && !to->underline          )
+      || (from->standout            && !to->standout           )
+      || (from->bold                && !to->bold               )
+      || (from->blink               && !to->blink              )
+      || (from->foreground != FALSE &&  to->foreground == FALSE)
+      || (from->background != FALSE &&  to->background == FALSE)
       ) {
     put_exit_attribute_mode();
     escseq = attr2escseq(to);
@@ -641,38 +653,27 @@ static void change_attr(const struct attribute_tag *from, const struct attribute
     escseq = attr2escseq(&attr);
   }
   if (escseq != NULL) {
-    putp(escseq);
+    my_putp(escseq);
+    *from = *to;
     debug(("change_attr %s\n", escseq));
   }
 }
 
 /*
- * uim⡼ɤ򳫻Ϥ
+ * ط°fromtoѹfrom <= to
+ * ʸѤʤ⤷ʤ
  */
-static void put_enter_uim_mode(void)
+static void change_background_attr(struct attribute_tag *from, struct attribute_tag to)
 {
-  if (!s_uim_mode) {
-    s_uim_mode = TRUE;
-    s_attr_pty = s_attr;
-    s_attr_uim.standout = s_attr.standout;
-    s_attr_uim.underline = s_attr.underline;
-    change_attr(&s_attr, &s_attr_uim);
-    debug(("<{uim>"));
-    s_attr = s_attr_uim;
-  }
-}
-
-/*
- * uim⡼ɤλ
- */
-static void put_exit_uim_mode(void)
-{
-  if (s_uim_mode) {
-    s_uim_mode = FALSE;
-    change_attr(&s_attr, &s_attr_pty);
-    debug(("<uim}>"));
-    s_attr = s_attr_pty;
+  if (!(    (from->underline           && !to.underline          )
+        ||  (from->standout            && !to.standout           )
+        ||  (from->blink               && !to.blink              )
+        ||  (from->background != FALSE &&  to.background == FALSE)
+      )) {
+    to.bold = from->bold;
+    to.foreground = from->foreground;
   }
+  change_attr(from, &to);
 }
 
 /*
@@ -740,28 +741,38 @@ static const char *attr2escseq(const struct attribute_tag *attr)
 
 /*
  * str°ѹ륨ץ󥹤õ
+ * len > 0
  */
-static void set_attr(const char *str)
+static void set_attr(const char *str, int len)
 {
   /* ץ󥹤γ */
   const char *start;
   const char *free_str = NULL;
   /* TRUEΤȤfree_strfreeɬפ */
   int must_free = FALSE;
+  const char *end = str + len;
   /* ΥХåե뤫 */
   if (s_escseq_buf != NULL) {
     const char *tmp_str = str;
+    /* s_escseq_buf ˤ'\0'ϴޤޤʤ */
+    int escseq_buf_len = strlen(s_escseq_buf);
     must_free = TRUE;
-    free_str = str = malloc(strlen(s_escseq_buf) + strlen(tmp_str) + 1);
-    sprintf((char *)str, "%s%s", s_escseq_buf, tmp_str);
+    free_str = str = malloc(escseq_buf_len + len + 1);
+    memcpy((char *)str, s_escseq_buf, escseq_buf_len);
+    memcpy((char *)str + escseq_buf_len, tmp_str, len);
+    ((char *)str)[escseq_buf_len + len] = '\0';
     free(s_escseq_buf);
     s_escseq_buf = NULL;
+    end = str + escseq_buf_len + len;
   }
 
-  while ((start = str = strstr(str, "\033")) != NULL) {
+  while ((start = str = memchr(str, ESCAPE_CODE, end - str)) != NULL) {
     str++;
-    if (str[0] == '\0') {
-      s_escseq_buf = strdup(start);
+    if (str == end) {
+      int escseq_buf_len = end - start;
+      s_escseq_buf = malloc(escseq_buf_len + 1);
+      memcpy(s_escseq_buf, start, escseq_buf_len);
+      s_escseq_buf[escseq_buf_len] = '\0';
     }
 
     else if (str[0] == '[') {
@@ -771,10 +782,10 @@ static void set_attr(const char *str)
       params[0] = 0;
       str++;
 
-      while (isdigit(str[0]) || str[0] == ';') {
-        if (isdigit(str[0])) {
+      while (isdigit((unsigned char)str[0]) || str[0] == ';') {
+        if (isdigit((unsigned char)str[0])) {
           int n = 0;
-          while (isdigit(str[0])) {
+          while (isdigit((unsigned char)str[0])) {
             n = n * 10 + str[0] - '0';
             str++;
           }
@@ -788,8 +799,11 @@ static void set_attr(const char *str)
         }
       }
 
-      if (str[0] == '\0') {
-        s_escseq_buf = strdup(start);
+      if (str == end) {
+        int escseq_buf_len = end - start;
+        s_escseq_buf = malloc(escseq_buf_len + 1);
+        memcpy(s_escseq_buf, start, escseq_buf_len);
+        s_escseq_buf[escseq_buf_len] = '\0';
       }
 
       else if (str[0] == 'm') {
@@ -834,6 +848,7 @@ static void set_attr(const char *str)
   if (must_free) {
     free((char *)free_str);
   }
+  s_attr_pty = s_attr;
 }
 
 /*
@@ -842,6 +857,7 @@ static void set_attr(const char *str)
  */
 void put_cursor_address(int row, int col)
 {
+  const char *tmp;
   if (row >= g_win->ws_row) {
     row = g_win->ws_row - 1;
   }
@@ -849,7 +865,8 @@ void put_cursor_address(int row, int col)
   if (row == s_cursor.row && col == s_cursor.col && col < g_win->ws_col - 2) {
     return;
   }
-  putp(tparm(cursor_address, row, col));
+  tmp = tparm(cursor_address, row, col);
+  my_putp(tmp);
   s_cursor.row = row;
   s_cursor.col = col;
   debug(("<go %d %d>", row, col));
@@ -865,10 +882,12 @@ void put_cursor_address_p(struct point_tag *p)
  */
 void put_insert(int n)
 {
+  const char *tmp;
   if (n <= 0) {
     return;
   }
-  putp(tparm(parm_ich, n));
+  tmp = tparm(parm_ich, n);
+  my_putp(tmp);
   debug(("<ins %d>", n));
 }
 
@@ -877,14 +896,20 @@ void put_insert(int n)
  */
 void put_delete(int n)
 {
+  const char *tmp;
+
   if (n <= 0) {
     return;
   }
+
   if (back_color_erase) {
-    put_exit_underline_mode();
-    put_exit_standout_mode();
+    s_attr_uim.standout = FALSE;
+    s_attr_uim.underline = FALSE;
+    change_background_attr(&s_attr, s_attr_uim);
   }
-  putp(tparm(parm_dch, n));
+
+  tmp = tparm(parm_dch, n);
+  my_putp(tmp);
   debug(("<del %d>", n));
 }
 
@@ -893,7 +918,7 @@ void put_delete(int n)
  */
 void put_crlf(void)
 {
-  printf("\r\n");
+  write(g_win_out, "\r\n", strlen("\r\n"));
   s_cursor.col = 0;
   s_cursor.row++;
   if (s_cursor.row >= g_win->ws_row) {
@@ -907,11 +932,13 @@ void put_crlf(void)
  */
 void put_goto_lastline(int col)
 {
+  const char *tmp;
   int row = g_win->ws_row;
   if (row == s_cursor.row && col == s_cursor.col) {
     return;
   }
-  putp(tparm(cursor_address, row, col));
+  tmp = tparm(cursor_address, row, col);
+  my_putp(tmp);
   s_cursor.row = row;
   s_cursor.col = col;
   debug(("<go %d %d>", row, col));
@@ -925,18 +952,32 @@ void put_erase(int n)
 {
   int i;
   char *spaces;
+
   if (n <= 0) {
     return;
   }
+
   spaces = malloc(n + 1);
-  put_exit_underline_mode();
-  put_exit_standout_mode();
   for (i = 0; i < n; i++) {
     spaces[i] = ' ';
   }
   spaces[n] = '\0';
-  put_uim_str(spaces);
+
+  if (s_escseq_buf != NULL) {
+    free(s_escseq_buf);
+    s_escseq_buf = NULL;
+  }
+
+  s_attr_uim.standout = FALSE;
+  s_attr_uim.underline = FALSE;
+  change_background_attr(&s_attr, s_attr_uim);
+
+  s_cursor.col += n;
+  assert(s_cursor.col <= g_win->ws_col || g_opt.no_report_cursor);
+  write(g_win_out, spaces, n);
+
   free(spaces);
+  debug(("<put erase %d>", n));
 }
 
 /*
@@ -945,16 +986,16 @@ void put_erase(int n)
  */
 void put_clear_to_end_of_line(int width)
 {
-  put_enter_uim_mode();
   if (s_attr_uim.background != FALSE && !back_color_erase) {
     put_erase(width);
     return;
   }
   if (back_color_erase) {
-    put_exit_standout_mode();
-    put_exit_underline_mode();
+    s_attr_uim.standout = FALSE;
+    s_attr_uim.underline = FALSE;
+    change_background_attr(&s_attr, s_attr_uim);
   }
-  putp(clr_eol);
+  my_putp(clr_eol);
   debug(("<clear>"));
 }
 
@@ -963,7 +1004,8 @@ void put_clear_to_end_of_line(int width)
  */
 void put_change_scroll_region(int start, int end)
 {
-  putp(tparm(change_scroll_region, start, end));
+  const char *tmp = tparm(change_scroll_region, start, end);
+  my_putp(tmp);
   s_cursor.row = s_cursor.col = 0;
   debug(("<region %d %d>", start, end));
 }
@@ -972,8 +1014,9 @@ void put_change_scroll_region(int start, int end)
  * strü˽Ϥ
  * ̤αüۤƤϤʤ
  * ץ󥹤ϴޤޤʤ
+ * ȿžattrǻ
  */
-void put_uim_str(const char *str)
+void put_uim_str(const char *str, int attr)
 {
   if (str[0] == '\0') {
     return;
@@ -982,22 +1025,49 @@ void put_uim_str(const char *str)
     free(s_escseq_buf);
     s_escseq_buf = NULL;
   }
-  put_enter_uim_mode();
+
+  s_attr_uim.standout = attr & UPreeditAttr_Reverse;
+  s_attr_uim.underline = attr & UPreeditAttr_UnderLine;
+  change_attr(&s_attr, &s_attr_uim);
+
   s_cursor.col += strwidth(str);
-  assert(s_cursor.col <= g_win->ws_col);
-  printf("%s", str);
+  assert(s_cursor.col <= g_win->ws_col || g_opt.no_report_cursor);
+  write(g_win_out, str, strlen(str));
   debug(("<put_uim_str \"%s\">", str));
 }
 
-/* strlenϤ
+/*
+ * strlenϤput_uim_str
  * const char *strȤʤäƤ뤬Ū˽񴹤Τstrʸ
  * ˤƤϤʤ
  */
-void put_uim_str_len(const char *str, int len)
+void put_uim_str_len(const char *str, int attr, int len)
 {
   char save_char = str[len];
   ((char *)str)[len] = '\0';
-  put_uim_str(str);
+  put_uim_str(str, attr);
+  ((char *)str)[len] = save_char;
+}
+
+/*
+ * str򿧤ʤǽϤput_uim_str
+ */
+void put_uim_str_no_color(const char *str, int attr)
+{
+  struct attribute_tag save_attr = s_attr_uim;
+  s_attr_uim = s_attr_none;
+  put_uim_str(str, attr);
+  s_attr_uim = save_attr;
+}
+
+/*
+ * strlenϤput_uim_str_no_color
+ */
+void put_uim_str_no_color_len(const char *str, int attr, int len)
+{
+  char save_char = str[len];
+  ((char *)str)[len] = '\0';
+  put_uim_str_no_color(str, attr);
   ((char *)str)[len] = save_char;
 }
 
@@ -1005,14 +1075,15 @@ void put_uim_str_len(const char *str, int len)
  * ptyνstrü˽Ϥ
  * ץ󥹤ޤ礬
  */
-void put_pty_str(const char *str)
+void put_pty_str(const char *str, int len)
 {
-  if (str[0] == '\0') {
+  if (len == 0) {
     return;
   }
-  put_exit_uim_mode();
-  printf("%s", str);
-  set_attr(str);
+  change_attr(&s_attr, &s_attr_pty);
+  /* put_exit_uim_mode(); */
+  write(g_win_out, str, len);
+  set_attr(str, len);
   g_commit = FALSE;
   s_cursor.row = s_cursor.col = UNDEFINED;
   debug(("<put_pty_str \"%s\">", str));
@@ -1034,7 +1105,7 @@ char *cut_padding(const char *escseq)
   for (i = 0, j = 0; escseq[i] != '\0';) {
     if (escseq[i] == '$' && escseq[i + 1] == '<') {
       int i2 = i + 1;
-      while (isdigit(escseq[++i2]));
+      while (isdigit((unsigned char)escseq[++i2]));
       if (escseq[i2] == '>') {
         i = i2 + 1;
         if (escseq[i] == '\0') {
@@ -1056,7 +1127,14 @@ void escseq_winch(void)
   s_cursor.row = s_cursor.col = UNDEFINED;
 }
 
-#if DEBUG > 1
+static int my_putchar(int c)
+{
+  char ch = c;
+  write(g_win_out, &ch, 1);
+  return c;
+}
+
+#if defined(DEBUG) && DEBUG > 1
 static void print_attr(struct attribute_tag *attr)
 {
   debug(("underline = %d standout = %d bold = %d blink = %d fore = %d back = %d\n",
