@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,78 +31,81 @@
 
 */
 
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 #ifndef DEBUG
 #define NDEBUG
 #endif
 #include <stdio.h>
-#if HAVE_ASSERT_H
+#ifdef HAVE_ASSERT_H
 #include <assert.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#if HAVE_SYS_TYPES_H
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
-#if HAVE_SIGNAL_H
+#ifdef HAVE_SIGNAL_H
 #include <signal.h>
 #endif
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_UIM_UIM_H
 #include <uim/uim.h>
-#endif
-#if HAVE_STRING_H
+#ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-#if HAVE_STRINGS_H
+#ifdef HAVE_STRINGS_H
 #include <strings.h>
 #endif
-#if HAVE_CURSES_H
+#ifdef HAVE_CURSES_H
 #include <curses.h>
 #endif
-#if HAVE_TERM_H
+#ifdef HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
 #endif
-#if HAVE_STDARG_H
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
 #endif
-#if HAVE_GETOPT_H
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #endif
-#if HAVE_PWD_H
+#ifdef HAVE_PWD_H
 #include <pwd.h>
 #endif
-#if HAVE_FCNTL_H
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
-#if HAVE_CTYPE_H
+#ifdef HAVE_CTYPE_H
 #include <ctype.h>
 #endif
-#if HAVE_SYS_STAT_H
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
-#if HAVE_SYS_TIME_H
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
-#if HAVE_STROPTS_H
+#ifdef HAVE_STROPTS_H
 #include <stropts.h>
 #endif
-#if HAVE_SYS_PARAM_H
+#ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
 
-#if HAVE_PTY_H
+#ifdef HAVE_PTY_H
 #include <pty.h>
 #endif
-#if HAVE_UTIL_H
+#ifdef HAVE_UTMP_H
+#include <utmp.h>
+#endif
+#ifdef HAVE_UTIL_H
 #include <util.h>
 #endif
-#if HAVE_LIBUTIL_H
+#ifdef HAVE_LIBUTIL_H
 #include <libutil.h>
 #endif
 
@@ -115,14 +118,25 @@
 #include "key.h"
 #include "read.h"
 
-/* global variables */
-struct winsize *g_win;
 #define DEFAULT_STATUS LASTLINE
-#define DEFAULT_BACKTICK_WIDTH 70
+
+/* global variables */
+struct opt_tag g_opt = {
+  DEFAULT_STATUS, /* status_type       */
+  FALSE,          /* ddskk             */
+  FALSE,          /* cursor_no_reverse */
+  FALSE,          /* use_civis         */
+  FALSE,          /* on_the_spot       */
+  UNDEFINED,      /* statusline_width  */
+  0,              /* timeout           */
+  FALSE,          /* no_report_cursor  */
+  FALSE           /* print_key        */
+};
+int g_win_in = STDIN_FILENO;
+int g_win_out = STDOUT_FILENO;
+struct winsize *g_win;
 
 static uim_context s_context;
-/* ơ饤μ */
-static int s_status_type = DEFAULT_STATUS;
 /* ümasterΥե뵭һ */
 static int s_master;
 /* ưü */
@@ -133,20 +147,19 @@ static struct termios s_save_tios;
 static char s_path_setmode[MAXPATHLEN];
 static char s_path_getmode[MAXPATHLEN];
 static int s_setmode_fd = -1;
-static int s_timeout = 0;
 
-static void init_agent(const char *engine, const char *enc);
+static void init_agent(const char *engine);
 static const char *get_default_im_name(void);
 static int make_color_escseq(const char *instr, struct attribute_tag *attr);
 static int colorname2n(const char *name);
-#ifndef HAVE_FORKPTY
-static pid_t forkpty(int *amaster, char *name, struct termios *termp, struct winsize *winp);
-#endif
+static pid_t my_forkpty(int *amaster, struct termios *termp, struct winsize *winp);
 static void main_loop(void);
-static void recover_loop();
+static void recover_loop(void);
 static struct winsize *get_winsize(void);
 static void set_signal_handler(void);
 static void recover(int sig_no);
+static void sigtstp_handler(int sig_no);
+static void sigcont_handler(int sig_no);
 static void sigwinch_handler(int sig_no);
 static void sigusr1_handler(int sig_no);
 static void sigusr2_handler(int sig_no);
@@ -157,15 +170,27 @@ static void version(void);
 /*
  * uim
  * engine Ѵ󥸥̾
- * enc 󥳡ǥ
  */
-static void init_agent(const char *engine, const char *enc)
+static void init_agent(const char *engine)
 {
+  int nr;
+  int i;
   if (uim_init() == -1) {
     printf("uim_init error\n");
     exit(EXIT_FAILURE);
   }
   s_context = uim_create_context(NULL, get_enc(), NULL, engine, uim_iconv, commit_cb);
+  nr = uim_get_nr_im(s_context);
+  for (i = 0; i < nr; i++) {
+    if (strcmp(engine, uim_get_im_name(s_context, i)) == 0) {
+      break;
+    }
+  }
+  if (i == nr) {
+    printf("%s is not a input method\n\n", engine);
+    usage();
+    exit(1);
+  }
 }
 
 static const char *get_default_im_name(void)
@@ -178,9 +203,7 @@ static const char *get_default_im_name(void)
     exit(EXIT_FAILURE);
   }
 
-  if ((engine = getenv("UIM_FEP")) == NULL) {
-    engine = uim_get_default_im_name("");  /* instructs native locale */
-  }
+  engine = uim_get_default_im_name("");  /* instructs native locale */
   return engine;
 }
 
@@ -190,72 +213,37 @@ int main(int argc, char **argv)
   const char **command = malloc(sizeof(const char *) * (argc + 1));
   const char *engine;
   char *sock_path = NULL; /* Socket for backtick */
-  pid_t child;
-  int use_civis = FALSE;
-  int cursor_no_reverse = FALSE;
-  int statusline_width = UNDEFINED;
-  int on_the_spot = FALSE;
+  int gnu_screen = FALSE;
   char *env_buf;
   struct attribute_tag attr_uim = {
     FALSE,     /* underline */
     FALSE,     /* standout */
     FALSE,     /* bold */
     FALSE,     /* blink */
-    FALSE,     /* foreground */
-    FALSE      /* background */
+    UNDEFINED, /* foreground */
+    UNDEFINED  /* background */
   };
+  FILE *fp;
 
-  tcflag_t save_iflag;
   int op;
 
-  /* exit if stdin is redirected */
-  if (!isatty(STDIN_FILENO)) {
-    printf("stdin is not a terminal\n");
+  if (getenv("UIM_FEP_PID")) {
+    puts("uim-fep is already running");
     return EXIT_FAILURE;
   }
 
-  tcgetattr(STDIN_FILENO, &s_save_tios);
-  setupterm(NULL, STDOUT_FILENO, NULL);
-
-  if (getenv("TMP")) {
-    struct passwd *pw = getpwuid(getuid());
-    /* Generate get mode filepath */
-    snprintf(s_path_getmode, sizeof(s_path_getmode), "%s/uim-fep-%s-%d-getmode", getenv("TMP"), pw->pw_name, getpid());
-    /* Generate set mode filepath */
-    snprintf(s_path_setmode, sizeof(s_path_setmode), "%s/uim-fep-%s-%d-setmode", getenv("TMP"), pw->pw_name, getpid());
-  } else {
-    struct passwd *pw = getpwuid(getuid());
-    /* Generate get mode filepath */
-    snprintf(s_path_getmode, sizeof(s_path_getmode), "/tmp/uim-fep-%s-%d-getmode", pw->pw_name, getpid());
-    /* Generate set mode filepath */
-    snprintf(s_path_setmode, sizeof(s_path_setmode), "/tmp/uim-fep-%s-%d-setmode", pw->pw_name, getpid());
-  }
-
-  env_buf = malloc(30);
-  snprintf(env_buf, 30, "UIM_FEP_PID=%d", getpid());
-  putenv(env_buf);
-  env_buf = malloc(strlen("UIM_FEP_GETMODE=") + strlen(s_path_getmode) + 1);
-  sprintf(env_buf, "UIM_FEP_GETMODE=%s", s_path_getmode);
-  putenv(env_buf);
-
-  if (mkfifo(s_path_setmode, 0600) != -1) {
-    env_buf = malloc(strlen("UIM_FEP_SETMODE=") + strlen(s_path_setmode) + 1);
-    sprintf(env_buf, "UIM_FEP_SETMODE=%s", s_path_setmode);
-    putenv(env_buf);
-    s_setmode_fd = open(s_path_setmode, O_RDONLY | O_NONBLOCK);
-  } else {
-    s_path_setmode[0] = '\0';
-    s_setmode_fd = -1;
-  }
-
-  if ((command[0] = getenv("SHELL")) == NULL) {
-    command[0] = "/bin/sh";
+  if ((command[0] = getenv("SHELL")) == NULL || *command[0] == '\0') {
+    struct passwd *pw;
+    if ((pw = getpwuid(getuid())) == NULL || *(command[0] = pw->pw_shell) == '\0') {
+      command[0] = "/bin/sh";
+    }
   }
   command[1] = NULL;
 
+  init_str();
   engine = get_default_im_name();
 
-  while ((op = getopt(argc, argv, "e:s:u:b:w:t:C:ciovh")) != -1) {
+  while ((op = getopt(argc, argv, "e:s:u:b:w:t:C:SciodKvh")) != -1) {
     int i;
     switch (op) {
       case 'e':
@@ -270,13 +258,13 @@ int main(int argc, char **argv)
 
       case 's':
         if (strncmp(optarg, "none", strlen(optarg)) == 0) {
-          s_status_type = NONE;
+          g_opt.status_type = NONE;
         }
         else if (strncmp(optarg, "backtick", strlen(optarg)) == 0) {
-          s_status_type = BACKTICK;
+          g_opt.status_type = BACKTICK;
         }
         else if (strncmp(optarg, "lastline", strlen(optarg)) == 0) {
-          s_status_type = LASTLINE;
+          g_opt.status_type = LASTLINE;
         }
         else {
           usage();
@@ -284,20 +272,35 @@ int main(int argc, char **argv)
         }
         break;
 
+      case 'S':
+        gnu_screen = TRUE;
+        g_opt.no_report_cursor = TRUE;
+        break;
+
+      case 'd':
+        g_opt.ddskk = TRUE;
+        break;
+
+      case 'K':
+        g_opt.print_key = TRUE;
+        g_opt.status_type = NONE;
+        g_opt.no_report_cursor = TRUE;
+        break;
+
       case 'u':
         engine = optarg;
         break;
 
       case 'c':
-        cursor_no_reverse = TRUE;
+        g_opt.cursor_no_reverse = TRUE;
         break;
 
       case 'i':
-        use_civis = TRUE;
+        g_opt.use_civis = TRUE;
         break;
 
       case 'o':
-        on_the_spot = TRUE;
+        g_opt.on_the_spot = TRUE;
         break;
 
       case 'b':
@@ -305,16 +308,16 @@ int main(int argc, char **argv)
         break;
 
       case 'w':
-        statusline_width = atoi(optarg);
-        if (statusline_width <= 0) {
+        g_opt.statusline_width = atoi(optarg);
+        if (g_opt.statusline_width <= 0) {
           usage();
           return EXIT_FAILURE;
         }
         break;
 
       case 't':
-        s_timeout = atof(optarg) * 1000000;
-        if (s_timeout <= 0) {
+        g_opt.timeout = atof(optarg) * 1000000;
+        if (g_opt.timeout <= 0) {
           usage();
           return EXIT_FAILURE;
         }
@@ -347,40 +350,116 @@ opt_end:
     return EXIT_FAILURE;
   }
 
-  g_win = get_winsize();
-  save_iflag = s_save_tios.c_iflag;
-  s_save_tios.c_iflag &= ~ISTRIP;
-  child = forkpty(&s_master, NULL, &s_save_tios, g_win);
-  s_save_tios.c_iflag = save_iflag;
-
-  if (child < 0) {
-    perror("fork");
-    return EXIT_FAILURE;
+  if (attr_uim.foreground == UNDEFINED) {
+    attr_uim.foreground = FALSE;
   }
-  if (child == 0) {
-    /* ҥץ */
-    execvp(command[0], (char *const *)command);
-    perror(command[0]);
-    done(EXIT_FAILURE);
+  if (attr_uim.background == UNDEFINED) {
+    attr_uim.background = FALSE;
+  }
+
+  if (gnu_screen) {
+    g_opt.status_type = BACKTICK;
+    s_master = PROC_FILENO;
+    g_win_in = WIN_IN_FILENO;
+    g_win_out = WIN_OUT_FILENO;
+  }
+
+  if (!isatty(g_win_in)) {
+    g_win_in = open("/dev/tty", O_RDONLY);
+  }
+
+  tcgetattr(g_win_in, &s_save_tios);
+  setupterm(NULL, g_win_out, NULL);
+
+  if (getenv("TMP")) {
+    struct passwd *pw = getpwuid(getuid());
+    /* Generate get mode filepath */
+    snprintf(s_path_getmode, sizeof(s_path_getmode), "%s/uim-fep-%s-%d-getmode", getenv("TMP"), pw->pw_name, getpid());
+    /* Generate set mode filepath */
+    snprintf(s_path_setmode, sizeof(s_path_setmode), "%s/uim-fep-%s-%d-setmode", getenv("TMP"), pw->pw_name, getpid());
+  } else {
+    struct passwd *pw = getpwuid(getuid());
+    /* Generate get mode filepath */
+    snprintf(s_path_getmode, sizeof(s_path_getmode), "/tmp/uim-fep-%s-%d-getmode", pw->pw_name, getpid());
+    /* Generate set mode filepath */
+    snprintf(s_path_setmode, sizeof(s_path_setmode), "/tmp/uim-fep-%s-%d-setmode", pw->pw_name, getpid());
+  }
+
+  env_buf = malloc(30);
+  snprintf(env_buf, 30, "UIM_FEP_PID=%d", getpid());
+  putenv(env_buf);
+
+  if ((fp = fopen(s_path_getmode, "wt")) != NULL) {
+    fclose(fp);
+    unlink(s_path_getmode);
+    env_buf = malloc(strlen("UIM_FEP_GETMODE=") + strlen(s_path_getmode) + 1);
+    sprintf(env_buf, "UIM_FEP_GETMODE=%s", s_path_getmode);
+    putenv(env_buf);
+  } else {
+    s_path_getmode[0] = '\0';
+  }
+
+  if (mkfifo(s_path_setmode, 0600) != -1) {
+    unlink(s_path_setmode);
+    env_buf = malloc(strlen("UIM_FEP_SETMODE=") + strlen(s_path_setmode) + 1);
+    sprintf(env_buf, "UIM_FEP_SETMODE=%s", s_path_setmode);
+    putenv(env_buf);
+  } else {
+    s_path_setmode[0] = '\0';
+    s_setmode_fd = -1;
+  }
+
+  g_win = get_winsize();
+  if (!gnu_screen && !g_opt.print_key) {
+    pid_t child = my_forkpty(&s_master, &s_save_tios, g_win);
+
+    if (child < 0) {
+      perror("fork");
+      return EXIT_FAILURE;
+    }
+    if (child == 0) {
+      /* ҥץ */
+      if (g_win_in != STDIN_FILENO) {
+        close(g_win_in);
+      }
+      execvp(command[0], (char *const *)command);
+      perror(command[0]);
+      done(EXIT_FAILURE);
+    }
   }
 
   free(command);
 
-  init_str();
-
-  if (s_status_type == BACKTICK && statusline_width > CANDSIZE / 2) {
-    statusline_width = CANDSIZE / 2;
+  if (g_opt.status_type == BACKTICK && g_opt.statusline_width > CANDSIZE / 2) {
+    g_opt.statusline_width = CANDSIZE / 2;
   }
 
-  setbuf(stdout, NULL);
-  if (s_status_type == BACKTICK) {
+  if (g_opt.status_type == BACKTICK) {
     init_sendsocket(sock_path);
   }
-  init_agent(engine, get_enc());
-  init_callbacks(s_context, s_status_type, cursor_no_reverse, statusline_width);
-  init_draw(s_context, on_the_spot, s_status_type, s_master, s_path_getmode);
-  init_escseq(use_civis, on_the_spot, s_status_type, &attr_uim);
+  init_agent(engine);
+  if (gnu_screen) {
+    uim_set_mode(s_context, 1);
+  }
+  init_callbacks(s_context);
+  init_draw(s_master, s_path_getmode);
+  init_escseq(&attr_uim);
   set_signal_handler();
+
+  if (s_path_setmode[0] != '\0' && mkfifo(s_path_setmode, 0600) != -1) {
+    s_setmode_fd = open(s_path_setmode, O_RDONLY | O_NONBLOCK);
+#ifndef __CYGWIN32__
+    open(s_path_setmode, O_WRONLY);
+#endif
+  } else {
+    s_path_setmode[0] = '\0';
+    s_setmode_fd = -1;
+  }
+
+  if (g_opt.print_key) {
+    printf("Press any key.\r\n");
+    printf("To exit the program, press 'q' key.\r\n");
+  }
   main_loop();
   done(EXIT_SUCCESS);
   return EXIT_SUCCESS;
@@ -474,8 +553,42 @@ static int colorname2n(const char *name)
   return UNDEFINED;
 }
 
-#ifndef HAVE_FORKPTY
-static pid_t forkpty(int *amaster, char *name, struct termios *termp, struct winsize *winp)
+#if defined(HAVE_FORKPTY)
+static pid_t my_forkpty(int *amaster, struct termios *termp, struct winsize *winp)
+{
+  pid_t pid;
+  int slave;
+
+  tcflag_t save_iflag = termp->c_iflag;
+  termp->c_iflag &= ~ISTRIP;
+
+  if (openpty(amaster, &slave, NULL, termp, g_win) == -1) {
+    perror("openpty");
+    return -1;
+  }
+
+  termp->c_iflag = save_iflag;
+
+  if ((pid = fork()) < 0) {
+    return -2;
+  }
+  if (pid == 0) {
+    /* ҥץ */
+    int redirected_stdin = dup(STDIN_FILENO);
+    close(*amaster);
+    login_tty(slave);
+    if (g_win_in != STDIN_FILENO) {
+      dup2(redirected_stdin, STDIN_FILENO);
+    }
+    close(redirected_stdin);
+    return 0;
+  } else {
+    close(slave);
+    return pid;
+  }
+}
+#elif defined(__svr4__) 
+static pid_t my_forkpty(int *amaster, struct termios *termp, struct winsize *winp)
 {
   pid_t pid;
   int slave;
@@ -508,7 +621,9 @@ static pid_t forkpty(int *amaster, char *name, struct termios *termp, struct win
       ioctl(slave, TIOCSWINSZ, winp);
     }
     close(*amaster);
-    dup2(slave, STDIN_FILENO);
+    if (g_win_in == STDIN_FILENO) {
+      dup2(slave, STDIN_FILENO);
+    }
     dup2(slave, STDOUT_FILENO);
     dup2(slave, STDERR_FILENO);
     close(slave);
@@ -525,16 +640,31 @@ static void main_loop(void)
   char buf[BUFSIZE];
   ssize_t len;
   fd_set fds;
-  int nfd = (s_master > s_setmode_fd ? s_master : s_setmode_fd) + 1;
+  int nfd;
   char *_clear_screen = cut_padding(clear_screen);
   char *_clr_eos = cut_padding(clr_eos);
 
+  if (g_win_in > s_master) {
+    if (g_win_in > s_setmode_fd) {
+      nfd = g_win_in;
+    } else {
+      nfd = s_setmode_fd;
+    }
+  } else {
+    if (s_master > s_setmode_fd) {
+      nfd = s_master;
+    } else {
+      nfd = s_setmode_fd;
+    }
+  }
+  nfd++;
+
   while (TRUE) {
     /* ߥåȤ줿Ȥ˥ץꥨǥåȤ뤫 */
     if (is_commit_and_preedit()) {
       struct timeval t;
       FD_ZERO(&fds);
-      FD_SET(STDIN_FILENO, &fds);
+      FD_SET(g_win_in, &fds);
       FD_SET(s_master, &fds);
       if (s_setmode_fd > 0) {
         FD_SET(s_setmode_fd, &fds);
@@ -549,7 +679,7 @@ static void main_loop(void)
       }
     }
     FD_ZERO(&fds);
-    FD_SET(STDIN_FILENO, &fds);
+    FD_SET(g_win_in, &fds);
     FD_SET(s_master, &fds);
     if (s_setmode_fd > 0) {
       FD_SET(s_setmode_fd, &fds);
@@ -563,16 +693,20 @@ static void main_loop(void)
     /* ⡼ɤѹ */
     if (s_setmode_fd > 0 && FD_ISSET(s_setmode_fd, &fds)) {
       int start, end;
-      if ((len = read(s_setmode_fd, buf, sizeof(buf))) <= 0) {
+#ifdef __CYGWIN32__
+      if ((len = read(s_setmode_fd, buf, sizeof(buf) - 1)) <= 0) {
         debug2(("pipe closed\n"));
         close(s_setmode_fd);
         s_setmode_fd = open(s_path_setmode, O_RDONLY | O_NONBLOCK);
       }
-      for (end = len - 1; end >= 0 && !isdigit(buf[end]); --end);
+#else
+      len = read(s_setmode_fd, buf, sizeof(buf) - 1);
+#endif
+      for (end = len - 1; end >= 0 && !isdigit((unsigned char)buf[end]); --end);
       /* ץꥨǥåȤԽǤʤХ⡼ɤѹ */
       if (end >= 0 && !g_start_preedit) {
         int mode;
-        for (start = end; start > 0 && isdigit(buf[start - 1]); --start);
+        for (start = end; start > 0 && isdigit((unsigned char)buf[start - 1]); --start);
         buf[end + 1] = '\0';
         mode = atoi(&buf[start]);
         if (mode != uim_get_current_mode(s_context)) {
@@ -586,21 +720,20 @@ static void main_loop(void)
 
 
     /* ܡ(stdin) */
-    if (FD_ISSET(STDIN_FILENO, &fds)) {
+    if (FD_ISSET(g_win_in, &fds)) {
       int i;
       int key;
       int key_state = 0;
       int key_len;
-      int raw;
 
-      if ((len = read_stdin(buf, sizeof(buf))) <= 0) {
+      if ((len = read_stdin(buf, sizeof(buf) - 1)) <= 0) {
         /* ˤϤʤȻפ */
         return;
       }
       buf[len] = '\0';
       debug(("read \"%s\"\n", buf));
 
-      if (len >= 10) {
+      if (len >= 10 && !g_opt.print_key) {
         /* ڡȤʤɤ̤Ϥ줿ȤѴʤ */
         if (!g_start_preedit) {
           write(s_master, buf, len);
@@ -609,41 +742,50 @@ static void main_loop(void)
 
         for (i = 0; i < len; i++) {
           key = tty2key(buf[i]);
-          if (key == UKey_Escape) {
+          key_state |= tty2key_state(buf[i]);
+          if (key == UKey_Escape && (key_state & UMod_Meta) == 0) {
+
             int *key_and_key_len = escape_sequence2key(buf + i);
             key = key_and_key_len[0];
+
             if (key == UKey_Escape) {
-              if (i + 1 < len) {
-                /* Alt+ */
-                key_state = UMod_Alt;
-                continue;
-              } else if (s_timeout > 0) {
+              int not_enough = key_and_key_len[1];
+              if (not_enough && g_opt.timeout > 0) {
                 struct timeval t;
                 FD_ZERO(&fds);
-                FD_SET(STDIN_FILENO, &fds);
+                FD_SET(g_win_in, &fds);
                 t.tv_sec = 0;
-                t.tv_usec = s_timeout;
+                t.tv_usec = g_opt.timeout;
                 if (my_select(nfd, &fds, &t) > 0) {
-                  len += read_stdin(buf + len, sizeof(buf) - len);
+                  len += read_stdin(buf + len, sizeof(buf) - len - 1);
                   buf[len] = '\0';
                   debug(("read_again \"%s\"\n", buf));
                   i--;
                   continue;
                 }
               }
+              if (i + 1 < len && key_state != UMod_Alt) {
+                key_state = UMod_Alt;
+                continue;
+              }
+              key_len = 1;
+            } else {
+              key_len = key_and_key_len[1];
             }
-            key_len = key_and_key_len[1];
           } else {
-            key_state += tty2key_state(buf[i]);
             key_len = 1;
           }
-          raw = press_key(key, key_state);
-          draw();
-          if (raw && !g_start_preedit) {
-            if (key_state & UMod_Alt) {
-              write(s_master, buf + i - 1, key_len + 1);
-            } else {
-              write(s_master, buf + i, key_len);
+          if (g_opt.print_key) {
+            print_key(key, key_state);
+          } else {
+            int raw = press_key(key, key_state);
+            draw();
+            if (raw && !g_start_preedit) {
+              if (key_state & UMod_Alt) {
+                write(s_master, buf + i - 1, key_len + 1);
+              } else {
+                write(s_master, buf + i, key_len);
+              }
             }
           }
           key_state = 0;
@@ -654,42 +796,32 @@ static void main_loop(void)
 
 
     /* input from pty (child process) */
-    if (FD_ISSET(s_master, &fds)) {
-      char *str1;
-      char *str2;
-      if ((len = read(s_master, buf, sizeof(buf))) <= 0) {
+    if (!g_opt.print_key && FD_ISSET(s_master, &fds)) {
+      if ((len = read(s_master, buf, sizeof(buf) - 1)) <= 0) {
         /* ҥץλ */
         return;
       }
       buf[len] = '\0';
 
-      /* if (g_start_preedit) { */
-        /* ץꥨǥåԽ */
-        /* put_exit_standout_mode(); */
-        /* put_exit_underline_mode(); */
-      /* } */
-
       /* ꥢ줿˥⡼ɤ褹 */
-      if (s_status_type == LASTLINE) {
-        str1 = rstrstr(buf, _clear_screen);
-        str2 = rstrstr(buf, _clr_eos);
+      if (g_opt.status_type == LASTLINE) {
+        char *str1 = rstrstr_len(buf, _clear_screen, len);
+        char *str2 = rstrstr_len(buf, _clr_eos, len);
         if (str1 != NULL || str2 != NULL) {
-          char save_char;
+          int str1_len;
           if (str2 > str1) {
             str1 = str2;
           }
+          str1_len = len - (str1 - buf);
           /* str1clear_screenclr_eosμʸؤƤ */
-          save_char = str1[0];
-          str1[0] = '\0';
-          put_pty_str(buf);
-          str1[0] = save_char;
+          put_pty_str(buf, len - str1_len);
           draw_statusline_force_restore();
-          put_pty_str(str1);
+          put_pty_str(str1, str1_len);
         } else {
-          put_pty_str(buf);
+          put_pty_str(buf, len);
         }
       } else {
-        put_pty_str(buf);
+        put_pty_str(buf, len);
       }
     }
   }
@@ -706,14 +838,14 @@ static void recover_loop(void)
 
   while (TRUE) {
     FD_ZERO(&fds);
-    FD_SET(STDIN_FILENO, &fds);
+    FD_SET(g_win_in, &fds);
     FD_SET(s_master, &fds);
     if (select(s_master + 1, &fds, NULL, NULL, NULL) <= 0) {
       /* signalǳޤ줿Ȥˤ롣select֤ͤ-1errno==EINTR */
       continue;
     }
-    if (FD_ISSET(STDIN_FILENO, &fds)) {
-      if ((len = read(STDIN_FILENO, buf, sizeof(buf))) <= 0) {
+    if (FD_ISSET(g_win_in, &fds)) {
+      if ((len = read(g_win_in, buf, sizeof(buf))) <= 0) {
         /* ˤϤʤȻפ */
         return;
       }
@@ -724,8 +856,7 @@ static void recover_loop(void)
         /* ҥץλ */
         return;
       }
-      buf[len] = '\0';
-      printf("%s", buf);
+      write(g_win_out, buf, len);
     }
   }
 }
@@ -737,8 +868,8 @@ static void recover_loop(void)
 static struct winsize *get_winsize(void)
 {
   struct winsize *win = malloc(sizeof(struct winsize));
-  ioctl(STDIN_FILENO, TIOCGWINSZ, win);
-  if (s_status_type == LASTLINE) {
+  ioctl(g_win_in, TIOCGWINSZ, win);
+  if (g_opt.status_type == LASTLINE) {
     win->ws_row--;
   }
   return win;
@@ -775,19 +906,52 @@ static void set_signal_handler(void)
 
   act.sa_handler = sigusr2_handler;
   sigaction(SIGUSR2, &act, NULL);
+
+  act.sa_handler = sigtstp_handler;
+  sigaction(SIGTSTP, &act, NULL);
+
+  act.sa_handler = sigcont_handler;
+  sigaction(SIGCONT, &act, NULL);
 }
 
 static void recover(int sig_no)
 {
   put_exit_attribute_mode();
-  put_exit_standout_mode();
-  put_exit_underline_mode();
   put_restore_cursor();
   put_cursor_normal();
   recover_loop();
   done(EXIT_SUCCESS);
 }
 
+static void sigtstp_handler(int sig_no)
+{
+  struct sigaction act;
+  sigset_t mask;
+
+  quit_escseq();
+  put_save_cursor();
+  tcsetattr(g_win_in, TCSAFLUSH, &s_save_tios);
+
+  sigemptyset(&act.sa_mask);
+  act.sa_flags = 0;
+  act.sa_handler = SIG_DFL;
+
+  sigemptyset(&mask);
+  sigaddset(&mask, SIGTSTP);
+  sigprocmask(SIG_UNBLOCK, &mask, NULL);
+
+  sigaction(SIGTSTP, &act, NULL);
+  kill(getpid(), SIGTSTP);
+  
+  act.sa_handler = sigtstp_handler;
+  sigaction(SIGTSTP, &act, NULL);
+}
+
+static void sigcont_handler(int sig_no)
+{
+  fixtty();
+}
+
 /*
  * üΥѤäȤ˲ü礭碌롣
  */
@@ -799,7 +963,7 @@ static void sigwinch_handler(int sig_no)
   escseq_winch();
   callbacks_winch();
   draw_winch();
-  if (s_status_type == LASTLINE) {
+  if (g_opt.status_type == LASTLINE) {
     put_save_cursor();
     put_cursor_invisible();
     if (g_win->ws_row > prev_win->ws_row) {
@@ -821,10 +985,10 @@ void done(int exit_value)
 {
   uim_quit();
   quit_escseq();
-  if (s_status_type == BACKTICK) {
+  if (g_opt.status_type == BACKTICK) {
     clear_backtick();
   }
-  tcsetattr(STDIN_FILENO, TCSAFLUSH, &s_save_tios);
+  tcsetattr(g_win_in, TCSAFLUSH, &s_save_tios);
   if (s_setmode_fd > 0) {
     close(s_setmode_fd);
   }
@@ -858,16 +1022,17 @@ static void usage(void)
   uim_context context;
   int nr;
   int i;
+  int max_im_name_len = 0;
 
   uim_init();
-  context = uim_create_context(NULL, NULL, NULL, NULL, uim_iconv, commit_cb);
+  context = uim_create_context(NULL, get_enc(), NULL, NULL, uim_iconv, commit_cb);
 
   printf("usage: uim-fep [OPTIONS]\n"
       "\n"
       "-u <input method>                         input method      [default=%s]\n"
       "-s <lastline/backtick/none>               statusline type   [default=%s]\n"
       "-b <file>                                 socket file       [default=%s]\n"
-      "-w <width>                                statusline width  [default=%d]\n"
+      "-w <width>                                statusline width\n"
       "%s"
       "-e command arg1 arg2 ...                  executed command  [default=%s]\n"
       "%s",
@@ -875,12 +1040,14 @@ static void usage(void)
       DEFAULT_STATUS == LASTLINE ? "lastline" :
       DEFAULT_STATUS == BACKTICK ? "backtick" : "none",
       usersockname(NULL),
-      DEFAULT_BACKTICK_WIDTH,
       "-t <sec>                                  key timeout\n"
       "-C [<foreground color>]:[<background color>]\n"
       "-c                                        reverse cursor\n"
       "-i                                        use cursor_invisible(civis)\n"
-      "-o                                        on the spot\n",
+      "-o                                        on the spot\n"
+      "-S                                        GNU screen mode\n"
+      "-d                                        ddskk like candidate style\n"
+      "-K                                        show key code\n",
       getenv("SHELL") != NULL ? getenv("SHELL") : "/bin/sh",
       "-h                                        display this help\n"
       "-v                                        display version\n"
@@ -889,13 +1056,41 @@ static void usage(void)
   printf("\n[input method]\n");
   nr = uim_get_nr_im(context);
   for (i = 0; i < nr; i++) {
-    printf("%s", uim_get_im_name(context, i));
-    printf(" (%s)", uim_get_im_language(context, i));
+    int im_name_len = strlen(uim_get_im_name(context, i));
+    im_name_len += strlen(uim_get_im_language(context, i));
+    im_name_len += strlen(" ()");
 #ifdef DEBUG
-    printf(" %s", uim_get_im_encoding(context, i));
+    im_name_len += strlen(" ");
+    im_name_len += strlen(uim_get_im_encoding(context, i));
 #endif
+    if (im_name_len > max_im_name_len) {
+      max_im_name_len = im_name_len;
+    }
+  }
+
+  for (i = 0; i < nr; i++) {
+    int j;
+    const char *im_name = uim_get_im_name(context, i);
+    int im_name_len = strlen(im_name);
+    printf("%s", im_name);
+
+    im_name = uim_get_im_language(context, i);
+    im_name_len += strlen(im_name);
+    printf(" (%s)", im_name);
+    im_name_len += strlen(" ()");
+#ifdef DEBUG
+    im_name = uim_get_im_encoding(context, i);
+    im_name_len += strlen(im_name);
+    printf(" %s", im_name);
+    im_name_len += strlen(" ");
+#endif
+    for (j = 0; j < max_im_name_len - im_name_len + 2; j++) {
+      putchar(' ');
+    }
+    printf("%s", uim_get_im_short_desc(context, i));
     printf("\n");
   }
+
   uim_quit();
 }
 
@@ -907,17 +1102,26 @@ static void version(void)
   printf("uim-fep %s\n", PACKAGE_VERSION);
 }
 
-#if DEBUG > 1
+#if defined(DEBUG) && DEBUG > 1
+static FILE *_log = NULL;
 void _debug(const char *fmt, ...)
 {
-  static FILE *log = NULL;
   va_list ap;
-  if (log == NULL) {
-    log = fopen("uim-fep-log", "w");
+  if (_log == NULL) {
+    _log = fopen("uim-fep-log", "w");
   }
   va_start(ap, fmt);
-  vfprintf(log, fmt, ap);
+  vfprintf(_log, fmt, ap);
   va_end(ap);
-  fflush(log);
+  fflush(_log);
+}
+
+void _debug_write(const char *str, int len)
+{
+  if (_log == NULL) {
+    _log = fopen("uim-fep-log", "w");
+  }
+  fwrite(str, 1, len, _log);
+  fflush(_log);
 }
 #endif
