@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -34,10 +34,10 @@
 #ifndef UIM_FEP_H
 #define UIM_FEP_H
 
-#if HAVE_SYS_IOCTL_H
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
-#if HAVE_TERMIOS_H
+#ifdef HAVE_TERMIOS_H
 #include <termios.h>
 #endif
 
@@ -47,26 +47,68 @@
 #define LASTLINE 3
 #define BACKTICK 4
 #define UNDEFINED -1729
+#define ESCAPE_CODE 27
+#define WIN_OUT_FILENO STDOUT_FILENO
+#define WIN_IN_FILENO  STDOUT_FILENO
+#define PROC_FILENO    STDIN_FILENO
 
+struct opt_tag {
+  /* ơ饤μ */
+  int status_type;
+  /* ddskk˻ɽ */
+  int ddskk;
+  /* TRUEʤ饫֤ȿžʤ */
+  int cursor_no_reverse;
+  /* ä */
+  int use_civis;
+  /* ץꥨǥåȤ뤫 */
+  int on_the_spot;
+  /* ơ饤 */
+  int statusline_width;
+  /* ESCθ˲ԤĤ */
+  int timeout;
+  /* ݡȥ뵡ǽʤ */
+  int no_report_cursor;
+  int print_key;
+};
+
+extern struct opt_tag g_opt;
+extern int g_win_in;
+extern int g_win_out;
 extern struct winsize *g_win;
+
 void done(int exit_value);
 
-#if DEBUG > 1
-#define debug(arg) _debug arg
-void _debug(const char *fmt, ...);
-#else
-#define debug(arg)
-#endif
+#ifdef DEBUG
 
 #if DEBUG > 2
 #define debug2(arg) _debug arg
+#define debug(arg) _debug arg
+#define debug_write2(str, len) _debug_write(str, len)
+#define debug_write(str, len) _debug_write(str, len)
+void _debug(const char *fmt, ...);
+
+#elif DEBUG == 2
+#define debug2(arg)
+#define debug(arg) _debug arg
+#define debug_write2(str, len)
+#define debug_write(str, len) _debug_write(str, len)
+
+void _debug(const char *fmt, ...);
 #else
 #define debug2(arg)
+#define debug(arg)
+#define debug_write2(str, len)
+#define debug_write(str, len)
 #endif
 
-#ifdef DEBUG
 #define return_if_fail(arg) if (!(arg)) { printf("assertion failed %s %d", __FILE__, __LINE__); return; }
+
 #else
+#define debug2(arg)
+#define debug(arg)
+#define debug_write2(str, len)
+#define debug_write(str, len)
 #define return_if_fail(arg) if (!(arg)) { return; }
 #endif
 
