@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -51,11 +51,13 @@
 #include "uim/uim-im-switcher.h"
 #include "uim/config.h"
 #include "uim/gettext.h"
+#include "uim/uim-compat-scm.h"
 #include "uim-cand-win-gtk.h"
+#include "caret-state-indicator.h"
 
 /* exported symbols */
-GtkIMContext *im_module_create (const gchar *context_id);
-void im_module_list(GtkIMContextInfo ***contexts, int *n_contexts);
+GtkIMContext *im_module_create(const gchar *context_id);
+void im_module_list(const GtkIMContextInfo ***contexts, int *n_contexts);
 void im_module_exit(void);
 void im_module_init(GTypeModule *type_module);
 
@@ -64,7 +66,7 @@ void im_module_init(GTypeModule *type_module);
 #define DEFAULT_SEPARATOR_STR "|"
 
 static int im_uim_fd = -1;
-static int read_tag;
+static unsigned int read_tag;
 static guint snooper_id;
 
 struct preedit_segment {
@@ -85,6 +87,7 @@ typedef struct _IMUIMContext {
   GtkWidget *menu;
   GdkWindow *win;
   GdkWindow *toplevel;
+  GtkWidget *caret_state_indicator;
   GdkRectangle preedit_pos; /* preedit_pos not always point the cursor location */
   /**/
   struct _IMUIMContext *prev, *next;
@@ -110,7 +113,8 @@ static void im_uim_init(IMUIMContext *uic);
 static void show_preedit(GtkIMContext *ic, GtkWidget *preedit_label);
 
 static void im_uim_helper_disconnect_cb(void);
-static void helper_read_cb(gpointer p, int fd, GdkInputCondition c);
+/* #ifndef GDK_DISABLE_DEPRECATED */
+static gboolean helper_read_cb(GIOChannel *channel, GIOCondition c, gpointer p);
 static GdkFilterReturn
 toplevel_window_candidate_cb(GdkXEvent *xevent, GdkEvent *ev, gpointer data);
 
@@ -139,13 +143,18 @@ static GType type_im_uim = 0;
 #define IM_UIM_CONTEXT(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj),type_im_uim,IMUIMContext))
 
 static const GtkIMContextInfo im_uim_info = {
-  "uim",
-  "UIM()",
-  "gtk+",
-  "",
-  ""
+  "uim", /* id */
+  "uim", /* human-readable name*/
+  "uim", /* domain for gettext*/
+  LOCALEDIR,
+  "ja:ko:zh:*"
 };
 
+static const GtkIMContextInfo *im_uim_info_list = {
+  &im_uim_info
+};
+
+
 static void
 im_uim_commit_string(void *ptr, const char *str)
 {
@@ -273,6 +282,25 @@ filter_keypress(GtkIMContext *ic,
 		GdkEventKey *key)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
+
+  /*** Hack for combination of xchat + GTK+ 2.6 ***/
+  if (snooper_installed == FALSE) {
+    int rv;
+    int kv = convert_keyval(key->keyval);
+    int mod = convert_modifier(key->state);
+
+    if (key->type == GDK_KEY_RELEASE) {
+      rv = uim_release_key(uic->uc, kv, mod);
+    } else {
+      rv = uim_press_key(uic->uc, kv, mod);
+    }
+    if (rv) {
+      return gtk_im_context_filter_keypress(uic->slave, key);
+    }
+    return TRUE;
+  }
+  /*** Hack for combination of xchat + GTK+ 2.6 ***/
+
   return gtk_im_context_filter_keypress(uic->slave, key);
 }
 
@@ -398,14 +426,9 @@ im_uim_set_cursor_location(GtkIMContext *ic,
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
 
-  if (uic->cwin->left &&  uic->nr_psegs > 1) {
-    return;
-  }
-
+  uic->preedit_pos = *area;
   uim_cand_win_gtk_set_cursor_location(uic->cwin, area);
-
-  uic->preedit_pos.x = area->x;
-  uic->preedit_pos.y = area->y;
+  caret_state_indicator_set_cursor_location(uic->caret_state_indicator, area);
 }
 
 
@@ -424,8 +447,11 @@ check_helper_connection()
   if (im_uim_fd < 0) {
     im_uim_fd = uim_helper_init_client_fd(im_uim_helper_disconnect_cb);
     if (im_uim_fd >= 0) {
-      read_tag = gdk_input_add(im_uim_fd, (GdkInputCondition)GDK_INPUT_READ,
-			       helper_read_cb, 0);
+      GIOChannel *channel;
+      channel = g_io_channel_unix_new(im_uim_fd);
+      read_tag = g_io_add_watch(channel, G_IO_IN | G_IO_HUP | G_IO_ERR,
+				helper_read_cb, NULL);
+      g_io_channel_unref(channel);
     }
   }
 }
@@ -439,14 +465,16 @@ focus_in(GtkIMContext *ic)
   focused_context = uic;
   disable_focused_context = FALSE;
 
-  /* XXX:This is not recommended way!! */
-  if(snooper_installed == FALSE) {
+  /* XXX:Use of key snooper is not recommended way!! */
+  if (snooper_installed == FALSE) {
     snooper_id = gtk_key_snooper_install((GtkKeySnoopFunc)uim_key_snoop, NULL );
     snooper_installed = TRUE;
   }
 
   check_helper_connection();
 
+  uim_helper_client_focus_in(uic->uc);
+
   uim_prop_list_update(uic->uc);
   uim_prop_label_update(uic->uc);
 
@@ -459,8 +487,6 @@ focus_in(GtkIMContext *ic)
   if (uic->cwin && uic->cwin_is_active) {
     gtk_widget_show(GTK_WIDGET(uic->cwin));
   }
-
-  uim_helper_client_focus_in(uic->uc);
 }
 
 static void
@@ -468,11 +494,13 @@ focus_out(GtkIMContext *ic)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
 
-  if(snooper_installed == TRUE) {
+  if (snooper_installed == TRUE) {
     gtk_key_snooper_remove(snooper_id);
     snooper_installed = FALSE;
   }
 
+  check_helper_connection();
+
   uim_helper_client_focus_out(uic->uc);
 
   if (uic->cwin) {
@@ -559,10 +587,11 @@ toplevel_window_candidate_cb(GdkXEvent *xevent, GdkEvent *ev, gpointer data)
     return GDK_FILTER_CONTINUE;
 
   if (uic->cwin && uic->cwin_is_active) {
-    gint topwin_x, topwin_y;
+    gint x, y, width, height, depth;
 
-    gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
-    uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+    gdk_window_get_geometry(uic->win, &x, &y, &width, &height, &depth);
+    gdk_window_get_origin(uic->win, &x, &y);
+    uim_cand_win_gtk_layout(uic->cwin, x, y, width, height);
   }
   return GDK_FILTER_CONTINUE;
 }
@@ -620,12 +649,12 @@ im_uim_finalize(GObject *obj)
 
   if (uic->menu) {
     gtk_widget_destroy(uic->menu);
+    uic->menu = NULL;
   }
   if (uic->cwin) {
     gtk_widget_destroy(GTK_WIDGET(uic->cwin));
     uic->cwin = NULL;
   }
-  uic->menu = NULL;
 
   uim_release_context(uic->uc);
 
@@ -662,13 +691,19 @@ im_uim_class_init(GtkIMContextClass *class)
 static void
 im_uim_class_finalize(GtkIMContextClass *class)
 {
-  uim_quit();
+
 }
 
 static void
 update_prop_list_cb(void *ptr, const char *str)
 {
-  GString *tmp = g_string_new("");
+  IMUIMContext *uic = (IMUIMContext *)ptr;
+  GString *tmp;
+
+  if (uic != focused_context)
+    return;
+
+  tmp = g_string_new("");
   g_string_printf(tmp, "prop_list_update\ncharset=UTF-8\n%s", str);
 
   uim_helper_send_message(im_uim_fd, tmp->str);
@@ -678,11 +713,25 @@ update_prop_list_cb(void *ptr, const char *str)
 static void
 update_prop_label_cb(void *ptr, const char *str)
 {
-  GString *tmp = g_string_new("");
+  IMUIMContext *uic = (IMUIMContext *)ptr;
+  GString *tmp;
+  gint x, y;
+  uim_bool show_state;
+
+  if (uic != focused_context)
+    return;
+
+  tmp = g_string_new("");
   g_string_printf(tmp, "prop_label_update\ncharset=UTF-8\n%s", str);
 
   uim_helper_send_message(im_uim_fd, tmp->str);
   g_string_free(tmp, TRUE);
+
+  show_state = uim_scm_symbol_value_bool("bridge-show-input-state?");
+  if (show_state == UIM_TRUE) {
+    gdk_window_get_origin(uic->win, &x, &y);
+    caret_state_indicator_update(uic->caret_state_indicator, x, y, str);
+  }
 }
 
 static void
@@ -690,12 +739,13 @@ im_uim_send_im_list(void)
 {
   int nr, i;
   GString *msg;
-  const char *current_im_name = uim_get_current_im_name(focused_context->uc);
-  if(focused_context) {
-    nr = uim_get_nr_im(focused_context->uc);
-  } else {
+  const char *current_im_name;
+
+  if (!focused_context)
     return;
-  }
+
+  nr = uim_get_nr_im(focused_context->uc);
+  current_im_name = uim_get_current_im_name(focused_context->uc);
 
   msg = g_string_new("im_list\ncharset=UTF-8\n");
   for (i = 0; i < nr; i++) {
@@ -710,13 +760,12 @@ im_uim_send_im_list(void)
 
     g_string_append(msg, name);
     g_string_append(msg, "\t");
-    if(lang) g_string_append(msg, lang);
+    if (lang) g_string_append(msg, lang);
     g_string_append(msg, "\t");
-    if(short_desc) g_string_append(msg, short_desc);
+    if (short_desc) g_string_append(msg, short_desc);
     g_string_append(msg, "\t");
-    if(strcmp(name, current_im_name) == 0) {
+    if (strcmp(name, current_im_name) == 0) {
       g_string_append(msg, "selected");
-      g_print("%s has been selected\n", name);
     }
     g_string_append(msg, "\n");
   }
@@ -728,7 +777,7 @@ static void
 cand_activate_cb(void *ptr, int nr, int display_limit)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  gint topwin_x, topwin_y;
+  gint x, y, width, height, depth;
   GSList *list = NULL;
   uim_candidate cand;
   gint i;
@@ -736,7 +785,7 @@ cand_activate_cb(void *ptr, int nr, int display_limit)
   uic->cwin_is_active = TRUE;
 
   for (i = 0; i < nr; i++) {
-    cand = uim_get_candidate(uic->uc, i, i % display_limit);
+    cand = uim_get_candidate(uic->uc, i, display_limit ? i % display_limit : i);
     list = g_slist_append(list, cand);
   }
 
@@ -745,8 +794,9 @@ cand_activate_cb(void *ptr, int nr, int display_limit)
   g_slist_foreach(list, (GFunc)uim_candidate_free, NULL);
   g_slist_free(list);
 
-  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
-  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  gdk_window_get_geometry(uic->win, &x, &y, &width, &height, &depth);
+  gdk_window_get_origin(uic->win, &x, &y);
+  uim_cand_win_gtk_layout(uic->cwin, x, y, width, height);
   gtk_widget_show(GTK_WIDGET(uic->cwin));
 
   if (uic->win) {
@@ -767,10 +817,11 @@ static void
 cand_select_cb(void *ptr, int index)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  gint topwin_x, topwin_y;
+  gint x, y, width, height, depth;
 
-  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
-  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  gdk_window_get_geometry(uic->win, &x, &y, &width, &height, &depth);
+  gdk_window_get_origin(uic->win, &x, &y);
+  uim_cand_win_gtk_layout(uic->cwin, x, y, width, height);
   g_signal_handlers_block_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
   uim_cand_win_gtk_set_index(uic->cwin, index);
   g_signal_handlers_unblock_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
@@ -780,10 +831,11 @@ static void
 cand_shift_page_cb(void *ptr, int direction)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  gint topwin_x, topwin_y;
+  gint x, y, width, height, depth;
 
-  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
-  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  gdk_window_get_geometry(uic->win, &x, &y, &width, &height, &depth);
+  gdk_window_get_origin(uic->win, &x, &y);
+  uim_cand_win_gtk_layout(uic->cwin, x, y, width, height);
   g_signal_handlers_block_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
   uim_cand_win_gtk_shift_page(uic->cwin, direction);
   uim_set_candidate_index(uic->uc, uic->cwin->candidate_index);
@@ -798,10 +850,12 @@ cand_deactivate_cb(void *ptr)
   uic->cwin_is_active = FALSE;
 
   if (uic->cwin) {
-    GdkWindow *toplevel;
-
     gtk_widget_hide(GTK_WIDGET(uic->cwin));
     uim_cand_win_gtk_clear_candidates(uic->cwin);
+  }
+  if (uic->win) {
+    GdkWindow *toplevel;
+
     toplevel = gdk_window_get_toplevel(uic->win);
     gdk_window_remove_filter(toplevel, toplevel_window_candidate_cb, uic);
   }
@@ -809,27 +863,22 @@ cand_deactivate_cb(void *ptr)
 
 
 GtkIMContext *
-im_module_create (const gchar *context_id)
+im_module_create(const gchar *context_id)
 {
-  GObject *obj = g_object_new(type_im_uim, NULL);
-  IMUIMContext *uic = IM_UIM_CONTEXT(obj);
-  const char *im_name = NULL;
+  GObject *obj;
+  IMUIMContext *uic;
+  const char *im_name;
 
   g_return_val_if_fail(context_id, NULL);
+  g_return_val_if_fail(!strcmp(context_id, "uim"), NULL);
 
-  if (!strncmp(context_id, "uim-", 4)) {
-    im_name = &context_id[4];
-  } else if (strcmp(context_id, "uim") == 0) {
-    im_name = uim_get_default_im_name(setlocale(LC_ALL, NULL));
-  } else {
-     g_return_val_if_fail(!strncmp(context_id, "uim", 3), NULL);
-  }
-
-  uim_init();
-  uic->uc =  uim_create_context(uic, "UTF-8",
-				NULL, im_name,
-				uim_iconv,
-				im_uim_commit_string);
+  obj = g_object_new(type_im_uim, NULL);
+  uic = IM_UIM_CONTEXT(obj);
+  im_name = uim_get_default_im_name(setlocale(LC_CTYPE, NULL));
+  uic->uc = uim_create_context(uic, "UTF-8",
+			       NULL, im_name,
+			       uim_iconv,
+			       im_uim_commit_string);
   if (uic->uc == NULL) {
     parent_class->finalize(obj);
     return NULL;
@@ -859,6 +908,8 @@ im_module_create (const gchar *context_id)
 			    NULL);
   g_signal_connect(G_OBJECT(uic->slave), "commit",
 		   G_CALLBACK(im_uim_commit_cb), uic);
+  
+  uic->caret_state_indicator = caret_state_indicator_new();
 
   /**/
   uic->next = context_list.next;
@@ -869,51 +920,11 @@ im_module_create (const gchar *context_id)
 }
 
 void
-im_module_list(GtkIMContextInfo ***contexts,
+im_module_list(const GtkIMContextInfo ***contexts,
 	       int *n_contexts)
 {
-  uim_context uc;
-  int nr, i;
-  static GtkIMContextInfo **info_list;
-
-  uim_init();
-  uc = uim_create_context(NULL, "UTF-8", NULL,
-			  NULL, NULL, NULL);
-  if (uc == NULL)
-    nr = 0;
-  else
-    nr = uim_get_nr_im(uc);
-
-  info_list = malloc(sizeof(GtkIMContextInfo*) * nr + 1);
-   
-  info_list[0] = malloc(sizeof(GtkIMContextInfo));
-  
-  info_list[0]->context_id      = "uim";
-  info_list[0]->context_name    = "uim";
-  info_list[0]->domain          = "gtk+";
-  info_list[0]->domain_dirname  = LOCALEDIR;
-  info_list[0]->default_locales = g_strdup("ja:ko:zh:*");
-  
-  for (i = 1; i < nr; i++) {
-    const char *name = uim_get_im_name(uc, i - 1);
-    const char *lang = uim_get_im_language(uc, i - 1);
-    char s[100];
-
-    info_list[i] = malloc(sizeof(GtkIMContextInfo));
-
-    sprintf(s, "uim-%s", name);
-    info_list[i]->context_id = g_strdup(s);
-
-    sprintf(s, "uim-%s (%s)", name, lang);
-    info_list[i]->context_name = g_strdup(s);
-    
-    info_list[i]->domain = "gtk+";
-    info_list[i]->domain_dirname = LOCALEDIR;
-    info_list[i]->default_locales = g_strdup(lang);
-  }
-  uim_release_context(uc);
-  *contexts = info_list;
-  *n_contexts = nr;
+  *contexts = &im_uim_info_list;
+  *n_contexts = 1;
 }
 
 static void
@@ -921,20 +932,24 @@ im_uim_parse_helper_str_im_change(const char *str)
 {
   IMUIMContext *cc;
   gchar **lines = g_strsplit(str, "\n", -1);
-  gchar* im_name = lines[1];
-  if(g_str_has_prefix(str, "im_change_this_text_area_only") == TRUE) {
-    if(focused_context) {
+  gchar *im_name = lines[1];
+  if (g_str_has_prefix(str, "im_change_this_text_area_only") == TRUE) {
+    if (focused_context) {
       uim_switch_im(focused_context->uc, im_name);
       uim_prop_list_update(focused_context->uc);
     }
-  } else if(g_str_has_prefix(str, "im_change_whole_desktop") == TRUE) {
+  } else if (g_str_has_prefix(str, "im_change_whole_desktop") == TRUE) {
     for (cc = context_list.next; cc != &context_list; cc = cc->next) {
       uim_switch_im(cc->uc, im_name);
+      if (focused_context && cc == focused_context)
+	uim_prop_list_update(cc->uc);
     }
-  } else if(g_str_has_prefix(str, "im_change_this_application_only") == TRUE) {
-    if(focused_context) {    
+  } else if (g_str_has_prefix(str, "im_change_this_application_only") == TRUE) {
+    if (focused_context) {    
       for (cc = context_list.next; cc != &context_list; cc = cc->next) {
 	uim_switch_im(cc->uc, im_name);
+	if (cc == focused_context)
+	  uim_prop_list_update(cc->uc);
       }
     }
   }
@@ -959,7 +974,7 @@ im_uim_parse_helper_str(const char *str)
       g_strfreev(lines);
     }
   } else if (focused_context && !disable_focused_context) {
-    if (g_str_has_prefix(str, "prop_list_get") == TRUE){
+    if (g_str_has_prefix(str, "prop_list_get") == TRUE) {
       uim_prop_list_update(focused_context->uc);
     } else if (g_str_has_prefix(str, "prop_label_get") == TRUE) {
       uim_prop_label_update(focused_context->uc);
@@ -973,7 +988,7 @@ im_uim_parse_helper_str(const char *str)
       im_uim_send_im_list();
     } else if (g_str_has_prefix(str, "commit_string")) {
       lines = g_strsplit(str, "\n", 0);
-      if (focused_context && lines && lines[0] && lines[1])
+      if (lines && lines[0] && lines[1])
 	g_signal_emit_by_name(focused_context, "commit", lines[1]);
     } else if (g_str_has_prefix(str, "focus_in") == TRUE) {
       disable_focused_context = TRUE;
@@ -983,22 +998,24 @@ im_uim_parse_helper_str(const char *str)
   }
 }
 
-static void
-helper_read_cb(gpointer p, int fd, GdkInputCondition c)
+static gboolean
+helper_read_cb(GIOChannel *channel, GIOCondition c, gpointer p)
 {
   char *tmp;
+  int fd = g_io_channel_unix_get_fd(channel);
   uim_helper_read_proc(fd);
   while ((tmp = uim_helper_get_message())) {
     im_uim_parse_helper_str(tmp);
     free(tmp);
   }
+  return TRUE;
 }
 
 static void
 im_uim_helper_disconnect_cb(void)
 {
   im_uim_fd = -1;
-  gdk_input_remove(read_tag);
+  g_source_remove(read_tag);
 }
 
 /* XXX:This is not a recommended way!! */
@@ -1020,12 +1037,17 @@ uim_key_snoop(GtkWidget *grab_widget, GdkEventKey *key, gpointer data)
     }
     return TRUE;
   }
+
   return FALSE;
 }
 
 void
 im_module_init(GTypeModule *type_module)
 {
+  if (uim_init() == -1) {
+    return;
+  }
+
   context_list.next = (IMUIMContext *)&context_list;
   context_list.prev = (IMUIMContext *)&context_list;
   type_im_uim =
