@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -48,6 +48,12 @@ G_BEGIN_DECLS
 typedef struct _UIMCandWinGtk      UIMCandWinGtk;
 typedef struct _UIMCandWinGtkClass UIMCandWinGtkClass;
 
+typedef enum {
+  UIM_CAND_WIN_POS_CARET,
+  UIM_CAND_WIN_POS_LEFT,
+  UIM_CAND_WIN_POS_RIGHT
+} UimCandWinPos;
+
 struct _UIMCandWinGtk {
   GtkWindow	 parent;
 
@@ -62,9 +68,17 @@ struct _UIMCandWinGtk {
   gint		 candidate_index;
   gint		 page_index;
 
-  gboolean	 left;
+  UimCandWinPos	 position;
 
   GdkRectangle	 cursor;
+
+  /* sub window */
+  struct sub_window {
+    GtkWidget	*window;
+    GtkWidget	*scrolled_window;
+    GtkWidget	*text_view;
+    gboolean     active;
+  } sub_window;
 };
 
 struct _UIMCandWinGtkClass {
@@ -75,7 +89,7 @@ struct _UIMCandWinGtkClass {
 };
 
 
-void		uim_cand_win_gtk_register_type		(GTypeModule *module);
+GType		uim_cand_win_gtk_register_type		(GTypeModule *module);
 GType		uim_cand_win_gtk_get_type		(void);
 UIMCandWinGtk  *uim_cand_win_gtk_new			(void);
 
@@ -100,7 +114,9 @@ void		uim_cand_win_gtk_set_scrollable		(UIMCandWinGtk *cwin,
 
 void		uim_cand_win_gtk_layout			(UIMCandWinGtk *cwin,
 							 gint topwin_x,
-							 gint topwin_y);
+							 gint topwin_y,
+							 gint topwin_width,
+							 gint topwin_height);
 void		uim_cand_win_gtk_set_cursor_location	(UIMCandWinGtk *cwin,
 							 GdkRectangle *area);
 
