@@ -1,4 +1,3 @@
-
 /*
 
   Copyright (c) 2004 uim Project http://uim.freedesktop.org/
@@ -18,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -45,50 +44,63 @@
 #include "uim/config.h"
 #include "uim/gettext.h"
 
+#include "dict-word-win-gtk.h"
+#include "dict-word-list-win-gtk.h"
 
-static int read_tag;
-static int uim_fd; /* file descriptor to connect helper message bus */
+static unsigned int read_tag;
+static int uim_fd;  /* file descriptor to connect helper message bus */
 static int ae_mode; /* add mode or edit mode */
 static int input_method;
 
-static enum{
+enum {
   MODE_EDIT,
-  MODE_ADD
-}MODE;
+  MODE_ADD,
+  NR_MODE
+};
 
-static enum {
+/* should be implemeted as loadable module */
+enum {
   IM_ANTHY,
+  IM_CANNA,
   IM_SKK,
-  IM_PRIME
-}IM;
+  IM_PRIME,
+  NR_IM
+};
 
 static void
 helper_disconnect_cb(void)
 {
   uim_fd = -1;
-  gdk_input_remove(read_tag);
+  g_source_remove(read_tag);
 }
 
-static void
-fd_read_cb(gpointer p, gint fd, GdkInputCondition c)
+static gboolean
+fd_read_cb(GIOChannel *channel, GIOCondition c, gpointer p)
 {
   char *tmp;
+  int fd = g_io_channel_unix_get_fd(channel);
   
   uim_helper_read_proc(fd);
   while ((tmp = uim_helper_get_message())) {
-    /*    parse_helper_str(tmp); */
-    g_free(tmp); tmp = NULL;
+    /* parse_helper_str(tmp); */
+    g_free(tmp);
+    tmp = NULL;
   }
+  return TRUE;
 }
 
 static void
 check_helper_connection(void)
 {
-  if(uim_fd < 0) {
+  if (uim_fd < 0) {
     uim_fd = uim_helper_init_client_fd(helper_disconnect_cb);
-    if(uim_fd > 0)
-      read_tag = gdk_input_add(uim_fd, (GdkInputCondition)GDK_INPUT_READ,
-			       fd_read_cb, NULL);
+    if (uim_fd > 0) {
+      GIOChannel *channel;
+      channel = g_io_channel_unix_new(uim_fd);
+      read_tag = g_io_add_watch(channel, G_IO_IN | G_IO_HUP | G_IO_ERR,
+				fd_read_cb, NULL);
+      g_io_channel_unref(channel);
+    }
   }
 }
 
@@ -96,90 +108,100 @@ static char *
 get_error_msg(void)
 {
   /* dummy */
+  return NULL;
 }
 
 static void
 parse_arg(int argc, char *argv[])
 {
-  int ch;
+  int ch, option_index;
 
-  ae_mode = MODE_ADD;
-  return;
-  /*
+  ae_mode = MODE_EDIT;
 
-  while((ch = getopt(argc, argv, "aei:")) != -1)
+  while ((ch = getopt(argc, argv, "aei:")) != -1)
   {
-      switch(ch) {
-      case 'a':
-          ae_mode = MODE_ADD;
-	  break;
-      case 'e':
-	  ae_mode = MODE_EDIT;
-	  break;
-      case 'i':
-	  if(strcmp(optarg, "anthy"))
-    	      im = IM_ANTHY;
-	  else if(strcmp(optarg, "prime"))
-	      im = IM_PRIME;
-	  else if (strcmp(optarg, "skk"))
-	      im = IM_SKK;
-	  else
-	      im = IM_ANTHY;
-	  break;
-      default:
-	  ae_mode = MODE_ADD;
-	  im = get_current_im();
-      }
+    switch (ch) {
+    case 'a':
+      ae_mode = MODE_ADD;
+      break;
+    case 'e':
+      ae_mode = MODE_EDIT;
+      break;
+    case 'i':
+      if (!strcmp(optarg, "anthy"))
+	input_method = IM_ANTHY;
+      else if (!strcmp(optarg, "canna"))
+	input_method = IM_CANNA;
+      else if (!strcmp(optarg, "prime"))
+	input_method = IM_PRIME;
+      else if (!strcmp(optarg, "skk"))
+	input_method = IM_SKK;
+      else
+	input_method = IM_ANTHY;
+      break;
+    default:
+      ae_mode = MODE_EDIT;
+      /* input_method = get_current_im(); */
+    }
   }
 
   argv += optind;
   argc -= optind;
-
-if(mode == edit) {
-ae_mode = EDIT;
-} else {
-ae_mode = ADD;
 }
 
-if(im == anthy) {
-
-input_method = anthy
-
-} else {
-input_method = get_current_im();
-}
-
-*/
-}
-
-static
+static GtkWidget *
 create_window_anthy(void)
 {
   GtkWidget *window;
-  GtkWidget *w;
-  /*  if(ae_mode == MODE_EDIT) {
-    w = anthy_edit_window_create();
-    } else {*/
-    w = anthy_add_window_create();
+  uim_dict *dict;
 
-  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
-  gtk_container_add(GTK_CONTAINER(window),w);
-  gtk_widget_show_all(window);
+  if (ae_mode == MODE_EDIT) {
+    window = word_list_window_new();
+  } else {
+    dict = uim_dict_open(N_("Anthy private dictionary"));
+    window = word_window_new(WORD_WINDOW_MODE_ADD, dict);
+    uim_dict_unref(dict);
+  }
+
+  gtk_widget_show(window);
+
+  return window;
+}
+
+static void
+window_destroy_cb(GtkWidget *widget, gpointer data)
+{
+  gtk_main_quit();
 }
 
 static int
 create_window(void)
 {
-  switch(input_method) {
+  GtkWidget *window = NULL;
+
+  switch (input_method) {
   case IM_ANTHY:
-    create_window_anthy();
+    window = create_window_anthy();
+    break;
+  case IM_CANNA:
+    /* window = create_window_canna(); */
     break;
   case IM_SKK:
-    /*    create_window_skk();*/
+    /* create_window_skk();*/
+    break;
+  case IM_PRIME:
+    /* create_window_prime();*/
     break;
   default:
     return -1;
   }
+
+  if (!window)
+    return -1;
+
+  g_signal_connect(G_OBJECT(window), "destroy",
+		   G_CALLBACK(window_destroy_cb), NULL);
+
   return 0;
 }
 
@@ -198,9 +220,8 @@ main(int argc, char *argv[])
 
   result = create_window();
 
-
-  if(result == -1) {
-    fprintf(stderr, "Error:%s\n", get_error_msg());
+  if (result == -1) {
+    g_printerr(_("Error:%s\n"), get_error_msg());
     exit(-1);
   }
 
@@ -209,5 +230,6 @@ main(int argc, char *argv[])
   check_helper_connection();  
 
   gtk_main ();
+
   return 0;
 }
