@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -44,6 +44,8 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#include "../gtk/caret-state-indicator.h"
+
 #define UIM_TYPE_CANDIDATE_WINDOW	(candidate_window_get_type())
 #define UIM_CANDIDATE_WINDOW(obj)	(G_TYPE_CHECK_INSTANCE_CAST ((obj), candidate_window_get_type(), UIMCandidateWindow))
 #define UIM_IS_CANDIDATE_WINDOW(obj)	(G_TYPE_CHECK_INSTANCE_TYPE ((obj), UIM_TYPE_CANDIDATE_WINDOW))
@@ -67,6 +69,13 @@ struct _UIMCandidateWindow {
   gint candidate_index;
   gint page_index;
 
+  gint pos_x;
+  gint pos_y;
+  gint width;
+  gint height;
+
+  GtkWidget *caret_state_indicator;
+
   gboolean is_active;
 };
 
@@ -87,6 +96,8 @@ static gint uim_cand_win_gtk_get_index(UIMCandidateWindow *cwin);
 static void uim_cand_win_gtk_set_index(UIMCandidateWindow *cwin, gint index);
 static void uim_cand_win_gtk_set_page(UIMCandidateWindow *cwin, gint page);
 
+static void uim_cand_win_gtk_layout(void);
+
 #define NR_CANDIDATES 10 /* FIXME! not used */
 #define CANDWIN_DEFAULT_WIDTH	80
 
@@ -113,6 +124,7 @@ static gboolean tree_selection_changed(GtkTreeSelection *selection,
 #if 0
 static gboolean tree_view_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
 #endif
+static gboolean configure_event_cb(GtkWidget *widget, GdkEventConfigure *event, gpointer data);
 
 static GType candidate_window_type = 0;
 static GTypeInfo const object_info = {
@@ -129,7 +141,7 @@ static GTypeInfo const object_info = {
 
 static gint cand_win_gtk_signals[NR_SIGNALS] = {0};
 
-static int read_tag;
+static unsigned int read_tag;
 
 static void init_candidate_win(void);
 static void candwin_activate(gchar **str);
@@ -211,15 +223,23 @@ tree_selection_changed(GtkTreeSelection *selection,
   idx = *indicies + cwin->display_limit * cwin->page_index;
 
   if (!path_currently_selected && cwin->candidate_index != idx) {
-    cwin->candidate_index = idx;
-    g_signal_emit(G_OBJECT(cwin),
-		  cand_win_gtk_signals[INDEX_CHANGED_SIGNAL], 0);
+    if (cwin->candidate_index >= 0) {
+      cwin->candidate_index = idx;
+      g_signal_emit(G_OBJECT(cwin),
+		    cand_win_gtk_signals[INDEX_CHANGED_SIGNAL], 0);
+    }
 
+    update_label(cwin);
+
+    if (cwin->candidate_index < 0)
+      return FALSE;
+    else
+      return TRUE;
+  } else {
+    update_label(cwin);
+
+    return TRUE;
   }
-
-  update_label(cwin);
-
-  return TRUE;
 }
 
 #if 0
@@ -276,6 +296,8 @@ init_candidate_win(void) {
   cwin = candidate_window_new();
   g_signal_connect(G_OBJECT(cwin), "index-changed",
 		   G_CALLBACK(index_changed_cb), NULL);
+  g_signal_connect(G_OBJECT(cwin), "configure_event",
+		   G_CALLBACK(configure_event_cb), NULL);
 }
 
 static void
@@ -344,7 +366,10 @@ candidate_window_init(UIMCandidateWindow *cwin)
   		   G_CALLBACK(tree_view_button_press), cwin);
 #endif
 
+  cwin->pos_x = 0;
+  cwin->pos_y = 0;
   cwin->is_active = FALSE;
+  cwin->caret_state_indicator = caret_state_indicator_new();
 
   gtk_widget_show(cwin->scrolled_window);
   gtk_widget_show(cwin->view);
@@ -359,7 +384,6 @@ candwin_activate(gchar **str)
   gint i, nr_stores = 1;
   guint j = 1;
   gchar *utf8_str;
-  gint candwin_width, candwin_height;
   const gchar *charset;
   guint display_limit;
   GSList *candidates = NULL;
@@ -374,11 +398,6 @@ candwin_activate(gchar **str)
     g_object_unref(G_OBJECT(store));
   }
 
-  gtk_window_get_size(GTK_WINDOW(cwin), &candwin_width, &candwin_height);
-  if (candwin_width > CANDWIN_DEFAULT_WIDTH)
-    gtk_window_resize(GTK_WINDOW(cwin),
-		    CANDWIN_DEFAULT_WIDTH, candwin_height);
-  
   if (!strncmp(str[1], "charset=", 8))
     charset = str[1] + 8;
   else
@@ -422,7 +441,7 @@ candwin_activate(gchar **str)
 
   /* create GtkListStores, and set candidates */
   for (i = 0; i < nr_stores; i++) {
-    GtkListStore *store = gtk_list_store_new(2, G_TYPE_UINT, G_TYPE_STRING);
+    GtkListStore *store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
     GSList *node;
 
     g_ptr_array_add(cwin->stores, store);
@@ -434,13 +453,15 @@ candwin_activate(gchar **str)
     {
       GtkTreeIter ti;
       if (node) {
-	gchar *cand = node->data;
+	gchar *str = node->data;
+	gchar **column = g_strsplit(str, "\t", 2);
 	gtk_list_store_append(store, &ti);
 	gtk_list_store_set(store, &ti,
-			   COLUMN_HEADING, j + 1,
-			   COLUMN_CANDIDATE, cand,
+			   COLUMN_HEADING, column[0],
+			   COLUMN_CANDIDATE, column[1],
 			   TERMINATOR);
-	g_free(cand);
+	g_strfreev(column);
+	g_free(str);
       } else {
 	/* No need to set any data for empty row. */
       }
@@ -448,6 +469,8 @@ candwin_activate(gchar **str)
   }
 
   uim_cand_win_gtk_set_page(cwin, 0);
+  update_label(cwin);
+
   gtk_widget_show_all(GTK_WIDGET(cwin));
   cwin->is_active = TRUE;
 }
@@ -464,32 +487,17 @@ candwin_update(gchar **str)
 static void
 candwin_move(char **str)
 {
-  int x, y;
-  int topwin_x, topwin_y;
-  int cw_wi, cw_he, sc_wi, sc_he;
+  GdkRectangle cursor_location;
 
-  sscanf(str[1], "%d", &topwin_x);
-  sscanf(str[2], "%d", &topwin_y);
+  sscanf(str[1], "%d", &cwin->pos_x);
+  sscanf(str[2], "%d", &cwin->pos_y);
 
-  gtk_window_get_size(GTK_WINDOW(cwin), &cw_wi, &cw_he);
+  cursor_location.x = 0;
+  cursor_location.y = 0;
+  cursor_location.height = 0;
 
-  sc_wi = gdk_screen_get_width(gdk_screen_get_default());
-  sc_he = gdk_screen_get_height(gdk_screen_get_default());
-
-  if (sc_wi < topwin_x + cw_wi) {
-    x = topwin_x - cw_wi;
-  } else {
-    x = topwin_x;
-  }
-
-  if (sc_he < topwin_y + cw_he) {
-    /* FIXME : How can I determine the preedit height? */
-    y = topwin_y - cw_he - 20;
-  } else {
-    y = topwin_y;
-  }
-
-  gtk_window_move(GTK_WINDOW(cwin), x, y);
+  uim_cand_win_gtk_layout();
+  caret_state_indicator_set_cursor_location(cwin->caret_state_indicator, &cursor_location);
 }
 
 static void
@@ -506,69 +514,91 @@ candwin_deactivate(void)
   cwin->is_active = FALSE;
 }
 
+static void
+caret_state_show(gchar **str)
+{
+  caret_state_indicator_update(cwin->caret_state_indicator, cwin->pos_x, cwin->pos_y, str[1]);
+  gtk_widget_show(GTK_WIDGET(cwin->caret_state_indicator));
+}
+
 static void str_parse(gchar *str)
 {
   gchar **tmp;
-  int i = 0;
+  gchar *command;
 
   tmp = g_strsplit(str, "\n", 0);
+  command = tmp[0];
 
-  while (tmp[i]) {
-    if (strcmp("activate", tmp[i]) == 0) {
+  if (command) {
+    if (strcmp("activate", command) == 0) {
       candwin_activate(tmp);
-    } else if (strcmp("select", tmp[i]) == 0) {
+    } else if (strcmp("select", command) == 0) {
       candwin_update(tmp);
-    } else if (strcmp("show", tmp[i]) == 0) {
+    } else if (strcmp("show", command) == 0) {
       candwin_show();
-    } else if (strcmp("hide", tmp[i]) == 0) {
+    } else if (strcmp("hide", command) == 0) {
       gtk_widget_hide_all(GTK_WIDGET(cwin));
-    } else if (strcmp("move", tmp[i]) == 0) {
+    } else if (strcmp("move", command) == 0) {
       candwin_move(tmp);
-    } else if (strcmp("deactivate", tmp[i]) == 0) {
+    } else if (strcmp("deactivate", command) == 0) {
       candwin_deactivate();
+    } else if (strcmp("show_caret_state", command) == 0) {
+      caret_state_show(tmp);
     }
-    i++;
   }
   g_strfreev(tmp);
 }
 
-static void
-read_cb(gpointer p, int fd, GdkInputCondition c)
+#define CANDIDATE_BUFFER_SIZE	4096
+static gboolean
+read_cb(GIOChannel *channel, GIOCondition c, gpointer p)
 {
-  char buf[1024];
+  char buf[CANDIDATE_BUFFER_SIZE];
+  char *read_buf = strdup("");
   int i = 0;
   int n;
   gchar **tmp;
+  int fd = g_io_channel_unix_get_fd(channel);
 
-  n = read(fd, buf, 1024 - 1);
-  if (n == 0) {
-    close(fd);
-    exit(-1);
+  while (uim_helper_fd_readable(fd) > 0) {
+    n = read(fd, buf, CANDIDATE_BUFFER_SIZE - 1);
+    if (n == 0) {
+      close(fd);
+      exit(-1);
+    }
+    if (n == -1)
+      return TRUE;
+    buf[n] = '\0';
+    read_buf = realloc(read_buf, strlen(read_buf) + n + 1);
+    strcat(read_buf, buf);
   }
-  if (n == -1)
-    return;
 
-  buf[n] = '\0';
-  tmp = g_strsplit(buf, "\n\n", 0);
+  tmp = g_strsplit(read_buf, "\n\n", 0);
 
   while (tmp[i]) {
     str_parse(tmp[i]);
     i++;
   }
   g_strfreev(tmp);
-  return;
+  free(read_buf);
+  return TRUE;
 }
 
 int
 main(int argc, char *argv[])
 {
+  GIOChannel *channel;
+
   gtk_set_locale();
   gtk_init(&argc, &argv);
 
   init_candidate_win();
 
-  read_tag = gdk_input_add(0, (GdkInputCondition)GDK_INPUT_READ,
-			   read_cb, 0);
+  channel = g_io_channel_unix_new(0);
+  read_tag = g_io_add_watch(channel, G_IO_IN | G_IO_HUP | G_IO_ERR,
+			    read_cb, 0);
+  g_io_channel_unref(channel);
+
   gtk_main();
   return 0;
 }
@@ -586,19 +616,19 @@ uim_cand_win_gtk_get_index(UIMCandidateWindow *cwin)
 static void
 uim_cand_win_gtk_set_index(UIMCandidateWindow *cwin, gint index)
 {
-  gint new_page = 0;
+  gint new_page;
 
   g_return_if_fail(UIM_IS_CANDIDATE_WINDOW(cwin));
 
-  if (index < 0)
-    cwin->candidate_index = cwin->nr_candidates - 1;
-  else if ((guint) index >= cwin->nr_candidates)
+  if (index >= (gint) cwin->nr_candidates)
     cwin->candidate_index = 0;
   else
     cwin->candidate_index = index;
 
-  if (cwin->display_limit)
+  if (cwin->candidate_index >= 0 && cwin->display_limit)
     new_page = cwin->candidate_index / cwin->display_limit;
+  else
+    new_page = cwin->page_index;
 
   if (cwin->page_index != new_page)
     uim_cand_win_gtk_set_page(cwin, new_page);
@@ -628,7 +658,7 @@ static void
 uim_cand_win_gtk_set_page(UIMCandidateWindow *cwin, gint page)
 {
   guint len, new_page;
-  guint new_index;
+  gint new_index;
 
   g_return_if_fail(UIM_IS_CANDIDATE_WINDOW(cwin));
   g_return_if_fail(cwin->stores);
@@ -658,13 +688,50 @@ uim_cand_win_gtk_set_page(UIMCandidateWindow *cwin, gint page)
       new_index
 	= (new_page * cwin->display_limit) + (cwin->candidate_index % cwin->display_limit);
     else
-      new_index = new_page * cwin->display_limit;
+      new_index = -1;
   } else {
     new_index = cwin->candidate_index;
   }
 
-  if (new_index >= cwin->nr_candidates)
+  if (new_index >= (gint) cwin->nr_candidates)
     new_index = cwin->nr_candidates - 1;
 
+ /* shrink the window */
+  gtk_window_resize(GTK_WINDOW(cwin), CANDWIN_DEFAULT_WIDTH, 1);
+
   uim_cand_win_gtk_set_index(cwin, new_index);
 }
+
+static void
+uim_cand_win_gtk_layout()
+{
+  int x, y;
+  int screen_width, screen_height;
+
+  screen_width = gdk_screen_get_width(gdk_screen_get_default());
+  screen_height = gdk_screen_get_height(gdk_screen_get_default());
+
+  if (screen_width < cwin->pos_x + cwin->width)
+    x = cwin->pos_x - cwin->width;
+  else
+    x = cwin->pos_x;
+
+  if (screen_height < cwin->pos_y + cwin->height)
+    y = cwin->pos_y - cwin->height - 20; /* FIXME: Preedit height is needed to
+					    be sent by uim-xim */
+  else
+    y = cwin->pos_y;
+
+  gtk_window_move(GTK_WINDOW(cwin), x, y);
+}
+
+static gboolean
+configure_event_cb(GtkWidget *widget, GdkEventConfigure *event, gpointer data)
+{
+  cwin->width = event->width;
+  cwin->height = event->height;
+
+  uim_cand_win_gtk_layout();
+
+  return FALSE;
+}
