@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2004-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -44,21 +44,22 @@
 #include "uim/gettext.h"
 
 
-static int read_tag;
+static unsigned int read_tag;
 static int uim_fd; /* file descriptor to connect helper message bus */
 static gchar *im_list_str_old; /* To compare new im_list_str */
 static GtkWidget *switcher_tree_view;
-static gboolean grouped;
 
 static void
 reload_im_list(GtkWindow *window, gpointer user_data);
 static void
-parse_helper_str(const char* sent_str);
+parse_helper_str(const char *sent_str);
 static void
-parse_helper_str_im_list(const char* im_list_str_new);
+parse_helper_str_im_list(const char *im_list_str_new);
 static void
 check_helper_connection(void);
 
+static gint changing_way;
+GtkWidget *radio0, *radio1, *radio2;
 
 
 /* TreeItem structure */
@@ -85,16 +86,16 @@ enum
    - Sorting way (input method, language, most recently used) (string)
  */
 
-static char*
+static char *
 get_next_line(FILE *fp)
 {
   char buf[1024];
   GString *line = g_string_new("");
   /* I don't want to depend on glib, but GString is too convenience... */
 
-  while(fgets(buf, sizeof(buf), fp) != NULL) {
+  while (fgets(buf, sizeof(buf), fp) != NULL) {
     g_string_append(line, buf);
-    if(line->str[line->len - 1] == '\n') {
+    if (line->str[line->len - 1] == '\n') {
       return g_string_free(line, FALSE);
     }
   }
@@ -107,13 +108,13 @@ parse_config_line_history(const char *line)
   int i;
   char **splitted1 = g_strsplit(line, "=", -1);
   char **splitted2;
-  if(splitted1 && splitted1[1] ) {
+  if (splitted1 && splitted1[1] ) {
     splitted2 = g_strsplit(splitted1[1], ",", -1);
     g_strfreev(splitted1);
   } else {
     return;
   }
-  for(i = 0 ;splitted2[i] != NULL; i++) {
+  for (i = 0 ;splitted2[i] != NULL; i++) {
   g_print("%s", splitted2[i]);
   }
     g_strfreev(splitted2);
@@ -133,11 +134,11 @@ parse_config_line_sorting_way(const char *line)
 static void
 parse_config_line(const char *line)
 {
-  if(g_strrstr(line, "history=") == line) {
+  if (g_strrstr(line, "history=") == line) {
     parse_config_line_history(line);
-  } else if(g_strrstr(line, "grouped=") == line) {
+  } else if (g_strrstr(line, "grouped=") == line) {
     parse_config_line_grouped(line);
-  } else if(g_strrstr(line, "sorting_way=") == line) {
+  } else if (g_strrstr(line, "sorting_way=") == line) {
     parse_config_line_sorting_way(line);
   }
 }
@@ -148,20 +149,20 @@ load_configration(const char *filename)
   const char *f;
   char *line;
   FILE *fp;
-  if(filename == NULL) {
-    f = "/home/tkng/uim.d/im-switcher";/* FIXME:later get correct file path here */
+  if (filename == NULL) {
+    return;/* FIXME:later get correct file path here */
   } else {
     f= filename;
   }
 
   fp = fopen(f, "r");
-  if(fp == NULL) {
+  if (fp == NULL) {
     fprintf(stderr, "%s\n", strerror(errno));
     return;
   }
 
  /* XXX: I want you to write more beaftiful loop */
-  while((line = get_next_line(fp)) != NULL) {
+  while ((line = get_next_line(fp)) != NULL) {
     parse_config_line(line);
     free(line);
   }
@@ -185,7 +186,7 @@ get_selected_im_name(void)
   GtkTreeStore *store;
   GtkTreeIter iter;
   gchar *str_data;
-  if(gtk_tree_selection_get_selected (sel, &model, &iter) == TRUE) {
+  if (gtk_tree_selection_get_selected (sel, &model, &iter) == TRUE) {
     store = GTK_TREE_STORE(model);
     gtk_tree_model_get (model, &iter, 
 			NAME_COLUMN, &str_data,
@@ -200,7 +201,7 @@ send_message_im_change(const gchar *type)
 {
   GString *msg = g_string_new(type);
   gchar *im_name = get_selected_im_name();
-  if(im_name == NULL) {
+  if (im_name == NULL) {
     g_string_free(msg, TRUE);
     return; /* Or should pop-up alert window here? */
   }
@@ -213,23 +214,26 @@ send_message_im_change(const gchar *type)
 }
 
 static void
-change_input_method(GtkButton *button, GtkWidget *radio)
+toggle_changing_way(GtkToggleButton *togglebutton, gpointer user_data)
 {
-  int i = 0;
-  GSList *group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radio));
-
-  while((group = g_slist_next(group)) != NULL) {
-    if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(group->data))) {
-      break;
-    }
-    i++;
+  if (gtk_toggle_button_get_active((GtkToggleButton *)radio0)) {
+    changing_way = 0;
+  } else if (gtk_toggle_button_get_active((GtkToggleButton *)radio1)) {
+    changing_way = 1;
+  } else  if (gtk_toggle_button_get_active((GtkToggleButton *)radio2)) {
+    changing_way = 2;
   }
-  printf("**%d**\n",i);
+}
+
+static void
+change_input_method(GtkButton *button, gpointer user_data)
+{
+  int i = changing_way;
   switch (i) {
-  case 1:
+  case 0:
     send_message_im_change("im_change_whole_desktop\n");
     break;
-  case 0:
+  case 1:
     send_message_im_change("im_change_this_application_only\n");
     break;
   case 2:
@@ -308,10 +312,10 @@ create_switcher(void)
 {
   GtkWidget *switcher_win;
   GtkWidget *scrolled_win; /* treeview container */
-  GtkWidget *hbox, *vbox1, *vbox2;
+  GtkWidget *hbox, *vbox1, *vbox2, *vbox3;
   GtkWidget *setting_button_box;
   GtkWidget *button;
-  GtkWidget *frame, *radio0, *radio1, *radio2;
+  GtkWidget *frame;
 
   switcher_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(switcher_win),
@@ -323,8 +327,10 @@ create_switcher(void)
   g_signal_connect(G_OBJECT(switcher_win), "focus-in-event",
 		   G_CALLBACK(reload_im_list), NULL);
 
-  vbox1 = gtk_vbox_new(FALSE, 8);
+  vbox1 = gtk_vbox_new(FALSE, 0);
+  vbox3 = gtk_vbox_new(FALSE, 0);
   hbox = gtk_hbox_new(FALSE, 8);
+  gtk_container_set_border_width(GTK_CONTAINER(vbox3), 4);
 
   scrolled_win = gtk_scrolled_window_new(NULL, NULL);
   gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_win),
@@ -339,37 +345,47 @@ create_switcher(void)
   gtk_box_pack_start (GTK_BOX (vbox1), hbox, TRUE, TRUE, 0);
 
 
-  frame = gtk_frame_new("Changing way");
-  gtk_box_pack_start(GTK_BOX(vbox1), frame, FALSE, FALSE, 3);
-  vbox2 = gtk_vbox_new(FALSE, 3);
+  frame = gtk_frame_new(_("Changing way"));
+  gtk_frame_set_label_align(GTK_FRAME(frame), 0.015, 0.5);
+
+  gtk_box_pack_start(GTK_BOX(vbox3), frame, FALSE, FALSE, 6);
+  vbox2 = gtk_vbox_new(FALSE, 8);
+  gtk_container_set_border_width(GTK_CONTAINER(vbox2), 10);
   gtk_container_add(GTK_CONTAINER(frame), vbox2);
 
   radio0 = gtk_radio_button_new_with_label(NULL, _("Change whole desktop"));
   radio1 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio0), _("Change this application only"));
   radio2 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio0), _("Change this text area only"));
 
-  gtk_box_pack_start(GTK_BOX(vbox2), radio0, FALSE, FALSE, 3);
-  gtk_box_pack_start(GTK_BOX(vbox2), radio1, FALSE, FALSE, 3);
-  gtk_box_pack_start(GTK_BOX(vbox2), radio2, FALSE, FALSE, 3);
+  gtk_box_pack_start(GTK_BOX(vbox2), radio0, FALSE, FALSE, 0);
+  gtk_box_pack_start(GTK_BOX(vbox2), radio1, FALSE, FALSE, 0);
+  gtk_box_pack_start(GTK_BOX(vbox2), radio2, FALSE, FALSE, 0);
 
+  g_signal_connect(G_OBJECT(radio0), "toggled", G_CALLBACK(toggle_changing_way), NULL);
+  g_signal_connect(G_OBJECT(radio1), "toggled", G_CALLBACK(toggle_changing_way), NULL);
+  g_signal_connect(G_OBJECT(radio2), "toggled", G_CALLBACK(toggle_changing_way), NULL);
+
+  /* set radio0 (Change whole desktop) as default */
+  gtk_toggle_button_set_active((GtkToggleButton *)radio0, TRUE);
 
   setting_button_box = gtk_hbutton_box_new();
   gtk_button_box_set_layout(GTK_BUTTON_BOX(setting_button_box), GTK_BUTTONBOX_END);
   gtk_box_set_spacing(GTK_BOX(setting_button_box), 8);
 
-  /* Canel button */
+  /* Cancel button */
   button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
   g_signal_connect(G_OBJECT(button), "clicked",
 		   G_CALLBACK(gtk_main_quit), NULL);
   gtk_box_pack_start(GTK_BOX(setting_button_box), button, TRUE, TRUE, 2);
 
-  /* Apply button */
+  /* OK button */
   button = gtk_button_new_from_stock(GTK_STOCK_OK);
   g_signal_connect(G_OBJECT(button), "clicked",
-		   G_CALLBACK(change_input_method), radio0);
+		   G_CALLBACK(change_input_method), NULL);
   gtk_box_pack_start(GTK_BOX(setting_button_box), button, TRUE, TRUE, 2);
 
-  gtk_box_pack_start(GTK_BOX(vbox1), setting_button_box, FALSE, FALSE, 2);
+  gtk_box_pack_start(GTK_BOX(vbox3), setting_button_box, FALSE, FALSE, 8);
+  gtk_box_pack_start(GTK_BOX(vbox1), vbox3, FALSE, FALSE, 0);
   gtk_container_add(GTK_CONTAINER(switcher_win), vbox1);
 
   {
@@ -393,24 +409,24 @@ reload_im_list(GtkWindow *window, gpointer user_data)
 }
 
 static void
-parse_helper_str(const char* sent_str)
+parse_helper_str(const char *sent_str)
 {
-  if(g_str_has_prefix(sent_str, "im_list") == TRUE) {
+  if (g_str_has_prefix(sent_str, "im_list") == TRUE) {
     parse_helper_str_im_list(sent_str);
   }
 }
 
-static char*
-get_text(const char* str)
+static char *
+get_text(const char *str)
 {
-  if(strcmp("", str) == 0)
+  if (strcmp("", str) == 0)
     return "-";
   else
     return gettext(str);
 }
 
 static void
-parse_helper_str_im_list(const char* im_list_str_new)
+parse_helper_str_im_list(const char *im_list_str_new)
 {
   gchar **lines;
   int i = 2;
@@ -418,7 +434,7 @@ parse_helper_str_im_list(const char* im_list_str_new)
   GtkTreeStore *tree_store;
   GtkTreeIter iter;
   GtkTreePath *path = NULL;
-  if(im_list_str_old && strcmp(im_list_str_new, im_list_str_old) == 0) {
+  if (im_list_str_old && strcmp(im_list_str_new, im_list_str_old) == 0) {
     return; /* No need to update */
   }
 
@@ -428,20 +444,20 @@ parse_helper_str_im_list(const char* im_list_str_new)
 				   G_TYPE_STRING,
 				   G_TYPE_STRING);
 
-  for(i=2; lines[i] != NULL; i++) {
+  for (i=2; lines[i] != NULL; i++) {
 
-    if(!lines[i] || strcmp(lines[i], "") == 0) {
+    if (!lines[i] || strcmp(lines[i], "") == 0) {
       break;
     }
     info = g_strsplit(lines[i], "\t", -1);
-    if(info && info[0] && info[1] && info[2]) {
+    if (info && info[0] && info[1] && info[2]) {
       gtk_tree_store_append(tree_store, &iter, NULL/* parent iter */);
       gtk_tree_store_set(tree_store, &iter,
 			 NAME_COLUMN, info[0],
 			 LANG_COLUMN, get_text(info[1]),
 			 DESC_COLUMN, get_text(info[2]),
 			 -1);
-      if(info[3] && (strcmp(info[3], "") != 0)) {
+      if (info[3] && (strcmp(info[3], "") != 0)) {
 	path = gtk_tree_model_get_path(GTK_TREE_MODEL(tree_store),
 				       &iter);
 	
@@ -452,7 +468,7 @@ parse_helper_str_im_list(const char* im_list_str_new)
   gtk_tree_view_set_model(GTK_TREE_VIEW(switcher_tree_view), 
 			  GTK_TREE_MODEL(tree_store));
   
-  if(path != NULL) {
+  if (path != NULL) {
     gtk_tree_view_set_cursor(GTK_TREE_VIEW(switcher_tree_view),
 			     path, NULL, FALSE);
   }
@@ -466,30 +482,36 @@ static void
 helper_disconnect_cb(void)
 {
   uim_fd = -1;
-  gdk_input_remove(read_tag);
+  g_source_remove(read_tag);
 }
 
-static void
-fd_read_cb(gpointer p, gint fd, GdkInputCondition c)
+static gboolean
+fd_read_cb(GIOChannel *channel, GIOCondition c, gpointer p)
 {
   char *tmp;
-  
+  int fd = g_io_channel_unix_get_fd(channel);
+
   uim_helper_read_proc(fd);
   while ((tmp = uim_helper_get_message())) {
     parse_helper_str(tmp);
     g_free(tmp); tmp = NULL;
   }
+  return TRUE;
 }
 
 
 static void
 check_helper_connection(void)
 {
-  if(uim_fd < 0) {
+  if (uim_fd < 0) {
     uim_fd = uim_helper_init_client_fd(helper_disconnect_cb);
-    if(uim_fd > 0)
-      read_tag = gdk_input_add(uim_fd, (GdkInputCondition)GDK_INPUT_READ,
-			       fd_read_cb, NULL);
+    if (uim_fd > 0) {
+      GIOChannel *channel;
+      channel = g_io_channel_unix_new(uim_fd);
+      read_tag = g_io_add_watch(channel, G_IO_IN | G_IO_HUP | G_IO_ERR,
+				fd_read_cb, NULL);
+      g_io_channel_unref(channel);
+    }
   }
 }
 
@@ -508,7 +530,7 @@ main(int argc, char *argv[])
 
   result = create_switcher();
 
-  if(result == -1) {
+  if (result == -1) {
     fprintf(stderr, "Error:%s\n", get_error_msg());
     exit(-1);
   }
