@@ -1,7 +1,7 @@
 ;;; anthy.scm: Anthy for uim.
 ;;; charset: EUC-JP
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -17,10 +17,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -34,71 +34,202 @@
 (require "ustr.scm")
 (require "japanese.scm")
 (require "japanese-kana.scm")
-(require "generic-key.scm")
+(require "japanese-azik.scm")
+(require-custom "generic-key-custom.scm")
+(require-custom "anthy-custom.scm")
+(require-custom "anthy-key-custom.scm")
 
-;; user configs
-(define anthy-use-candidate-window? #t)
-(define anthy-candidate-op-count 1)
-(define anthy-nr-candidate-max 10)
-(define anthy-show-segment-separator? #f)
-(define anthy-segment-separator "|")
 
-;; key defs
-(define-key anthy-on-key? '("<Control>j" "<Control>J" generic-on-key?))
-(define-key anthy-latin-key? '("<Control>j" "<Control>J" generic-off-key?))
-(define-key anthy-wide-latin-key? '())
-;(define-key anthy-hankaku-kana-key? '("<Control>q" "<Control>Q"))
-(define-key anthy-hankaku-kana-key? '())
-(define-key anthy-kana-toggle-key? '())
-(define-key anthy-commit-key? 'generic-commit-key?)
-(define-key anthy-commit-as-opposite-kana-key? '())
-(define-key anthy-commit-as-katakana-key? "F7")
-(define-key anthy-commit-as-hankana-key? "F8")
-(define-key anthy-commit-as-wide-latin-key? "F9")
-(define-key anthy-commit-as-latin-key? "F10")
-(define-key anthy-begin-conv-key? 'generic-begin-conv-key?)
-(define-key anthy-extend-segment-key? '("<Control>o" "<Control>O" "<Shift>right"))
-(define-key anthy-shrink-segment-key? '("<Control>i" "<Control>I" "<Shift>left"))
-(define-key anthy-next-candidate-key? 'generic-next-candidate-key?)
-(define-key anthy-prev-candidate-key? 'generic-prev-candidate-key?)
-(define-key anthy-next-page-key? 'generic-next-page-key?)
-(define-key anthy-prev-page-key? 'generic-prev-page-key?)
-(define-key anthy-cancel-key? 'generic-cancel-key?)
-(define-key anthy-backspace-key? 'generic-backspace-key?)
-(define-key anthy-delete-key? 'generic-delete-key?)
-(define-key anthy-kill-key? 'generic-kill-key?)
-(define-key anthy-kill-backward-key? 'generic-kill-backward-key?)
-(define-key anthy-go-left-key? 'generic-go-left-key?)
-(define-key anthy-go-right-key? 'generic-go-right-key?)
-(define-key anthy-beginning-of-preedit-key? 'generic-beginning-of-preedit-key?)
-(define-key anthy-end-of-preedit-key? 'generic-end-of-preedit-key?)
-(define-key anthy-next-segment-key? 'generic-go-right-key?)
-(define-key anthy-prev-segment-key? 'generic-go-left-key?)
-
-;; implementations
+;;; implementations
 
 (define anthy-lib-initialized? #f)
 
-(define anthy-type-hiragana 0)
-(define anthy-type-katakana 1)
-(define anthy-type-hankana 2)
-
-(define anthy-mode-direct 0)
-(define anthy-mode-hiragana 1)
-(define anthy-mode-katakana 2)
-(define anthy-mode-wide-latin 3)
-(define anthy-mode-hankana 4)
+(define anthy-type-hiragana   0)
+(define anthy-type-katakana   1)
+(define anthy-type-hankana    2)
+(define anthy-type-latin      3)
+(define anthy-type-wide-latin 4)
 
 (define anthy-input-rule-roma 0)
 (define anthy-input-rule-kana 1)
 (define anthy-input-rule-azik 2)
 
+(define anthy-prepare-activation
+  (lambda (ac)
+    (anthy-flush ac)
+    (anthy-update-preedit ac)))
+
+(register-action 'action_anthy_hiragana
+;;		 (indication-alist-indicator 'action_anthy_hiragana
+;;					     anthy-input-mode-indication-alist)
+		 (lambda (ac) ;; indication handler
+		   '(figure_ja_hiragana
+		     ""
+		     "Ҥ餬"
+		     "Ҥ餬ϥ⡼"))
+
+		 (lambda (ac) ;; activity predicate
+		   (and (anthy-context-on ac)
+			(= (anthy-context-kana-mode ac)
+			   anthy-type-hiragana)))
+
+		 (lambda (ac) ;; action handler
+		   (anthy-prepare-activation ac)
+		   (anthy-context-set-on! ac #t)
+		   (anthy-context-set-kana-mode! ac anthy-type-hiragana)))
+
+(register-action 'action_anthy_katakana
+;;		 (indication-alist-indicator 'action_anthy_katakana
+;;					     anthy-input-mode-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_katakana
+		     ""
+		     ""
+		     "ϥ⡼"))
+		 (lambda (ac)
+		   (and (anthy-context-on ac)
+			(= (anthy-context-kana-mode ac)
+			   anthy-type-katakana)))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (anthy-context-set-on! ac #t)
+
+		   (if (= anthy-input-rule-kana
+			  (anthy-context-input-rule ac))
+		       (rk-context-set-rule! (anthy-context-rkc ac)
+					     ja-kana-katakana-rule))
+		   (anthy-context-set-kana-mode! ac anthy-type-katakana)))
+
+(register-action 'action_anthy_hankana
+;;		 (indication-alist-indicator 'action_anthy_hankana
+;;					     anthy-input-mode-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_hankana
+		     ""
+		     "Ⱦѥ"
+		     "Ⱦѥϥ⡼"))
+		 (lambda (ac)
+		   (and (anthy-context-on ac)
+			(= (anthy-context-kana-mode ac)
+			   anthy-type-hankana)))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (anthy-context-set-on! ac #t)
+		   (anthy-context-set-kana-mode! ac anthy-type-hankana)))
+
+(register-action 'action_anthy_direct
+;;		 (indication-alist-indicator 'action_anthy_direct
+;;					     anthy-input-mode-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_direct
+		     "a"
+		     "ľ"
+		     "ľ(̵Ѵ)ϥ⡼"))
+		 (lambda (ac)
+		   (and (not (anthy-context-on ac))
+			(not (anthy-context-wide-latin ac))))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (anthy-context-set-on! ac #f)
+		   (anthy-context-set-wide-latin! ac #f)))
+
+(register-action 'action_anthy_zenkaku
+;;		 (indication-alist-indicator 'action_anthy_zenkaku
+;;					     anthy-input-mode-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_zenkaku
+		     ""
+		     "ѱѿ"
+		     "ѱѿϥ⡼"))
+		 (lambda (ac)
+		   (and (not (anthy-context-on ac))
+			(anthy-context-wide-latin ac)))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (anthy-context-set-on! ac #f)
+		   (anthy-context-set-wide-latin! ac #t)))
+
+(register-action 'action_anthy_roma
+;;		 (indication-alist-indicator 'action_anthy_roma
+;;					     anthy-kana-input-method-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_roma
+		     ""
+		     "޻"
+		     "޻ϥ⡼"))
+		 (lambda (ac)
+		   (= (anthy-context-input-rule ac)
+		      anthy-input-rule-roma))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (rk-context-set-rule! (anthy-context-rkc ac)
+					 ja-rk-rule)
+		   (anthy-context-set-input-rule! ac anthy-input-rule-roma)))
+
+(register-action 'action_anthy_kana
+;;		 (indication-alist-indicator 'action_anthy_kana
+;;					     anthy-kana-input-method-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_kana
+		     ""
+		     ""
+		     "ϥ⡼"))
+		 (lambda (ac)
+		   (= (anthy-context-input-rule ac)
+		      anthy-input-rule-kana))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (rk-context-set-rule! (anthy-context-rkc ac)
+					 (if (= (anthy-context-kana-mode ac)
+						anthy-type-katakana)
+						ja-kana-katakana-rule
+						ja-kana-hiragana-rule))
+
+		   (anthy-context-set-input-rule! ac anthy-input-rule-kana)
+		   ;;(define-key anthy-kana-toggle-key? "")
+		   ;;(define-key anthy-latin-key? generic-on-key?)
+		   ;;(define-key anthy-wide-latin-key? "")
+		   ))
+
+(register-action 'action_anthy_azik
+;;		 (indication-alist-indicator 'action_anthy_azik
+;;					     anthy-kana-input-method-indication-alist)
+		 (lambda (ac)
+		   '(figure_ja_azik
+		     ""
+		     "AZIK"
+		     "AZIKĥ޻ϥ⡼"))
+		 (lambda (ac)
+		   (= (anthy-context-input-rule ac)
+		      anthy-input-rule-azik))
+		 (lambda (ac)
+		   (anthy-prepare-activation ac)
+		   (rk-context-set-rule! (anthy-context-rkc ac)
+					 ja-azik-rule)
+		   (anthy-context-set-input-rule! ac anthy-input-rule-azik)))
+
+;; Update widget definitions based on action configurations. The
+;; procedure is needed for on-the-fly reconfiguration involving the
+;; custom API
+(define anthy-configure-widgets
+  (lambda ()
+    (register-widget 'widget_anthy_input_mode
+		     (activity-indicator-new anthy-input-mode-actions)
+		     (actions-new anthy-input-mode-actions))
+
+    (register-widget 'widget_anthy_kana_input_method
+		     (activity-indicator-new anthy-kana-input-method-actions)
+		     (actions-new anthy-kana-input-method-actions))
+    (context-list-replace-widgets! 'anthy anthy-widgets)))
+
 (define anthy-context-rec-spec
   (append
    context-rec-spec
    (list
     (list 'on                 #f)
     (list 'converting         #f)
+    (list 'transposing        #f)
+    (list 'transposing-type    0)
     (list 'ac-id              #f) ;; anthy-context-id
     (list 'preconv-ustr       #f) ;; preedit strings
     (list 'rkc                #f)
@@ -121,12 +252,11 @@
 	 (set! anthy-lib-initialized? (anthy-lib-init)))
      (if anthy-lib-initialized?
 	 (anthy-context-set-ac-id! ac (anthy-lib-alloc-context)))
+     (anthy-context-set-widgets! ac anthy-widgets)
      (anthy-context-set-rkc! ac rkc)
      (anthy-context-set-preconv-ustr! ac (ustr-new))
      (anthy-context-set-raw-ustr! ac (ustr-new))
      (anthy-context-set-segments! ac (ustr-new))
-     (anthy-flush ac)
-     (anthy-context-set-on! ac #f)
 
      ;; 2004-08-26 Takuro Ashie <ashie@homa.ne.jp>
      ;;   * I think load-kana-table should be marked as depracated.
@@ -158,8 +288,11 @@
 	   (residual-kana (rk-peek-terminal-match rkc))
 	   (rule (anthy-context-input-rule ac))
 	   (preconv-str (anthy-context-preconv-ustr ac))
-	   (extract-kana (lambda (entry)
-			   (nth kana entry))))
+	   (extract-kana
+	    (if (= rule anthy-input-rule-kana)
+		(lambda (entry) (car entry))
+		(lambda (entry) (nth kana entry)))))
+
       (string-append
        (string-append-map-ustr-former extract-kana preconv-str)
        (if convert-pending-into-kana?
@@ -195,65 +328,36 @@
        (ja-raw-string-list-to-valid-roma
         (append left-str
                 (if (null? residual-kana)
-                    right-str
+		    (begin
+		      (if (null? right-str)
+			  (list pending)
+			  (append right-str (list pending))))
                     (begin
                       (rk-flush rkc)
                       (if (null? right-str)
                           (list pending)
-                          (append (list pending) right-str))))))
+			  (append right-str (list pending)))))))
        wide?))))
 
 (define anthy-init-handler
   (lambda (id im arg)
-    (let ((ac (anthy-context-new id im)))
-      (im-clear-mode-list ac)
-      (im-pushback-mode-list ac "ľ")
-      (im-pushback-mode-list ac "Ҥ餬")
-      (im-pushback-mode-list ac "")
-      (im-pushback-mode-list ac "ѱѿ")
-      (im-pushback-mode-list ac "Ⱦѥ")
-      (im-update-mode-list ac)
-      (im-update-mode ac 0)
-      (anthy-update-prop-list ac)
-      ac)))
+    (anthy-context-new id im)))
 
 (define anthy-release-handler
   (lambda (ac)
     (let ((ac-id (anthy-context-ac-id ac)))
       (anthy-lib-free-context ac-id))))
 
-(define anthy-mode-handler
-  (lambda (ac mode)
-    (anthy-flush ac)
-    (cond
-     ((= mode anthy-mode-direct)
-      (anthy-context-set-on! ac #f)
-      (anthy-context-set-wide-latin! ac #f))
-     ((= mode anthy-mode-hiragana)
-      (anthy-context-set-on! ac #t)
-      (anthy-context-set-kana-mode! ac anthy-type-hiragana))
-     ((= mode anthy-mode-katakana)
-      (anthy-context-set-on! ac #t)
-      (anthy-context-set-kana-mode! ac anthy-type-katakana))
-     ((= mode anthy-mode-wide-latin)
-      (anthy-context-set-on! ac #f)
-      (anthy-context-set-wide-latin! ac #t))
-     ((= mode anthy-mode-hankana)
-      (anthy-context-set-on! ac #t)
-      (anthy-context-set-kana-mode! ac anthy-type-hankana)))
-    (anthy-update-preedit ac)
-    (anthy-update-prop-label ac)
-    (anthy-update-prop-list ac))
-  ())
-
 (define anthy-flush
   (lambda (ac)
     (rk-flush (anthy-context-rkc ac))
     (ustr-clear! (anthy-context-preconv-ustr ac))
     (ustr-clear! (anthy-context-raw-ustr ac))
     (ustr-clear! (anthy-context-segments ac))
-    (anthy-context-set-on! ac #t)
+    (anthy-context-set-transposing! ac #f)
     (anthy-context-set-converting! ac #f)
+    (if (anthy-context-candidate-window ac)
+	  (im-deactivate-candidate-selector ac))
     (anthy-context-set-candidate-window! ac #f)
     (anthy-context-set-candidate-op-count! ac 0)))
 
@@ -267,9 +371,11 @@
   (lambda (ac)
     (if (not (anthy-context-commit-raw ac))
 	(let ((segments (if (anthy-context-on ac)
-			    (if (anthy-context-converting ac)
-				(anthy-converting-state-preedit ac)
-				(anthy-input-state-preedit ac))
+			    (if (anthy-context-transposing ac)
+				(anthy-context-transposing-state-preedit ac)
+				(if (anthy-context-converting ac)
+				    (anthy-converting-state-preedit ac)
+				    (anthy-input-state-preedit ac)))
 			    ())))
 	  (context-update-preedit ac segments))
 	(anthy-context-set-commit-raw! ac #f))))
@@ -277,10 +383,7 @@
 (define anthy-proc-raw-state
   (lambda (ac key key-state)
     (if (anthy-on-key? key key-state)
-	(begin
-	  (anthy-begin-input ac)
-	  (anthy-update-mode ac)
-	  (anthy-update-prop-label ac))
+	(anthy-begin-input ac)
 	(anthy-commit-raw ac))))
 
 (define anthy-begin-conv
@@ -317,17 +420,13 @@
 	(begin
 	  (anthy-flush ac)
 	  (anthy-context-set-on! ac #f)
-	  (anthy-context-set-wide-latin! ac #t)
-	  (anthy-update-mode ac)
-	  (anthy-update-prop-label ac)))
+	  (anthy-context-set-wide-latin! ac #t)))
 	  
        ((anthy-latin-key? key key-state)
 	   (begin
 	     (anthy-flush ac)
 	     (anthy-context-set-on! ac #f)
-	     (anthy-context-set-wide-latin! ac #f)
-	     (anthy-update-mode ac)
-	     (anthy-update-prop-label ac)))
+	     (anthy-context-set-wide-latin! ac #f)))
 
        ((anthy-backspace-key? key key-state)
 	(anthy-commit-raw ac))
@@ -336,16 +435,10 @@
 	(anthy-commit-raw ac))
        
        ((anthy-hankaku-kana-key? key key-state)
-	   (begin 
-	     (anthy-context-set-kana-mode! ac anthy-type-hankana)
-	     (anthy-update-mode ac)
-	     (anthy-update-prop-label ac)))
+	(anthy-context-set-kana-mode! ac anthy-type-hankana))
 
        ((anthy-kana-toggle-key? key key-state)
-	(begin 
-	  (anthy-context-kana-toggle ac)
-	  (anthy-update-mode ac)
-	  (anthy-update-prop-label ac)))
+	(anthy-context-kana-toggle ac))
 
        ;; modifiers (except shift) => ignore
        ((and (modifier-key-mask key-state)
@@ -379,6 +472,31 @@
     (or (not (ustr-empty? (anthy-context-preconv-ustr ac)))
 	(> (length (rk-pending (anthy-context-rkc ac))) 0))))
 
+(define anthy-proc-transposing-state
+  (lambda (ac key key-state)
+    (cond
+     ((anthy-transpose-as-katakana-key? key key-state)
+      (anthy-context-set-transposing-type! ac anthy-type-katakana))
+
+     ((anthy-transpose-as-hankana-key? key key-state)
+      (anthy-context-set-transposing-type! ac anthy-type-hankana))
+
+     ((anthy-transpose-as-latin-key? key key-state)
+      (anthy-context-set-transposing-type! ac anthy-type-latin))
+
+     ((anthy-transpose-as-wide-latin-key? key key-state)
+      (anthy-context-set-transposing-type! ac anthy-type-wide-latin))
+
+     (else
+      (begin
+	; commit
+	(im-commit ac (anthy-transposing-text ac))
+	(anthy-flush ac)
+	(if (not (anthy-commit-key? key key-state))
+	    (begin 
+	      (anthy-context-set-transposing! ac #f)
+	      (anthy-proc-input-state ac key key-state))))))))
+
 (define anthy-proc-input-state-with-preedit
   (lambda (ac key key-state)
     (let ((preconv-str (anthy-context-preconv-ustr ac))
@@ -387,16 +505,12 @@
 	  (kana (anthy-context-kana-mode ac))
 	  (rule (anthy-context-input-rule ac)))
       (cond
-       
+
        ;; begin conversion
        ((anthy-begin-conv-key? key key-state)
 	(anthy-begin-conv ac))
        
        ;; backspace
-       ;;
-       ;; 2004-08-27 Takuro Ashie <ashie@homa.ne.jp>
-       ;;   We should restore pending state of rk-context when the input-rule
-       ;;   is kana mode.
        ((anthy-backspace-key? key key-state)
 	(if (not (rk-backspace rkc))
             (begin
@@ -428,48 +542,23 @@
 	   (anthy-make-whole-string ac #t (multi-segment-opposite-kana kana)))
 	  (anthy-flush ac)))
 
-       ;; ʥ⡼ɤǤʤꤹ
-       ((anthy-commit-as-katakana-key? key key-state)
+       ;; Transposing֤ذܹ
+       ((or (anthy-transpose-as-katakana-key?   key key-state)
+	    (anthy-transpose-as-hankana-key?    key key-state)
+	    (anthy-transpose-as-latin-key?      key key-state)
+	    (anthy-transpose-as-wide-latin-key? key key-state))
 	(begin
-	  (im-commit
-	   ac
-	   (anthy-make-whole-string ac #t multi-segment-type-katakana))
-	  (anthy-flush ac)))
+	  (anthy-context-set-transposing! ac #t)
+	  (anthy-proc-transposing-state ac key key-state)))
 
-       ;; Ⱦѥʥ⡼ɤǤʤꤹ
-       ((anthy-commit-as-hankana-key? key key-state)
-	(begin
-	  (im-commit
-	   ac
-	   (anthy-make-whole-string ac #t multi-segment-type-hankana))
-	  (anthy-flush ac)))
-
-       ;; ʤѿᤷƳꤹ
-       ((anthy-commit-as-latin-key? key key-state)
-	(begin
-	  (im-commit
-	   ac
-           (anthy-make-whole-raw-string ac #f))
-	  (anthy-flush ac)))
-
-       ;; ʤѱѿᤷƳꤹ
-       ((anthy-commit-as-wide-latin-key? key key-state)
-	(begin
-	  (im-commit
-	   ac
-           (anthy-make-whole-raw-string ac #t))
-	  (anthy-flush ac)))
-
-       ;; ߤΤʤ塢Ҥ餬/ʥ⡼ɤڤ괹
+       ;; Commit current preedit string, then toggle hiragana/katakana mode.
        ((anthy-kana-toggle-key? key key-state)
 	(begin
 	  (im-commit
 	   ac
 	   (anthy-make-whole-string ac #t kana))
 	  (anthy-flush ac)
-	  (anthy-context-kana-toggle ac)
-	  (anthy-update-mode ac)
-	  (anthy-update-prop-label ac)))
+	  (anthy-context-kana-toggle ac)))
 
        ;; cancel
        ((anthy-cancel-key? key key-state)
@@ -484,47 +573,38 @@
 	  (anthy-flush ac)))
 
        ;; left
-       ;;
        ;; 2004-08-27 Takuro Ashie <ashie@homa.ne.jp>
        ;;   * We should restore pending state of rk-context when the input-rule
        ;;     is kana mode.
-       ;;   * Behaviour of setting left and right string should be changed on
-       ;;     kana mode.
        ((anthy-go-left-key? key key-state)
-        (begin
-	  (ustr-cursor-move-backward! preconv-str)
-	  (ustr-cursor-move-backward! raw-str)))
+	(anthy-context-confirm-kana! ac)
+	(ustr-cursor-move-backward! preconv-str)
+	(ustr-cursor-move-backward! raw-str))
+
        ;; right
-       ;;
        ;; 2004-08-27 Takuro Ashie <ashie@homa.ne.jp>
        ;;   * We should restore pending state of rk-context when the input-rule
        ;;     is kana mode.
-       ;;   * Behaviour of setting left and right string should be changed on
-       ;;     kana mode.
        ((anthy-go-right-key? key key-state)
-        (begin
-	  (ustr-cursor-move-forward! preconv-str)
-	  (ustr-cursor-move-forward! raw-str)))
+	(anthy-context-confirm-kana! ac)
+	(ustr-cursor-move-forward! preconv-str)
+	(ustr-cursor-move-forward! raw-str))
 
        ;; beginning-of-preedit
-       ;;
        ;; 2004-08-27 Takuro Ashie <ashie@homa.ne.jp>
-       ;;   * Behaviour of setting left and right string should be changed on
-       ;;     kana mode.
+       ;;   * We should restore pending state of rk-context when the input-rule
+       ;;     is kana mode.
        ((anthy-beginning-of-preedit-key? key key-state)
+	(anthy-context-confirm-kana! ac)
 	(ustr-cursor-move-beginning! preconv-str))
 
        ;; end-of-preedit
-       ;;
        ;; 2004-08-27 Takuro Ashie <ashie@homa.ne.jp>
        ;;   * We should restore pending state of rk-context when the input-rule
        ;;     is kana mode.
-       ;;   * Behaviour of setting left and right string should be changed on
-       ;;     kana mode.
        ((anthy-end-of-preedit-key? key key-state)
-	(ustr-cursor-move-end! preconv-str)
-	;;(rk-flush rkc)
-	)
+	(anthy-context-confirm-kana! ac)
+	(ustr-cursor-move-end! preconv-str))
 
        ;; modifiers (except shift) => ignore
        ((and (modifier-key-mask key-state)
@@ -542,16 +622,27 @@
 	  (if (and res
 		   (or (list? (car res))
 		       (not (string=? (car res) ""))))
-	      ;;
               (let ((next-pend (rk-pending rkc)))
-		(ustr-insert-elem! preconv-str res)
+		(if (list? (car res))
+		    (ustr-insert-seq!  preconv-str res)
+		    (ustr-insert-elem! preconv-str res))
 		(if (and next-pend
 			 (not (string=? next-pend "")))
 		    (ustr-insert-elem! raw-str pend)
 		    (ustr-insert-elem! raw-str (string-append pend key-str))))
-	      ;;
-              (if (= rule anthy-input-rule-kana)
-		  (ustr-insert-elem! preconv-str (list pend "" ""))))))))))
+	      )))))))
+
+(define anthy-context-confirm-kana!
+  (lambda (ac)
+    (if (= (anthy-context-input-rule ac)
+	   anthy-input-rule-kana)
+	(let* ((preconv-str (anthy-context-preconv-ustr ac))
+	       (rkc (anthy-context-rkc ac))
+	       (residual-kana (rk-peek-terminal-match rkc)))
+	    (if residual-kana
+		(begin
+		  (ustr-insert-elem! preconv-str residual-kana)
+		  (rk-flush rkc)))))))
 
 (define anthy-proc-input-state
   (lambda (ac key key-state)
@@ -567,28 +658,49 @@
 	  (cons attr anthy-segment-separator)
 	  #f))))
 
+(define anthy-context-transposing-state-preedit
+  (lambda (ac)
+    (let* ((transposing-text (anthy-transposing-text ac)))
+      (list (cons preedit-underline transposing-text)
+	    (cons preedit-cursor "")))))
+
+(define anthy-transposing-text
+  (lambda (ac)
+    (let* ((transposing-type (anthy-context-transposing-type ac)))
+      (cond
+       ((= transposing-type anthy-type-katakana)
+	(anthy-make-whole-string ac #t multi-segment-type-katakana))
+
+       ((= transposing-type anthy-type-hankana)
+	(anthy-make-whole-string ac #t multi-segment-type-hankana))
+
+       ((= transposing-type anthy-type-latin)
+	(anthy-make-whole-raw-string ac #f))
+
+       ((= transposing-type anthy-type-wide-latin)
+	(anthy-make-whole-raw-string ac #t))
+       ))))
+
 (define anthy-converting-state-preedit
   (lambda (ac)
     (let* ((ac-id (anthy-context-ac-id ac))
 	   (segments (anthy-context-segments ac))
 	   (cur-seg (ustr-cursor-pos segments))
 	   (separator (anthy-separator ac)))
-      (apply
-       append
-       (map (lambda (seg-idx cand-idx)
-	      (let* ((attr (if (= seg-idx cur-seg)
-			       (bit-or preedit-reverse
-				       preedit-cursor)
-			       preedit-underline))
-		     (cand (anthy-lib-get-nth-candidate ac-id
-							seg-idx cand-idx))
-		     (seg (list (cons attr cand))))
-		(if (and separator
-			 (< 0 seg-idx))
-		    (cons separator seg)
-		    seg)))
-	    (iota (ustr-length segments))
-	    (ustr-whole-seq segments))))))
+      (append-map
+       (lambda (seg-idx cand-idx)
+	 (let* ((attr (if (= seg-idx cur-seg)
+			  (bit-or preedit-reverse
+				  preedit-cursor)
+			  preedit-underline))
+		(cand (anthy-lib-get-nth-candidate ac-id seg-idx cand-idx))
+		(seg (list (cons attr cand))))
+	   (if (and separator
+		    (< 0 seg-idx))
+	       (cons separator seg)
+	       seg)))
+       (iota (ustr-length segments))
+       (ustr-whole-seq segments)))))
 
 (define anthy-input-state-preedit
   (lambda (ac)
@@ -596,8 +708,12 @@
 	   (rkc (anthy-context-rkc ac))
 	   (pending (rk-pending rkc))
 	   (kana (anthy-context-kana-mode ac))
-	   (extract-kana (lambda (entry)
-			   (nth kana entry))))
+	   (rule (anthy-context-input-rule ac))
+	   (extract-kana
+	    (if (= rule anthy-input-rule-kana)
+		(lambda (entry) (car entry))
+		(lambda (entry) (nth kana entry)))))
+
       (list
        (and (not (ustr-cursor-at-beginning? preconv-str))
 	    (cons preedit-underline
@@ -614,11 +730,10 @@
   (lambda (ac)
     (let ((ac-id (anthy-context-ac-id ac))
 	  (segments (anthy-context-segments ac)))
-      (apply string-append
-	     (map (lambda (seg-idx cand-idx)
-		    (anthy-lib-get-nth-candidate ac-id seg-idx cand-idx))
-		  (iota (ustr-length segments))
-		  (ustr-whole-seq segments))))))
+      (string-append-map (lambda (seg-idx cand-idx)
+			   (anthy-lib-get-nth-candidate ac-id seg-idx cand-idx))
+			 (iota (ustr-length segments))
+			 (ustr-whole-seq segments)))))
 
 (define anthy-commit-string
   (lambda (ac)
@@ -631,9 +746,9 @@
 
 (define anthy-do-commit
   (lambda (ac)
-    (anthy-reset-candidate-window ac)
     (im-commit ac (anthy-get-commit-string ac))
     (anthy-commit-string ac)
+    (anthy-reset-candidate-window ac)
     (anthy-flush ac)))
 
 (define anthy-correct-segment-cursor
@@ -686,6 +801,33 @@
       (if (anthy-context-candidate-window ac)
 	  (im-select-candidate ac compensated-n)))))
 
+(define anthy-move-candidate-in-page
+  (lambda (ac numeralc)
+    (let* ((ac-id (anthy-context-ac-id ac))
+	   (segments (anthy-context-segments ac))
+	   (cur-seg (ustr-cursor-pos segments))
+	   (max (anthy-lib-get-nr-candidates ac-id cur-seg))
+	   (n (ustr-cursor-frontside segments))
+	   (cur-page (if (= anthy-nr-candidate-max 0)
+	   		 0
+			 (quotient n anthy-nr-candidate-max)))
+	   (pageidx (- (numeral-char->number numeralc) 1))
+	   (compensated-pageidx (cond
+				 ((< pageidx 0) ; pressing key_0
+				  (+ pageidx 10))
+				 (else
+				  pageidx)))
+	   (idx (+ (* cur-page anthy-nr-candidate-max) compensated-pageidx))
+	   (compensated-idx (cond
+			     ((>= idx max)
+			      (- max 1))
+			     (else
+			      idx)))
+	   (new-op-count (+ 1 (anthy-context-candidate-op-count ac))))
+      (ustr-cursor-set-frontside! segments compensated-idx)
+      (anthy-context-set-candidate-op-count! ac new-op-count)
+      (im-select-candidate ac compensated-idx))))
+
 (define anthy-reset-candidate-window
   (lambda (ac)
     (if (anthy-context-candidate-window ac)
@@ -697,6 +839,14 @@
 (define anthy-proc-converting-state
   (lambda (ac key key-state)
     (cond
+     ((anthy-prev-page-key? key key-state)
+      (if (anthy-context-candidate-window ac)
+	  (im-shift-page-candidate ac #f)))
+
+     ((anthy-next-page-key? key key-state)
+      (if (anthy-context-candidate-window ac)
+	  (im-shift-page-candidate ac #t)))
+
      ((anthy-commit-key? key key-state)
       (anthy-do-commit ac))
      
@@ -735,13 +885,10 @@
      ((anthy-cancel-key? key key-state)
       (anthy-cancel-conv ac))
 
-     ((anthy-prev-page-key? key key-state)
-      (if (anthy-context-candidate-window ac)
-	  (im-shift-page-candidate ac #f)))
-
-     ((anthy-next-page-key? key key-state)
-      (if (anthy-context-candidate-window ac)
-	  (im-shift-page-candidate ac #t)))
+     ((and anthy-select-candidate-by-numeral-key?
+	   (numeral-char? key)
+	   (anthy-context-candidate-window ac))
+      (anthy-move-candidate-in-page ac key))
 
      ;; don't discard shift-modified keys. Some of them ("?", "~",
      ;; etc) are used to implicit commit. Reported by [Anthy-dev 745]
@@ -766,8 +913,7 @@
       (cond
        ((anthy-on-key? key key-state)
 	(anthy-flush ac)
-	(anthy-update-mode ac)
-	(anthy-update-prop-label ac))
+	(anthy-context-set-on! ac #t))
        ((and (modifier-key-mask key-state)
 	     (not (shift-key-mask key-state)))
 	(anthy-commit-raw ac))
@@ -782,14 +928,17 @@
     (if (control-char? key)
 	(im-commit-raw ac)
 	(if (anthy-context-on ac)
-	    (if (anthy-context-converting ac)
-		(anthy-proc-converting-state ac key key-state)
-		(anthy-proc-input-state ac key key-state))
+	    (if (anthy-context-transposing ac)
+		(anthy-proc-transposing-state ac key key-state)
+		(if (anthy-context-converting ac)
+		    (anthy-proc-converting-state ac key key-state)
+		    (anthy-proc-input-state ac key key-state)))
 	    (if (anthy-context-wide-latin ac)
 		(anthy-proc-wide-latin ac key key-state)
 		(anthy-proc-raw-state ac key key-state))))
     ;; preedit
-    (anthy-update-preedit ac)))
+    (anthy-update-preedit ac)
+))
 
 
 (define anthy-release-key-handler
@@ -813,7 +962,7 @@
     (let* ((ac-id (anthy-context-ac-id ac))
 	   (cur-seg (ustr-cursor-pos (anthy-context-segments ac)))
 	   (cand (anthy-lib-get-nth-candidate ac-id cur-seg idx)))
-      (list cand (digit->string (+ idx 1))))))
+      (list cand (digit->string (+ idx 1)) ""))))
 
 (define anthy-set-candidate-index-handler
   (lambda (ac idx)
@@ -821,176 +970,22 @@
 ;    (anthy-move-segment ac 1)
     (anthy-update-preedit ac)))
 
-(define anthy-prop-handler
-  (lambda (ac message)
-    (let* ((rkc (anthy-context-rkc ac)))
-      (anthy-flush ac)
-      (anthy-update-preedit ac)
-      (cond
-       ((string=? message
-		       "prop_anthy_hiragana")
-	(begin
-	  (anthy-context-set-on! ac #t)
-	  (anthy-context-set-kana-mode! ac anthy-type-hiragana)))
-       ((string=? message
-		       "prop_anthy_katakana")
-	(begin
-	  (anthy-context-set-on! ac #t)
-	  (anthy-context-set-kana-mode! ac anthy-type-katakana)))
-       ((string=? message
-		       "prop_anthy_hankana")
-	(begin
-	  (anthy-context-set-on! ac #t)
-	  (anthy-context-set-kana-mode! ac anthy-type-hankana)))
-       ((string=? message
-		       "prop_anthy_direct")
-	(begin
-	  (anthy-context-set-on! ac #f)
-	  (anthy-context-set-wide-latin! ac #f)))
-       ((string=? message
-		       "prop_anthy_zenkaku")
-	(begin
-	  (anthy-context-set-on! ac #f)
-	  (anthy-context-set-wide-latin! ac #t)))
-       ((string=? message
-		       "prop_anthy_roma")
-	(begin
-	  (rk-context-set-rule! rkc ja-rk-rule)
-	  (anthy-context-set-input-rule! ac anthy-input-rule-roma)))
-       ((string=? message
-		       "prop_anthy_kana")
-	(begin
-	  (rk-context-set-rule! rkc ja-kana-rule)
-	  (anthy-context-set-input-rule! ac anthy-input-rule-kana)
-;	  (define-key anthy-kana-toggle-key? "")
-;	  (define-key anthy-latin-key? generic-on-key?)
-;	  (define-key anthy-wide-latin-key? "")
-	  )))
-      (anthy-update-mode ac)
-      (anthy-update-prop-label ac))))
-
-(define anthy-update-prop-label
-  (lambda (ac)
-    (let ((str "")
-	  (kana (anthy-context-kana-mode ac))
-	  (rule (anthy-context-input-rule ac)))
-      (set! str
-	    (if (anthy-context-on ac)
-		(cond
-		 ((= kana anthy-type-hiragana)
-		  "\tҤ餬\n")
-		 ((= kana anthy-type-katakana)
-		  "\t\n")
-		  ((= kana anthy-type-hankana)
-		   "\tȾѥ\n"))
-		(if (anthy-context-wide-latin ac)
-		    "\tѱѿ\n"
-		    "a\tľ\n")))
-      (set! str
-	    (cond
-	     ((= rule anthy-input-rule-roma)
-	      (string-append str "\t޻\n"))
-	     ((= rule anthy-input-rule-kana)
-	      (string-append str "\t\n"))))
-      (im-update-prop-label ac str))))
-
-(define anthy-update-mode
-  (lambda (ac)
-    (if (anthy-context-on ac)
-	(let ((kana (anthy-context-kana-mode ac)))
-	  (cond
-	   ((= kana anthy-type-hiragana)
-	    (im-update-mode ac anthy-mode-hiragana))
-	   ((= kana anthy-type-katakana)
-	    (im-update-mode ac anthy-mode-katakana))
-	   ((= kana anthy-type-hankana)
-	    (im-update-mode ac anthy-mode-hankana))))
-	(if (anthy-context-wide-latin ac)
-	    (im-update-mode ac anthy-mode-wide-latin)
-	    (im-update-mode ac anthy-mode-direct)))
-    (anthy-update-prop-label ac)
-    (anthy-update-prop-list ac)))
-
-(define anthy-update-prop-list
-  (lambda (ac)
-    (let* ((kana (anthy-context-kana-mode ac))
-           (rule (anthy-context-input-rule ac))
-	   (str "branch\t"))
-      (set! str (string-append str
-			       (if (anthy-context-on ac)
-				   (cond
-				       ((= kana anthy-type-hiragana)  
-				           "\tҤ餬\n")
-				       ((= kana anthy-type-katakana)
-				           "\t\n")
-				       ((= kana anthy-type-hankana)
-					   "\tȾѥ\n"))
-				   (if (anthy-context-wide-latin ac)
-				       "\tѱѿ\n"
-				       "a\tľ\n"))))
-      (set! str (string-append
-		 str
-		 "leaf\t\tҤ餬\tҤ餬ʤǤ\tprop_anthy_hiragana\t"
-		 (if (and (anthy-context-on ac)
-			  (= kana anthy-type-hiragana))
-		     "*\n"
-		     "\n")
-
-		 "leaf\t\t\tʤϤǤޤ\tprop_anthy_katakana\t"
-		 (if (and (anthy-context-on ac)
-			  (= kana anthy-type-katakana))
-		     "*\n"
-		     "\n")
-
-		 "leaf\t\tȾѥ\tȾѥʤϤǤޤ\tprop_anthy_hankana\t"
-		 (if (and (anthy-context-on ac)
-			  (= kana anthy-type-hankana))
-		     "*\n"
-		     "\n")
-
-		 "leaf\tA\tľ\t쥯ȤǤ\tprop_anthy_direct\t"
-		 (if (and (not (anthy-context-on ac))
-			  (not (anthy-context-wide-latin ac)))
-                     "*\n"
-                     "\n")
-
-		 "leaf\t\tѱѿ\tѱѿ⡼\tprop_anthy_zenkaku\t"
-		 (if (and (not (anthy-context-on ac))
-			  (anthy-context-wide-latin ac))
-                     "*\n"
-                     "\n")
-
-		 "branch\t"
-		 (if (= rule anthy-input-rule-roma)
-		     "\t޻\n"
-		     "\t\n")
-
-		 "leaf\t\t޻\t޻\tprop_anthy_roma\t"
-		 (if (= rule anthy-input-rule-roma)
-		     "*\n"
-		     "\n")
-
-		 "leaf\t\t\t\tprop_anthy_kana\t"
-		 (if (= rule anthy-input-rule-kana)
-		     "*\n"
-		     "\n")
-		 ))
-      (im-update-prop-list ac str)
-      )))
+(anthy-configure-widgets)
 
 (register-im
  'anthy
  "ja"
  "EUC-JP"
- (N_ "Japanese Kana Kanji Conversion Engine, Anthy")
+ anthy-im-name-label
+ anthy-im-short-desc
  #f
  anthy-init-handler
  anthy-release-handler
- anthy-mode-handler
+ context-mode-handler
  anthy-press-key-handler
  anthy-release-key-handler
  anthy-reset-handler
  anthy-get-candidate-handler
  anthy-set-candidate-index-handler
- anthy-prop-handler
+ context-prop-activate-handler
 )
