@@ -1,6 +1,6 @@
 ;;; canna.scm: Canna for uim.
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -16,10 +16,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,57 +31,167 @@
 
 (require "japanese.scm")
 (require "japanese-kana.scm")
-(require "generic-key.scm")
+(require "japanese-azik.scm")
+(require-custom "generic-key-custom.scm")
+(require-custom "canna-custom.scm")
+(require-custom "canna-key-custom.scm")
+
+
+;;; implementations
 
-; variables
 (define canna-init-lib-ok? #f)
-(define canna-use-candidate-window? #t)
-(define canna-candidate-op-count 1)
-(define canna-nr-candidate-max 10)
-(define canna-show-segment-separator? #f)
-(define canna-segment-separator "|")
-; TODO: support cannaserver on other host
-(define canna-server-name #f)
-;(define canna-server-name "localhost")
-;(define canna-server-name "127.0.0.1")
-;;; Key definition
-(define-key canna-on-key? '("<Control>\\" generic-on-key?))
-(define-key canna-off-key? '("<Control>\\" generic-off-key?))
-(define-key canna-latin-key? '("<Control>\\" generic-off-key?))
-;(define-key canna-latin-key '("l" generic-on-key?))
-(define-key canna-wide-latin-key? "L")
-(define-key canna-hankaku-kana-key? '("<Control>q" "<Control>Q"))
-(define-key canna-kana-toggle-key? "q")
-(define-key canna-commit-as-opposite-kana-key? "Q")
-(define-key canna-begin-conv-key? '(generic-begin-conv-key? generic-on-key?))
-(define-key canna-commit-key? 'generic-commit-key?)
-(define-key canna-extend-segment-key? '("<Control>o" "<Control>O" "<Shift>right"))
-(define-key canna-shrink-segment-key? '("<Control>i" "<Control>I" "<Shift>left"))
-(define-key canna-next-candidate-key? 'generic-next-candidate-key?)
-(define-key canna-prev-candidate-key? 'generic-prev-candidate-key?)
-;(define-key canna-next-page-key? 'generic-next-page-key?)
-;(define-key canna-prev-page-key? 'generic-prev-page-key?)
-(define-key canna-cancel-key? 'generic-cancel-key?)
-(define-key canna-backspace-key? 'generic-backspace-key?)
-(define-key canna-delete-key? 'generic-delete-key?)
-;(define-key canna-kill-key? 'generic-kill-key?)
-;(define-key canna-kill-backward-key? 'generic-kill-backward-key?)
-(define-key canna-go-left-key? 'generic-go-left-key?)
-(define-key canna-go-right-key? 'generic-go-right-key?)
-(define-key canna-beginning-of-preedit-key? 'generic-beginning-of-preedit-key?)
-(define-key canna-end-of-preedit-key? 'generic-end-of-preedit-key?)
-(define-key canna-next-segment-key? 'generic-go-right-key?)
-(define-key canna-prev-segment-key? 'generic-go-left-key?)
 
 (define canna-input-rule-roma 0)
 (define canna-input-rule-kana 1)
 (define canna-input-rule-azik 2)
 
+(define canna-prepare-activation
+  (lambda (cc)
+    (canna-flush cc)
+    (canna-update-preedit cc)))
+
+(register-action 'action_canna_hiragana
+		 (lambda (cc) ;; indication handler
+		   '(figure_ja_hiragana
+		     ""
+		     "Ҥ餬"
+		     "Ҥ餬ϥ⡼"))
+
+		 (lambda (cc) ;; activity predicate
+		   (and (canna-context-on cc)
+			(= (canna-context-kana-mode cc)
+			   multi-segment-type-hiragana)))
+
+		 (lambda (cc) ;; action handler
+		   (canna-prepare-activation cc)
+		   (canna-context-set-on! cc #t)
+		   (canna-context-set-kana-mode! cc
+						 multi-segment-type-hiragana)))
+
+(register-action 'action_canna_katakana
+		 (lambda (cc)
+		   '(figure_ja_katakana
+		     ""
+		     ""
+		     "ϥ⡼"))
+		 (lambda (cc)
+		   (and (canna-context-on cc)
+			(= (canna-context-kana-mode cc)
+			   multi-segment-type-katakana)))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (canna-context-set-on! cc #t)
+		   (canna-context-set-kana-mode! cc
+						 multi-segment-type-katakana)))
+
+(register-action 'action_canna_hankana
+		 (lambda (cc)
+		   '(figure_ja_hankana
+		     ""
+		     "Ⱦѥ"
+		     "Ⱦѥϥ⡼"))
+		 (lambda (cc)
+		   (and (canna-context-on cc)
+			(= (canna-context-kana-mode cc)
+			   multi-segment-type-hankana)))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (canna-context-set-on! cc #t)
+		   (canna-context-set-kana-mode! cc
+						 multi-segment-type-hankana)))
+
+(register-action 'action_canna_direct
+		 (lambda (cc)
+		   '(figure_ja_direct
+		     "a"
+		     "ľ"
+		     "ľ(̵Ѵ)ϥ⡼"))
+		 (lambda (cc)
+		   (and (not (canna-context-on cc))
+			(not (canna-context-wide-latin cc))))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (canna-context-set-on! cc #f)
+		   (canna-context-set-wide-latin! cc #f)))
+
+(register-action 'action_canna_zenkaku
+		 (lambda (cc)
+		   '(figure_ja_zenkaku
+		     ""
+		     "ѱѿ"
+		     "ѱѿϥ⡼"))
+		 (lambda (cc)
+		   (and (not (canna-context-on cc))
+			(canna-context-wide-latin cc)))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (canna-context-set-on! cc #f)
+		   (canna-context-set-wide-latin! cc #t)))
+
+(register-action 'action_canna_roma
+		 (lambda (cc)
+		   '(figure_ja_roma
+		     ""
+		     "޻"
+		     "޻ϥ⡼"))
+		 (lambda (cc)
+		   (= (canna-context-input-rule cc)
+		      canna-input-rule-roma))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (rk-context-set-rule! (canna-context-rkc cc)
+					 ja-rk-rule)
+		   (canna-context-set-input-rule! cc canna-input-rule-roma)))
+
+(register-action 'action_canna_kana
+		 (lambda (cc)
+		   '(figure_ja_kana
+		     ""
+		     ""
+		     "ϥ⡼"))
+		 (lambda (cc)
+		   (= (canna-context-input-rule cc)
+		      canna-input-rule-kana))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (rk-context-set-rule! (canna-context-rkc cc)
+					 ja-kana-hiragana-rule)
+		   (canna-context-set-input-rule! cc canna-input-rule-kana)))
+
+(register-action 'action_canna_azik
+		 (lambda (cc)
+		   '(figure_ja_azik
+		     ""
+		     "AZIK"
+		     "AZIKĥ޻ϥ⡼"))
+		 (lambda (cc)
+		   (= (canna-context-input-rule cc)
+		      canna-input-rule-azik))
+		 (lambda (cc)
+		   (canna-prepare-activation cc)
+		   (rk-context-set-rule! (canna-context-rkc cc)
+					 ja-azik-rule)
+		   (canna-context-set-input-rule! cc canna-input-rule-azik)))
+
+;; Update widget definitions based on action configurations. The
+;; procedure is needed for on-the-fly reconfiguration involving the
+;; custom API
+(define canna-configure-widgets
+  (lambda ()
+    (register-widget 'widget_canna_input_mode
+		     (activity-indicator-new canna-input-mode-actions)
+		     (actions-new canna-input-mode-actions))
+
+    (register-widget 'widget_canna_kana_input_method
+		     (activity-indicator-new canna-kana-input-method-actions)
+		     (actions-new canna-kana-input-method-actions))
+    (context-list-replace-widgets! 'canna canna-widgets)))
+
 (define canna-context-rec-spec
   (append
    context-rec-spec
    (list
-    (list 'on                 ()) ;; #f
+    (list 'on                 #f)
     (list 'state              ())
     (list 'cc-id              ()) ;; canna-context-id
     (list 'left-string        ()) ;; preedit strings in the left of cursor
@@ -91,10 +201,9 @@
     (list 'cur-seg            ())
     (list 'candidate-window   ())
     (list 'candidate-op-count ())
-    (list 'wide-latin         ()) ;; #f
+    (list 'wide-latin         #f)
     (list 'kana-mode          multi-segment-type-hiragana)
     (list 'commit-raw         #t)
-    ;;(list 'unused-13th        #f)  ;; see anthy.scm
     (list 'input-rule         canna-input-rule-roma))))
 (define-record 'canna-context canna-context-rec-spec)
 (define canna-context-new-internal canna-context-new)
@@ -105,9 +214,8 @@
 ;    (canna-context-set-cc-id! cc (if canna-init-lib-ok?
 ;				      (canna-lib-alloc-context) ()))
     (canna-context-set-cc-id! cc (canna-lib-alloc-context))
+    (canna-context-set-widgets! cc canna-widgets)
     (canna-context-set-rkc! cc rkc)
-    (canna-flush cc)
-    (canna-context-set-on! cc #f)
     (if using-kana-table?
         (canna-context-set-input-rule! cc canna-input-rule-kana)
         (canna-context-set-input-rule! cc canna-input-rule-roma))
@@ -132,33 +240,10 @@
 	 (opposite-kana (canna-opposite-kana kana)))
     (canna-context-set-kana-mode! cc opposite-kana)))
 
-(define (canna-mode-handler cc mode)
-  (canna-flush cc)
-  (cond
-   ((= mode multi-segment-mode-direct)
-    (canna-context-set-on! cc #f)
-    (canna-context-set-wide-latin! cc #f))
-   ((= mode multi-segment-mode-hiragana)
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-hiragana))
-   ((= mode multi-segment-mode-katakana)
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-katakana))
-   ((= mode multi-segment-mode-wide-latin)
-    (canna-context-set-on! cc #f)
-    (canna-context-set-wide-latin! cc #t))
-   ((= mode multi-segment-mode-hankana)
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-hankana)))
-  (canna-update-prop-list cc)
-  (canna-update-preedit cc)
-  ())
-
 (define (canna-flush cc)
   (rk-flush (canna-context-rkc cc))
   (canna-context-set-left-string! cc '())
   (canna-context-set-right-string! cc '())
-  (canna-context-set-on! cc #t)
   (canna-context-set-state! cc #f)
   (canna-context-set-index-list! cc ())
   (canna-context-set-candidate-window! cc #f)
@@ -223,17 +308,13 @@
       (begin
 	(canna-flush cc)
 	(canna-context-set-on! cc #f)
-	(canna-context-set-wide-latin! cc #t)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc)))
+	(canna-context-set-wide-latin! cc #t)))
      
      ((canna-latin-key? key key-state)
       (begin
 	(canna-flush cc)
 	(canna-context-set-on! cc #f)
-	(canna-context-set-wide-latin! cc #f)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc)))
+	(canna-context-set-wide-latin! cc #f)))
      
      ((canna-backspace-key? key key-state)
       (canna-commit-raw cc))
@@ -242,16 +323,10 @@
       (canna-commit-raw cc))
      
      ((canna-hankaku-kana-key? key key-state)
-      (begin 
-	(canna-context-set-kana-mode! cc multi-segment-type-hankana)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc)))
+      (canna-context-set-kana-mode! cc multi-segment-type-hankana))
      
      ((canna-kana-toggle-key? key key-state)
-      (begin 
-	(canna-context-kana-toggle cc)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc)))
+      (canna-context-kana-toggle cc))
      
      ;; modifiers (except shift) => ignore
      ((and (modifier-key-mask key-state)
@@ -331,12 +406,10 @@
 	 (string-append
 	  (multi-segment-make-left-string (canna-context-left-string cc)
 					  kana)
-	  (multi-semgnet-make-right-string (canna-context-right-string cc)
+	  (multi-segment-make-right-string (canna-context-right-string cc)
 					   kana)))
 	(canna-flush cc)
-	(canna-context-kana-toggle cc)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc)))
+	(canna-context-kana-toggle cc)))
      ;; cancel
      ((canna-cancel-key? key key-state)
       (canna-flush cc))
@@ -501,44 +574,12 @@
     (canna-flush cc))
 
 (define (canna-init-handler id im arg)
-  (let ((cc (canna-context-new id im)))
-    (im-clear-mode-list cc)
-    (im-pushback-mode-list cc "ľ")
-    (im-pushback-mode-list cc "Ҥ餬")
-    (im-pushback-mode-list cc "")
-    (im-pushback-mode-list cc "ѱѿ")
-    (im-pushback-mode-list cc "Ⱦѥ")
-    (im-update-mode-list cc)
-    (im-update-mode cc 0)
-    (canna-update-prop-list cc)
-    cc))
+  (canna-context-new id im))
 
 (define (canna-release-handler cc)
   (let ((cc-id (canna-context-cc-id cc)))
     (canna-lib-release-context cc-id)))
 
-(define (canna-mode-handler cc mode)
-  (canna-flush cc)
-  (cond
-   ((= mode multi-segment-mode-direct)  ; 'direct'
-    (canna-context-set-on! cc #f)
-    (canna-context-set-wide-latin! cc #f))
-   ((= mode multi-segment-mode-hiragana) ; 'hiragana'
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-hiragana))
-   ((= mode multi-segment-mode-katakana) ; 'katakana'
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-katakana))
-   ((= mode multi-segment-mode-wide-latin) ; 'wide-latin'
-    (canna-context-set-on! cc #f)
-    (canna-context-set-wide-latin! cc #t))
-   ((= mode multi-segment-mode-hankana)
-    (canna-context-set-on! cc #t)
-    (canna-context-set-kana-mode! cc multi-segment-type-hankana)))  
-  (canna-update-preedit cc)
-  (canna-update-prop-label cc)
-  ())
-
 (define (canna-move-segment cc dir)
   (let ((pos (+ (canna-context-cur-seg cc) dir))
 	(nseg (length (canna-context-index-list cc))))
@@ -653,8 +694,7 @@
     (cond
      ((canna-on-key? key key-state)
       (canna-flush cc)
-      (canna-update-mode cc)
-      (canna-update-prop-label cc))
+      (canna-context-set-on! cc #t))
      ((and (modifier-key-mask key-state)
 	   (not (shift-key-mask key-state)))
       (canna-commit-raw cc))
@@ -693,7 +733,7 @@
   (let* ((cc-id (canna-context-cc-id cc))
 	 (cand (canna-lib-get-nth-candidate
 		cc-id (canna-context-cur-seg cc) idx)))
-    (list cand (digit->string (+ idx 1)))))
+    (list cand (digit->string (+ idx 1)) "")))
 
 (define (canna-set-candidate-index-handler cc idx)
   (let* ((seg (canna-context-cur-seg cc))
@@ -701,178 +741,10 @@
     (set-car! (nthcdr seg (canna-context-index-list cc)) idx)
     (canna-update-preedit cc)))
 
-;;;
-
-(define (canna-prop-handler cc message)
-  (let ((rkc (canna-context-rkc cc)))
-    (canna-flush cc)
-    (canna-update-preedit cc)
-    (cond
-     ((string=? message
-		"prop_canna_hiragana")
-      (begin
-	(canna-context-set-on! cc #t)
-	(canna-context-set-kana-mode! cc multi-segment-type-hiragana)))
-     ((string=? message
-		"prop_canna_katakana")
-      (begin
-	(canna-context-set-on! cc #t)
-	(canna-context-set-kana-mode! cc multi-segment-type-katakana)))
-     ((string=? message
-		"prop_canna_hankana")
-      (begin
-	(canna-context-set-on! cc #t)
-	(canna-context-set-kana-mode! cc multi-segment-type-hankana)))
-     ((string=? message
-		"prop_canna_direct")
-      (begin
-	(canna-context-set-on! cc #f)
-	(canna-context-set-wide-latin! cc #f)))
-     ((string=? message
-		"prop_canna_zenkaku")
-      (begin
-	(canna-context-set-on! cc #f)
-	(canna-context-set-wide-latin! cc #t)))
-     ((string=? message
-		"prop_canna_roma")
-      (begin
-	(rk-context-set-rule! rkc ja-rk-rule)
-	(canna-context-set-input-rule! cc canna-input-rule-roma)))
-     ((string=? message
-		"prop_canna_kana")
-      (begin
-	(rk-context-set-rule! rkc ja-kana-rule)
-	(canna-context-set-input-rule! cc canna-input-rule-kana)
-	)))
-    (canna-update-mode cc)
-    (canna-update-prop-label cc)))
-;;;
-
 (define (canna-proc-raw-state cc key key-state)
   (if (canna-on-key? key key-state)
-      (begin
-	(canna-begin-input cc)
-	(canna-update-mode cc)
-	(canna-update-prop-label cc))
-      (im-commit-raw cc)
-      (canna-update-preedit)))
-
-(define (canna-update-prop-label cc)
-  (let ((str "")
-	(kana (canna-context-kana-mode cc))
-	(rule (canna-context-input-rule cc)))
-    (set! str
-	  (if (canna-context-on cc)
-	      (cond
-	       ((= kana multi-segment-type-hiragana)
-		"\tҤ餬\n")
-	       ((= kana multi-segment-type-katakana)
-		"\t\n")
-	       ((= kana multi-segment-type-hankana)
-		"\tȾѥ\n"))
-	      (if (canna-context-wide-latin cc)
-		  "\tѱѿ\n"
-		  "a\tľ\n")))
-    (set! str
-	  (cond
-	   ((= rule canna-input-rule-roma)
-	    (string-append str "\t޻\n"))
-	   ((= rule canna-input-rule-kana)
-	    (string-append str "\t\n"))))
-    (im-update-prop-label cc str)))
-
-(define (canna-update-mode cc)
-  (if (canna-context-on cc)
-      (let ((kana (canna-context-kana-mode cc)))
-	(cond
-	 ((= kana multi-segment-type-hiragana)
-	  (im-update-mode cc multi-segment-mode-hiragana))
-	 ((= kana multi-segment-type-katakana)
-	  (im-update-mode cc multi-segment-mode-katakana))
-	 ((= kana multi-segment-type-hankana)
-	  (im-update-mode cc multi-segment-mode-hankana))))
-      (if (canna-context-wide-latin cc)
-	  (im-update-mode cc multi-segment-mode-wide-latin)
-	  (im-update-mode cc multi-segment-mode-direct)))
-  (canna-update-prop-label cc)
-  (canna-update-prop-list cc))
-
-(define (canna-update-prop-list cc)
-  (let ((str "branch\t")
-	(kana (canna-context-kana-mode cc))
-	(rule (canna-context-input-rule cc)))
-    (set! str
-	  (string-append str
-			 (if (canna-context-on cc)
-			     (cond
-			         ((= kana multi-segment-type-hiragana)
-				     "\tҤ餬\n")
-				 ((= kana multi-segment-type-katakana)
-				     "\t\n")
-				 ((= kana multi-segment-type-hankana)
-				     "\tȾѥ\n"))
-			     (if (canna-context-wide-latin cc)
-				 "\tѱѿ\n"
-				 "a\tľ\n"))))
-    (set! str
-	  (string-append
-	   str
-	   "leaf\t\tҤ餬\tҤ餬ʤǤ\tprop_canna_hiragana\t"
-	   (if (and (canna-context-on cc)
-		    (= kana multi-segment-type-hiragana))
-	       "*\n"
-	       "\n")
-
-	   "leaf\t\t\tʤϤǤޤ\tprop_canna_katakana\t"
-	   (if (and (canna-context-on cc)
-		    (= kana multi-segment-type-katakana))
-	       "*\n"
-	       "\n")
-
-	   "leaf\t\tȾѥ\tȾѥʤϤǤޤ\tprop_canna_hankana\t"
-	   (if (and (canna-context-on cc)
-		    (= kana multi-segment-type-hankana))
-	       "*\n"
-	       "\n")
-
-	   "leaf\tA\tľ\t쥯ȤǤ\tprop_canna_direct\t"
-	   (if (and (not (canna-context-on cc))
-		    (not (canna-context-wide-latin cc)))
-	       "*\n"
-	       "\n")
-
-	   "leaf\t\tѱѿ\tѱѿ⡼\tprop_canna_zenkaku\t"
-	   (if (and (not (canna-context-on cc))
-		    (canna-context-wide-latin cc))
-	       "*\n"
-	       "\n")
-
-	   "branch\t"
-	   (if (= rule canna-input-rule-roma)
-	       "\t޻\n"
-	       "\t\n")
-
-	   "leaf\t\t޻\t޻\tprop_canna_roma\t"
-	   (if (= rule canna-input-rule-roma)
-	       "*\n"
-	       "\n")
-
-	   "leaf\t\t\t\tprop_canna_kana\t"
-	   (if (= rule canna-input-rule-kana)
-	       "*\n"
-	       "\n")))
-    (im-update-prop-list cc str)))
-
-(define canna-prop-list
-  '(
-    ("branch" "Ͼ")
-    ("leaf" "" "Ҥ餬" "Ҥ餬ʤϤǤޤ" "prop_canna_hiragana")
-    ("leaf" "" "" "" "prop_canna_katakana")
-    ("leaf" "" "ѱѿ" "ѱѿ⡼" "prop_canna_zenkaku")
-    ("leaf" "a" "ľ" "ե٥åȤϤǤޤ" "prop_canna_direct")
-    ("branch" "ϥ⡼")
-    ("leaf" "" "޻" "޻" "prop_canna_roma")
-    ("leaf" "" "" "" "prop_canna_kana")))
+      (canna-begin-input cc)
+      (im-commit-raw cc)))
 
 (if (and
      (symbol-bound? 'canna-lib-init)
@@ -882,18 +754,21 @@
 (if (and
      (symbol-bound? 'canna-lib-init)
      (= canna-init-lib-ok? #t))
-    (register-im
-     'canna
-     "ja"
-     "EUC-JP"
-     (N_ "Canna")
-     #f
-     canna-init-handler
-     canna-release-handler
-     canna-mode-handler
-     canna-press-key-handler
-     canna-release-key-handler
-     canna-reset-handler
-     canna-get-candidate-handler
-     canna-set-candidate-index-handler
-     canna-prop-handler))
+    (begin
+      (canna-configure-widgets)
+      (register-im
+       'canna
+       "ja"
+       "EUC-JP"
+       canna-im-name-label
+       canna-im-short-desc
+       #f
+       canna-init-handler
+       canna-release-handler
+       context-mode-handler
+       canna-press-key-handler
+       canna-release-key-handler
+       canna-reset-handler
+       canna-get-candidate-handler
+       canna-set-candidate-index-handler
+       context-prop-activate-handler)))
