@@ -1,5 +1,5 @@
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -15,10 +15,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -28,20 +28,84 @@
 ;;; SUCH DAMAGE.
 ;;;;
 
+(require "util.scm")
 (require "rk.scm")
-(require "generic-key.scm")
-;;
+(require-custom "generic-key-custom.scm")
+(require-custom "generic-custom.scm")
 
-(define generic-use-candidate-window? #t)
-(define generic-candidate-op-count 1)
-(define generic-nr-candidate-max 10)
+
+;; widgets and actions
+
+;; widgets
+(define generic-widgets '(widget_generic_input_mode))
+
+;; default activity for each widgets
+(define default-widget_generic_input_mode 'action_generic_off)
+
+;; actions of widget_generic_input_mode
+(define generic-input-mode-actions
+  '(action_generic_off
+    action_generic_on))
+
+
+;;; implementations
+
+(define ascii-rule
+  (map (compose (lambda (entry)
+		  (list (list entry) entry))
+		list
+		charcode->string)
+       (iota 127 32)))
+
+(define generic-prepare-activation
+  (lambda (gc)
+    (let ((rkc (generic-context-rk-context gc)))
+      (rk-flush rkc)
+      (generic-update-preedit gc))))
+
+(register-action 'action_generic_off
+		 (lambda (gc)
+		   (list
+		    'figure_generic_off
+		    "-"
+		    (N_ "off")
+		    (N_ "Direct Input Mode")))
+		 (lambda (gc)
+		   (not (generic-context-on gc)))
+		 (lambda (gc)
+		   (generic-prepare-activation gc)
+		   (generic-context-set-on! gc #f)))
+
+(register-action 'action_generic_on
+		 (lambda (gc)
+		   (let* ((im (generic-context-im gc))
+			  (name (symbol->string (im-name im))))
+		     (list
+		      'figure_generic_on
+		      "O"
+		      (N_ "on")
+		      (string-append name (N_ " Mode")))))
+		 (lambda (gc)
+		   (generic-context-on gc))
+		 (lambda (gc)
+		   (generic-prepare-activation gc)
+		   (generic-context-set-on! gc #t)))
+
+;; Update widget definitions based on action configurations. The
+;; procedure is needed for on-the-fly reconfiguration involving the
+;; custom API
+(define generic-configure-widgets
+  (lambda ()
+    (register-widget 'widget_generic_input_mode
+		     (activity-indicator-new generic-input-mode-actions)
+		     (actions-new generic-input-mode-actions))))
 
 (define generic-context-rec-spec
   (append
    context-rec-spec
-   '((rk-context         ()) ;; #f?
+   '((rk-context         #f)
      (rk-nth             0)
-     (on                 ()) ;; #f?
+     (on                 #f)
      (candidate-op-count 0)
      (raw-commit         #f)
      (converting         #f))))
@@ -50,17 +114,18 @@
 
 (define generic-context-new
   (lambda (id im rule back)
-    (let ((pc (generic-context-new-internal id im)))
-      (generic-context-set-rk-context!
-       pc (rk-context-new rule #f back))
-      pc)))
+    (let ((gc (generic-context-new-internal id im))
+	  (rkc (rk-context-new rule #f back)))
+      (generic-context-set-widgets! gc generic-widgets)
+      (generic-context-set-rk-context! gc rkc)
+      gc)))
+
 (define generic-context-flush
   (lambda (pc)
     (generic-context-set-rk-nth! pc 0)
     (generic-context-set-candidate-op-count! pc 0)
     (generic-context-set-converting! pc #f)))
 
-
 (define generic-update-preedit
   (lambda (pc)
     (if (generic-context-raw-commit pc)
@@ -81,6 +146,42 @@
     (im-commit-raw pc)
     (generic-context-set-raw-commit! pc #t)))
 
+(define generic-commit
+  (lambda (pc)
+    (let* ((rkc (generic-context-rk-context pc))
+	   (cs (rk-current-seq rkc)))
+      (if (> (length (cadr cs)) 0)
+	  (begin
+	    (im-commit pc (nth (generic-context-rk-nth pc) (cadr cs)))
+	    (im-deactivate-candidate-selector pc)
+	    (rk-flush rkc)
+	    (generic-context-flush pc))
+	  (begin
+	    (im-commit-raw pc)
+	    (rk-flush rkc)
+	    (im-update-preedit pc))))))
+
+(define generic-commit-by-numkey
+  (lambda (pc key)
+    (let* ((rkc (generic-context-rk-context pc))
+	   (cs (rk-current-seq rkc))
+	   (n (generic-context-rk-nth pc) (cadr cs))
+	   (cur-page (if (= generic-nr-candidate-max 0)
+			 0
+			 (quotient n generic-nr-candidate-max)))
+	   (pageidx (- (numeral-char->number key) 1))
+	   (compensated-pageidx (cond
+				 ((< pageidx 0) ; pressing key_0
+				  (+ pageidx 10))
+				 (else
+				  pageidx)))
+	   (idx (+ (* cur-page generic-nr-candidate-max) compensated-pageidx)))
+      (im-commit pc (nth idx (cadr cs)))
+      (im-deactivate-candidate-selector pc)
+      (rk-flush rkc))))
+
+;; XXX: This function is too long, we should split into several functions..
+;; 2004-11-05 TOKUNAGA Hiroyuki
 (define generic-proc-input-state
   (lambda (pc key state)
     (let* ((rkc (generic-context-rk-context pc))
@@ -92,9 +193,6 @@
 	   (begin
 	     (rk-flush rkc)
 	     (generic-context-set-on! pc #f)
-	     (generic-update-prop-label pc)
-	     (generic-update-prop-list pc)
-	     (im-update-mode pc 0)
 	     #f)
 	   #t)
        (if (generic-prev-candidate-key? key state)
@@ -123,17 +221,8 @@
 	     #f)
 	   #t)
        (if (generic-commit-key? key state)
-	   (let ((cs (rk-current-seq rkc)))
-	     (if (> (length (cadr cs)) 0)
-		 (begin
-		   (im-commit pc (nth (generic-context-rk-nth pc) (cadr cs)))
-		   (im-deactivate-candidate-selector pc)
-		   (rk-flush rkc)
-		   (generic-context-flush pc))
-		 (begin
-		   (im-commit-raw pc)
-		   (rk-flush rkc)
-		   (im-update-preedit pc)))
+	   (begin
+	     (generic-commit pc)
 	     #f)
 	   #t)
        (if (symbol? key)
@@ -250,24 +339,24 @@
 	     #f)
 	   #t)
        (if (generic-commit-key? key state)
-	   (let ((cs (rk-current-seq rkc)))
-	     (if (> (length (cadr cs)) 0)
-		 (begin
-		   (im-commit pc (nth (generic-context-rk-nth pc) (cadr cs)))
-		   (generic-context-flush pc)
-		   (rk-flush rkc)
-		   (im-deactivate-candidate-selector pc))
-		 (begin
-		   (im-commit-raw pc)
-		   (rk-flush rkc)
-		   (im-update-preedit pc)))
+	   (begin
+	     (generic-commit pc)
 	     #f)
 	   #t)
        (if (symbol? key)
+	   (if (>= (generic-context-candidate-op-count pc)
+		   generic-candidate-op-count)
+	       #f
+	       (begin
+		 (rk-flush rkc)
+		 (generic-commit-raw pc)
+		 (generic-context-set-rk-nth! pc 0)
+		 #f))
+	   #t)
+       (if (and generic-commit-candidate-by-numeral-key?
+		(numeral-char? key))
 	   (begin
-	     (rk-flush rkc)
-	     (generic-commit-raw pc)
-	     (generic-context-set-rk-nth! pc 0)
+	     (generic-commit-by-numkey pc key)
 	     #f)
 	   #t)
        (if (and (modifier-key-mask state)
@@ -293,9 +382,6 @@
      (if (generic-on-key? key state)
 	 (begin
 	   (generic-context-set-on! pc #t)
-	   (generic-update-prop-label pc)
-	   (generic-update-prop-list pc)
-	   (im-update-mode pc 1)
 	   #f)
 	 #t)
      ;;
@@ -325,22 +411,12 @@
     (let ((rkc (generic-context-rk-context pc)))
       (rk-flush rkc))))
 
-(define generic-mode-handler
-  (lambda (pc mode)
-    (let ((rkc (generic-context-rk-context pc)))
-      (if (= mode 0)
-	  (generic-context-set-on! pc #f)
-	  (generic-context-set-on! pc #t))
-      (rk-flush rkc)
-      (generic-update-preedit pc)
-      (generic-update-prop-label pc))))
-
 (define generic-get-candidate-handler
   (lambda (pc idx accel-enum-hint)
     (let* ((rkc (generic-context-rk-context pc))
 	   (cs (cadr (rk-current-seq rkc)))
 	   (cand (car (nthcdr idx cs))))
-      (list cand (digit->string (+ idx 1))))))
+      (list cand (digit->string (+ idx 1)) ""))))
 
 (define generic-set-candidate-index-handler
   (lambda (pc idx)
@@ -349,75 +425,25 @@
 
 (define generic-init-handler
   (lambda (id im init-handler)
-    (let ((pc (init-handler id im #f)))
-      (im-clear-mode-list pc)
-      (im-pushback-mode-list pc "RAW")
-      (im-pushback-mode-list pc (im-name im))
-      (im-update-mode-list pc)
-      (im-update-mode pc 0)
-      (generic-update-prop-list pc)
-      pc)))
-
-(define generic-prop-handler
-  (lambda (pc message)
-    (let ((rkc (generic-context-rk-context pc)))
-
-      (rk-flush rkc)
-      (generic-update-preedit pc)
-      
-      (if (string=? message
-		    "prop_generic_raw")
-	  (begin
-	    (im-update-mode pc 0)
-	    (generic-context-set-on! pc #f)))
-      (if (string=? message
-		    "prop_generic_on")
-	  (begin
-	    (im-update-mode pc 1)
-	    (generic-context-set-on! pc #t)))
-      (generic-update-prop-label pc id)
-      (generic-update-prop-list id))))
-
-(define generic-update-prop-label
-  (lambda (pc)
-    (let*  ((state (generic-context-on pc))
-	    (str ""))
-      (if state
-	  (set! str "O\ton\n")
-	  (set! str "o\toff\n"))
-      (im-update-prop-label pc str))))
-
-(define generic-update-prop-list
-  (lambda (pc)
-    (let* ((state (generic-context-on pc))
-	   (name (im-name (generic-context-im pc)))
-	   (str ""))
-      (if state
-	  (set! str "O\ton\n")
-	  (set! str "o\toff\n"))
-      (set! str (string-append
-		 "branch\t" str
-		 "leaf\to\tRaw Input Mode\tRaw input mode\tprop_generic_raw\t"
-		 (if state
-		     "\n"
-		     "*\n")
-		 "leaf\tO\t" name " Mode\t " name " mode\tprop_generic_on\t"
-		 (if state
-		     "*\n"
-		     "\n")))
-      (im-update-prop-list pc str)
-      )))
-
+    (init-handler id im #f)))
 
 (define generic-register-im
-  (lambda (name lang code short-desc init)
+  (lambda (name lang code name-label short-desc init-arg)
     (register-im
-     name lang code short-desc init generic-init-handler #f
-     generic-mode-handler
+     name
+     lang
+     code
+     name-label
+     short-desc
+     init-arg
+     generic-init-handler
+     #f  ;; release-handler
+     context-mode-handler
      generic-key-press-handler
      generic-key-release-handler
      generic-reset-handler
      generic-get-candidate-handler
      generic-set-candidate-index-handler
-     generic-prop-handler
-     )))
+     context-prop-activate-handler)))
+
+(generic-configure-widgets)
