@@ -1,5 +1,5 @@
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -15,10 +15,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -53,17 +53,23 @@
       (generic-key-press-handler-with-this-env gc key state))))
 
 (define hangul-register-im
-  (lambda (name lang code short-desc init)
+  (lambda (name lang code name-label short-desc init-arg)
     (register-im
-     name lang code short-desc init generic-init-handler #f
-     generic-mode-handler
+     name
+     lang
+     code
+     name-label
+     short-desc
+     init-arg
+     generic-init-handler
+     #f ;; reset-handler
+     context-mode-handler
      hangul-key-press-handler
      generic-key-release-handler
      generic-reset-handler
      generic-get-candidate-handler
      generic-set-candidate-index-handler
-     generic-prop-handler
-     )))
+     context-prop-activate-handler)))
     
 (define hangul2-init-handler
   (lambda (id im arg)
@@ -81,11 +87,26 @@
     (generic-context-new id im romaja-rule #t)))
 
 (hangul-register-im
- 'hangul2 "ko" "UTF-8" (N_ "hangul2") hangul2-init-handler)
+ 'hangul2
+ "ko"
+ "UTF-8"
+ (N_ "Hangul (2-bul)")
+ (N_ "2-bul style hangul input method")
+ hangul2-init-handler)
 
 ;; hangul3 IM does not require generic-keys disabled
 (generic-register-im
- 'hangul3 "ko" "UTF-8" (N_ "hangul3") hangul3-init-handler)
+ 'hangul3
+ "ko"
+ "UTF-8"
+ (N_ "Hangul (3-bul)")
+ (N_ "3-bul style hangul input method")
+ hangul3-init-handler)
 
 (hangul-register-im
- 'romaja "ko" "UTF-8" (N_ "romaja") romaja-init-handler)
+ 'romaja
+ "ko"
+ "UTF-8"
+ (N_ "Hangul (Romaja)")
+ (N_ "Romaja input style hangul input method")
+ romaja-init-handler)
