@@ -1,5 +1,5 @@
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -15,10 +15,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -33,97 +33,11 @@
 ;; config
 (define enable-emacs-like-key-prefix? #t)
 
-;;
-(define valid-key-symbols
-  '(backspace
-    delete
-    escape
-    return
-    tab
-    left
-    up
-    right
-    down
-    prior
-    next
-    home
-    end
-    zenkaku-hankaku
-    Multi_key
-    Mode_switch
-    Henkan_Mode
-    Muhenkan
-    F1
-    F2
-    F3
-    F4
-    F5
-    F6
-    F7
-    F8
-    F9
-    F10
-    F11
-    F12
-    F13
-    F14
-    F15
-    F16
-    F17
-    F18
-    F19
-    F20
-    F21
-    F22
-    F23
-    F24
-    F25
-    F26
-    F27
-    F28
-    F29
-    F30
-    F31
-    F32
-    F33
-    F34
-    F35
-    Private1
-    Private2
-    Private3
-    Private4
-    Private5
-    Private6
-    Private7
-    Private8
-    Private9
-    Private10
-    Private11
-    Private12
-    Private13
-    Private14
-    Private15
-    Private16
-    Private17
-    Private18
-    Private19
-    Private20
-    Private21
-    Private22
-    Private23
-    Private24
-    Private25
-    Private26
-    Private27
-    Private28
-    Private29
-    Private30
-    Shift_key
-    Alt_key
-    Control_key
-    Meta_key
-    Super_key
-    Hyper_key))
+;; valid-key-symbols is defined in uim-key.c
+
+(define key-symbol?
+  (lambda (sym)
+    (member sym valid-key-symbols)))
 
 (define intern-key-symbol
   (lambda (key-str)
@@ -425,16 +339,6 @@
       (set! enable-emacs-like-key-prefix? saved-enable-eprefix?)
       res)))
 
-;(define set-key-binding
-;  (lambda (im state flavor)
-;    #f
-;    ))
-
-;(define regist-key-binding 
-;  (lambda (bind-name bind-table)
-;    #f
-;))
-
 ;;
 (define-key left-key? "left")
 (define-key right-key? "right")
