@@ -1,5 +1,5 @@
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -15,10 +15,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -28,14 +28,67 @@
 ;;; SUCH DAMAGE.
 ;;;;
 
-;;
-;;
+(require-custom "generic-key-custom.scm")
 
-(define-key m17nlib-on-key? generic-on-key?)
-(define-key m17nlib-off-key? generic-off-key?)
+;;; user configs
 
 (define m17nlib-candidate-max 10)
 
+;; key defs
+(define-key m17nlib-on-key? 'generic-on-key?)
+(define-key m17nlib-off-key? 'generic-off-key?)
+
+;; widgets and actions
+
+;; widgets
+(define m17nlib-widgets '(widget_m17nlib_input_mode))
+
+;; default activity for each widgets
+(define default-widget_m17nlib_input_mode 'action_m17nlib_off)
+
+;; actions of widget_m17nlib_input_mode
+(define m17nlib-input-mode-actions
+  '(action_m17nlib_off
+    action_m17nlib_on))
+
+
+;;; implementations
+
+(register-action 'action_m17nlib_off
+		 (lambda (mc)
+		   (list
+		    'figure_m17nlib_off
+		    "-"
+		    (N_ "off")
+		    (N_ "Direct Input Mode")))
+		 (lambda (mc)
+		   (not (m17nlib-context-on mc)))
+		 (lambda (mc)
+		   (m17nlib-context-set-on! mc #f)))
+
+(register-action 'action_m17nlib_on
+		 (lambda (mc)
+		   (let* ((im (m17nlib-context-im mc))
+			  (name (symbol->string (im-name im))))
+		     (list
+		      'figure_m17nlib_on
+		      "O"
+		      (N_ "on")
+		      (string-append name (N_ " Mode")))))
+		 (lambda (mc)
+		   (m17nlib-context-on mc))
+		 (lambda (mc)
+		   (m17nlib-context-set-on! mc #t)))
+
+;; Update widget definitions based on action configurations. The
+;; procedure is needed for on-the-fly reconfiguration involving the
+;; custom API
+(define m17nlib-configure-widgets
+  (lambda ()
+    (register-widget 'widget_m17nlib_input_mode
+		     (activity-indicator-new m17nlib-input-mode-actions)
+		     (actions-new m17nlib-input-mode-actions))))
+
 (define m17nlib-context-rec-spec
   (append
    context-rec-spec
@@ -43,14 +96,16 @@
    '((mc-id             #f)
      (on                #f)
      (showing-candidate #f))))
+
 (define-record 'm17nlib-context m17nlib-context-rec-spec)
 (define m17nlib-context-new-internal m17nlib-context-new)
 
 (define m17nlib-context-new
   (lambda (id im name)
-    (let ((mc (m17nlib-context-new-internal id im)))
-      (m17nlib-context-set-mc-id! mc
-				  (m17nlib-lib-alloc-context name))
+    (let ((mc (m17nlib-context-new-internal id im))
+	  (mc-id (m17nlib-lib-alloc-context name)))
+      (m17nlib-context-set-widgets! mc m17nlib-widgets)
+      (m17nlib-context-set-mc-id! mc mc-id)
       mc)))
 
 (define m17nlib-update-preedit
@@ -85,28 +140,39 @@
 
 (define m17nlib-update-candidate
   (lambda (mc)
+    (m17nlib-lib-fill-new-candidates! (m17nlib-context-mc-id mc))
     (let* ((mid (m17nlib-context-mc-id mc))
 	   (max (m17nlib-lib-get-nr-candidates mid))
-	   (showing-candidate? (m17nlib-context-showing-candidate mc)))
-      (if showing-candidate?
-	  (if (m17nlib-lib-candidate-show? mid)
-	      (begin
-		(im-activate-candidate-selector
-		 mc
-		 max m17nlib-candidate-max)
-		(im-select-candidate mc
-				     (m17nlib-lib-get-candidate-index mid)))
-	      (begin
-		(m17nlib-context-set-showing-candidate! mc #f)
-		(im-deactivate-candidate-selector mc mid))) ;end showing candidate
-	  (if (m17nlib-lib-candidate-show? mid)
-	      (begin
-		(im-activate-candidate-selector
-		 mc
-		 max m17nlib-candidate-max)
-		(im-select-candidate mc (m17nlib-lib-get-candidate-index mid))
-		(m17nlib-context-set-showing-candidate! mc #t)))
-	  ))));nothing to do
+	   (showing-candidate? (m17nlib-context-showing-candidate mc))
+	   (candidates-changed? (m17nlib-lib-candidates-changed? mid)))
+
+      ;; FIXME: Rewrite this seriese of if with cond.
+      ;; close candidate window
+      (if (or
+	   (and showing-candidate? candidates-changed?)
+	   (and showing-candidate? (not (m17nlib-lib-candidate-show? mid))))
+	  (begin
+	    (im-deactivate-candidate-selector mc mid)
+	    (m17nlib-context-set-showing-candidate! mc #f)))
+
+      (if (and
+	   (or
+	    candidates-changed?
+	    (and 
+	     (not showing-candidate?)
+	     (m17nlib-lib-candidate-show? mid)))
+	   (not (= max 0)))
+	  (begin
+	    (im-activate-candidate-selector
+	     mc max m17nlib-candidate-max)
+	    (im-select-candidate
+	     mc (m17nlib-lib-get-candidate-index mid))
+	    (m17nlib-context-set-showing-candidate! mc #t)))
+
+      (if (and showing-candidate?
+	       (m17nlib-lib-candidate-show? mid))
+	  (im-select-candidate mc (m17nlib-lib-get-candidate-index mid))))))
+
 
 (define m17nlib-append-modifiers
   (lambda (key key-state key-str)
@@ -184,10 +250,6 @@
   (lambda (mc)
     #f))
 
-(define m17nlib-mode-handler
-  (lambda (mc mode)
-    #f))
-
 (define m17nlib-push-key
   (lambda (mc key key-state)
     (let* ((mid (m17nlib-context-mc-id mc)))
@@ -207,12 +269,16 @@
       (if (m17nlib-context-on mc)
 	  (if (m17nlib-push-key mc key key-state)
 	      #f ;; Discard key event
-	      (let ((commit-str (m17nlib-lib-get-commit-string mid)))
+	      (let* ((result (m17nlib-lib-get-result mid))
+		     (consumed? (car result))
+		     (commit-str (cdr result)))
 		(if (string=? commit-str "")
 		    (im-commit-raw mc)
 		    (begin
 		      (im-commit mc commit-str)
-		      (m17nlib-lib-commit mid)))))
+		      (m17nlib-lib-commit mid)
+		      (if (not consumed?)
+			  (im-commit-raw mc))))))
 	  (m17nlib-proc-direct-state mc key key-state))
       (m17nlib-update-preedit mc)
       (m17nlib-update-candidate mc))))
@@ -230,38 +296,58 @@
     (let* ((mid (m17nlib-context-mc-id mc))
 	   (cand (m17nlib-lib-get-nth-candidate
 		  mid idx)))
-      (list cand (digit->string (+ idx 1))))))
+      (list cand (digit->string (+ idx 1)) ""))))
 
 (define m17nlib-set-candidate-index-handler
   (lambda (mc idx)
     #f))
 
-(define m17nlib-prop-handler
-  (lambda (mc message)
-    #f))
+;; Developer specified IM rejection should be completely withdrawn
+;; after we got flexible install-time IM preference list. These
+;; redundant IMs should be "disabled by default, but can be enabled"
+;; under the feature. But we should invest our time for more valuable
+;; issues.  -- YamaKen 2005-01-25
+(define duplicated-im-list
+  '("m17n-ja-anthy"
+    "m17n-ja-tcode"
+    "m17n-zh-pinyin"
+    "m17n-zh-py"))
+
+;; At now, simply enable all IMs. Although they are redundant and
+;; unconfortable, they can be disabled by uim-pref.
+;;   -- YamaKen 2005-01-25
+(define duplicated-im?
+  (if #t
+      (lambda (name)
+	#f)
+      (lambda (name)
+	(member name duplicated-im-list))))
 
 (define m17nlib-register
   (lambda (i nr-im)
     (if (> nr-im i)
 	(begin
-	  (register-im
-	   (m17nlib-lib-nth-input-method-name i)
-	   (m17nlib-lib-nth-input-method-lang i)
-	   "UTF-8"
-	   (N_ "An input method provided by the m17n library")
-	   (m17nlib-lib-nth-input-method-name i)
-	   m17nlib-init-handler
-	   m17nlib-release-handler
-	   m17nlib-mode-handler
-	   m17nlib-press-key-handler
-	   m17nlib-release-key-handler
-	   m17nlib-reset-handler
-	   m17nlib-get-candidate-handler
-	   m17nlib-set-candidate-index-handler
-	   m17nlib-prop-handler
-	   )
-	  (m17nlib-register (+ i 1) nr-im))
+	  (if (not (duplicated-im? (m17nlib-lib-nth-input-method-name i)))
+	      (register-im
+	       (string->symbol (m17nlib-lib-nth-input-method-name i))
+	       (m17nlib-lib-nth-input-method-lang i)
+	       "UTF-8"
+	       (m17nlib-lib-nth-input-method-name i)
+	       (N_ "An input method provided by the m17n library")
+	       (m17nlib-lib-nth-input-method-name i)
+	       m17nlib-init-handler
+	       m17nlib-release-handler
+	       context-mode-handler
+	       m17nlib-press-key-handler
+	       m17nlib-release-key-handler
+	       m17nlib-reset-handler
+	       m17nlib-get-candidate-handler
+	       m17nlib-set-candidate-index-handler
+	       context-prop-activate-handler
+	       ))
+	      (m17nlib-register (+ i 1) nr-im))
 	())))
 
 (m17nlib-lib-init)
 (m17nlib-register 0 (m17nlib-lib-nr-input-methods))
+(m17nlib-configure-widgets)
