@@ -15,10 +15,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -36,45 +36,16 @@
 ;
 ; back match is mainly used for Hangul
 
+(define-record 'rk-context
+  '((rule             ())
+    (seq              ())
+    (immediate-commit #f)
+    (back-match       #f)))
+(define rk-context-new-internal rk-context-new)
 
-;(rule seq immediate-flag back)
 (define rk-context-new
   (lambda (rule immediate-commit back)
-    (let ((res (copy-list '(() () () ()))))
-      (rk-context-set-rule! res rule)
-      (rk-context-set-immediate-commit! res
-					immediate-commit)
-      (rk-context-set-back-match! res back)
-      res)))
-
-;; accessor
-(define rk-context-rule
-  (lambda (rkc)
-    (car (nthcdr 0 rkc))))
-(define rk-context-set-rule!
-  (lambda (rkc rule)
-    (set-car! (nthcdr 0 rkc) rule)))
-
-(define rk-context-seq
-  (lambda (rkc)
-    (car (nthcdr 1 rkc))))
-(define rk-context-set-seq!
-  (lambda (rkc seq)
-    (set-car! (nthcdr 1 rkc) seq)))
-
-(define rk-context-immediate-commit
-  (lambda (rkc)
-    (car (nthcdr 2 rkc))))
-(define rk-context-set-immediate-commit!
-  (lambda (rkc flag)
-    (set-car! (nthcdr 2 rkc) flag)))
-
-(define rk-context-back-match
-  (lambda (rkc)
-    (car (nthcdr 3 rkc))))
-(define rk-context-set-back-match!
-  (lambda (rkc flag)
-    (set-car! (nthcdr 3 rkc) flag)))
+    (rk-context-new-internal rule () immediate-commit back)))
 
 ;; back match
 (define rk-find-longest-back-match
@@ -83,7 +54,7 @@
 	(if (rk-lib-find-seq seq rule)
 	    seq
 	    (rk-find-longest-back-match rule (cdr seq)))
-	nil)))
+	#f)))
 ;; back match
 (define rk-find-longest-head
   (lambda (rseq rule)
@@ -92,7 +63,7 @@
 	  seq
 	  (if rseq
 	      (rk-find-longest-head (cdr rseq) rule)
-	      nil)))))
+	      #f)))))
 ;; back match
 (define rk-check-back-commit
   (lambda (rkc rule rseq)
@@ -107,7 +78,7 @@
 	   (tail-partial
 	    (if longest-tail
 		(rk-lib-find-partial-seq longest-tail rule)
-		nil))
+		#f))
 	   (c (rk-lib-find-seq longest-tail rule))
 	   (t (rk-lib-find-seq seq rule))
 	   (res ()))
@@ -184,6 +155,21 @@
        #t)
      #f)))
 
+
+; Merges two strings that have been converted, for example
+; ("" "" "") ("1" "1" "1") --> ("1" "1" "1").
+; SEQ1 and SEQ2 must be proper lists having the same length.
+; A disgusting hack for implementing ("n" "1") --> ("1").
+; Anyone with the time, skill and passion, please clean this up :-(
+(define rk-merge-seqs
+  (lambda (seq1 seq2)
+    (if (and (pair? seq1) (pair? seq2))
+	(cons (string-append (car seq1) (car seq2))
+	      (rk-merge-seqs (cdr seq1) (cdr seq2)))
+	; This should be () when we reach the end of the lists, or
+	; whatever passed as SEQ1 if SEQ2 is #f
+	seq1)))
+
 ;; front match
 (define rk-proc-tail
   (lambda (context seq)
@@ -191,12 +177,20 @@
 	   (old-seq
 	    (rk-lib-find-seq
 	     (reverse (rk-context-seq context)) rule))
-	   (res nil))
+	   (res #f))
       (if old-seq
 	  (begin
 	    (rk-flush context)
+	    ;; Comment out the code using rk-merge-seqs for the
+	    ;; moment because of rk-backspace problem -- ekato
+	    ;;(set! res
+	    ;;	  (rk-merge-seqs
+	    ;;	   (cadr old-seq)
+	    ;;	   (rk-push-key! context (car seq)))))
+	    ;;
 	    (rk-push-key! context (car seq))
 	    (set! res (cadr old-seq)))
+	    ;;
 	  (if (rk-context-seq context)
 	      (begin
 		(rk-flush context)
@@ -213,7 +207,7 @@
 	  (cadr seq))
 	(begin
 	  (rk-context-set-seq! context s)
-	  nil))))
+	  #f))))
 ;; API
 (define rk-expect
   (lambda (rkc)
@@ -247,14 +241,26 @@
        (if (rk-partial-seq? rkc s)
 	   (begin
 	     (rk-context-set-seq! rkc s)
-	     nil)
+	     #f)
 	   (if seq
 	       (rk-proc-end-seq rkc seq s)
 	       (rk-proc-tail rkc s))))
       res)))
 
+;; API
+;; returns the rule entry that exactly matches with current pending
+;; key sequence. rkc will not be altered.
+(define rk-peek-terminal-match
+  (lambda (rkc)
+    (let ((rule-entry (rk-current-seq rkc)))
+      (and rule-entry
+	   (cadr rule-entry)))))
 
+;; API
 ;;ĶŬrk.scm򤹤뤫ȤǤä߷פʤ
+;;
+;; The procedure name is confusable. I suggest rk-terminate-input!
+;;  -- YamaKen 2004-10-25
 (define rk-push-key-last!
   (lambda (rkc)
     (let*
@@ -264,13 +270,16 @@
       (rk-proc-end-seq rkc seq s)
       )))
 
-
+;; API
+;; returns string list or #f
 (define rk-push-key!
   (lambda (rkc key)
     (if (rk-context-back-match rkc)
+	;; mainly for Hangul
 	(rk-push-key-back-match rkc key)
+	;; for other languages
 	(rk-push-key-front-match rkc key))))
-;;
+;; API
 (define rk-pending
   (lambda (c)
     (string-list-concat
