@@ -1,6 +1,6 @@
 ;;; util.scm: Utility functions for uim.
 ;;;
-;;; Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 ;;;
 ;;; All rights reserved.
 ;;;
@@ -16,10 +16,10 @@
 ;;;    may be used to endorse or promote products derived from this software
 ;;;    without specific prior written permission.
 ;;;
-;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
 ;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 ;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
 ;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 ;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 ;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -29,54 +29,107 @@
 ;;; SUCH DAMAGE.
 ;;;;
 
-;;
-(define control-char?
+;; Current uim implementation treats char as integer
+
+(define char-control?
   (lambda (c)
     (and (integer? c)
-	 (or (< c 32)
+	 (or (<= c 31)
 	     (= c 127)))))
-;;
-(define alphabet-char?
+
+(define char-upper-case?
   (lambda (c)
     (and (integer? c)
-	 (or
-	  (and (>= c 65) (<= c 90))
-	  (and (>= c 97) (<= c 122))))))
-;;
-(define usual-char?
+	 (>= c 65)
+	 (<= c 90))))
+
+(define char-lower-case?
   (lambda (c)
     (and (integer? c)
-	 (and (> c 32) (< c 127)))))
-;;
-(define numeral-char?
+	 (>= c 97)
+	 (<= c 122))))
+
+(define char-alphabetic?
   (lambda (c)
+    (or (char-upper-case? c)
+	(char-lower-case? c))))
+
+(define char-numeric?
+  (lambda (c)
+    (and (integer? c)
+	 (>= c 48)
+	 (<= c 57))))
+
+(define char-printable?
+  (lambda (c)
     (and (integer? c)
-	  (and (>= c 48)
-	       (<= c 57)))))
+	 (<= c 127)
+	 (not (char-control? c)))))
+
+(define char-graphic?
+  (lambda (c)
+    (and (char-printable? c)
+	 (not (= c 32)))))
 
-;;
 (define numeral-char->number
   (lambda (c)
-    (if (numeral-char? c)
+    (if (char-numeric? c)
 	(- c 48)
 	c)))
-;;
-(define to-lower-char
+
+(define char-downcase
   (lambda (c)
-    (if (and (alphabet-char? c) (< c 91))
+    (if (char-upper-case? c)
 	(+ c 32)
 	c)))
-;;
+
+(define char-upcase
+  (lambda (c)
+    (if (char-lower-case? c)
+	(- c 32)
+	c)))
+
+;; backward compatibility
+(define control-char? char-control?)
+(define alphabet-char? char-alphabetic?)
+(define numeral-char? char-numeric?)
+(define usual-char? char-graphic?)
+(define to-lower-char char-downcase)
+
+;; TODO: write test
+(define string-escape
+  (lambda (s)
+    (let ((buf (string-append "\"\"" s s)))
+      (print-to-string s buf))))
+
+;; TODO: write test
+(define string->char
+  (lambda (str)
+    (and (= (string-length str)
+	    1)
+	 (string->charcode str))))
+
+;; TODO: write test
+(define string->printable-char
+  (lambda (str)
+    (let ((c (string->char str)))
+      (and (char-printable? c)
+	   c))))
+
+(define string->letter
+  (lambda (str)
+    (let ((c (string->printable-char str)))
+      (and (char-alphabetic? c)
+	   c))))
+
 (define string-list-concat
   (lambda (lst)
     (apply string-append (reverse lst))))
 
-;;
 (define string-find
   (lambda (lst str)
     (member str lst)))
 
-;;
 (define truncate-list
   (lambda (lst n)
     (if (or (< (length lst)
@@ -121,27 +174,54 @@
 ;; local procedure. don't use in outside of util.scm
 (define iterate-lists
   (lambda (mapper state lists)
-    (let ((runs-out? (apply proc-or (map null? lists))))
+    (let ((runs-out? (apply proc-or (mapcar null? lists))))
       (if runs-out?
 	  (cdr (mapper state ()))
-	  (let* ((elms (map car lists))
-		 (rests (map cdr lists))
+	  (let* ((elms (mapcar car lists))
+		 (rests (mapcar cdr lists))
 		 (pair (mapper state elms))
 		 (terminate? (car pair))
 		 (new-state (cdr pair)))
 	    (if terminate?
 		new-state
-		(apply iterate-lists (list mapper new-state rests))))))))
+		(iterate-lists mapper new-state rests)))))))
 
-;; not yet tested -- YamaKen 2004-10-30
-(define alist-replace!
+(define alist-replace
   (lambda (kons alist)
     (let* ((id (car kons))
 	   (preexisting (assoc id alist)))
       (if preexisting
-	  (set-cdr! preexisting (cdr kons))
-	  (set! alist
-		(cons kons alist))))))
+	  (begin
+	    (set-cdr! preexisting (cdr kons))
+	    alist)
+	  (cons kons alist)))))
+
+(define join
+  (lambda (sep list)
+    (let ((len (length list)))
+      (if (= len 0)
+	  ()
+	  (cdr (apply append (zip (make-list len sep)
+				  list)))))))
+
+(define string-join
+  (lambda (sep str-list)
+    (apply string-append (join sep str-list))))
+
+(define string-append-map
+  (lambda args
+    (apply string-append (apply map args))))
+
+;; only accepts single-arg functions
+;; (define caddr (compose car cdr cdr))
+(define compose
+  (lambda funcs
+    (fold (lambda (f g)
+	    (lambda (arg)
+	      (f (g arg))))
+	  (lambda (arg)
+	    arg)
+	  (reverse funcs))))
 
 ;;
 ;; R5RS procedures (don't expect 100% compatibility)
@@ -165,14 +245,26 @@
 	(and (pair? x)
 	     (list? (cdr x))))))
 
+(define number->string integer->string)
+(define string->number string->integer)
 (define string->symbol intern)
 
-;; accepts up to 2 lists
-(define map mapcar)
+(define map
+  (lambda args
+    (let ((f (car args))
+	  (lists (cdr args)))
+      (if (<= (length lists) 3)  ;; uim's siod accepts up to 3 lists
+	  (apply mapcar args)    ;; faster native processing
+	  (iterate-lists (lambda (state elms)
+			   (if (null? elms)
+			       (cons #t (reverse state))
+			       (let ((mapped (apply f elms)))
+				 (cons #f (cons mapped state)))))
+			 () lists)))))
 
-;; accepts up to 2 lists
-;; process order is guaranteed by siod's mapcar implementation
-(define for-each mapcar)
+(define for-each map)
+
+(define quotient /)	;; / in siod is quotient actually
 
 ;;(define list-tail
 ;;  (lambda (lst n)
@@ -197,7 +289,6 @@
 ;;(define drop-right)
 ;;(define split-at)
 ;;(define last)
-;;(define unfold)
 
 (define list-tabulate
   (lambda (n init-proc)
@@ -219,12 +310,16 @@
 		     fill))))
 
 ;; This procedure does not conform to the SRFI-1 specification. The
-;; optional arguments 'start' and 'step' is not supported.
+;; optional argument 'step' is not supported.
 (define iota
-  (lambda (count)
-    (list-tabulate count
-		   (lambda (i)
-		     i))))
+  (lambda args
+    (let ((count (car args))
+	  (start (if (not (null? (cdr args)))
+		     (cadr args)
+		     0)))
+      (list-tabulate (- count start)
+		     (lambda (i)
+		       (+ start i))))))
     
 (define zip
   (lambda lists
@@ -235,6 +330,10 @@
 		   (rests (map cdr lists)))
 	      (cons elms (apply zip rests)))))))
 
+(define append-map
+  (lambda args
+    (apply append (apply map args))))
+
 (define append-reverse
   (lambda (rev-head tail)
     (fold cons tail rev-head)))
@@ -258,7 +357,7 @@
 			   '(#t . #f)
 			   (let ((res (apply pred elms)))
 			     (cons res res))))
-	       #f lists))))	       
+		     #f lists))))       
 
 (define every
   (lambda args
@@ -269,7 +368,7 @@
 			   '(#t . #t)
 			   (let ((res (apply pred elms)))
 			     (cons (not res) res))))
-	       #f lists))))	       
+		     #f lists))))	       
 
 (define fold
   (lambda args
@@ -282,6 +381,21 @@
 			   (cons #f (apply kons (append elms (list state))))))
 		     knil lists))))
 
+(define unfold
+  (lambda args
+    (let ((term? (nth 0 args))
+	  (kar (nth 1 args))
+	  (kdr (nth 2 args))
+	  (seed (nth 3 args))
+	  (tail-gen (if (= (length args)
+			   5)
+			(nth 4 args)
+			(lambda (x) ()))))
+      (if (term? seed)
+	  (tail-gen seed)
+	  (cons (kar seed)
+		(unfold term? kar kdr (kdr seed) tail-gen))))))
+
 (define filter
   (lambda args
     (let ((pred (car args))
@@ -314,12 +428,24 @@
 	      (not (pred elm)))
 	    lst)))
 
+;; TODO: write test
+(define delete
+  (lambda args
+    (let ((x (car args))
+	  (lst (cadr args))
+	  (val=? (if (null? (cddr args))
+		     equal?
+		     (car (cddr args)))))
+      (filter (lambda (elm)
+		(not (val=? elm x)))
+	      lst))))
+
 (define alist-delete
   (lambda args
     (let ((key (car args))
 	  (alist (cadr args))
 	  (key=? (if (null? (cddr args))
-		     =
+		     equal?
 		     (car (cddr args)))))
     (remove (lambda (elm)
 	      (key=? (car elm)
@@ -330,6 +456,30 @@
 ;; uim-specific utilities
 ;;
 
+;; TODO: write test
+(define make-scm-pathname
+  (lambda (file)
+    (or (and (= (string->charcode file)
+		(string->charcode "/"))
+	     file)
+	(string-append (load-path) "/" file))))
+
+;; TODO: write test
+;; returns succeeded or not
+(define try-load
+  (lambda (file)
+    (and (file-readable? (make-scm-pathname file))
+	 (not (*catch 'errobj (begin (load file)
+				     #f))))))
+
+;; TODO: write test
+;; returns succeeded or not
+(define try-require
+  (lambda (file)
+    (and (file-readable? (make-scm-pathname file))
+	 (eq? (symbolconc '* (string->symbol file) '-loaded*)
+	      (*catch 'errobj (require file))))))
+
 ;; for eval
 (define toplevel-env ())
 
@@ -412,6 +562,16 @@
       (if (not (null? (cdr style)))
 	  (update-style style-spec (cdr style))))))
 
+;; for backward compatibility
+(define uim-symbol-value-str
+  (lambda (sym)
+    (let ((val (if (symbol-bound? sym)
+		   (symbol-value sym)
+		   "")))
+      (if (symbol? val)
+	  (symbol->string val)
+	  val))))
+
 ;;
 ;; Preedit color related configurations and functions.
 ;;
@@ -464,12 +624,6 @@
 (define multi-segment-type-katakana 1)
 (define multi-segment-type-hankana 2)
 
-(define multi-segment-mode-direct 0)
-(define multi-segment-mode-hiragana 1)
-(define multi-segment-mode-katakana 2)
-(define multi-segment-mode-wide-latin 3)
-(define multi-segment-mode-hankana 4)
-
 (define multi-segment-make-index-list
   (lambda (n old-lst)
     (let ((old-lst-len (length old-lst)))
@@ -518,4 +672,17 @@
       multi-segment-type-hiragana))))
 
 
-(define candidate-window-position "caret")
+;; FIXME: write test.
+(define ucs-to-utf8-string
+  (lambda (ucs)
+    (let ((utf-8
+	   (if (< ucs 128)
+	       (list ucs)		; ASCII
+	       (let enc ((to-be-split ucs)
+			 (threshold 64))
+		 (if (< to-be-split threshold)
+		     (list (bit-or to-be-split
+				   (bit-xor 255 (- (* 2 threshold) 1))))
+		     (cons (bit-or 128 (bit-and 63 to-be-split))
+			   (enc (/ to-be-split 64) (/ threshold 2))))))))
+      (string-append-map charcode->string (reverse utf-8)))))
