@@ -1,6 +1,35 @@
 #!/usr/bin/env gosh
 
-;; This file is tested with revision 1404
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
+;;;
+;;; All rights reserved.
+;;;
+;;; Redistribution and use in source and binary forms, with or without
+;;; modification, are permitted provided that the following conditions
+;;; are met:
+;;; 1. Redistributions of source code must retain the above copyright
+;;;    notice, this list of conditions and the following disclaimer.
+;;; 2. Redistributions in binary form must reproduce the above copyright
+;;;    notice, this list of conditions and the following disclaimer in the
+;;;    documentation and/or other materials provided with the distribution.
+;;; 3. Neither the name of authors nor the names of its contributors
+;;;    may be used to endorse or promote products derived from this software
+;;;    without specific prior written permission.
+;;;
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
+;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
+;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+;;; SUCH DAMAGE.
+;;;;
+
+;; This file is tested with revision 327 of new repository
 
 (use test.unit)
 
@@ -19,51 +48,160 @@
 (define-uim-test-case "testcase im im-management"
   (setup
    (lambda ()
-     (uim '(define prev-im #f))
-     (uim '(define prev-nr-ims (length im-list)))
-     (uim '(define test-im-init-args #f))
      (uim '(begin
+	     (require-module "anthy")
+	     (require-module "canna")
+	     (require-module "skk")
+	     (require-module "tcode")
+	     (set! enabled-im-list (append enabled-im-list
+					   '(test-im test-im2)))
+	     (for-each require-module installed-im-module-list)
+	     (define prev-im #f)
+	     (define prev-nr-ims (length im-list))
+	     (define test-im-init-args #f)
+	     (define test-im-alt-init-args #f)
+	     ;; temporary workaround to cheat on revised register-im
+	     ;; TODO: rewrite tests in accordance with enabled-im-list
+	     (define custom-full-featured? #t)
 	     (set! test-im-init-args (list 'test-im
 					   "ja"
 					   "UTF-8"
+					   "a label"
 					   "a short description"
 					   #f
-					   default-init-handler
-					   default-release-handler
-					   default-mode-handler
-					   default-key-press-handler
-					   default-key-release-handler
-					   default-reset-handler
-					   default-get-candidate-handler
-					   default-set-candidate-index-handler
-					   default-prop-handler))
+					   direct-init-handler
+					   direct-release-handler
+					   context-mode-handler
+					   direct-key-press-handler
+					   direct-key-release-handler
+					   direct-reset-handler
+					   direct-get-candidate-handler
+					   direct-set-candidate-index-handler
+					   context-prop-activate-handler))
+	     (set! test-im-alt-init-args (list 'test-im
+					       "en"
+					       "en_US.UTF-8"
+					       "an alternative label"
+					       "an alternative short desc"
+					       'alt-arg
+					       'alt-init-handler
+					       'alt-release-handler
+					       'alt-mode-handler
+					       'alt-key-press-handler
+					       'alt-key-release-handler
+					       'alt-reset-handler
+					       'alt-get-candidate-handler
+					       'alt-set-candidate-index-handler
+					       'alt-prop-activate-handler))
 	     #t))))
 
-  ("test register-im"
-   (uim '(begin
-	   (apply register-im test-im-init-args)
-	   #t))
+  ("test normalize-im-list"
+   (uim '(set! im-list (remove (lambda (im)
+				 (not (eq? (im-name im)
+					   'direct)))
+			       im-list)))
+   (assert-equal '(direct)
+		 (uim '(map im-name im-list)))
+   ;; direct IM always remains at head
+   (assert-true  (uim-bool '(apply register-im test-im-init-args)))
+   (assert-equal '(direct test-im)
+		 (uim '(map im-name im-list)))
+   ;; other IMs are cons'ed at right of direct
+   (assert-true  (uim-bool '(apply register-im (cons 'test-im2
+						     (cdr test-im-init-args)))))
+   (assert-equal '(direct test-im2 test-im)
+		 (uim '(map im-name im-list)))
+   ;; direct IM can be registered to null list
+   (uim '(set! im-list ()))
+   ;; second time im-register-im for 'direct returns #f
+   (assert-false (uim-bool '(apply register-im (cons 'direct
+						     (cdr test-im-init-args)))))
+   (assert-equal '(direct)
+		 (uim '(map im-name im-list)))
+   ;; ordinary IM can be registered to null list
+   (uim '(set! im-list ()))
+   ;; second time im-register-im for 'test-im returns #f
+   (assert-false (uim-bool '(apply register-im test-im-init-args)))
+   (assert-equal '(test-im)
+		 (uim '(map im-name im-list))))
 
+  ("test register-im"
+   (assert-true  (uim-bool '(apply register-im test-im-init-args)))
    (assert-equal (+ (uim 'prev-nr-ims) 1)
 		 (uim '(length im-list)))
    (assert-equal 'test-im
 		 (uim '(im-name (retrieve-im 'test-im #f))))
-   (assert-equal 14
+   (assert-false (uim-bool '(im-module-name (retrieve-im 'test-im #f))))
+   (assert-equal 16
 		 (uim '(length (retrieve-im 'test-im #f))))
+   (uim '(im-set-module-name! (retrieve-im 'test-im #f) "foo"))
+   (assert-equal "foo"
+		 (uim '(im-module-name (retrieve-im 'test-im #f))))
 
-   ;; duplicate register will be rejected
-   (assert-false (uim-bool '(apply register-im test-im-init-args)))
+   ;; duplicate register overwrites preexisting one
+   (assert-false (uim-bool '(apply register-im test-im-alt-init-args)))
    (assert-equal (+ (uim 'prev-nr-ims) 1)
-		 (uim '(length im-list))))
+		 (uim '(length im-list)))
+   (assert-equal '(test-im
+		   "en"
+		   "en_US.UTF-8"
+		   "an alternative label"
+		   "an alternative short desc"
+		   alt-arg
+		   alt-init-handler
+		   alt-release-handler
+		   alt-mode-handler
+		   alt-key-press-handler
+		   alt-key-release-handler
+		   alt-reset-handler
+		   alt-get-candidate-handler
+		   alt-set-candidate-index-handler
+		   alt-prop-activate-handler
+		   ())  ;; replace with #f for R5RS compliant interpreter
+		 (uim '(retrieve-im 'test-im #f)))
+   ;; subsequent registration that has different im-name will be
+   ;; registered as another IM
+   (assert-true  (uim-bool '(apply register-im
+				   (cons 'test-im2
+					 (cdr test-im-alt-init-args)))))
+   (assert-equal (+ (uim 'prev-nr-ims) 2)
+		 (uim '(length im-list)))
+   (assert-equal '(test-im2
+		   "en"
+		   "en_US.UTF-8"
+		   "an alternative label"
+		   "an alternative short desc"
+		   alt-arg
+		   alt-init-handler
+		   alt-release-handler
+		   alt-mode-handler
+		   alt-key-press-handler
+		   alt-key-release-handler
+		   alt-reset-handler
+		   alt-get-candidate-handler
+		   alt-set-candidate-index-handler
+		   alt-prop-activate-handler
+		   ())  ;; replace with #f for R5RS compliant interpreter
+		 (uim '(retrieve-im 'test-im2 #f))))
+
+  ("test register-im (module-name)"
+   (assert-true  (uim-bool '(apply register-im test-im-init-args)))
+   (assert-false (uim-bool '(im-module-name (retrieve-im 'test-im #f))))
+
+   (uim '(set! currently-loading-module-name "foo"))
+   (assert-true  (uim-bool '(apply register-im (cons 'test-im2
+						     (cdr test-im-init-args)))))
+   (assert-equal "foo"
+		 (uim '(im-module-name (retrieve-im 'test-im2 #f)))))
 
   ("test retrieve-im"
    (assert-false (uim-bool '(retrieve-im 'nonexistent)))
-   (assert-equal 'default
-		 (uim '(im-name (retrieve-im 'default))))
+   (assert-equal 'direct
+		 (uim '(im-name (retrieve-im 'direct))))
    (assert-equal 'anthy
 		 (uim '(im-name (retrieve-im 'anthy))))
-   (assert-equal 'ipa
-		 (uim '(im-name (retrieve-im 'ipa))))
+   (assert-equal 'ipa-x-sampa
+		 (uim '(im-name (retrieve-im 'ipa-x-sampa))))
    (assert-false (uim-bool '(retrieve-im 'test-im)))
    (uim '(begin
 	   (apply register-im test-im-init-args)
@@ -89,7 +227,7 @@
    (assert-equal 'anthy
 		 (uim '(im-name (default-im-for-debug))))
    ;; default-im-name does not affect default-im-for-debug
-   (uim '(set! default-im-name 'ipa))
+   (uim '(set! default-im-name 'ipa-x-sampa))
    (assert-equal 'anthy
 		 (uim '(im-name (default-im-for-debug)))))
 
@@ -107,39 +245,39 @@
    (assert-true  (uim-bool '(memq 'hangul3 (map car im-list))))
    (assert-true  (uim-bool '(memq 'romaja (map car im-list))))
    (assert-true  (uim-bool '(memq 'viqr (map car im-list))))
-   (assert-true  (uim-bool '(memq 'ipa (map car im-list))))
-   (assert-true  (uim-bool '(memq 'default (map car im-list))))
+   (assert-true  (uim-bool '(memq 'ipa-x-sampa (map car im-list))))
+   (assert-true  (uim-bool '(memq 'direct (map car im-list))))
    ;; unsupported or direct-input languages
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "C"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "POSIX"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "en"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "en_US"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "en_US.US-ASCII"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "en_US.UTF-8"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "de"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "fr"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "de"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "pt"))))
    ;; non-existent language code
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "xx"))))
    ;; default-im-name does not affect find-im-for-locale
-   (uim '(set! default-im-name 'ipa))
-   (assert-equal 'default
+   (uim '(set! default-im-name 'ipa-x-sampa))
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "pt"))))
    ;; default-im-for-debug does not affect find-im-for-locale
-   (uim '(setenv "UIM_IM_ENGINE" "ipa" #t))
-   (assert-equal 'default
+   (uim '(setenv "UIM_IM_ENGINE" "ipa-x-sampa" #t))
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "pt"))))
    (uim '(unsetenv "UIM_IM_ENGINE"))
    ;; Chinese
@@ -151,7 +289,7 @@
 		 (uim '(im-name (find-im-for-locale "zh.Big5"))))
    (assert-equal 'm17n-zh-py
 		 (uim '(im-name (find-im-for-locale "zh.UTF-8"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "zh_US"))))
    (assert-equal 'py
 		 (uim '(im-name (find-im-for-locale "zh_CN"))))
@@ -205,15 +343,15 @@
    ;; native locale
    (uim '(unsetenv "LC_ALL"))
    (uim '(unsetenv "LANG"))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale ""))))
    (uim '(setenv "LC_ALL" "C" #t))
    (uim '(unsetenv "LANG"))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale ""))))
    (uim '(unsetenv "LC_ALL"))
    (uim '(setenv "LANG" "C" #t))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale ""))))
    (uim '(setenv "LC_ALL" "ja_JP.EUC-JP" #t))
    (uim '(unsetenv "LANG"))
@@ -222,18 +360,18 @@
    ;; nonexistent native locale
    (uim '(setenv "LC_ALL" "xx" #t))
    (uim '(unsetenv "LANG"))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im-for-locale "")))))
 
   ("test find-default-im"
    ;; find by find-im-for-locale
    (uim '(unsetenv "UIM_IM_ENGINE"))
    (uim '(set! default-im-name #f))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-default-im "C"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-default-im "en"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-default-im "pt"))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-default-im "ja_JP.EUC-JP"))))
@@ -242,10 +380,10 @@
    (uim '(set! default-im-name #f))
    (uim '(unsetenv "LC_ALL"))
    (uim '(unsetenv "LANG"))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-default-im ""))))
    (uim '(setenv "LC_ALL" "C" #t))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-default-im ""))))
    (uim '(setenv "LC_ALL" "ja_JP.EUC-JP" #t))
    (assert-equal 'anthy
@@ -253,12 +391,12 @@
    (uim '(unsetenv "LC_ALL"))
    ;; default-im-name precedes the locale specified by arg
    (uim '(unsetenv "UIM_IM_ENGINE"))
-   (uim '(set! default-im-name 'ipa))
-   (assert-equal 'ipa
+   (uim '(set! default-im-name 'ipa-x-sampa))
+   (assert-equal 'ipa-x-sampa
 		 (uim '(im-name (find-default-im "en"))))
-   (assert-equal 'ipa
+   (assert-equal 'ipa-x-sampa
 		 (uim '(im-name (find-default-im "pt"))))
-   (assert-equal 'ipa
+   (assert-equal 'ipa-x-sampa
 		 (uim '(im-name (find-default-im "ja_JP.EUC-JP"))))
    ;; default-im-for-debug precedes the locale specified by arg
    (uim '(setenv "UIM_IM_ENGINE" "py" #t))
@@ -271,7 +409,7 @@
 		 (uim '(im-name (find-default-im "ja_JP.EUC-JP"))))
    ;; default-im-for-debug precedes default-im-name
    (uim '(setenv "UIM_IM_ENGINE" "py" #t))
-   (uim '(set! default-im-name 'ipa))
+   (uim '(set! default-im-name 'ipa-x-sampa))
    (assert-equal 'py
 		 (uim '(im-name (find-default-im "en"))))
    (assert-equal 'py
@@ -283,9 +421,9 @@
    ;; explicit IM specification
    (uim '(unsetenv "UIM_IM_ENGINE"))
    (uim '(set! default-im-name #f))
-   (assert-equal 'default
-		 (uim '(im-name (find-im 'default #f))))
-   (assert-equal 'default
+   (assert-equal 'direct
+		 (uim '(im-name (find-im 'direct #f))))
+   (assert-equal 'direct
 		 (uim '(im-name (find-im 'nonexistent #f))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-im 'anthy #f))))
@@ -302,19 +440,19 @@
    (assert-equal 'pyunihan
 		 (uim '(im-name (find-im 'pyunihan #f))))
    ;; implicit selection by locale information
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f #f))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f ""))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f "C"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f "en"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f "pt"))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-im #f "ja_JP.EUC-JP"))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (find-im #f "xx"))))
    ;; im-name always precedes locale
    (assert-equal 'py
@@ -342,8 +480,8 @@
 		 (uim '(im-name (find-im #f #f))))
    (assert-equal 'pyunihan
 		 (uim '(im-name (find-im #f "ja_JP.EUC-JP"))))
-   (assert-equal 'default
-		 (uim '(im-name (find-im 'default #f))))
+   (assert-equal 'direct
+		 (uim '(im-name (find-im 'direct #f))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-im 'anthy #f))))
    (assert-equal 'skk
@@ -386,8 +524,8 @@
 		 (uim '(im-name (find-im #f #f))))
    (assert-equal 'pyunihan
 		 (uim '(im-name (find-im #f "ja_JP.EUC-JP"))))
-   (assert-equal 'default
-		 (uim '(im-name (find-im 'default #f))))
+   (assert-equal 'direct
+		 (uim '(im-name (find-im 'direct #f))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-im 'anthy #f))))
    (assert-equal 'skk
@@ -430,8 +568,8 @@
 		 (uim '(im-name (find-im #f #f))))
    (assert-equal 'pyunihan
 		 (uim '(im-name (find-im #f "ja_JP.EUC-JP"))))
-   (assert-equal 'default
-		 (uim '(im-name (find-im 'default #f))))
+   (assert-equal 'direct
+		 (uim '(im-name (find-im 'direct #f))))
    (assert-equal 'anthy
 		 (uim '(im-name (find-im 'anthy #f))))
    (assert-equal 'skk
@@ -471,6 +609,7 @@
 (define-uim-test-case "testcase im im-switching"
   (setup
    (lambda ()
+     (uim '(for-each require-module installed-im-module-list))
      (uim '(define test-im-anthy #f))
      (uim '(define test-im-skk #f))
      (uim '(define test-im-tcode #f))
@@ -526,11 +665,12 @@
 (define-uim-test-case "testcase im context management"
   (setup
    (lambda ()
+     (uim '(for-each require-module installed-im-module-list))
      ;; define as hand-made data to avoid that implementation of
      ;; register-context affect other tests
      (uim '(begin
 	     (set! context-list (list (im-new 1 (retrieve-im 'tcode))
-				      (im-new 2 (retrieve-im 'default))
+				      (im-new 2 (retrieve-im 'direct))
 				      (im-new 3 (retrieve-im 'skk))
 				      (im-new 4 (retrieve-im 'anthy))))
 	     #t))))
@@ -549,7 +689,7 @@
   ("test context-im"
    (assert-equal 'tcode
 		 (uim '(im-name (context-im (nth 0 context-list)))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (nth 1 context-list)))))
    (assert-equal 'skk
 		 (uim '(im-name (context-im (nth 2 context-list)))))
@@ -559,7 +699,7 @@
   ("test find-context"
    (assert-equal 'tcode
 		 (uim '(im-name (context-im (find-context 1)))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (find-context 2)))))
    (assert-equal 'skk
 		 (uim '(im-name (context-im (find-context 3)))))
@@ -575,7 +715,7 @@
 		 (uim '(length context-list)))
    (assert-equal 'tcode
 		 (uim '(im-name (context-im (find-context 1)))))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (find-context 2)))))
    (assert-false (uim-bool '(find-context 3)))
    (assert-equal 'anthy
@@ -585,7 +725,7 @@
    (assert-equal 2
 		 (uim '(length context-list)))
    (assert-false (uim-bool '(find-context 1)))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (find-context 2)))))
    (assert-false (uim-bool '(find-context 3)))
    (assert-equal 'anthy
@@ -596,7 +736,7 @@
    (assert-equal 2
 		 (uim '(length context-list)))
    (assert-false (uim-bool '(find-context 1)))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (find-context 2)))))
    (assert-false (uim-bool '(find-context 3)))
    (assert-equal 'anthy
@@ -606,7 +746,7 @@
    (assert-equal 1
 		 (uim '(length context-list)))
    (assert-false (uim-bool '(find-context 1)))
-   (assert-equal 'default
+   (assert-equal 'direct
 		 (uim '(im-name (context-im (find-context 2)))))
    (assert-false (uim-bool '(find-context 3)))
    (assert-false (uim-bool '(find-context 4)))
@@ -707,6 +847,25 @@
    )
   )
 
-;; TODO
-(define-uim-test-case "test im default im"
-  )
+(define-uim-test-case "testcase im im-custom"
+  (setup
+   (lambda ()
+     (uim '(begin
+	     (require-module "anthy")
+	     (require-module "canna")
+	     (require-module "skk")
+	     (require-module "tcode")))))
+
+  ("test custom-im-list-as-choice-rec"
+   (assert-equal '((canna "Canna" "A multi-segment kana-kanji conversion engine")
+		   (skk "SKK" "uim version of SKK input method")
+		   (anthy "Anthy" "A multi-segment kana-kanji conversion engine"))
+		 (uim '(custom-im-list-as-choice-rec
+			(map retrieve-im '(canna skk anthy)))))
+   (assert-equal '((tcode
+		    "T-Code"
+		    "A kanji direct input method"))
+		 (uim '(custom-im-list-as-choice-rec
+			(map retrieve-im '(tcode)))))
+   (assert-equal ()
+		 (uim '(custom-im-list-as-choice-rec ())))))
