@@ -1,5 +1,36 @@
 #!/usr/bin/env gosh
 
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
+;;;
+;;; All rights reserved.
+;;;
+;;; Redistribution and use in source and binary forms, with or without
+;;; modification, are permitted provided that the following conditions
+;;; are met:
+;;; 1. Redistributions of source code must retain the above copyright
+;;;    notice, this list of conditions and the following disclaimer.
+;;; 2. Redistributions in binary form must reproduce the above copyright
+;;;    notice, this list of conditions and the following disclaimer in the
+;;;    documentation and/or other materials provided with the distribution.
+;;; 3. Neither the name of authors nor the names of its contributors
+;;;    may be used to endorse or promote products derived from this software
+;;;    without specific prior written permission.
+;;;
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
+;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
+;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+;;; SUCH DAMAGE.
+;;;;
+
+;; This file is tested with revision 362 of new repository
+
 (use test.unit)
 
 (require "test/uim-test-utils")
@@ -163,6 +194,7 @@
    (assert-equal 126 (uim '(string->charcode "~"))))
 
   ("test digit->string"
+   (assert-equal "-2147483648" (uim '(digit->string -2147483648)))
    (assert-equal "-10"  (uim '(digit->string -10)))
    (assert-equal "-2"   (uim '(digit->string -2)))
    (assert-equal "-1"   (uim '(digit->string -1)))
@@ -187,7 +219,8 @@
    (assert-equal "18"   (uim '(digit->string 18)))
    (assert-equal "19"   (uim '(digit->string 19)))
    (assert-equal "100"  (uim '(digit->string 100)))
-   (assert-equal "1000" (uim '(digit->string 1000))))
+   (assert-equal "1000" (uim '(digit->string 1000)))
+   (assert-equal "2147483647" (uim '(digit->string 2147483647))))
 
   ;; compare string sequence
   ("test str-seq-equal?"
