@@ -1,92 +1,217 @@
 #!/usr/bin/env gosh
 
-;; This file is tested with revision 1231
+;;; Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
+;;;
+;;; All rights reserved.
+;;;
+;;; Redistribution and use in source and binary forms, with or without
+;;; modification, are permitted provided that the following conditions
+;;; are met:
+;;; 1. Redistributions of source code must retain the above copyright
+;;;    notice, this list of conditions and the following disclaimer.
+;;; 2. Redistributions in binary form must reproduce the above copyright
+;;;    notice, this list of conditions and the following disclaimer in the
+;;;    documentation and/or other materials provided with the distribution.
+;;; 3. Neither the name of authors nor the names of its contributors
+;;;    may be used to endorse or promote products derived from this software
+;;;    without specific prior written permission.
+;;;
+;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
+;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
+;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+;;; SUCH DAMAGE.
+;;;;
+
+;; This file is tested with revision 268 (new repository)
 
 (use test.unit)
 
 (require "test/uim-test-utils")
 
 (define-uim-test-case "testcase util character predicates"
+  ("test char-control?"
+   (assert-false (uim-bool '(char-control? 'symbol)))
+   (assert-false (uim-bool '(char-control? "string")))
+   (assert-false (uim-bool '(char-control? '(0 1 2))))
+   (assert-false (uim-bool '(char-control? car)))
+   (assert-true  (uim-bool '(char-control? 0)))    ; NUL
+   (assert-true  (uim-bool '(char-control? 1)))    ; SOH
+   (assert-true  (uim-bool '(char-control? 31)))   ; US
+   (assert-false (uim-bool '(char-control? 32)))   ; SPACE
+   (assert-false (uim-bool '(char-control? 33)))   ; !
+   (assert-false (uim-bool '(char-control? 47)))   ; /
+   (assert-false (uim-bool '(char-control? 48)))   ; 0
+   (assert-false (uim-bool '(char-control? 57)))   ; 9
+   (assert-false (uim-bool '(char-control? 58)))   ; :
+   (assert-false (uim-bool '(char-control? 64)))   ; @
+   (assert-false (uim-bool '(char-control? 65)))   ; A
+   (assert-false (uim-bool '(char-control? 90)))   ; Z
+   (assert-false (uim-bool '(char-control? 91)))   ; [
+   (assert-false (uim-bool '(char-control? 96)))   ; `
+   (assert-false (uim-bool '(char-control? 97)))   ; a
+   (assert-false (uim-bool '(char-control? 122)))  ; z
+   (assert-false (uim-bool '(char-control? 123)))  ; {
+   (assert-false (uim-bool '(char-control? 126)))  ; ~
+   (assert-true  (uim-bool '(char-control? 127)))) ; DEL
+  ("test char-upper-case?"
+   (assert-false (uim-bool '(char-upper-case? 'symbol)))
+   (assert-false (uim-bool '(char-upper-case? "string")))
+   (assert-false (uim-bool '(char-upper-case? '(0 1 2))))
+   (assert-false (uim-bool '(char-upper-case? car)))
+   (assert-false (uim-bool '(char-upper-case? 0)))    ; NUL
+   (assert-false (uim-bool '(char-upper-case? 1)))    ; SOH
+   (assert-false (uim-bool '(char-upper-case? 31)))   ; US
+   (assert-false (uim-bool '(char-upper-case? 32)))   ; SPACE
+   (assert-false (uim-bool '(char-upper-case? 33)))   ; !
+   (assert-false (uim-bool '(char-upper-case? 47)))   ; /
+   (assert-false (uim-bool '(char-upper-case? 48)))   ; 0
+   (assert-false (uim-bool '(char-upper-case? 57)))   ; 9
+   (assert-false (uim-bool '(char-upper-case? 58)))   ; :
+   (assert-false (uim-bool '(char-upper-case? 64)))   ; @
+   (assert-true  (uim-bool '(char-upper-case? 65)))   ; A
+   (assert-true  (uim-bool '(char-upper-case? 90)))   ; Z
+   (assert-false (uim-bool '(char-upper-case? 91)))   ; [
+   (assert-false (uim-bool '(char-upper-case? 96)))   ; `
+   (assert-false (uim-bool '(char-upper-case? 97)))   ; a
+   (assert-false (uim-bool '(char-upper-case? 122)))  ; z
+   (assert-false (uim-bool '(char-upper-case? 123)))  ; {
+   (assert-false (uim-bool '(char-upper-case? 126)))  ; ~
+   (assert-false (uim-bool '(char-upper-case? 127)))) ; DEL
+  ("test char-lower-case?"
+   (assert-false (uim-bool '(char-lower-case? 'symbol)))
+   (assert-false (uim-bool '(char-lower-case? "string")))
+   (assert-false (uim-bool '(char-lower-case? '(0 1 2))))
+   (assert-false (uim-bool '(char-lower-case? car)))
+   (assert-false (uim-bool '(char-lower-case? 0)))    ; NUL
+   (assert-false (uim-bool '(char-lower-case? 1)))    ; SOH
+   (assert-false (uim-bool '(char-lower-case? 31)))   ; US
+   (assert-false (uim-bool '(char-lower-case? 32)))   ; SPACE
+   (assert-false (uim-bool '(char-lower-case? 33)))   ; !
+   (assert-false (uim-bool '(char-lower-case? 47)))   ; /
+   (assert-false (uim-bool '(char-lower-case? 48)))   ; 0
+   (assert-false (uim-bool '(char-lower-case? 57)))   ; 9
+   (assert-false (uim-bool '(char-lower-case? 58)))   ; :
+   (assert-false (uim-bool '(char-lower-case? 64)))   ; @
+   (assert-false (uim-bool '(char-lower-case? 65)))   ; A
+   (assert-false (uim-bool '(char-lower-case? 90)))   ; Z
+   (assert-false (uim-bool '(char-lower-case? 91)))   ; [
+   (assert-false (uim-bool '(char-lower-case? 96)))   ; `
+   (assert-true  (uim-bool '(char-lower-case? 97)))   ; a
+   (assert-true  (uim-bool '(char-lower-case? 122)))  ; z
+   (assert-false (uim-bool '(char-lower-case? 123)))  ; {
+   (assert-false (uim-bool '(char-lower-case? 126)))  ; ~
+   (assert-false (uim-bool '(char-lower-case? 127)))) ; DEL
+  ("test char-alphabetic?"
+   (assert-false (uim-bool '(char-alphabetic? 'symbol)))
+   (assert-false (uim-bool '(char-alphabetic? "string")))
+   (assert-false (uim-bool '(char-alphabetic? '(0 1 2))))
+   (assert-false (uim-bool '(char-alphabetic? car)))
+   (assert-false (uim-bool '(char-alphabetic? 0)))    ; NUL
+   (assert-false (uim-bool '(char-alphabetic? 1)))    ; SOH
+   (assert-false (uim-bool '(char-alphabetic? 31)))   ; US
+   (assert-false (uim-bool '(char-alphabetic? 32)))   ; SPACE
+   (assert-false (uim-bool '(char-alphabetic? 33)))   ; !
+   (assert-false (uim-bool '(char-alphabetic? 47)))   ; /
+   (assert-false (uim-bool '(char-alphabetic? 48)))   ; 0
+   (assert-false (uim-bool '(char-alphabetic? 57)))   ; 9
+   (assert-false (uim-bool '(char-alphabetic? 58)))   ; :
+   (assert-false (uim-bool '(char-alphabetic? 64)))   ; @
+   (assert-true  (uim-bool '(char-alphabetic? 65)))   ; A
+   (assert-true  (uim-bool '(char-alphabetic? 90)))   ; Z
+   (assert-false (uim-bool '(char-alphabetic? 91)))   ; [
+   (assert-false (uim-bool '(char-alphabetic? 96)))   ; `
+   (assert-true  (uim-bool '(char-alphabetic? 97)))   ; a
+   (assert-true  (uim-bool '(char-alphabetic? 122)))  ; z
+   (assert-false (uim-bool '(char-alphabetic? 123)))  ; {
+   (assert-false (uim-bool '(char-alphabetic? 126)))  ; ~
+   (assert-false (uim-bool '(char-alphabetic? 127)))) ; DEL
+  ("test char-numeric?"
+   (assert-false (uim-bool '(char-numeric? 'symbol)))
+   (assert-false (uim-bool '(char-numeric? "string")))
+   (assert-false (uim-bool '(char-numeric? '(0 1 2))))
+   (assert-false (uim-bool '(char-numeric? car)))
+   (assert-false (uim-bool '(char-numeric? 0)))     ; NUL
+   (assert-false (uim-bool '(char-numeric? 1)))     ; SOH
+   (assert-false (uim-bool '(char-numeric? 31)))    ; US
+   (assert-false (uim-bool '(char-numeric? 32)))    ; SPACE
+   (assert-false (uim-bool '(char-numeric? 33)))    ; !
+   (assert-false (uim-bool '(char-numeric? 47)))    ; /
+   (assert-true  (uim-bool '(char-numeric? 48)))    ; 0
+   (assert-true  (uim-bool '(char-numeric? 57)))    ; 9
+   (assert-false (uim-bool '(char-numeric? 58)))    ; :
+   (assert-false (uim-bool '(char-numeric? 64)))    ; @
+   (assert-false (uim-bool '(char-numeric? 65)))    ; A
+   (assert-false (uim-bool '(char-numeric? 90)))    ; Z
+   (assert-false (uim-bool '(char-numeric? 91)))    ; [
+   (assert-false (uim-bool '(char-numeric? 96)))    ; `
+   (assert-false (uim-bool '(char-numeric? 97)))    ; a
+   (assert-false (uim-bool '(char-numeric? 122)))   ; z
+   (assert-false (uim-bool '(char-numeric? 123)))   ; {
+   (assert-false (uim-bool '(char-numeric? 126)))   ; ~
+   (assert-false (uim-bool '(char-numeric? 127)))) ; DEL
+  ("test char-printable?"
+   (assert-false (uim-bool '(char-printable? 'symbol)))
+   (assert-false (uim-bool '(char-printable? "string")))
+   (assert-false (uim-bool '(char-printable? '(0 1 2))))
+   (assert-false (uim-bool '(char-printable? car)))
+   (assert-false (uim-bool '(char-printable? 0)))    ; NUL
+   (assert-false (uim-bool '(char-printable? 1)))    ; SOH
+   (assert-false (uim-bool '(char-printable? 31)))   ; US
+   (assert-true  (uim-bool '(char-printable? 32)))   ; SPACE
+   (assert-true  (uim-bool '(char-printable? 33)))   ; !
+   (assert-true  (uim-bool '(char-printable? 47)))   ; /
+   (assert-true  (uim-bool '(char-printable? 48)))   ; 0
+   (assert-true  (uim-bool '(char-printable? 57)))   ; 9
+   (assert-true  (uim-bool '(char-printable? 58)))   ; :
+   (assert-true  (uim-bool '(char-printable? 64)))   ; @
+   (assert-true  (uim-bool '(char-printable? 65)))   ; A
+   (assert-true  (uim-bool '(char-printable? 90)))   ; Z
+   (assert-true  (uim-bool '(char-printable? 91)))   ; [
+   (assert-true  (uim-bool '(char-printable? 96)))   ; `
+   (assert-true  (uim-bool '(char-printable? 97)))   ; a
+   (assert-true  (uim-bool '(char-printable? 122)))  ; z
+   (assert-true  (uim-bool '(char-printable? 123)))  ; {
+   (assert-true  (uim-bool '(char-printable? 126)))  ; ~
+   (assert-false (uim-bool '(char-printable? 127)))) ; DEL
+  ("test char-graphic?"
+   (assert-false (uim-bool '(char-graphic? 'symbol)))
+   (assert-false (uim-bool '(char-graphic? "string")))
+   (assert-false (uim-bool '(char-graphic? '(0 1 2))))
+   (assert-false (uim-bool '(char-graphic? car)))
+   (assert-false (uim-bool '(char-graphic? 0)))    ; NUL
+   (assert-false (uim-bool '(char-graphic? 1)))    ; SOH
+   (assert-false (uim-bool '(char-graphic? 31)))   ; US
+   (assert-false (uim-bool '(char-graphic? 32)))   ; SPACE
+   (assert-true  (uim-bool '(char-graphic? 33)))   ; !
+   (assert-true  (uim-bool '(char-graphic? 47)))   ; /
+   (assert-true  (uim-bool '(char-graphic? 48)))   ; 0
+   (assert-true  (uim-bool '(char-graphic? 57)))   ; 9
+   (assert-true  (uim-bool '(char-graphic? 58)))   ; :
+   (assert-true  (uim-bool '(char-graphic? 64)))   ; @
+   (assert-true  (uim-bool '(char-graphic? 65)))   ; A
+   (assert-true  (uim-bool '(char-graphic? 90)))   ; Z
+   (assert-true  (uim-bool '(char-graphic? 91)))   ; [
+   (assert-true  (uim-bool '(char-graphic? 96)))   ; `
+   (assert-true  (uim-bool '(char-graphic? 97)))   ; a
+   (assert-true  (uim-bool '(char-graphic? 122)))  ; z
+   (assert-true  (uim-bool '(char-graphic? 123)))  ; {
+   (assert-true  (uim-bool '(char-graphic? 126)))  ; ~
+   (assert-false (uim-bool '(char-graphic? 127)))) ; DEL
   ("test control-char?"
-   (assert-true  (uim-bool '(control-char? 0)))    ; NUL
-   (assert-true  (uim-bool '(control-char? 1)))    ; SOH
-   (assert-true  (uim-bool '(control-char? 31)))   ; US
-   (assert-false (uim-bool '(control-char? 32)))   ; SPACE
-   (assert-false (uim-bool '(control-char? 33)))   ; !
-   (assert-false (uim-bool '(control-char? 47)))   ; /
-   (assert-false (uim-bool '(control-char? 48)))   ; 0
-   (assert-false (uim-bool '(control-char? 57)))   ; 9
-   (assert-false (uim-bool '(control-char? 58)))   ; :
-   (assert-false (uim-bool '(control-char? 64)))   ; @
-   (assert-false (uim-bool '(control-char? 65)))   ; A
-   (assert-false (uim-bool '(control-char? 90)))   ; Z
-   (assert-false (uim-bool '(control-char? 91)))   ; [
-   (assert-false (uim-bool '(control-char? 96)))   ; `
-   (assert-false (uim-bool '(control-char? 97)))   ; a
-   (assert-false (uim-bool '(control-char? 122)))  ; z
-   (assert-false (uim-bool '(control-char? 123)))  ; {
-   (assert-false (uim-bool '(control-char? 126)))  ; ~
-   (assert-true  (uim-bool '(control-char? 127)))) ; DEL
+   (assert-true  (uim-bool '(eq? control-char? char-control?))))
   ("test alphabet-char?"
-   (assert-false (uim-bool '(alphabet-char? 0)))    ; NUL
-   (assert-false (uim-bool '(alphabet-char? 1)))    ; SOH
-   (assert-false (uim-bool '(alphabet-char? 31)))   ; US
-   (assert-false (uim-bool '(alphabet-char? 32)))   ; SPACE
-   (assert-false (uim-bool '(alphabet-char? 33)))   ; !
-   (assert-false (uim-bool '(alphabet-char? 47)))   ; /
-   (assert-false (uim-bool '(alphabet-char? 48)))   ; 0
-   (assert-false (uim-bool '(alphabet-char? 57)))   ; 9
-   (assert-false (uim-bool '(alphabet-char? 58)))   ; :
-   (assert-false (uim-bool '(alphabet-char? 64)))   ; @
-   (assert-true  (uim-bool '(alphabet-char? 65)))   ; A
-   (assert-true  (uim-bool '(alphabet-char? 90)))   ; Z
-   (assert-false (uim-bool '(alphabet-char? 91)))   ; [
-   (assert-false (uim-bool '(alphabet-char? 96)))   ; `
-   (assert-true  (uim-bool '(alphabet-char? 97)))   ; a
-   (assert-true  (uim-bool '(alphabet-char? 122)))  ; z
-   (assert-false (uim-bool '(alphabet-char? 123)))  ; {
-   (assert-false (uim-bool '(alphabet-char? 126)))  ; ~
-   (assert-false (uim-bool '(alphabet-char? 127)))) ; DEL
-  ("test usual-char?"
-   (assert-false (uim-bool '(usual-char? 0)))    ; NUL
-   (assert-false (uim-bool '(usual-char? 1)))    ; SOH
-   (assert-false (uim-bool '(usual-char? 31)))   ; US
-   (assert-false (uim-bool '(usual-char? 32)))   ; SPACE
-   (assert-true  (uim-bool '(usual-char? 33)))   ; !
-   (assert-true  (uim-bool '(usual-char? 47)))   ; /
-   (assert-true  (uim-bool '(usual-char? 48)))   ; 0
-   (assert-true  (uim-bool '(usual-char? 57)))   ; 9
-   (assert-true  (uim-bool '(usual-char? 58)))   ; :
-   (assert-true  (uim-bool '(usual-char? 64)))   ; @
-   (assert-true  (uim-bool '(usual-char? 65)))   ; A
-   (assert-true  (uim-bool '(usual-char? 90)))   ; Z
-   (assert-true  (uim-bool '(usual-char? 91)))   ; [
-   (assert-true  (uim-bool '(usual-char? 96)))   ; `
-   (assert-true  (uim-bool '(usual-char? 97)))   ; a
-   (assert-true  (uim-bool '(usual-char? 122)))  ; z
-   (assert-true  (uim-bool '(usual-char? 123)))  ; {
-   (assert-true  (uim-bool '(usual-char? 126)))  ; ~
-   (assert-false (uim-bool '(usual-char? 127)))) ; DEL
+   (assert-true  (uim-bool '(eq? alphabet-char? char-alphabetic?))))
   ("test numeral-char?"
-   (assert-false (uim-bool '(numeral-char? 0)))     ; NUL
-   (assert-false (uim-bool '(numeral-char? 1)))     ; SOH
-   (assert-false (uim-bool '(numeral-char? 31)))    ; US
-   (assert-false (uim-bool '(numeral-char? 32)))    ; SPACE
-   (assert-false (uim-bool '(numeral-char? 33)))    ; !
-   (assert-false (uim-bool '(numeral-char? 47)))    ; /
-   (assert-true  (uim-bool '(numeral-char? 48)))    ; 0
-   (assert-true  (uim-bool '(numeral-char? 57)))    ; 9
-   (assert-false (uim-bool '(numeral-char? 58)))    ; :
-   (assert-false (uim-bool '(numeral-char? 64)))    ; @
-   (assert-false (uim-bool '(numeral-char? 65)))    ; A
-   (assert-false (uim-bool '(numeral-char? 90)))    ; Z
-   (assert-false (uim-bool '(numeral-char? 91)))    ; [
-   (assert-false (uim-bool '(numeral-char? 96)))    ; `
-   (assert-false (uim-bool '(numeral-char? 97)))    ; a
-   (assert-false (uim-bool '(numeral-char? 122)))   ; z
-   (assert-false (uim-bool '(numeral-char? 123)))   ; {
-   (assert-false (uim-bool '(numeral-char? 126)))   ; ~
-   (assert-false (uim-bool '(numeral-char? 127))))) ; DEL
+   (assert-true  (uim-bool '(eq? numeral-char? char-numeric?))))
+  ("test usual-char?"
+   (assert-true  (uim-bool '(eq? usual-char? char-graphic?)))))
 
 (define-uim-test-case "test util character conversion procedures"
   ("test numeral-char->number"
@@ -117,26 +242,72 @@
    (assert-true  (uim-bool '(integer? (numeral-char->number 123))))  ; {
    (assert-true  (uim-bool '(integer? (numeral-char->number 126))))  ; ~
    (assert-true  (uim-bool '(integer? (numeral-char->number 127))))) ; DEL
+  ("test char-downcase"
+   (assert-equal 0   (uim '(char-downcase 0)))     ; NUL
+   (assert-equal 1   (uim '(char-downcase 1)))     ; SOH
+   (assert-equal 31  (uim '(char-downcase 31)))    ; US
+   (assert-equal 32  (uim '(char-downcase 32)))    ; SPACE
+   (assert-equal 33  (uim '(char-downcase 33)))    ; !
+   (assert-equal 47  (uim '(char-downcase 47)))    ; /
+   (assert-equal 48  (uim '(char-downcase 48)))    ; 0
+   (assert-equal 57  (uim '(char-downcase 57)))    ; 9
+   (assert-equal 58  (uim '(char-downcase 58)))    ; :
+   (assert-equal 64  (uim '(char-downcase 64)))    ; @
+   (assert-equal 97  (uim '(char-downcase 65)))    ; A
+   (assert-equal 122 (uim '(char-downcase 90)))    ; Z
+   (assert-equal 91  (uim '(char-downcase 91)))    ; [
+   (assert-equal 96  (uim '(char-downcase 96)))    ; `
+   (assert-equal 97  (uim '(char-downcase 97)))    ; a
+   (assert-equal 122 (uim '(char-downcase 122)))   ; z
+   (assert-equal 123 (uim '(char-downcase 123)))   ; {
+   (assert-equal 126 (uim '(char-downcase 126)))   ; ~
+   (assert-equal 127 (uim '(char-downcase 127))))  ; DEL
+  ("test char-upcase"
+   (assert-equal 0   (uim '(char-upcase 0)))     ; NUL
+   (assert-equal 1   (uim '(char-upcase 1)))     ; SOH
+   (assert-equal 31  (uim '(char-upcase 31)))    ; US
+   (assert-equal 32  (uim '(char-upcase 32)))    ; SPACE
+   (assert-equal 33  (uim '(char-upcase 33)))    ; !
+   (assert-equal 47  (uim '(char-upcase 47)))    ; /
+   (assert-equal 48  (uim '(char-upcase 48)))    ; 0
+   (assert-equal 57  (uim '(char-upcase 57)))    ; 9
+   (assert-equal 58  (uim '(char-upcase 58)))    ; :
+   (assert-equal 64  (uim '(char-upcase 64)))    ; @
+   (assert-equal 65  (uim '(char-upcase 65)))    ; A
+   (assert-equal 90  (uim '(char-upcase 90)))    ; Z
+   (assert-equal 91  (uim '(char-upcase 91)))    ; [
+   (assert-equal 96  (uim '(char-upcase 96)))    ; `
+   (assert-equal 65  (uim '(char-upcase 97)))    ; a
+   (assert-equal 90  (uim '(char-upcase 122)))   ; z
+   (assert-equal 123 (uim '(char-upcase 123)))   ; {
+   (assert-equal 126 (uim '(char-upcase 126)))   ; ~
+   (assert-equal 127 (uim '(char-upcase 127))))  ; DEL
+  ("test string->letter"
+   (assert-false (uim-bool '(string->letter "")))    ; NUL
+   (assert-false (uim-bool '(string->letter "")))  ; SOH
+   (assert-false (uim-bool '(string->letter "")))  ; US
+   (assert-false (uim-bool '(string->letter " ")))   ; SPACE
+   (assert-false (uim-bool '(string->letter "!")))   ; !
+   (assert-false (uim-bool '(string->letter "/")))   ; /
+   (assert-false (uim-bool '(string->letter "0")))   ; 0
+   (assert-false (uim-bool '(string->letter "9")))   ; 9
+   (assert-false (uim-bool '(string->letter ":")))   ; :
+   (assert-false (uim-bool '(string->letter "@")))   ; @
+   (assert-false (uim-bool '(string->letter "AA")))  ; AA
+   (assert-equal 65   (uim '(string->letter "A")))   ; A
+   (assert-equal 90   (uim '(string->letter "Z")))   ; Z
+   (assert-false (uim-bool '(string->letter "ZZ")))  ; ZZ
+   (assert-false (uim-bool '(string->letter "[")))   ; [
+   (assert-false (uim-bool '(string->letter "`")))   ; `
+   (assert-false (uim-bool '(string->letter "aa")))  ; aa
+   (assert-equal 97   (uim '(string->letter "a")))   ; a
+   (assert-equal 122  (uim '(string->letter "z")))   ; z
+   (assert-false (uim-bool '(string->letter "zz")))  ; zz
+   (assert-false (uim-bool '(string->letter "{")))   ; {
+   (assert-false (uim-bool '(string->letter "~")))   ; ~
+   (assert-false (uim-bool '(string->letter "")))) ; DEL
   ("test to-lower-char"
-   (assert-equal 0   (uim '(to-lower-char 0)))     ; NUL
-   (assert-equal 1   (uim '(to-lower-char 1)))     ; SOH
-   (assert-equal 31  (uim '(to-lower-char 31)))    ; US
-   (assert-equal 32  (uim '(to-lower-char 32)))    ; SPACE
-   (assert-equal 33  (uim '(to-lower-char 33)))    ; !
-   (assert-equal 47  (uim '(to-lower-char 47)))    ; /
-   (assert-equal 48  (uim '(to-lower-char 48)))    ; 0
-   (assert-equal 57  (uim '(to-lower-char 57)))    ; 9
-   (assert-equal 58  (uim '(to-lower-char 58)))    ; :
-   (assert-equal 64  (uim '(to-lower-char 64)))    ; @
-   (assert-equal 97  (uim '(to-lower-char 65)))    ; A
-   (assert-equal 122 (uim '(to-lower-char 90)))    ; Z
-   (assert-equal 91  (uim '(to-lower-char 91)))    ; [
-   (assert-equal 96  (uim '(to-lower-char 96)))    ; `
-   (assert-equal 97  (uim '(to-lower-char 97)))    ; a
-   (assert-equal 122 (uim '(to-lower-char 122)))   ; z
-   (assert-equal 123 (uim '(to-lower-char 123)))   ; {
-   (assert-equal 126 (uim '(to-lower-char 126)))   ; ~
-   (assert-equal 127 (uim '(to-lower-char 127))))) ; DEL
+   (assert-true  (uim-bool '(eq? to-lower-char char-downcase)))))
 
 (define-uim-test-case "test util string list procedures"
   ("test string-list-concat"
@@ -260,7 +431,9 @@
    (assert-equal '(1 "2" three (4) 5 six "7" (8 8) -9)
 		 (uim '(list-head lst 9)))
    (assert-error (lambda ()
-		   (uim '(list-head lst 10)))))
+		   (uim '(list-head lst 10))))
+   (assert-error (lambda ()
+		   (uim '(list-head lst -1)))))
   ("test iterate-lists"
    (assert-equal '(("o" . "O") ("l" . "L") ("l" . "L") ("e" . "E") ("h" . "H"))
 		 (uim '(iterate-lists (lambda (state elms)
@@ -270,7 +443,133 @@
 							   state))))
 				      ()
 				      '(("h" "e" "l" "l" "o")
-					("H" "E" "L" "L" "O" "!")))))))
+					("H" "E" "L" "L" "O" "!"))))))
+
+  ("test alist-replace"
+   (uim '(define alist ()))
+   (uim '(set! alist (alist-replace '(first 1 "1") alist)))
+   (assert-equal '((first 1 "1"))
+		 (uim 'alist))
+   (uim '(set! alist (alist-replace '(second 2 "2") alist)))
+   (assert-equal '((second 2 "2") (first 1 "1"))
+		 (uim 'alist))
+   (uim '(set! alist (alist-replace '(third 3 "3") alist)))
+   (assert-equal '((third 3 "3") (second 2 "2") (first 1 "1"))
+		 (uim 'alist))
+   (uim '(set! alist (alist-replace '(second two "two") alist)))
+   (assert-equal '((third 3 "3") (second two "two") (first 1 "1"))
+		 (uim 'alist)))
+
+  ("test join"
+   (assert-equal ()
+		 (uim '(join () ())))
+   (assert-equal '(())
+		 (uim '(join () '(()))))
+   (assert-equal '(1)
+		 (uim '(join () '(1))))
+   (assert-equal '(() () ())
+		 (uim '(join () '(() ()))))
+   (assert-equal '(1 () 2)
+		 (uim '(join () '(1 2))))
+   (assert-equal '(1 () 2 () 3)
+		 (uim '(join () '(1 2 3))))
+   (assert-equal '(one () two () three)
+		 (uim '(join () '(one two three))))
+   (assert-equal '("1" () "2" () "3")
+		 (uim '(join () '("1" "2" "3"))))
+   (assert-equal '(() () () () ())
+		 (uim '(join () '(() () ()))))
+
+   (assert-equal ()
+		 (uim '(join "/" ())))
+   (assert-equal '(())
+		 (uim '(join "/" '(()))))
+   (assert-equal '(1)
+		 (uim '(join "/" '(1))))
+   (assert-equal '(() "/" ())
+		 (uim '(join "/" '(() ()))))
+   (assert-equal '(1 "/" 2)
+		 (uim '(join "/" '(1 2))))
+   (assert-equal '(1 "/" 2 "/" 3)
+		 (uim '(join "/" '(1 2 3))))
+   (assert-equal '(one "/" two "/" three)
+		 (uim '(join "/" '(one two three))))
+   (assert-equal '("1" "/" "2" "/" "3")
+		 (uim '(join "/" '("1" "2" "3"))))
+   (assert-equal '(() "/" () "/" ())
+		 (uim '(join "/" '(() () ())))))
+
+  ("test string-join"
+   (assert-equal ""
+		 (uim '(string-join () ())))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(())))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(1)))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(() ())))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(1 2)))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(1 2 3)))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(one two three)))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '("1" "2" "3")))))
+   (assert-error (lambda ()
+		   (uim '(string-join () '(() () ())))))
+
+   (assert-equal ""
+		 (uim '(string-join "/" ())))
+   (assert-equal ""
+		 (uim '(string-join "/" '(""))))
+   (assert-equal "1"
+		 (uim '(string-join "/" '("1"))))
+   (assert-equal "1/2"
+		 (uim '(string-join "/" '("1" "2"))))
+   (assert-equal "1/2/3"
+		 (uim '(string-join "/" '("1" "2" "3"))))
+
+   (assert-equal ""
+		 (uim '(string-join "-sep-" ())))
+   (assert-equal ""
+		 (uim '(string-join "-sep-" '(""))))
+   (assert-equal "1"
+		 (uim '(string-join "-sep-" '("1"))))
+   (assert-equal "1-sep-2"
+		 (uim '(string-join "-sep-" '("1" "2"))))
+   (assert-equal "1-sep-2-sep-3"
+		 (uim '(string-join "-sep-" '("1" "2" "3")))))
+
+  ("test string-append-map"
+   (assert-equal ""
+		 (uim '(string-append-map car ())))
+   (assert-equal "c"
+		 (uim '(string-append-map car '(("c" "C")))))
+   (assert-equal "ca"
+		 (uim '(string-append-map car '(("c" "C") ("a" "A")))))
+   (assert-equal "car"
+		 (uim '(string-append-map car '(("c" "C") ("a" "A") ("r" "R")))))))
+
+(define-uim-test-case "testcase util misc"
+  ("test compose"
+   (uim '(define test-list '(0 1 2 3 4 5)))
+   (assert-true  (uim-bool '(procedure? (compose))))
+   (assert-true  (uim-bool '(procedure? (compose car))))
+   (assert-true  (uim-bool '(procedure? (compose car cdr))))
+   (assert-true  (uim-bool '(procedure? (compose car cdr list))))
+   (assert-equal '(0 1 2 3 4 5)
+		 (uim '((compose) test-list)))
+   (assert-equal 0
+		 (uim '((compose car) test-list)))
+   (assert-equal 1
+		 (uim '((compose car cdr) test-list)))
+   (assert-equal 2
+		 (uim '((compose car cdr cdr) test-list)))
+   (assert-equal 4
+		 (uim '((compose car cdr reverse) test-list)))
+   (assert-equal 3
+		 (uim '((compose car cdr cdr reverse) test-list)))))
 
 (define-uim-test-case "testcase util R5RS procedures"
   (setup
@@ -385,7 +684,52 @@
    (assert-equal '(5 7 9)
 		 (uim '(map +
 			    '(1 2 3)
-			    '(4 5 6)))))
+			    '(4 5 6))))
+   (assert-equal '()
+		 (uim '(map +
+			    '()
+			    '()
+			    '())))
+   (assert-equal '(12)
+		 (uim '(map +
+			    '(1)
+			    '(4)
+			    '(7))))
+   (assert-equal '(12 15)
+		 (uim '(map +
+			    '(1 2)
+			    '(4 5)
+			    '(7 8))))
+   (assert-equal '(12 15 18)
+		 (uim '(map +
+			    '(1 2 3)
+			    '(4 5 6)
+			    '(7 8 9))))
+   (assert-equal '()
+		 (uim '(map +
+			    '()
+			    '()
+			    '()
+			    '())))
+   (assert-equal '(22)
+		 (uim '(map +
+			    '(1)
+			    '(4)
+			    '(7)
+			    '(10))))
+   (assert-equal '(22 26)
+		 (uim '(map +
+			    '(1 2)
+			    '(4 5)
+			    '(7 8)
+			    '(10 11))))
+   (assert-equal '(22 26 30)
+		 (uim '(map +
+			    '(1 2 3)
+			    '(4 5 6)
+			    '(7 8 9)
+			    '(10 11 12)))))
+
   ("test for-each"
    (assert-equal 3
 		 (uim '(let ((i 0))
@@ -431,7 +775,9 @@
    (assert-equal '()
 		 (uim '(list-tail lst 9)))
    (assert-error (lambda ()
-		   (uim '(list-tail lst 10))))))
+		   (uim '(list-tail lst 10))))
+   (assert-error (lambda ()
+		   (uim '(list-tail lst -1))))))
 
 (define-uim-test-case "test util SRFI procedures"
   (setup
@@ -460,6 +806,18 @@
 			       (five "5" 5)
 			       (three "3" 3))))))
 
+  ("test list-tabulate"
+   (assert-equal ()
+		 (uim '(list-tabulate 0 (lambda (x) x))))
+   (assert-equal '(0)
+		 (uim '(list-tabulate 1 (lambda (x) x))))
+   (assert-equal '(0 1 2 3 4)
+		 (uim '(list-tabulate 5 (lambda (x) x))))
+   (assert-equal '(0 1 4 9 16)
+		 (uim '(list-tabulate 5 (lambda (x) (* x x)))))
+   (assert-error (lambda ()
+		   (uim '(list-tabulate -1 (lambda (x) x))))))
+
   ("test make-list"
    (assert-equal '(fill fill fill)
 		 (uim '(make-list 3 'fill)))
@@ -476,7 +834,48 @@
    (assert-equal '()
 		 (uim '(make-list 0 ())))
    (assert-equal '()
-		 (uim '(make-list 0 'fill))))
+		 (uim '(make-list 0 'fill)))
+   (assert-error (lambda ()
+		   (uim '(make-list -1 'fill)))))
+
+  ("test iota"
+   (assert-equal ()
+		 (uim '(iota 0)))
+   (assert-equal '(0)
+		 (uim '(iota 1)))
+   (assert-equal '(0 1 2 3 4)
+		 (uim '(iota 5)))
+   (assert-error (lambda ()
+		   (uim '(iota -1))))
+
+   (assert-equal ()
+		 (uim '(iota 0 0)))
+   (assert-equal '(0)
+		 (uim '(iota 1 0)))
+   (assert-equal '(0 1 2 3 4)
+		 (uim '(iota 5 0)))
+   (assert-error (lambda ()
+		   (uim '(iota -1 0))))
+
+   (assert-error (lambda ()
+		   (uim '(iota 0 1))))
+   (assert-equal '()
+		 (uim '(iota 1 1)))
+   (assert-equal '(1 2 3 4)
+		 (uim '(iota 5 1)))
+   (assert-error (lambda ()
+		   (uim '(iota -1 1))))
+
+   (assert-error (lambda ()
+		   (uim '(iota 1 3))))
+   (assert-equal '(3 4)
+		 (uim '(iota 5 3)))
+   (assert-error (lambda ()
+		   (uim '(iota -1 3))))
+
+   (assert-equal '()
+		 (uim '(iota 5 5))))
+
   ("test zip"
    (assert-equal '((1) (2) (3) (4) (5))
 		 (uim '(zip '(1 2 3 4 5))))
@@ -494,6 +893,17 @@
 			    '(one two three))))
    (assert-equal ()
 		 (uim '(zip ()))))
+
+  ("test append-map"
+   (assert-equal '()
+		 (uim '(append-map car ())))
+   (assert-equal '(c)
+		 (uim '(append-map car '(((c) (C))))))
+   (assert-equal '(c a)
+		 (uim '(append-map car '(((c) (C)) ((a) (A))))))
+   (assert-equal '(c a r)
+		 (uim '(append-map car '(((c) (C)) ((a) (A)) ((r) (R)))))))
+
   ("test append-reverse"
    (assert-equal '("5" "4" "3" "2" "1" six seven eight)
 		 (uim '(append-reverse '("1" "2" "3" "4" "5")
@@ -564,6 +974,114 @@
 		 (uim '(fold * 1 '(1 2 3 4 5))))
    (assert-equal 14400
 		 (uim '(fold * 1 '(1 2 3 4 5) '(1 2 3 4 5)))))
+  ("test unfold"
+   ;; immediate term
+   (assert-equal '()
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5)))
+   (assert-equal '(5)
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5
+			       (lambda (rest)
+				 (list rest)))))
+   (assert-equal '(-1)
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5
+			       (lambda (rest)
+				 '(-1)))))
+   ;; 5 times
+   (assert-equal '(5 4 3 2 1)
+		 (uim '(unfold (lambda (x)
+				 (= x 0))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5)))
+   (assert-equal '(5 4 3 2 1 0)
+		 (uim '(unfold (lambda (x)
+				 (= x 0))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5
+			       (lambda (x)
+				 (list x)))))
+   (assert-equal '(5 4 3 2 1 -1)
+		 (uim '(unfold (lambda (x)
+				 (= x 0))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (- rest 1))
+			       5
+			       (lambda (x)
+				 '(-1)))))
+   ;; 5 times, reversed
+   (assert-equal '(0 1 2 3 4)
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (+ rest 1))
+			       0)))
+   (assert-equal '(0 1 2 3 4 5)
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (+ rest 1))
+			       0
+			       (lambda (x)
+				 (list x)))))
+   (assert-equal '(0 1 2 3 4 -1)
+		 (uim '(unfold (lambda (x)
+				 (= x 5))
+			       (lambda (rest)
+				 rest)
+			       (lambda (rest)
+				 (+ rest 1))
+			       0
+			       (lambda (x)
+				 '(-1)))))
+   ;; restruct same list
+   (assert-equal '(0 1 2 3 4 5)
+		 (uim '(unfold null?
+			       car
+			       cdr
+			       '(0 1 2 3 4 5))))
+   (assert-equal '(0 1 2 3 4 5)
+		 (uim '(unfold null?
+			       car
+			       cdr
+			       '(0 1 2 3 4 5)
+			       (lambda (x)
+				 x))))
+   (assert-equal '(0 1 2 3 4 5 . -1)
+		 (uim '(unfold null?
+			       car
+			       cdr
+			       '(0 1 2 3 4 5)
+			       (lambda (x)
+				 -1)))))
   ("test filter"
    (assert-equal ()
 		 (uim '(filter not ())))
@@ -577,7 +1095,9 @@
    (assert-equal '(3 8 10)
 		 (uim '(filter integer?
 			       '(one "2" 3 #f (5) six "7" 8 (9) 10)))))
+
   ("test filter-map"
+   ;; single list
    (assert-equal ()
 		 (uim '(filter-map not ())))
    (assert-equal '(5 6 4)
@@ -596,18 +1116,39 @@
    (assert-equal (uim '(list (integer? 0) (integer? 0) (integer? 0)))
 		 (uim '(filter-map integer?
 				   '(one "2" 3 #f (5) six "7" 8 (9) 10))))
-;   (assert-equal '()
-;		 (uim '(filter-map +
-;				   ()
-;				   ())))
-;   (assert-equal '(8 11 10 8)
-;		 (uim '(filter-map (lambda (x y)
-;				     (let ((sum (+ x y)))
-;				       (and (< 6 sum)
-;					    sum)))
-;				   '(3 5 2 6 4 1)
-;				   '(1 3 9 4 2 7))))
-   )
+   ;; multiple lists
+   (assert-equal '()
+		 (uim '(filter-map +
+				   ()
+				   ())))
+   (assert-equal '(8 11 10 8)
+		 (uim '(filter-map (lambda (x y)
+				     (let ((sum (+ x y)))
+				       (and (< 6 sum)
+					    sum)))
+				   '(3 5 2 6 4 1)
+				   '(1 3 9 4 2 7))))
+   (assert-equal ()
+		 (uim '(filter-map (lambda (x y)
+				     (let ((sum (+ x y)))
+				       (and (< 6 sum)
+					    sum)))
+				   ()
+				   '(1 3 9 4 2 7))))
+   (assert-equal '(8 11 10)
+		 (uim '(filter-map (lambda (x y)
+				     (let ((sum (+ x y)))
+				       (and (< 6 sum)
+					    sum)))
+				   '(3 5 2 6 4 1)
+				   '(1 3 9 4))))
+   (assert-equal '("aAa1" "bBb2" "cCc3" "dDd4")
+		 (uim '(filter-map string-append
+				   '("a" "b" "c" "d")
+				   '("A" "B" "C" "D")
+				   '("a" "b" "c" "d")
+				   '("1" "2" "3" "4" "5")))))
+
   ("test remove"
    (assert-equal '(1 three (4) 5 six (8 8) -9)
 		 (uim '(remove string? lst)))
@@ -792,3 +1333,32 @@
 			     (test-rec-set-fourth! trec 44)
 			     (equal? (test-rec-fourth trec)
 				     44))))))
+
+(define-uim-test-case "testcase util multi-segment utils"
+  ("test multi-segment-make-index-list"
+   (uim '(define old-lst '(0 1 2 3 4)))
+   (assert-false (uim-bool '(multi-segment-make-index-list -1 old-lst)))
+   (assert-equal ()
+		 (uim '(multi-segment-make-index-list 0 old-lst)))
+   (assert-equal '(0)
+		 (uim '(multi-segment-make-index-list 1 old-lst)))
+   (assert-equal '(0 1)
+		 (uim '(multi-segment-make-index-list 2 old-lst)))
+   (assert-equal '(0 1 2 3 4)
+		 (uim '(multi-segment-make-index-list 5 old-lst)))
+   (assert-equal '(0 1 2 3 4 0)
+		 (uim '(multi-segment-make-index-list 6 old-lst)))
+   (assert-equal '(0 1 2 3 4 0 0)
+		 (uim '(multi-segment-make-index-list 7 old-lst))))
+
+  ("test multi-segment-opposite-kana"
+   (assert-equal (uim 'multi-segment-type-katakana)
+		 (uim '(multi-segment-opposite-kana
+			multi-segment-type-hiragana)))
+   (assert-equal (uim 'multi-segment-type-hiragana)
+		 (uim '(multi-segment-opposite-kana
+			multi-segment-type-katakana)))
+   (assert-equal (uim 'multi-segment-type-hiragana)
+		 (uim '(multi-segment-opposite-kana
+			multi-segment-type-hankana))))
+)
