@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003,2004,2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -32,11 +32,12 @@
 */
 
 #include "config.h"
-#ifdef HAVE_M17NLIB
 #include <stdlib.h>
 #include <string.h>
 #include <m17n.h>
+#include "uim-scm.h"
 #include "context.h"
+#include "plugin.h"
 
 static int m17nlib_ok;
 static MConverter *converter;
@@ -54,6 +55,9 @@ static struct im_ {
 static int max_input_contexts;
 static struct ic_ {
   MInputContext *mic;
+  char **old_candidates; /* FIXME: ugly hack for perfomance... */
+  char **new_candidates; /* FIXME: ugly hack for perfomance... */
+  int  nr_candidates;
 } *ic_array;
 
 static char *
@@ -126,13 +130,15 @@ pushback_input_method(MInputMethod *im,
 		      char *lib_lang, char *name)
 {
   char *lang = remap_lang_name(lib_lang);
-  if (!lang) {
-    return ;
-  }
 
   im_array = realloc(im_array, 
 		     sizeof(struct im_) * (nr_input_methods + 1));
-  im_array[nr_input_methods].lang = strdup(lang);
+
+  if (lang != NULL) {
+    im_array[nr_input_methods].lang = strdup(lang);
+  } else {
+    im_array[nr_input_methods].lang = NULL;
+  }
   im_array[nr_input_methods].name = strdup(name);
   im_array[nr_input_methods].im = im;
   nr_input_methods++;
@@ -192,26 +198,22 @@ candidates_done_cb(MInputContext *ic, MSymbol command)
   fprintf(stderr,"candidate done\n");
 }
 
-static MPlist *
-register_callbacks(MPlist *callback_list)
+static void
+register_callbacks(void)
 {
-  if(!callback_list)
-    callback_list = mplist();
-
-  mplist_add(callback_list, Minput_preedit_start, (void *)preedit_start_cb);
-  mplist_add(callback_list, Minput_preedit_draw,  (void *)preedit_draw_cb);
-  mplist_add(callback_list, Minput_preedit_done,  (void *)preedit_done_cb);
-  mplist_add(callback_list, Minput_status_start,  (void *)status_start_cb);
-  mplist_add(callback_list, Minput_status_draw,   (void *)status_draw_cb);
-  mplist_add(callback_list, Minput_status_done,   (void *)status_done_cb);
-  mplist_add(callback_list, Minput_candidates_start, (void *)candidates_start_cb);
-  mplist_add(callback_list, Minput_candidates_draw,  (void *)candidates_draw_cb);
-  mplist_add(callback_list, Minput_candidates_done,  (void *)candidates_done_cb);
-
-  return callback_list;
+  /*
+  mplist_add(minput_default_driver.callback_list, Minput_preedit_start, (void *)preedit_start_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_preedit_draw,  (void *)preedit_draw_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_preedit_done,  (void *)preedit_done_cb);
+    mplist_add(minput_default_driver.callback_list, Minput_status_start,  (void *)status_start_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_status_draw,   (void *)status_draw_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_status_done,   (void *)status_done_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_candidates_start, (void *)candidates_start_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_candidates_draw,  (void *)candidates_draw_cb);
+  mplist_add(minput_default_driver.callback_list, Minput_candidates_done,  (void *)candidates_done_cb);*/
 }
 
-static LISP
+static uim_lisp
 init_m17nlib()
 {
   MPlist *imlist, *elm;
@@ -225,7 +227,7 @@ init_m17nlib()
   imlist = mdatabase_list(msymbol("input-method"), Mnil, Mnil, Mnil);
   if (!imlist) {
     /* maybe user forgot to install m17n-db */
-    return NIL;
+    return uim_scm_f();
   }
   for (elm = imlist; mplist_key(elm) != Mnil; elm = mplist_next(elm)) {
     MDatabase *mdb = mplist_value(elm);
@@ -237,32 +239,17 @@ init_m17nlib()
 	pushback_input_method(im, msymbol_name(lang),
 			      msymbol_name(im->name));
 	
-	im->driver.callback_list = register_callbacks(im->driver.callback_list);
       }
     }
   }
+  /*  register_callbacks();*/
   m17n_object_unref(imlist);
   converter = mconv_buffer_converter(utf8, NULL, 0);
   if (!converter) {
-    return NIL;
+    return uim_scm_f();
   }
   m17nlib_ok = 1;
-  return siod_true_value();
-}
-
-void
-uim_quit_m17nlib(void)
-{
-  if (converter) {
-    mconv_free_converter(converter);
-    converter = NULL;
-  }
-  if (m17nlib_ok) {
-    M17N_FINI();
-    m17nlib_ok = 0;
-  }
-  free(im_array);
-  free(ic_array);
+  return uim_scm_t();
 }
 
 static char *
@@ -275,210 +262,220 @@ convert_mtext2str(MText *mtext)
   return strdup(buffer_for_converter);
 }
 
-static LISP
-compose_modep(LISP id_)
+static uim_lisp
+compose_modep(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   /* range check of id might need. */
   MInputContext *ic = ic_array[id].mic;
-  if(!ic) {
-    return NIL;
+  if (!ic) {
+    return uim_scm_f();
   }
-  if(ic->candidate_from == ic->candidate_to) {
-    return NIL;
+  if (ic->candidate_from == ic->candidate_to ||
+     ic->candidate_from > ic->candidate_to) {
+    return uim_scm_f();
   } else {
-    return siod_true_value();
+    return uim_scm_t();
   }
 }
 
-static LISP
-preedit_changedp(LISP id_)
+static uim_lisp
+preedit_changedp(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   /* range check of id might need. */
   MInputContext *ic = ic_array[id].mic;
-  if(!ic) {
-    return siod_false_value();
+  if (!ic) {
+    return uim_scm_f();
   }
-  if(ic->preedit_changed == 1) {
-    return siod_true_value();
+  if (ic->preedit_changed == 1) {
+    return uim_scm_t();
   } else {
-    return siod_false_value();
+    return uim_scm_f();
   }
 }
 
-static LISP
-get_left_of_cursor(LISP id_)
+static uim_lisp
+get_left_of_cursor(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int buflen;
   int i;
-  LISP buf_;
+  uim_lisp buf_;
   char *buf;
   char *p;
   MInputContext *ic = ic_array[id].mic;
   if (!ic) {
-    return strcons(0, "");
+    return uim_scm_make_str("");
   }
-  if(ic->cursor_pos == 0) {
-    return strcons(0, "");
+  if (ic->cursor_pos == 0) {
+    return uim_scm_make_str("");
   }
   buf = convert_mtext2str(ic->preedit);
   p = (char *)buf;
 
-  for(i=0; i<ic->cursor_pos ;i++) {
+  for (i=0; i<ic->cursor_pos ;i++) {
     p = m17nlib_utf8_find_next_char(p);
   }
   *p = 0;
 
   buflen = strlen((char *)buf);
-  buf_ = strcons(buflen, (char *)buf);
+  buf_ = uim_scm_make_str((char *)buf);
   return buf_;
 }
 
-static LISP
-get_right_of_cursor(LISP id_)
+static uim_lisp
+get_right_of_cursor(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int buflen;
   int i;
-  LISP buf_;
+  uim_lisp buf_;
   char *buf;
-  unsigned char *p;
+  char *p;
   MInputContext *ic = ic_array[id].mic;
   if (!ic) {
-    return strcons(0, "");
+    return uim_scm_make_str("");
   }
 
   buf = convert_mtext2str(ic->preedit);
   p = buf;
 
-  for(i=0; i<ic->cursor_pos ;i++) {
-    p = (unsigned char *)strcons(0, "");
+  for (i=0; i<ic->cursor_pos ;i++) {
+    p = m17nlib_utf8_find_next_char(p);
   }
   buflen = strlen((char *)p);
-  buf_ = strcons(buflen, (char *)p);
+  buf_ = uim_scm_make_str((char *)p);
   return buf_;
 }
 
-static LISP
-get_left_of_candidate(LISP id_)
+static uim_lisp
+get_left_of_candidate(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int buflen;
   int i;
-  LISP buf_;
-  unsigned char *buf;
+  uim_lisp buf_;
+  char *buf;
   char *p;
   MInputContext *ic = ic_array[id].mic;
   if (!ic) {
-    return strcons(0, "");
+    return uim_scm_make_str("");
   }
-  if(ic->candidate_from == 0) {
-    return strcons(0, "");
+  if (ic->candidate_from == 0) {
+    return uim_scm_make_str("");
   }
   buf = convert_mtext2str(ic->preedit);
-  p = (char *)buf;
+  p = buf;
 
-  for(i=0; i<ic->candidate_from ;i++) {
+  for (i=0; i<ic->candidate_from ;i++) {
     p = m17nlib_utf8_find_next_char(p);
   }
   *p = 0;
   buflen = strlen((char *)buf);
-  buf_ = strcons(buflen, (char *)buf);
+  buf_ = uim_scm_make_str((char *)buf);
   free(buf);
   return buf_;
 }
 
-static LISP
-get_selected_candidate(LISP id_)
+static uim_lisp
+get_selected_candidate(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int buflen;
   int i;
-  LISP buf_;
-  unsigned char *buf;
-  unsigned char *p;
-  unsigned char *start;
+  uim_lisp buf_;
+  char *buf;
+  char *p;
+  char *start;
   MInputContext *ic = ic_array[id].mic;
   if (!ic) {
-    return strcons(0, "");
+    return uim_scm_make_str("");
   }
   buf = convert_mtext2str(ic->preedit);
   p = buf;
 
-  if(!p) {
-    return strcons(0, "");
+  if (!p) {
+    return uim_scm_make_str("");
   }
 
-  for(i=0; i<ic->candidate_from ;i++) {
-    p = (unsigned char *)m17nlib_utf8_find_next_char((char *)p);
+  for (i=0; i<ic->candidate_from ;i++) {
+    p = m17nlib_utf8_find_next_char(p);
   }
   start = p;
 
-  for(i=0; i<ic->candidate_to - ic->candidate_from ;i++) {
-    p = (unsigned char *)m17nlib_utf8_find_next_char((char *)p);
+  for (i=0; i<ic->candidate_to - ic->candidate_from ;i++) {
+    p = m17nlib_utf8_find_next_char(p);
   }
   *p = 0;
 
-  buflen = strlen((char *)start);
-  buf_ = strcons(buflen, (char *)start);
+  buflen = strlen(start);
+  buf_ = uim_scm_make_str(start);
   free(buf);
   return buf_;
 }
 
-static LISP
-get_right_of_candidate(LISP id_)
+static uim_lisp
+get_right_of_candidate(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int buflen;
   int i;
-  LISP buf_;
-  unsigned char *buf;
-  unsigned char *p;
+  uim_lisp buf_;
+  char *buf;
+  char *p;
   MInputContext *ic = ic_array[id].mic;
   if (!ic) {
-    return strcons(0, "");
+    return uim_scm_make_str("");
   }
   buf = convert_mtext2str(ic->preedit);
   p = buf;
 
-  for(i=0; i<ic->candidate_to ;i++) {
-    p = (unsigned char *)m17nlib_utf8_find_next_char((char *)p);
+  for (i=0; i<ic->candidate_to ;i++) {
+    p = m17nlib_utf8_find_next_char(p);
   }
-  buflen = strlen((char *)p);
-  buf_ = strcons(buflen, (char *)p);
+  buflen = strlen(p);
+  buf_ = uim_scm_make_str(p);
   free(buf);
   return buf_;
 }
 
-static LISP
+static uim_lisp
 get_nr_input_methods()
 {
-  return intcons(nr_input_methods);
+  return uim_scm_make_int(nr_input_methods);
 }
 
-static LISP
-get_input_method_name(LISP nth_)
+static uim_lisp
+get_input_method_name(uim_lisp nth_)
 {
-  int nth = get_c_int(nth_);
+  int nth = uim_scm_c_int(nth_);
   if (nth < nr_input_methods) {
     char *name = alloca(strlen(im_array[nth].name) + 20);
-    sprintf(name, "m17n-%s-%s", im_array[nth].lang, im_array[nth].name);
-    return rintern(name);
+
+    if (im_array[nth].lang != NULL)
+      sprintf(name, "m17n-%s-%s", im_array[nth].lang, im_array[nth].name);
+    else
+      sprintf(name, "m17n-%s", im_array[nth].name);
+
+    return uim_scm_make_str(name);
   }
-  return NIL;
+  return uim_scm_f();
 }
 
-static LISP
-get_input_method_lang(LISP nth_)
+static uim_lisp
+get_input_method_lang(uim_lisp nth_)
 {
-  int nth = get_c_int(nth_);
+  int nth = uim_scm_c_int(nth_);
   if (nth < nr_input_methods) {
     char *lang = im_array[nth].lang;
-    return strcons(strlen(lang), lang);
+    if (lang != NULL) {
+      return uim_scm_make_str(lang);
+    } else {
+      return uim_scm_make_str("unknown");
+    }
   }
-  return NIL;
+  return uim_scm_f();
 }
 
 static MInputMethod *
@@ -491,7 +488,11 @@ find_im_by_name(char *name)
   name = &name[5];
   for (i = 0; i < nr_input_methods; i++) {
     char buf[100];
-    sprintf(buf, "%s-%s", im_array[i].lang, im_array[i].name);
+    if (im_array[i].lang == NULL) {
+      snprintf(buf, 100, "%s", im_array[i].name);
+    } else {
+      snprintf(buf, 100, "%s-%s", im_array[i].lang, im_array[i].name);
+    }
     if (!strcmp(name, buf)) {
       return im_array[i].im;
     }
@@ -499,23 +500,25 @@ find_im_by_name(char *name)
   return NULL;
 }
 
-static LISP
-alloc_id(LISP name_)
+static uim_lisp
+alloc_id(uim_lisp name_)
 {
   int id = unused_ic_id();
-  char *name = uim_get_c_string(name_);
+  char *name = uim_scm_c_str(name_);
   MInputMethod *im = find_im_by_name(name);
   if (im) {
     ic_array[id].mic = minput_create_ic(im, NULL);
   }
+  ic_array[id].old_candidates = NULL;
+  ic_array[id].new_candidates = NULL;
   free(name);
-  return intcons(id);
+  return uim_scm_make_int(id);
 }
 
-static LISP
-free_id(LISP id_)
+static uim_lisp
+free_id(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   if (id < max_input_contexts) {
     struct ic_ *ic = &ic_array[id];
     if (ic->mic) {
@@ -523,7 +526,7 @@ free_id(LISP id_)
       ic->mic = NULL;
     }
   }
-  return NIL;
+  return uim_scm_f();
 }
 
 static MSymbol
@@ -538,102 +541,109 @@ get_key_sym(int ch)
   return Mnil;
 }
 
-static LISP
-push_symbol_key(LISP id_, LISP key_)
+static uim_lisp
+push_symbol_key(uim_lisp id_, uim_lisp key_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   MSymbol key;
   MInputContext *ic = ic_array[id].mic;
-  /*  printf("%s\n",get_c_string(key_)); */
-  key = msymbol(get_c_string(key_));
+  /*  printf("%s\n",uim_scm_c_str(key_)); */
+  key = msymbol(uim_scm_c_str(key_));
   if (key == Mnil) {
-    return siod_true_value();
+    return uim_scm_t();
   }
   if (minput_filter(ic, key, NULL) == 1) {
-    return siod_true_value();
+    return uim_scm_t();
   }
 
-  return NIL;
+  return uim_scm_f();
 }
 
-static LISP
-push_key(LISP id_, LISP key_, LISP mod_)
+static uim_lisp
+push_key(uim_lisp id_, uim_lisp key_, uim_lisp mod_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   int ch;
   MSymbol key;
   MInputContext *ic = ic_array[id].mic;
-  if NINTNUMP(key_)
-    return siod_true_value();  
-  ch = get_c_int(key_);
+
+  ch = uim_scm_c_int(key_);
   key = get_key_sym(ch);
   if (key == Mnil) {
-    return siod_true_value();
+    return uim_scm_t();
   }
   if (minput_filter(ic, key, NULL) == 1) {
-    return siod_true_value();
+    return uim_scm_t();
   }
 
-  return NIL;
+  return uim_scm_f();
 }
 
-static LISP
-get_commit_string(LISP id_)
+static uim_lisp
+get_result(uim_lisp id_)
 {
   MText *produced;
-  unsigned char *buf;
-  int id = get_c_int(id_);
+  char *commit_string;
+  int consumed;
+  int id = uim_scm_c_int(id_);
   MInputContext *ic = ic_array[id].mic;
-  LISP buf_;
+  uim_lisp  consumed_, commit_string_;
 
-  produced = mtext();
-  minput_lookup(ic, NULL, NULL, produced);
-  buf = convert_mtext2str(produced);
+  produced  = mtext();
+
+  consumed  = minput_lookup(ic, NULL, NULL, produced);
+
+  if (consumed == -1) {
+    consumed_ = uim_scm_f();
+  } else {
+    consumed_ = uim_scm_t();
+  }
+
+  commit_string = convert_mtext2str(produced);
   m17n_object_unref(produced);
-  buf_ = strcons(strlen(buf), buf);
-  free(buf);
-  return buf_;
+  commit_string_ = uim_scm_make_str(commit_string);
+  free(commit_string);
+
+  return uim_scm_cons(consumed_, commit_string_);
 }
 
-static LISP
-commit(LISP id_)
+static uim_lisp
+commit(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   MInputContext *ic = ic_array[id].mic;
 
 /* To avoid a bug of m17n-lib */
   ic->candidate_to   = 0;
-  return NIL;
+  return uim_scm_f();
 }
 
-static LISP
-candidate_showp(LISP id_)
+static uim_lisp
+candidate_showp(uim_lisp id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   MInputContext *ic = ic_array[id].mic;
 
   if (ic->candidate_show == 1) {
-    return siod_true_value();
+    return uim_scm_t();
   }
-  return NIL;
+  return uim_scm_f();
 }
 
-static LISP
-get_nr_candidates(LISP id_)
+static int
+calc_cands_num(int id)
 {
-  MPlist *group;
-  MPlist *elm;
   int result = 0;
-  int id = get_c_int(id_);
+  MPlist *group; 
   MInputContext *ic = ic_array[id].mic;
 
-  if(!ic || !ic->candidate_list)
-    return siod_false_value();
+  if (!ic || !ic->candidate_list)
+    return 0;
 
   group = ic->candidate_list;
 
-  while(mplist_value(group) != Mnil) {
-    if(mplist_key(group) == Mtext) {
+  while (mplist_value(group) != Mnil) {
+    if (mplist_key(group) == Mtext) {
       for (; mplist_key(group) != Mnil; group = mplist_next(group)) {
         result += mtext_len(mplist_value(group));
       }
@@ -643,66 +653,136 @@ get_nr_candidates(LISP id_)
       }
     }
   }
-
-  return intcons(result);
+  return result;
 }
 
-static LISP
-get_nth_candidate(LISP id_, LISP nth_)
+static void
+old_cands_free(char **old_cands)
+{
+  int i = 0;
+  if (old_cands) {
+    for (i=0; old_cands[i]; i++) {
+      free(old_cands[i]);
+    }
+    free(old_cands);
+  }
+}
+
+static uim_lisp
+fill_new_candidates(uim_lisp id_)
 {
   MText *produced = NULL; /* Quiet gcc */
   MPlist *group;
   MPlist *elm;
   int i;
-  unsigned char *buf = NULL; /* Quiet gcc */
-  int id = get_c_int(id_);
-  int nth = get_c_int(nth_);
+  char *buf = NULL; /* Quiet gcc */
+  int id = uim_scm_c_int(id_);
   MInputContext *ic = ic_array[id].mic;
-  LISP buf_;
+  int cands_num = calc_cands_num(id);
+  char **new_cands;
+  uim_lisp buf_;
 
-  if(!ic || !ic->candidate_list)
-    return NIL;
+  if (!ic || !ic->candidate_list)
+    return uim_scm_f();
 
   group = ic->candidate_list;
 
-  if(mplist_key(group) == Mtext) {
+  old_cands_free(ic_array[id].old_candidates);
+  ic_array[id].old_candidates = ic_array[id].new_candidates;
+
+  new_cands = malloc (cands_num * sizeof(char *) + 2);
+
+  if (mplist_key(group) == Mtext) {
+
     for (i=0; mplist_key(group) != Mnil; group = mplist_next(group)) {
       int j;
       for (j=0; j < mtext_len(mplist_value(group)); j++, i++) {
-        if(i == nth) {
           produced = mtext();
           mtext_cat_char(produced, mtext_ref_char(mplist_value(group), j));
-          buf = convert_mtext2str(produced);
+          new_cands[i] = convert_mtext2str(produced);
           m17n_object_unref(produced);
-        }
       }
     }
   } else {
     for (i=0; mplist_key(group) != Mnil; group = mplist_next(group)) {
+
       for (elm = mplist_value(group); mplist_key(elm) != Mnil; elm = mplist_next(elm),i++) {
-        if(i == nth) {
-          produced = mplist_value(elm);
-          buf = convert_mtext2str(produced);
-        }
+	produced = mplist_value(elm);
+	new_cands[i] = convert_mtext2str(produced);
       }
     }
   }
+  new_cands[i] = NULL;
 
-  if(!buf) {
-    return strcons(0, "");
+  ic_array[id].new_candidates = new_cands;
+  ic_array[id].nr_candidates = i;
+  
+  return uim_scm_t();
+
+  if (!buf) {
+    return uim_scm_make_str("");
   } else {
-    buf_ = strcons(strlen(buf), buf);
+    buf_ = uim_scm_make_str(buf);
     free(buf);
     return buf_;
   }
 }
 
-static LISP
-get_candidate_index(LISP id_)
+static uim_bool
+same_candidatesp(const char **old, const char **new)
 {
-  int id = get_c_int(id_);
+  int i;
+  if (!old)
+    return UIM_FALSE;
+
+  for (i=0; old[i] && new[i]; i++) {
+    if (strcmp(old[i], new[i]) != 0) {
+      return UIM_FALSE;
+    }
+  }
+  return UIM_TRUE;
+}
+
+static uim_lisp
+candidates_changedp(uim_lisp id_)
+{
+  int id = uim_scm_c_int(id_);
+  char **old_cands = ic_array[id].old_candidates;
+  char **new_cands = ic_array[id].new_candidates;
+
+  if (!same_candidatesp(old_cands, new_cands)) {
+    return uim_scm_t();
+  }
+  return uim_scm_f();
+}
+
+static uim_lisp
+get_nr_candidates(uim_lisp id_)
+{
+  int id = uim_scm_c_int(id_);
+  return uim_scm_make_int(calc_cands_num(id));
+}
+
+static uim_lisp
+get_nth_candidate(uim_lisp id_, uim_lisp nth_)
+{
+  int id = uim_scm_c_int(id_);
+  int nth = uim_scm_c_int(nth_);
+  int nr = ic_array[id].nr_candidates;
+  
+  if (nr >= nth) {
+    return uim_scm_make_str(ic_array[id].new_candidates[nth]);
+  } else {
+    return uim_scm_make_str("");
+  }
+}
+
+static uim_lisp
+get_candidate_index(uim_lisp id_)
+{
+  int id = uim_scm_c_int(id_);
   MInputContext *ic = ic_array[id].mic;
-  return intcons(ic->candidate_index);
+  return uim_scm_make_int(ic->candidate_index);
 }
 
 /* Utility function */
@@ -717,29 +797,44 @@ m17nlib_utf8_find_next_char(const char *p)
 }
 
 void
-uim_init_m17nlib(void)
+uim_plugin_instance_init(void)
 {
-  init_subr_0("m17nlib-lib-init", init_m17nlib);
-  init_subr_0("m17nlib-lib-nr-input-methods", get_nr_input_methods);
-  init_subr_1("m17nlib-lib-nth-input-method-lang", get_input_method_lang);
-  init_subr_1("m17nlib-lib-nth-input-method-name", get_input_method_name);
-  init_subr_1("m17nlib-lib-alloc-context", alloc_id);
-  init_subr_1("m17nlib-lib-free-context", free_id);
-  init_subr_3("m17nlib-lib-push-key", push_key);
-  init_subr_2("m17nlib-lib-push-symbol-key", push_symbol_key);
-  init_subr_1("m17nlib-lib-compose-mode?", compose_modep);
-  init_subr_1("m17nlib-lib-preedit-changed?", preedit_changedp);
-  init_subr_1("m17nlib-lib-get-left-of-cursor",     get_left_of_cursor);
-  init_subr_1("m17nlib-lib-get-right-of-cursor",    get_right_of_cursor);
-  init_subr_1("m17nlib-lib-get-left-of-candidate",  get_left_of_candidate);
-  init_subr_1("m17nlib-lib-get-selected-candidate", get_selected_candidate);
-  init_subr_1("m17nlib-lib-get-right-of-candidate", get_right_of_candidate);
-  init_subr_1("m17nlib-lib-get-commit-string", get_commit_string);
-  init_subr_1("m17nlib-lib-commit", commit);
-  init_subr_1("m17nlib-lib-candidate-show?", candidate_showp);
-  init_subr_1("m17nlib-lib-get-nr-candidates", get_nr_candidates);
-  init_subr_2("m17nlib-lib-get-nth-candidate", get_nth_candidate);
-  init_subr_1("m17nlib-lib-get-candidate-index", get_candidate_index);
+  uim_scm_init_subr_0("m17nlib-lib-init", init_m17nlib);
+  uim_scm_init_subr_0("m17nlib-lib-nr-input-methods", get_nr_input_methods);
+  uim_scm_init_subr_1("m17nlib-lib-nth-input-method-lang", get_input_method_lang);
+  uim_scm_init_subr_1("m17nlib-lib-nth-input-method-name", get_input_method_name);
+  uim_scm_init_subr_1("m17nlib-lib-alloc-context", alloc_id);
+  uim_scm_init_subr_1("m17nlib-lib-free-context", free_id);
+  uim_scm_init_subr_3("m17nlib-lib-push-key", push_key);
+  uim_scm_init_subr_2("m17nlib-lib-push-symbol-key", push_symbol_key);
+  uim_scm_init_subr_1("m17nlib-lib-compose-mode?", compose_modep);
+  uim_scm_init_subr_1("m17nlib-lib-preedit-changed?", preedit_changedp);
+  uim_scm_init_subr_1("m17nlib-lib-get-left-of-cursor",     get_left_of_cursor);
+  uim_scm_init_subr_1("m17nlib-lib-get-right-of-cursor",    get_right_of_cursor);
+  uim_scm_init_subr_1("m17nlib-lib-get-left-of-candidate",  get_left_of_candidate);
+  uim_scm_init_subr_1("m17nlib-lib-get-selected-candidate", get_selected_candidate);
+  uim_scm_init_subr_1("m17nlib-lib-get-right-of-candidate", get_right_of_candidate);
+  uim_scm_init_subr_1("m17nlib-lib-get-result", get_result);
+  uim_scm_init_subr_1("m17nlib-lib-commit", commit);
+  uim_scm_init_subr_1("m17nlib-lib-candidate-show?", candidate_showp);
+  uim_scm_init_subr_1("m17nlib-lib-fill-new-candidates!", fill_new_candidates);
+  uim_scm_init_subr_1("m17nlib-lib-candidates-changed?", candidates_changedp);
+  uim_scm_init_subr_1("m17nlib-lib-get-nr-candidates", get_nr_candidates);
+  uim_scm_init_subr_2("m17nlib-lib-get-nth-candidate", get_nth_candidate);
+  uim_scm_init_subr_1("m17nlib-lib-get-candidate-index", get_candidate_index);
 }
 
-#endif /* HAVE_M17NLIB */
+void
+uim_plugin_instance_quit(void)
+{
+  if (converter) {
+    mconv_free_converter(converter);
+    converter = NULL;
+  }
+  if (m17nlib_ok) {
+    M17N_FINI();
+    m17nlib_ok = 0;
+    free(im_array);
+    free(ic_array);
+  }
+}
