@@ -17,12 +17,22 @@
  *  (2) C context
  *       No scheme context on the stack
  * Yusuke.
+ *
+ * The nested Scheme evaluation feature (NESTED_REPL_C_STRING) will
+ * remove this limitation. The feature will be enabled by default once
+ * tested enough.  -- YamaKen 2004-12-31
  */
 #ifndef __SIOD_H__
 #define __SIOD_H__
 
+#include "config.h"
+
 #include <stdio.h>
 
+#ifndef NESTED_REPL_C_STRING
+#define NESTED_REPL_C_STRING 1
+#endif
+
 struct obj
   {
     short gc_mark;
@@ -114,6 +124,16 @@ struct obj
 	    char *name;
 	  }
 	c_file;
+	struct
+	  {
+	    void *data;
+	  }
+	c_pointer;
+	struct
+	  {
+	    void (*func)(void);
+	  }
+	c_func_pointer;
       }
     storage_as;
 #if DEBUG_SCM
@@ -169,107 +189,30 @@ struct obj
 #define tc_subr_5 20
 #define tc_subr_2n 21
 #define tc_user_min 50
+#define tc_c_pointer 50
+#define tc_c_func_pointer 51
 #define tc_user_max 100
 
 #define tc_table_dim 100
 
 typedef struct obj *LISP;
 typedef LISP (*SUBR_FUNC) (void);
+typedef void (*C_FUNC) (void);
 
 #define CONSP(x)   TYPEP(x,tc_cons)
 #define INTNUMP(x) TYPEP(x,tc_intnum)
 #define SYMBOLP(x) TYPEP(x,tc_symbol)
 #define STRINGP(x) TYPEP(x,tc_string)
+#define POINTERP(x) TYPEP(x,tc_c_pointer)
+#define FPOINTERP(x) TYPEP(x,tc_c_func_pointer)
 
 #define NCONSP(x)   NTYPEP(x,tc_cons)
 #define NINTNUMP(x) NTYPEP(x,tc_intnum)
 #define NSYMBOLP(x) NTYPEP(x,tc_symbol)
 #define NSTRINGP(x) NTYPEP(x,tc_string)
+#define NPOINTERP(x) NTYPEP(x,tc_c_pointer)
+#define NFPOINTERP(x) NTYPEP(x,tc_c_func_pointer)
 
 #define TKBUFFERN 5120
 
-void siod_init (int argc, char **argv, int warnflag, FILE *);
-void siod_quit (void);
-
-void set_repl_hooks (void (*puts_f) (char *),
-		     LISP (*read_f) (void),
-		     LISP (*eval_f) (LISP),
-		     void (*print_f) (LISP));
-char *get_c_string (LISP x);
-char *get_c_string_dim (LISP x, long *);
-int get_c_int (LISP x);
-long nlength(LISP x);
-
-LISP cons (LISP x, LISP y);
-LISP car (LISP x);
-LISP cdr (LISP x);
-LISP setcar (LISP cell, LISP value);
-LISP intcons (int x);
-LISP eql (LISP x, LISP y);
-LISP symcons (char *pname, LISP vcell);
-LISP symbol_boundp (LISP x, LISP env);
-LISP symbol_value (LISP x, LISP env);
-LISP symbol_to_string (LISP x, LISP env);
-LISP rintern (const char *name);
-LISP closure (LISP env, LISP code);
-
-void init_subr (char *name, long type, SUBR_FUNC fcn);
-void init_subr_0 (char *name, LISP (*fcn) (void));
-void init_subr_1 (char *name, LISP (*fcn) (LISP));
-void init_subr_2 (char *name, LISP (*fcn) (LISP, LISP));
-void init_subr_2n (char *name, LISP (*fcn) (LISP, LISP));
-void init_subr_3 (char *name, LISP (*fcn) (LISP, LISP, LISP));
-void init_subr_4 (char *name, LISP (*fcn) (LISP, LISP, LISP, LISP));
-void init_subr_5 (char *name, LISP (*fcn) (LISP, LISP, LISP, LISP, LISP));
-void init_lsubr (char *name, LISP (*fcn) (LISP));
-void init_fsubr (char *name, LISP (*fcn) (LISP, LISP));
-void init_msubr (char *name, LISP (*fcn) (LISP *, LISP *));
-
-LISP delq (LISP elem, LISP l);
-void set_eval_hooks (long type, LISP (*fcn) (LISP, LISP *, LISP *));
-LISP leval (LISP x, LISP env);
-LISP symbolconc (LISP args);
-LISP lprin1f (LISP exp, FILE * f);
-LISP lread (LISP);
-LISP lreadtk (char *, long j);
-LISP lreadf (FILE * f);
-LISP require (LISP fname);
-LISP strcons (long length, const char *data);
-LISP equal (LISP, LISP);
-void set_fatal_exit_hook (void (*fcn) (void));
-LISP intern (LISP x);
-long repl_c_string (char *, long want_init, long want_print);
-LISP reverse (LISP);
-LISP nreverse (LISP);
-LISP number2string (LISP, LISP, LISP, LISP);
-LISP string2number (LISP, LISP);
-LISP cadr (LISP);
-LISP caar (LISP);
-LISP cdar (LISP);
-LISP cddr (LISP);
-LISP caaar (LISP);
-LISP caadr (LISP);
-LISP cadar (LISP);
-LISP caddr (LISP);
-LISP cdaar (LISP);
-LISP cdadr (LISP);
-LISP cddar (LISP);
-LISP cdddr (LISP);
-LISP siod_true_value (void);
-LISP siod_false_value (void);
-LISP lapply (LISP fcn, LISP args);
-LISP listn (long n,...);
-char *must_malloc (unsigned long size);
-FILE *get_c_file (LISP p, FILE * deflt);
-char *last_c_errmsg (int);
-LISP llast_c_errmsg (int);
-void siod_c_provide(const char *);
-
-LISP funcall1 (LISP, LISP);
-LISP funcall2 (LISP, LISP, LISP);
-LISP apply1 (LISP, LISP, LISP);
-
-extern long siod_verbose_level;
-void siod_set_lib_path(const char *);
-
 #endif /* __SIOD_H__ */
