@@ -1,7 +1,6 @@
-
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -18,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -40,23 +39,24 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <errno.h>
 #include <sys/stat.h>
 #include <unistd.h>
 #include "uim.h"
 #include "uim-helper.h"
 #include "context.h"
 
+
+#define RECV_BUFFER_SIZE 1024
+
+/*Common buffer for some functions's temporary buffer.
+  Pay attention for use.*/
+static char uim_recv_buf[RECV_BUFFER_SIZE];
+static char *uim_read_buf;
+
 static int uim_fd = -1;
 static void (*uim_disconnect_cb)(void);
 
-static void uim_helper_client_focus(uim_context uc, int flg);
-
-#define BUFFER_SIZE (32 * 1024)
-/*Common buffer for some functions's temporary buffer.
-  Pay attention for use.*/
-static char uim_help_buf[BUFFER_SIZE];
-static int uim_read_buf_size;
-static char *uim_read_buf;
 
 static char *
 get_server_command(void)
@@ -64,7 +64,6 @@ get_server_command(void)
   return "uim-helper-server";
 }
 
-
 int uim_helper_init_client_fd(void (*disconnect_cb)(void))
 {
   int fd;
@@ -73,7 +72,7 @@ int uim_helper_init_client_fd(void (*disconnect_cb)(void))
   
   uim_fd = -1;
   
-  if(!path)
+  if (!path)
     return -1;
 
   bzero(&server, sizeof(server));
@@ -97,23 +96,23 @@ int uim_helper_init_client_fd(void (*disconnect_cb)(void))
 #endif
 
 
-  if(connect(fd, (struct sockaddr *)&server,sizeof(server)) == -1){
+  if (connect(fd, (struct sockaddr *)&server,sizeof(server)) == -1) {
     int serv_pid = 0;
     FILE *serv_r = NULL, *serv_w = NULL;
     char buf[128];
     
     serv_pid = uim_ipc_open_command(serv_pid, &serv_r, &serv_w, get_server_command());
     
-    if(serv_pid == 0) {
+    if (serv_pid == 0) {
       return -1;
     }
     
     while (fgets (buf, sizeof(buf), serv_r ) != NULL ) {
-      if(strcmp( buf, "\n" ) == 0)
+      if (strcmp( buf, "\n" ) == 0)
 	break;
     }
     
-    if(connect(fd, (struct sockaddr *)&server,sizeof(server)) == -1){
+    if (connect(fd, (struct sockaddr *)&server,sizeof(server)) == -1) {
       return -1;
     }
   }
@@ -145,90 +144,41 @@ uim_helper_close_client_fd(int fd)
 void
 uim_helper_client_focus_in(uim_context uc)
 {
-  uim_helper_client_focus(uc, 0);
+  if (uc)
+    uim_helper_send_message(uim_fd, "focus_in\n");
 }
 
 void
 uim_helper_client_focus_out(uim_context uc)
 {
-  uim_helper_client_focus(uc, 1);
+  if (uc)
+    uim_helper_send_message(uim_fd, "focus_out\n");
 }
 
-static void
-uim_helper_client_focus(uim_context uc, int flg)
-{
-  if(uim_fd < 0)
-    return;
-
-  if(!uc)
-    return;
-
-  if(flg == 0)
-    snprintf(uim_help_buf, BUFFER_SIZE, "focus_in\n");
-  else
-    snprintf(uim_help_buf, BUFFER_SIZE, "focus_out\n");
-
-  uim_helper_send_message(uim_fd, uim_help_buf);
-}
-
-
-
 void
 uim_helper_client_get_prop_list(void)
 {
-  snprintf(uim_help_buf, BUFFER_SIZE, "prop_list_get\n");
-  uim_helper_send_message(uim_fd, uim_help_buf);
+  uim_helper_send_message(uim_fd, "prop_list_get\n");
 }
 
 void
 uim_helper_read_proc(int fd)
 {
-  char buf[BUFFER_SIZE];
   int rc;
 
-  while(uim_helper_fd_readable(fd) > 0) {
-    
-    rc = read(fd, buf, sizeof(buf)-1);
-    buf[rc] = '\0';
-    
-    if(rc == 0) {
-      if (uim_disconnect_cb) {
-	uim_disconnect_cb();
-      }
-      uim_fd = -1;
+  while (uim_helper_fd_readable(fd) > 0) {
+    rc = read(fd, uim_recv_buf, sizeof(uim_recv_buf));
+    if (rc == 0 || (rc < 0 && errno != EAGAIN)) {
+      uim_helper_close_client_fd(fd);
       return;
+    } else if (rc > 0) {
+      uim_read_buf = uim_helper_buffer_append(uim_read_buf, uim_recv_buf, rc);
     }
-    uim_read_buf = (char *)realloc(uim_read_buf, strlen(uim_read_buf) + strlen(buf)+1);
-    strcat(uim_read_buf, buf);
   }
-  uim_read_buf_size = strlen(uim_read_buf);
-  return;
-}
-
-static void
-shift_read_buffer(int count)
-{
-  memmove(uim_read_buf, &uim_read_buf[count],
-	  uim_read_buf_size - count);
-  uim_read_buf_size -= count;
-  uim_read_buf[uim_read_buf_size] = '\0';
 }
 
 char *
 uim_helper_get_message(void)
 {
-  int i;
-  char *buf;
-
-  for (i = 0; i < uim_read_buf_size - 1; i++) {
-    if (uim_read_buf[i] == '\n' &&
-	uim_read_buf[i+1] == '\n') {
-      buf = (char *)malloc(i+2);
-      memcpy(buf, uim_read_buf, i+1);
-      buf[i+1] = '\0';
-      shift_read_buffer(i+2);
-      return buf;
-    }
-  }
-  return NULL;
+  return uim_helper_buffer_get_message(uim_read_buf);
 }
