@@ -1,7 +1,7 @@
 /*
   uim-helper-server.c: This file will be renamed uim-helper-messagebus.c later.
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -18,10 +18,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -33,6 +33,7 @@
 */
 
 #include <sys/types.h>
+#include <pwd.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h>
@@ -43,9 +44,18 @@
 #include <errno.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <signal.h>
+#include <fcntl.h>
 #include "uim.h"
 #include "uim-helper.h"
 
+
+struct client {
+  int fd;
+  char *rbuf;
+  char *wbuf;
+};
+
 #define MAX_CLIENT 32
 #define BUFFER_SIZE 1024
 
@@ -54,11 +64,14 @@
   (sizeof(*(su)) - sizeof((su)->sun_path) + strlen((su)->sun_path))
 #endif
 
+static fd_set s_fdset_read;
+static fd_set s_fdset_write;
+static int s_max_fd;
+
 static int nr_client_slots;
-static struct client {
-  int fd;
-  char *rbuf;
-} *clients;
+static struct client *clients;
+
+static char read_buf[BUFFER_SIZE];
 
 /*
   prepare file descriptor.
@@ -68,7 +81,10 @@ init_serv_fd(char *path)
 {
   int foo;
   int fd;
+  int flag;
   struct sockaddr_un myhost;
+  struct passwd *pw;
+  char *logname;
 
   fd = socket(PF_UNIX, SOCK_STREAM, 0);
   if (fd < 0) {
@@ -88,17 +104,37 @@ init_serv_fd(char *path)
 
   chmod(path, S_IRUSR|S_IWUSR);
 
+  logname = getenv("LOGNAME");
+  if (logname) {
+    pw = getpwnam(logname);
+    if (pw)
+      chown(path, pw->pw_uid, -1);
+  }
+
+  if ((flag = fcntl(fd, F_GETFL)) == -1) {
+    close(fd);
+    return -1;
+  }
+
+  flag |= O_NONBLOCK;
+  if (fcntl(fd, F_SETFL, flag) == -1) {
+    close(fd);
+    return -1;
+  }
+
   foo = listen(fd, 5);
   if (foo == -1) {
     perror("failed in listen()");
     return -1;
   }
   /*  fprintf(stderr,"listen at %s\n",path);*/
+  FD_SET(fd, &s_fdset_read);
+  s_max_fd = fd;
   return fd;
 }
 
 static struct client *
-get_unused_client()
+get_unused_client(void)
 {
   int i;
   for (i = 0; i < nr_client_slots; i++) {
@@ -106,9 +142,10 @@ get_unused_client()
       return &clients[i];
     }
   }
-  nr_client_slots ++;
+  nr_client_slots++;
   clients = realloc(clients, sizeof(struct client) * nr_client_slots);
   clients[nr_client_slots - 1].rbuf = strdup("");
+  clients[nr_client_slots - 1].wbuf = strdup("");
   return &clients[nr_client_slots - 1];
 }
 
@@ -119,83 +156,83 @@ free_client(struct client *cl)
     free(cl->rbuf);
     cl->rbuf = strdup("");
   }
+  if (cl->wbuf) {
+    free(cl->wbuf);
+    cl->wbuf = strdup("");
+  }
   cl->fd = -1;
 }
 
-
 static void
-parse_content(char *content, struct client *cl)
+distribute_message(char *msg, struct client *cl)
 {
   int i;
+  size_t msg_len;
 
-  for (i = 0; i < nr_client_slots; i ++){
-    if(clients[i].fd != -1 && 
-       clients[i].fd != cl->fd && 
-       uim_helper_fd_writable(clients[i].fd)) {
-      write(clients[i].fd, content, strlen(content));
+  msg_len = strlen(msg);
+
+  for (i = 0; i < nr_client_slots; i++) {
+    if (clients[i].fd != -1 && clients[i].fd != cl->fd) {
+      clients[i].wbuf = uim_helper_buffer_append(clients[i].wbuf, msg, msg_len);
+      FD_SET(clients[i].fd, &s_fdset_write);
     }
   }
 }
 
 static int
-proc_func(struct client *cl)
+reflect_message_fragment(struct client *cl)
 {
   int rc;
-  char buf[BUFFER_SIZE];
+  char *msg;
 
   /* do read */
-  rc = read(cl->fd, buf, BUFFER_SIZE - 1);
-  if (rc == 0) {
-    close(cl->fd);
+  rc = read(cl->fd, read_buf, sizeof(read_buf));
+  if (rc <= 0) {
+    if (rc < 0 && (errno == EAGAIN || errno == EINTR))
+      return 0;
     return -1;
   }
 
-  if (rc < 0) {
-    return -1;
+  cl->rbuf = uim_helper_buffer_append(cl->rbuf, read_buf, rc);
+
+  while ((msg = uim_helper_buffer_get_message(cl->rbuf))) {
+    distribute_message(msg, cl);
+    free(msg);
   }
-
-  buf[rc] = '\0';
-
-  cl->rbuf = (char *)realloc(cl->rbuf, strlen(cl->rbuf) + strlen(buf)+1);
-  strcat(cl->rbuf, buf);
-
-  if (uim_helper_str_terminated(cl->rbuf)) {
-    /* process */
-    parse_content(cl->rbuf, cl);
-    free(cl->rbuf);
-    cl->rbuf = strdup("");
-  }
-
   return 1;
 }
 
+static uim_bool
+check_session_alive(void)
+{
+  /* If there's no connection, we can assume user logged out. */
+  int i;
+  for (i = 0; i < nr_client_slots; i++) {
+    if (clients[i].fd != -1) {
+      return UIM_TRUE;
+    }
+  }
+  return UIM_FALSE; /* User already logged out */
+}
+
+
+/* FIXME: This function is too long to read... */
 static void
 uim_helper_server_process_connection(int serv_fd)
 {
   int i;
-  int fd_biggest = 0;
   fd_set readfds;
-
-  fd_biggest = serv_fd;
+  fd_set writefds;
 
   while (1) {
-    /* setup readfds */
-    FD_ZERO(&readfds);
-    FD_SET(serv_fd, &readfds);  
-    for (i = 0; i < nr_client_slots; i ++) {
-      int fd = clients[i].fd;
-      if (fd == -1) {
-	continue;
-      }
-      FD_SET(fd, &readfds);
-      if (fd > fd_biggest) {
-	fd_biggest = fd;
-      }
-    }
+    memcpy(&readfds, &s_fdset_read, sizeof(fd_set));
+    memcpy(&writefds, &s_fdset_write, sizeof(fd_set));
 
     /* call select(), waiting until a file descriptor readable */
-    if (select(fd_biggest+1, &readfds, NULL, NULL, NULL) < 0) {
-      perror("select faild");
+    if (select(s_max_fd + 1, &readfds, &writefds, NULL, NULL) <= 0) {
+      perror("uim-helper_server select(2) failed");
+      sleep(3);
+      continue;
     }
 
     /* for accept new connection */
@@ -203,14 +240,26 @@ uim_helper_server_process_connection(int serv_fd)
       struct sockaddr_un clientsoc;
       socklen_t len = sizeof(clientsoc);
       int new_fd;
+      int flag;
       struct client *cl;
       new_fd = accept(serv_fd, (struct sockaddr *)&clientsoc, &len);
 
-      if(new_fd < 0) {
+      if (new_fd < 0) {
 	perror("accpet failed");
 	continue;
       }
 
+      if ((flag = fcntl(new_fd, F_GETFL)) == -1) {
+	close(new_fd);
+	continue;
+      }
+
+      flag |= O_NONBLOCK;
+      if (fcntl(new_fd, F_SETFL, flag) == -1) {
+	close(new_fd);
+	continue;
+      }
+
       cl = get_unused_client();
       if (!cl) {
 	close(new_fd);
@@ -223,22 +272,70 @@ uim_helper_server_process_connection(int serv_fd)
 	write(cl->fd, buf, 1);
       }
 #endif
-    }    
-    
-    /* check data from clients reached */
-    for (i = 0; i < nr_client_slots; i ++) {
-      if (clients[i].fd != -1 &&
-	  FD_ISSET(clients[i].fd, &readfds)) {
-	int result;
-	/* actual process */
-	result = proc_func(&clients[i]);
+      FD_SET(cl->fd, &s_fdset_read);
+      if (cl->fd > s_max_fd)
+	s_max_fd = cl->fd;
+    } else {
+      /* check data to write and from clients reached */
+      for (i = 0; i < nr_client_slots; i++) {
+	if (clients[i].fd != -1 && FD_ISSET(clients[i].fd, &writefds)) {
+	  int ret, message_len, out_len;
+	  char *out;
 
-	if (result < 0) {
-	  close(clients[i].fd);
-	  free_client(&clients[i]);
+	  out = clients[i].wbuf;
+	  message_len = out_len = strlen(clients[i].wbuf);
+	  while (out_len > 0) {
+	    if ((ret = write(clients[i].fd, out, out_len)) < 0) {
+	      if (errno == EAGAIN) {
+#if 0
+		fprintf(stderr, "EAGAIN: fd = %d\n", clients[i].fd);
+#endif
+	      } else {
+		perror("uim-helper_server write(2) failed");
+		if (errno == EPIPE) {
+		  fprintf(stderr, "fd = %d\n", clients[i].fd);
+		  FD_CLR(clients[i].fd, &s_fdset_read);
+		  FD_CLR(clients[i].fd, &s_fdset_write);
+		  if (clients[i].fd == s_max_fd)
+		    s_max_fd--;
+		  close(clients[i].fd);
+		  free_client(&clients[i]);
+		}
+	      }
+	      break;
+	    } else {
+	      out += ret;
+	      out_len -= ret;
+	    }
+	  }
+	  if (out_len == 0) {
+	    free(clients[i].wbuf);
+	    clients[i].wbuf = strdup("");
+	    FD_CLR(clients[i].fd, &s_fdset_write);
+	  } else {
+	    uim_helper_buffer_shift(clients[i].wbuf, message_len - out_len);
+	  }
+	}
+	if (clients[i].fd != -1 && FD_ISSET(clients[i].fd, &readfds)) {
+	  int result;
+	  /* actual process */
+	  result = reflect_message_fragment(&clients[i]);
+
+	  if (result < 0) {
+	    FD_CLR(clients[i].fd, &s_fdset_read);
+	    FD_CLR(clients[i].fd, &s_fdset_write);
+	    if (clients[i].fd == s_max_fd)
+	      s_max_fd--;
+	    close(clients[i].fd);
+	    free_client(&clients[i]);
+	  }
 	}
       }
     }
+
+    if(check_session_alive() == UIM_FALSE) {
+      return;
+    }
   }
 }
 
@@ -252,6 +349,9 @@ main(int argc, char **argv)
   clients = NULL;
   nr_client_slots = 0;
 
+  FD_ZERO(&s_fdset_read);
+  FD_ZERO(&s_fdset_write);
+  s_max_fd = 0;
   serv_fd = init_serv_fd(path);
 
   printf("waiting\n\n");
@@ -267,6 +367,8 @@ main(int argc, char **argv)
 
   free(path);
 
+  signal(SIGPIPE, SIG_IGN);
+
   uim_helper_server_process_connection(serv_fd);
 
   return 0;
