@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,6 +31,8 @@
 
 */
 
+#include "config.h"
+
 #include <unistd.h>
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -42,10 +44,15 @@
 #include <stdio.h>
 #include <string.h>
 #include <pwd.h>
+#include <signal.h>
 #include <errno.h>
 #include "context.h"
 #include "uim-helper.h"
 
+#ifndef HAVE_SIG_T
+typedef void (*sig_t)(int);
+#endif
+
 enum RorW
   {
     READ,
@@ -59,17 +66,17 @@ uim_helper_fd(int fd, enum RorW rw)
     fd_set fds;
     struct timeval tv;
 
-    /*   if(!fd || fd < 0)
+    /*   if (!fd || fd < 0)
 	 return -1;*/
 
     FD_ZERO(&fds);
     FD_SET(fd, &fds);
     tv.tv_sec = tv.tv_usec = 0;
 
-    if(rw == READ)
-      rc = select(fd+1, &fds, NULL, NULL, &tv);
+    if (rw == READ)
+      rc = select(fd + 1 , &fds, NULL, NULL, &tv);
     else    
-      rc = select(fd+1, NULL, &fds, NULL, &tv);
+      rc = select(fd + 1, NULL, &fds, NULL, &tv);
 
     if (rc < 0)
       return -1;
@@ -81,24 +88,34 @@ void
 uim_helper_send_message(int fd, const char *message)
 {
   int res;
+  int len, out_len;
+  sig_t old_sigpipe;
+  char *buf, *bufp;
 
-  if (fd < 0)
+  if (fd < 0 || !message)
     return;
 
-  if (!message)
-    return;
+  len = strlen(message) + 1;
+  buf = malloc(len + 1);
+  snprintf(buf, len + 1, "%s\n", message);
 
-  /* readable and cannot read any character, means disconnected.
-     so we should read here and proc such condition. */
+  old_sigpipe = signal(SIGPIPE, SIG_IGN);
 
-  if (uim_helper_fd(fd, WRITE)) {
-    int len = strlen(message);
-    char *buf = malloc(len + 2);
-    snprintf(buf, len + 2,"%s\n", message);
-    res = write(fd, buf, len + 1);
-    free(buf);
+  out_len = len;
+  bufp = buf;
+  while (out_len > 0) {
+    if ((res = write(fd, bufp, out_len)) < 0) {
+      if (errno == EAGAIN || errno == EINTR)
+	continue;
+      fprintf(stderr, "uim_helper_send_message(): unknown error\n");
+      break;
+    }
+
+    bufp += res;
+    out_len -= res;
   }
-
+  free(buf);
+  signal(SIGPIPE, old_sigpipe);
   return;
 }
 
@@ -109,7 +126,7 @@ uim_helper_get_pathname(void)
   char *login = NULL;
   struct passwd *pw = NULL;
  
-  if(is_setugid() == 0) {
+  if (is_setugid() == 0) {
     login = getenv("LOGNAME");
   }
 
@@ -139,17 +156,59 @@ int uim_helper_fd_writable(int fd)
 
 int uim_helper_str_terminated(const char *str)
 {
-  if(!str)
+  if (!str)
     return 0;
 
-  if(strlen(str) > 2&&
-     str[strlen(str)-1] == '\n' &&
-     str[strlen(str)-2] == '\n' )
+  if (strlen(str) > 2 &&
+     str[strlen(str) - 1] == '\n' &&
+     str[strlen(str) - 2] == '\n')
     return 1;
 
   return 0;
 }
 
+char *
+uim_helper_buffer_append(char *buf, const char *fragment, size_t fragment_size)
+{
+  size_t buf_size, extended_size;
+
+  buf_size = strlen(buf);
+  extended_size = buf_size + fragment_size + 1;
+  buf = (char *)realloc(buf, extended_size);
+  if (buf) {
+    memcpy(&buf[buf_size], fragment, fragment_size);
+    buf[extended_size - 1] = '\0';
+  }
+
+  return buf;
+}
+
+void
+uim_helper_buffer_shift(char *buf, int count)
+{
+  int len = strlen(buf);
+  memmove(buf, &buf[count], len - count);
+  buf[len - count] = '\0';
+}
+
+char *
+uim_helper_buffer_get_message(char *buf)
+{
+  size_t msg_size;
+  char *msg, *msg_term;
+
+  msg_term = strstr(buf, "\n\n");
+  if (msg_term) {
+    msg_size = msg_term + 2 - buf;
+    msg = (char *)malloc(msg_size + 1);
+    memcpy(msg, buf, msg_size);
+    msg[msg_size] = '\0';
+    uim_helper_buffer_shift(buf, msg_size);
+    return msg;
+  }
+  return NULL;
+}
+
 int
 is_setugid(void)
 {
