@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -35,15 +35,17 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "uim.h"
+#include "uim-scm.h"
+#include "uim-compat-scm.h"
 #include "context.h"
 
-/* valid-key-symbols in key.scm has also to be updated if you add new
- * key definition to key_tab
- */
-static struct {
+struct key_entry {
   int key;
-  char *str;
-} key_tab[] = {
+  const char *str;
+};
+
+static struct key_entry key_tab[] = {
   {UKey_Backspace, "backspace"},
   {UKey_Delete, "delete"},
   {UKey_Escape, "escape"},
@@ -57,6 +59,7 @@ static struct {
   {UKey_Next, "next"},
   {UKey_Home, "home"},
   {UKey_End, "end"},
+  {UKey_Insert, "insert"},
   {UKey_Zenkaku_Hankaku, "zenkaku-hankaku"},
   {UKey_Multi_key, "Multi_key"},
   {UKey_Mode_switch, "Mode_switch"},
@@ -137,10 +140,13 @@ static struct {
   {0,0}
 };
 
+#if 0
+int uim_key_sym_to_int(uim_lisp sym);
+
 int
-uim_key_sym_to_int(LISP sym_)
+uim_key_sym_to_int(uim_lisp sym_)
 {
-  char *sym = uim_get_c_string(sym_);
+  char *sym = uim_scm_refer_c_str(sym_);
   int res = 0;
   int i;
   for (i = 0; key_tab[i].key; i++) {
@@ -148,15 +154,28 @@ uim_key_sym_to_int(LISP sym_)
       res = key_tab[i].key;
     }
   }
-  free(sym);
   return res;
 }
+#endif
 
-static char *
+static void
+define_valid_key_symbols(void)
+{
+  int i;
+
+  UIM_EVAL_STRING(NULL, "(define valid-key-symbols ())");
+  for (i = 0; key_tab[i].key; i++) {
+    UIM_EVAL_FSTRING1(NULL,
+		      "(set! valid-key-symbols (cons '%s valid-key-symbols))",
+		      key_tab[i].str);
+  }
+}
+
+static const char *
 get_sym(int key)
 {
   int i;
-  char *res = NULL;
+  const char *res = NULL;
   for (i = 0; key_tab[i].key; i++) {
     if (key_tab[i].key == key) {
       res = key_tab[i].str;
@@ -168,7 +187,7 @@ get_sym(int key)
 static int
 keycode_to_sym(int key, char *buf)
 {
-  char *s = get_sym(key);
+  char *s = (char *)get_sym(key);
   if (!s) {
     if (key > 128) {
       return -1;
@@ -181,7 +200,7 @@ keycode_to_sym(int key, char *buf)
 }
 
 static void
-handle_key(uim_context uc, char *p, int key, int state)
+handle_key(uim_context uc, const char *p, int key, int state)
 {
   char keybuf[20];
   int rv;
@@ -236,21 +255,23 @@ uim_release_key(uim_context uc, int key, int state)
   return uc->commit_raw_flag;
 }
 
-static LISP
-define_key(LISP args, LISP env)
+static uim_lisp
+define_key(uim_lisp args, uim_lisp env)
 {
-  LISP define_key_internal, predicate_sym, sources;
+  uim_lisp form, predicate_sym, sources;
 
-  define_key_internal = leval(rintern("define-key-internal"), env);
-  predicate_sym = car(args);
-  sources = NULLP(args) ? NIL : car(CDR(args));
+  predicate_sym = uim_scm_car(args);
+  sources = uim_scm_nullp(args) ? uim_scm_null_list() : uim_scm_cadr(args);
+  form = uim_scm_list3(uim_scm_make_symbol("define-key-internal"),
+		       uim_scm_quote(predicate_sym),
+		       sources);
 
-  sources = leval(sources, env);
-  return funcall2(define_key_internal, predicate_sym, sources);
+  return uim_scm_eval(form);
 }
 
 void
 uim_init_key_subrs(void)
 {
-  init_fsubr("define-key",  define_key);
+  define_valid_key_symbols();
+  uim_scm_init_fsubr("define-key", define_key);
 }
