@@ -2,7 +2,7 @@
   uim-sh.c: uim interactive shell for debugging, batch processing and
             serving as generic inferior process
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -19,10 +19,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -35,6 +35,7 @@
 
 #include "uim.h"
 #include "uim-scm.h"
+#include "uim-compat-scm.h"
 
 #ifdef UIM_SH_USE_EXIT_HOOK
 extern int uim_siod_fatal;
@@ -42,6 +43,7 @@ extern int uim_siod_fatal;
 
 int
 main(int argc, char *argv[]) {
+  long verbose;
   uim_lisp args, form;
 
   uim_scm_set_output(stdout);
@@ -49,21 +51,24 @@ main(int argc, char *argv[]) {
   /* TODO: be able to suppress ordinary initialization process */
   uim_init();
 
+  verbose = uim_scm_get_verbose_level();
   uim_scm_set_verbose_level(1);
-  uim_scm_load_file("uim-sh.scm");
+  uim_scm_require_file("uim-sh.scm");
 #ifdef UIM_SH_USE_EXIT_HOOK
-  /* is not working even if uim_siod_fatal is accessible. outermost
-   * *catch affects me?
-   */
+  /*
+     is not working even if uim_siod_fatal is accessible. outermost
+     *catch affects me?
+  */
   if (uim_siod_fatal)
     return 1;
 #endif
 
-  /* keep verbose level 1 to make (require "foo.scm") returns
-   * *foo.scm-loaded* only. This is required to run uim-test-utils.scm
-   * correctly.
-   */
-  /* uim_scm_set_verbose_level(5); */
+  /*
+    verbose level must be greater than or equal to 1 to print anything
+  */
+  if (verbose < 1)
+    verbose = 1;
+  uim_scm_set_verbose_level(verbose);
 
   args = uim_scm_c_strs_into_list(argc, (const char *const *)argv);
   form = uim_scm_list2(uim_scm_intern_c_str("uim-sh"),
