@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003,2004,2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,21 +31,20 @@
 
 */
 
+#include "config.h"
 #include <pwd.h>
 #include <unistd.h>
 #include <sys/types.h>
-#include <stdio.h>
+#include <sys/stat.h>
 #include <stdlib.h>
 #include <string.h>
-#include <ctype.h>
-#include "config.h"
 #include "uim.h"
 #include "uim-im-switcher.h"
 #include "uim-scm.h"
+#include "uim-custom.h"
 #include "context.h"
 #include "gettext.h"
 
-extern long siod_verbose_level;
 extern char *uim_return_str;
 extern char *uim_return_str_list[10];
 /* duplicate definition */
@@ -54,33 +53,11 @@ extern char *uim_return_str_list[10];
 
 char *uim_last_client_encoding;
 
-#ifdef HAVE_CANNA_RK_H
-extern void uim_init_canna();
-extern void uim_quit_canna();
-#endif /* HAVE_CANNA_RK_H */
-
-#ifdef HAVE_M17NLIB
-extern void uim_init_m17nlib();
-extern void uim_quit_m17nlib();
-#endif /* HAVE_M17NLIB */
-
 #define CONTEXT_ARRAY_SIZE 512
 static uim_context context_array[CONTEXT_ARRAY_SIZE];
 struct uim_im *uim_im_array;
 int uim_nr_im;
 static int uim_initialized;
-static int uim_siod_fatal;
-static FILE *uim_output = NULL;
-
-FILE *
-uim_scm_get_output(void) {
-  return uim_output;
-}
-
-void
-uim_scm_set_output(FILE *fp) {
-  uim_output = fp;
-}
 
 void
 uim_set_preedit_cb(uim_context uc,
@@ -106,6 +83,7 @@ get_context_id(uim_context uc)
       return;
     }
   }
+  uc->id = -1;
 }
 
 static void
@@ -124,11 +102,16 @@ uim_create_context(void *ptr,
 {
   uim_context uc;
 
+  if (!uim_initialized) {
+    fprintf(stderr, "uim_create_context() before uim_init()\n");
+    return 0;
+  }
+
   if (!conv) {
     conv = uim_iconv;
   }
 
-  if (uim_siod_fatal || !conv) {
+  if (!uim_scm_is_alive() || !conv) {
     return NULL;
   }
 
@@ -137,6 +120,9 @@ uim_create_context(void *ptr,
     return NULL;
   }
   get_context_id(uc);
+  if (uc->id == -1)
+    return NULL;
+
   uc->ptr = ptr;
   uc->is_enable = 1;
   uc->commit_cb = commit_cb;
@@ -177,10 +163,6 @@ uim_create_context(void *ptr,
   uc->nr_candidates = 0;
   uc->candidate_index = 0;
   /**/
-  uc->cb_q.first_cb = NULL;
-  uc->cb_q.tail_cb = NULL;
-  uc->cb_q.flushing = 0;
-  /**/
   uc->psegs = NULL;
   uc->nr_psegs = 0;
 
@@ -221,7 +203,8 @@ uim_reset_context(uim_context uc)
 }
 
 void
-uim_switch_im(uim_context uc, const char *engine) {
+uim_switch_im(uim_context uc, const char *engine)
+{
   /* related to the commit log of r1400:
 
      We should not add the API uim_destroy_context(). We should move
@@ -232,8 +215,13 @@ uim_switch_im(uim_context uc, const char *engine) {
      immodule API. We should follow its design to make our API simple.
      -- 2004-10-05 YamaKen
   */
-  uim_reset_context(uc);
-  UIM_EVAL_FSTRING2(uc, "(create-context %d #f '%s)", uc->id, engine);
+  int id = uc->id;
+  uim_reset_context(uc); /* FIXME: reset should be called here? */
+
+  UIM_EVAL_FSTRING1(uc, "(release-context %d)", uc->id);
+  uim_release_preedit_segments(uc);
+
+  UIM_EVAL_FSTRING2(uc, "(create-context %d #f '%s)", id, engine);
   if (uc->current_im_name)
     free(uc->current_im_name);
   uc->current_im_name = strdup(engine);
@@ -248,6 +236,13 @@ uim_switch_im(uim_context uc, const char *engine) {
   uc->short_desc = uim_return_str;
   uim_return_str = NULL;  /* ownership has been transferred */
 #endif
+
+  /*
+     To remember the IM selected by uim-toolbar, we record the latest
+     IM to default-im-name. This makes subsequent contexts the
+     selected one. -- YamaKen 2004-11-15
+  */
+  UIM_EVAL_FSTRING1(uc, "(set! default-im-name '%s)", engine);
 }
 
 void
@@ -255,7 +250,7 @@ uim_release_context(uim_context uc)
 {
   int i;
 
-  if(!uc)
+  if (!uc)
     return;
 
   UIM_EVAL_FSTRING1(uc, "(release-context %d)", uc->id);
@@ -266,12 +261,14 @@ uim_release_context(uim_context uc)
   uim_release_preedit_segments(uc);
   for (i = 0; i < uc->nr_modes; i++) {
     free(uc->modes[i]);
+    uc->modes[i] = NULL;
   }
   free(uc->propstr);
   free(uc->proplabelstr);
   free(uc->modes);
   free(uc->short_desc);
   free(uc->encoding);
+  free(uc->current_im_name);
   free(uc);
 }
 
@@ -323,10 +320,10 @@ uim_set_prop_label_update_cb(uim_context uc,
 void
 uim_prop_activate(uim_context uc, const char *str)
 {
-  if(!str)
+  if (!str)
     return;
       
-  UIM_EVAL_FSTRING2(uc, "(prop-handler %d 'prop-activate \"%s\")",
+  UIM_EVAL_FSTRING2(uc, "(prop-activate-handler %d \"%s\")",
 		    uc->id, str);
 }
 
@@ -340,10 +337,10 @@ uim_prop_activate(uim_context uc, const char *str)
 void
 uim_prop_update_custom(uim_context uc, const char *custom, const char *val)
 {
-  if(!custom || !val)
+  if (!custom || !val)
     return;
 
-  UIM_EVAL_FSTRING3(uc, "(prop-handler %d 'prop-update-custom '%s %s)",
+  UIM_EVAL_FSTRING3(uc, "(custom-set-handler %d '%s %s)",
 		    uc->id, custom, val);
 }
 
@@ -370,14 +367,14 @@ uim_set_mode_cb(uim_context uc, void (*update_cb)(void *ptr,
 void
 uim_prop_list_update(uim_context uc)
 {
-  if(uc && uc->propstr)
+  if (uc && uc->propstr)
     uc->prop_list_update_cb(uc->ptr, uc->propstr);
 }
 
 void
 uim_prop_label_update(uim_context uc)
 {
-  if(uc && uc->proplabelstr)
+  if (uc && uc->proplabelstr)
     uc->prop_label_update_cb(uc->ptr, uc->proplabelstr);
 }
 
@@ -386,7 +383,7 @@ uim_get_nr_im(uim_context uc)
 {
   int i, nr = 0;
 
-  if(!uc)
+  if (!uc)
     return 0;
 
   for (i = 0; i < uim_nr_im; i++) {
@@ -415,7 +412,7 @@ get_nth_im(uim_context uc, int nth)
 const char *
 uim_get_current_im_name(uim_context uc)
 {
-  if(uc) {
+  if (uc) {
     return uc->current_im_name;
   }
   return NULL;
@@ -426,7 +423,7 @@ uim_get_im_name(uim_context uc, int nth)
 {
   struct uim_im *im = get_nth_im(uc, nth);
 
-  if(im) {
+  if (im) {
     return im->name;
   }
   return NULL;
@@ -437,7 +434,7 @@ uim_get_im_language(uim_context uc, int nth)
 {
   struct uim_im *im = get_nth_im(uc, nth);
 
-  if(im) {
+  if (im) {
     return im->lang;
   }
   return NULL;
@@ -464,7 +461,7 @@ uim_get_im_encoding(uim_context uc, int nth)
 {
   struct uim_im *im = get_nth_im(uc, nth);
 
-  if(im) {
+  if (im) {
     return im->encoding;
   }
   return NULL;
@@ -474,7 +471,11 @@ static const char *
 uim_check_im_exist(const char *im_engine_name)
 {
   int i;
-  for(i = 0; i < uim_nr_im; i++) {
+
+  if (im_engine_name == NULL)
+    return NULL;
+
+  for (i = 0; i < uim_nr_im; i++) {
     struct uim_im *im = &uim_im_array[i];
     if (strcmp(im_engine_name, im->name) == 0) {
       return im->name;
@@ -491,50 +492,20 @@ uim_get_default_im_name(const char *localename)
   valid_default_im_name = uim_check_im_exist(uim_return_str);
 
   if (!valid_default_im_name)
-    valid_default_im_name = "default";  /* never happen */
+    valid_default_im_name = "direct";  /* never happen */
   return valid_default_im_name;
 }
 
-
-static int
-load_conf()
+const char *
+uim_get_im_name_for_locale(const char *localename)
 {
-  struct passwd *pw;
-  char *fn = NULL, *cmd;
-  int cmd_size;
-  FILE *fp;
-  long svl = siod_verbose_level;
-  long ret;
+  const char *valid_im_name;
+  UIM_EVAL_FSTRING1(NULL, "(uim-get-im-name-for-locale \"%s\")", localename);
+  valid_im_name = uim_check_im_exist(uim_return_str);
 
-  if(is_setugid() != 0)
-    return -1;
-
-  fn = getenv("LIBUIM_USER_SCM_FILE");
-  if (!fn) {
-    pw = getpwuid(getuid());
-    fn = malloc(strlen(pw->pw_dir) + sizeof("/.uim"));
-    sprintf(fn, "%s/.uim", pw->pw_dir);
-  } else {
-    fn = strdup(fn);
-  }
-
-  fp = fopen(fn, "r");
-  if (fp == NULL) {
-    return -1;
-  }
-  fclose(fp);
-
-  if (siod_verbose_level < 1) {
-    siod_verbose_level = 1;	/* to show error message of SIOD */
-  }
-  cmd_size = uim_sizeof_sexp_str("(*catch 'errobj (load \"%s\" #f #f))", fn);
-  cmd = malloc(cmd_size);
-  sprintf(cmd, "(*catch 'errobj (load \"%s\" #f #f))", fn);
-  ret = repl_c_string(cmd, 0, 0);
-  siod_verbose_level = svl;
-  free(fn);
-  free(cmd);
-  return ret;
+  if (!valid_im_name)
+    valid_im_name = "direct";  /* never happen */
+  return valid_im_name;
 }
 
 int uim_set_candidate_selector_cb(uim_context uc,
@@ -555,6 +526,7 @@ uim_candidate
 uim_get_candidate(uim_context uc, int index, int accel_enumeration_hint)
 {
   uim_candidate cand = malloc(sizeof(*cand));
+  memset(cand, 0, sizeof(*cand));
   UIM_EVAL_FSTRING3(uc, "(get-candidate %d %d %d)",
 		    uc->id, index, accel_enumeration_hint);
 
@@ -566,6 +538,10 @@ uim_get_candidate(uim_context uc, int index, int accel_enumeration_hint)
     cand->heading_label = NULL;
   }
 
+  if (uim_return_str_list[2]) {
+    cand->annotation = uc->conv_if->convert(uc->conv, uim_return_str_list[2]);
+  }
+
   return cand;
 }
 
@@ -581,12 +557,20 @@ uim_candidate_get_heading_label(uim_candidate cand)
   return cand->heading_label;
 }
 
+const char *
+uim_candidate_get_annotation_str(uim_candidate cand)
+{
+  return cand->annotation;
+}
+
 void
 uim_candidate_free(uim_candidate cand)
 {
   free(cand->str);
   free(cand->heading_label);
-  /* free(cand->annotation); */
+  if (cand->annotation)
+    free(cand->annotation);
+
   free(cand);
 }
 
@@ -617,81 +601,61 @@ uim_set_surrounding_text(uim_context uc, const char *text,
 {
 }
 
-static void
-exit_hook(void)
-{
-  uim_siod_fatal = 1;
-}
-
 static void
 uim_init_scm()
 {
   int i;
   char *scm_files = NULL;
-  char *siod_argv[] =
-    {
-      "siod",
-      "-h100000:10",  /* heap_size(unit: lisp objects):nheaps */
-      "-o1000",       /* obarray_dim */
-      "-s200000",     /* stack_size (unit: bytes) */
-      "-n128",        /* inums_dim (preallocated fixnum objects) */
-      "-v0",          /* siod_verbose_level */
-    };
-  char verbose_argv[4] = "-v4";
-  char *env;
+  char *env = NULL;
 
-  if(is_setugid() == 0) {
-    if ((env = getenv("LIBUIM_VERBOSE"))) {
-      if (isdigit(*env)) {
-	if (isdigit(*(env+1)))
-	  verbose_argv[2] = '9';	/* SIOD's max verbose level is 5 */
-	else
-	  verbose_argv[2] = *env;
-      }
-      siod_argv[5] = verbose_argv;
-    }
-  }
-  /* init siod */
-  siod_init (6, siod_argv, 1, uim_output);
-  set_fatal_exit_hook(exit_hook);
-  /**/
-  uim_init_im_subrs();
+  /*  if (is_setugid() == 0) {*/
+    env = getenv("LIBUIM_VERBOSE");
+    /*  }*/
+  uim_scm_init(env);  /* init Scheme interpreter */
+
+#ifdef UIM_COMPAT_SCM
+  uim_init_compat_scm_subrs();
+#endif
+  uim_init_intl_subrs();
   uim_init_util_subrs();
+  uim_init_plugin();
+  uim_init_im_subrs();
   uim_init_key_subrs();
+#ifdef UIM_COMPAT_TABLE
   uim_init_table_subrs();
-  uim_init_anthy();
-  uim_init_prime();
-  uim_init_skk_dic();
-#ifdef HAVE_CANNA_RK_H
-  uim_init_canna();
-#endif /* HAVE_CANNA_RK_H */
-#ifdef HAVE_M17NLIB
-  uim_init_m17nlib();
-#endif /* HAVE_M17NLIB */
+#endif
 
-  if(is_setugid() == 0) {
+  
+  if (is_setugid() == 0) {
     scm_files = getenv("LIBUIM_SCM_FILES");
   }
-  if (scm_files) {
-    siod_set_lib_path(scm_files);
-  } else {
-    siod_set_lib_path(SCM_FILES);
-  }
+  uim_scm_set_lib_path((scm_files) ? scm_files : SCM_FILES);
 
-  uim_scm_load_file("im.scm");
-  uim_scm_load_file("loader.scm");
-  uim_scm_load_file("default-im.scm");  /* must be loaded at last of IMs */
-
-  if(is_setugid() != 0) {
-    uim_scm_load_file("default.scm");
-  } else if (getenv("LIBUIM_VANILLA") || load_conf() == -1) {
-    uim_scm_load_file("default.scm");
-  }
+  uim_scm_require_file("init.scm");
 
   uim_return_str = NULL;
-  for (i = 0; i < (int)UIM_RETURN_STR_LIST_SIZE; i++){
+  for (i = 0; i < (int)UIM_RETURN_STR_LIST_SIZE; i++) {
     uim_return_str_list[i] = NULL;
   }
+
+#if 0
+  /*
+    Current libuim implementation has the gettext encoding problem. It
+    requires library-wide default encoding configurability rather than
+    per context encoding.  -- YamaKen 2005-01-31
+  */
+#ifdef ENABLE_NLS
+ {
+   const char *client_enc;
+
+   /* portable equivalent of nl_langinfo(CODESET) */
+   UIM_EVAL_FSTRING1(NULL, "(locale-lang (locale-new \"%s\"))",
+		     setlocale(LC_CTYPE, NULL));
+   client_enc = uim_scm_refer_c_str(uim_scm_return_value());
+   uim_last_client_encoding = strdup(client_enc);
+ }
+#endif
+#endif
 }
 
 int
@@ -700,9 +664,6 @@ uim_init(void)
   if (uim_initialized) {
     return 0;
   }
-  if (!uim_output) {
-    uim_output = stderr;
-  }
   uim_last_client_encoding = NULL;
   uim_im_array = NULL;
   uim_nr_im = 0;
@@ -727,17 +688,9 @@ uim_quit(void)
     }
   }
   /**/
-  uim_quit_skk_dic();
-  uim_quit_anthy();
-  uim_quit_prime();
-#ifdef HAVE_CANNA_RK_H
-  uim_quit_canna();
-#endif /* HAVE_CANNA_RK_H */
-#ifdef HAVE_M17NLIB
-  uim_quit_m17nlib();
-#endif /* HAVE_M17NLIB */
-  siod_quit();
+  uim_quit_plugin();
+  uim_scm_quit();
+  free(uim_last_client_encoding);
   uim_last_client_encoding = NULL;
-  uim_output = NULL;
   uim_initialized = 0;
 }
