@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -30,34 +30,41 @@
   SUCH DAMAGE.
 */
 
-#include <config.h>
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <stdlib.h>
+#include <stdio.h>
+#include <sys/types.h>
+#include <unistd.h>
+#include <errno.h>
+
 #include "uim/uim.h"
-#include "uim/uim-helper.h"
 
 #include "ximserver.h"
 #include "xim.h"
 #include "convdisp.h"
 #include "canddisp.h"
-#include <unistd.h>
+#include "util.h"
+
 
 static FILE *candwin_r = NULL, *candwin_w = NULL;
 static int candwin_pid = 0;
 static Canddisp *disp;
-static char *command;
+static const char *command;
 
-// fd dispatch
-#define READ_OK 1
-extern void add_fd_watch(int fd, int mask, void (*fn)(int, int));
-extern void remove_current_fd_watch(int fd);
 static void candwin_read_cb(int fd, int ev);
 
-static char *candwin_command(void)
+static const char *candwin_command(void)
 {
     char *candwin_prog = NULL;
 
     candwin_prog = getenv("UIM_CANDWIN_PROG");
     if (candwin_prog == NULL) {
-#ifdef USE_GTK2
+#if defined(USE_QT_CANDWIN)
+	return "uim-candwin-qt";
+#elif defined(USE_GTK_CANDWIN) && defined(USE_GTK2)
 	return "uim-candwin-gtk";
 #else
 	return NULL;
@@ -102,6 +109,7 @@ void Canddisp::activate(std::vector<const char *> candidates, int display_limit)
     }
     fprintf(candwin_w, "\n");
     fflush(candwin_w);
+    check_connection();
 }
 
 void Canddisp::select(int index)
@@ -111,6 +119,7 @@ void Canddisp::select(int index)
     fprintf(candwin_w, "select\n");
     fprintf(candwin_w, "%d\n\n", index);
     fflush(candwin_w);
+    check_connection();
 }
 
 void Canddisp::deactivate()
@@ -119,6 +128,7 @@ void Canddisp::deactivate()
 	return;
     fprintf(candwin_w, "deactivate\n\n");
     fflush(candwin_w);
+    check_connection();
 }
 
 void Canddisp::show()
@@ -127,6 +137,7 @@ void Canddisp::show()
 	return;
     fprintf(candwin_w, "show\n\n");
     fflush(candwin_w);
+    check_connection();
 }
 
 void Canddisp::hide()
@@ -135,6 +146,7 @@ void Canddisp::hide()
 	return;
     fprintf(candwin_w, "hide\n\n");
     fflush(candwin_w);
+    check_connection();
 }
 
 void Canddisp::move(int x, int y)
@@ -146,6 +158,28 @@ void Canddisp::move(int x, int y)
     fprintf(candwin_w, "%d\n", y);
     fprintf(candwin_w, "\n");
     fflush(candwin_w);
+    check_connection();
+}
+
+void Canddisp::show_caret_state(const char *str)
+{
+    if (candwin_w == NULL)
+	return;
+    fprintf(candwin_w, "show_caret_state\n");
+    fprintf(candwin_w, "%s", str);
+    fprintf(candwin_w, "\n");
+    fflush(candwin_w);
+    check_connection();
+}
+
+void Canddisp::check_connection()
+{
+    if (errno == EBADF || errno == EPIPE) {
+	disp = NULL;
+	int fd = fileno(candwin_r);
+	remove_current_fd_watch(fd);
+	delete this;
+    }
 }
 
 static void candwin_read_cb(int fd, int ev)
@@ -198,3 +232,22 @@ static void candwin_read_cb(int fd, int ev)
     }
     return;
 }
+
+void terminate_canddisp_connection()
+{
+    int fd_r, fd_w;
+
+    if (candwin_r) {
+	fd_r = fileno(candwin_r);
+	close(fd_r);
+	remove_current_fd_watch(fd_r);
+    }
+    if (candwin_w) {
+	fd_w = fileno(candwin_w);
+	close(fd_w);
+    }
+
+    disp = NULL;
+
+    return;
+}
