@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -32,6 +32,10 @@
 
 // connection management
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -39,10 +43,8 @@
 #include <list>
 #include <map>
 #include <X11/Xatom.h>
-#ifndef __GNUC__
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
-# endif
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
 #endif
 
 #define BUF_SIZE 1024
@@ -243,28 +245,100 @@ bool XConnection::readToBuf(XClientMessageEvent *ev)
     return true;
 }
 
-void XConnection::writeProc()
+void XConnection::writePendingPacket()
 {
     std::list<TxPacket *>::iterator i, j;
-    std::list<TxPacket *> tmp;
+    bool sent_preedit_done = false;
     int major;
 
-    if (!mTxQ.size() && !mPTxQ.size())
-	return;
+    while (mPendingTxQ.size()) {
+	if (hasSyncFlag() || hasPreeditStartSyncFlag() ||
+			hasPreeditCaretSyncFlag())
+	    break;
 
-    OnSend(); // add XIM_COMMIT packet to passive queue
-
-    // handle passive (i.e. precedent) packets
-    bool sent_preedit_done = false;
-    while (mPTxQ.size()) {
-	i = mPTxQ.begin();
+	i = mPendingTxQ.begin();
 	major = (*i)->get_major();
 
-	if (major == XIM_COMMIT)
+	switch (major) {
+	case XIM_COMMIT:
+	case XIM_FORWARD_EVENT:
 	    setSyncFlag();
-
-	if (major == XIM_PREEDIT_DONE)
+	    break;
+	case XIM_PREEDIT_START:
+	    setPreeditStartSyncFlag();
+	    break;
+	case XIM_PREEDIT_CARET:
+	    setPreeditCaretSyncFlag();
+	    break;
+	case XIM_PREEDIT_DONE:
 	    sent_preedit_done = true;
+	default:
+	    break;
+	}
+
+	doSend(*i, true);
+	delete *i;
+	mPendingTxQ.pop_front();
+
+ 	// XIM_PREEDIT_START/DRAW just after XIM_PREEDIT_DONE maybe
+	// need to wait XIM_COMMIT and its XIM_SYNC_REPLY
+	if (sent_preedit_done == true) {
+	    // push first XIM_COMMIT into the head of the queue
+	    std::list<TxPacket *> tmp;
+	    bool first = true;
+	    while (mPendingTxQ.size()) {
+		j = mPendingTxQ.begin();
+		major = (*j)->get_major();
+		if (major == XIM_COMMIT && first == true) {
+		    tmp.push_front(*j);
+		    first = false;
+		} else
+		    tmp.push_back(*j);
+
+		mPendingTxQ.pop_front();
+	    }
+	    mPendingTxQ = tmp;
+	    sent_preedit_done = false;
+	}
+    }
+}
+
+void XConnection::writePassivePacket()
+{
+    std::list<TxPacket *>::iterator i, j;
+    bool sent_preedit_done = false;
+    int major;
+
+    while (mPTxQ.size()) {
+	if (mPendingTxQ.size())
+	    break;
+
+	i = mPTxQ.begin();
+
+	if (hasSyncFlag() || hasPreeditStartSyncFlag() ||
+			hasPreeditCaretSyncFlag()) {
+	    mPendingTxQ.push_back(*i);
+	    mPTxQ.pop_front();
+	    break;
+	}
+
+	major = (*i)->get_major();
+	switch (major) {
+	case XIM_COMMIT:
+	    setSyncFlag();
+	    break;
+	case XIM_PREEDIT_START:
+	    setPreeditStartSyncFlag();
+	    break;
+	case XIM_PREEDIT_CARET:
+	    setPreeditCaretSyncFlag();
+	    break;
+	case XIM_PREEDIT_DONE:
+	    sent_preedit_done = true;
+	    break;
+	default:
+	    break;
+	}
 
 	doSend(*i, true);
 	delete *i;
@@ -274,6 +348,7 @@ void XConnection::writeProc()
 	// need to wait XIM_COMMIT and its XIM_SYNC_REPLY
 	if (sent_preedit_done == true) {
 	    // push first XIM_COMMIT into the head of passive queue
+	    std::list<TxPacket *> tmp;
 	    bool first = true;
 	    while (mPTxQ.size()) {
 		j = mPTxQ.begin();
@@ -290,8 +365,13 @@ void XConnection::writeProc()
 	    sent_preedit_done = false;
 	}
     }
+}
+
+void XConnection::writeNormalPacket()
+{
+    std::list<TxPacket *>::iterator i;
+    int major;
 
-    // handle normal packets
     while (mTxQ.size()) {
 	i = mTxQ.begin();
 	major = (*i)->get_major();
@@ -308,6 +388,15 @@ void XConnection::writeProc()
 	delete *i;
 	mTxQ.pop_front();
     }
+}
+
+void XConnection::writeProc()
+{
+    OnSend(); // add XIM_COMMIT packet to passive queue
+
+    writePendingPacket();
+    writePassivePacket();
+    writeNormalPacket();
 
     XFlush(XimServer::gDpy);
     if (mIsCloseWait) {
