@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -36,9 +36,9 @@
 #include "xim.h"
 #include "xdispatch.h"
 
-class XConnection :
-    public Connection , public WindowIf
-{
+int connection_setup();
+
+class XConnection: public Connection, public WindowIf {
 public:
     XConnection(XimServer *svr, Window clientWin, Window commWin);
     virtual ~XConnection();
@@ -47,6 +47,9 @@ public:
 
     void readProc(XClientMessageEvent *);
     void writeProc();
+    void writePendingPacket();
+    void writePassivePacket();
+    void writeNormalPacket();
     bool isValid() {return mIsValid;};
 private:
     bool readToBuf(XClientMessageEvent *);
@@ -59,7 +62,7 @@ private:
     struct {
 	char *buf;
 	int len;
-    }mBuf;
+    } mBuf;
 };
 
 #endif
