@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -32,9 +32,16 @@
 
 // classes for preedit draw
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/extensions/shape.h>
+#if HAVE_XFT_UTF8_STRING
+#include <X11/Xft/Xft.h>
+#endif
 #include <stdlib.h>
 #include <locale.h>
 #include "xim.h"
@@ -42,8 +49,12 @@
 #include "convdisp.h"
 #include "canddisp.h"
 #include "xdispatch.h"
+#include "util.h"
+
+#include "uim/uim-compat-scm.h"
 
 #define UNDERLINE_HEIGHT	2
+#define DEFAULT_FONT_SIZE	16
 // Temporal hack for flashplayer plugin's broken over-the-spot XIM style
 #define FLASHPLAYER_WORKAROUND
 
@@ -52,93 +63,115 @@
 // PeLineWin: Child class for root window style preedit window
 // PeOvWin:   Child class for over the spot style preedit window
 
-static XFontSet gFontset_zhCN, gFontset_zhTW, gFontset_ja, gFontset_ko;
 #ifdef FLASHPLAYER_WORKAROUND
 static Window getTopWindow(Display *, Window);
 #endif
 
-// Very dirty...
-void init_convdisp()
-{
-    char **missing, *def;
-    int nr_missing;
-    char *orig_locale;
-    
-    orig_locale = strdup(setlocale(LC_CTYPE, NULL)); // Copy the string for Mac
-    						     // OS X 10.3's setlocale.
+const char *fontset_zhCN = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -isas-fangsong ti-medium-r-normal--16-160-72-72-c-160-gb2312.1980-0";
+const char *fontset_zhTW = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -taipei-fixed-medium-r-normal--16-150-75-75-c-160-big5-0";
+const char *fontset_ja = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -jis-fixed-medium-r-normal--16-*-75-75-c-160-jisx0208.1983-0, -sony-fixed-medium-r-normal--16-*-*-*-c-80-jisx0201.1976-0";
+const char *fontset_ko = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -daewoo-gothic-medium-r-normal--16-120-100-100-c-160-ksc5601.1987-0";
 
-    setlocale(LC_CTYPE, "ja_JP.eucJP");
-    gFontset_ja = 
-	XCreateFontSet(XimServer::gDpy,
-		       "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1,"
-		       "-jis-fixed-medium-r-normal--16-*-75-75-c-160-jisx0208.1983-0,"
-		       "-sony-fixed-medium-r-normal--16-*-*-*-c-80-jisx0201.1976-0",
-		       &missing, &nr_missing, &def);
-    if (missing)
-	XFreeStringList(missing);
-    if (!gFontset_ja)
-	fprintf(stderr, "Warning: Default fontset for LANG=ja is missing.\n");
 
-    setlocale(LC_CTYPE, "zh_CN.eucCN");
-    gFontset_zhCN = 
-	XCreateFontSet(XimServer::gDpy,
-		       "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1,"
-		       "-isas-fangsong ti-medium-r-normal--16-160-72-72-c-160-gb2312.1980-0",
-		       &missing, &nr_missing, &def);
-    if (missing)
-	XFreeStringList(missing);
-    if (!gFontset_zhCN) {
-	fprintf(stderr, "Warning: Default fontset for LANG=zh_CN is missing.\n");
-	gFontset_zhCN = gFontset_ja;
+#if HAVE_XFT_UTF8_STRING
+XftFont *gXftFont;
+char *gXftFontName;
+char *gXftFontLocale;
+
+void
+init_default_xftfont() {
+    char *fontname = uim_scm_symbol_value_str("uim-xim-xft-font-name");
+    gXftFontName = fontname;
+
+    gXftFont = XftFontOpen(XimServer::gDpy, DefaultScreen(XimServer::gDpy),
+		    XFT_FAMILY, XftTypeString, fontname,
+		    XFT_PIXEL_SIZE, XftTypeDouble, (double)DEFAULT_FONT_SIZE,
+		    NULL);
+    gXftFontLocale = strdup(setlocale(LC_CTYPE, NULL));
+    // maybe not needed, but in case it return NULL...
+    if (!gXftFont) {
+	gXftFont = XftFontOpen(XimServer::gDpy, DefaultScreen(XimServer::gDpy),
+			XFT_FAMILY, XftTypeString, "Sans",
+			XFT_PIXEL_SIZE, XftTypeDouble, (double)DEFAULT_FONT_SIZE,
+			NULL);
     }
-
-    setlocale(LC_CTYPE, "zh_TW.big5");
-    gFontset_zhTW =
-	XCreateFontSet(XimServer::gDpy,
-		       "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1,"
-		       "-taipei-fixed-medium-r-normal--16-150-75-75-c-160-big5-0",
-		       &missing, &nr_missing, &def);
-    if (missing)
-	XFreeStringList(missing);
-    if (!gFontset_zhTW) {
-	fprintf(stderr, "Warning: Default fontset for LANG=zh_TW is missing.\n");
-	gFontset_zhTW = gFontset_ja;
-    }
-
-    setlocale(LC_CTYPE, "ko_KR.eucKR");
-    gFontset_ko =
-	XCreateFontSet(XimServer::gDpy,
-		       "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1,"
-		       "-daewoo-gothic-medium-r-normal--16-120-100-100-c-160-ksc5601.1987-0",
-		       &missing, &nr_missing, &def);
-    if (missing)
-	XFreeStringList(missing);
-    if (!gFontset_ko) {
-	fprintf(stderr, "Warning: Default fontset for LANG=ko is missing.\n");
-	gFontset_ko = gFontset_ja;
-    }
-
-    setlocale(LC_CTYPE, orig_locale);
-    free(orig_locale);
 }
 
-XFontSet choose_default_fontset (const char *lang) {
+static char *
+dequote(const char *str)
+{
+   char *ret = NULL;
+
+   if (str) {
+	int len = strlen(str);
+	if (str[0] == '"' && str[len - 1] == '"') {
+	    ret = strdup(++str);
+	    ret[len - 2] = '\0';
+	} else
+	    ret = strdup(str);
+   }
+
+   return ret;
+}
+
+void
+update_default_xftfont(const char *s) {
+    char *fontname = dequote(s);
+
+    if (fontname) {
+	XftFont *xftfont = XftFontOpen(XimServer::gDpy,
+			DefaultScreen(XimServer::gDpy),
+			XFT_FAMILY, XftTypeString, fontname,
+			XFT_PIXEL_SIZE, XftTypeDouble, (double)DEFAULT_FONT_SIZE,
+			NULL);
+	if (xftfont) {
+	    XftFontClose(XimServer::gDpy, gXftFont);
+	    free(gXftFontName);
+	    free(gXftFontLocale);
+	    gXftFont = xftfont;
+	    gXftFontName = fontname;
+	    gXftFontLocale = strdup(setlocale(LC_CTYPE, NULL));
+	}
+    }
+}
+#endif
+
+static XFontSet
+create_default_fontset(const char *im_lang, const char *locale) {
+    char *orig_locale;
+    const char *name;
     XFontSet ret;
 
-    if (!strcmp(lang, "ja"))
-	ret = gFontset_ja;
-    else if (!strcmp(lang, "ko"))
-	ret = gFontset_ko;
-    else if (!strcmp(lang, "zh_CN"))
-	ret = gFontset_zhCN;
-    else if (!strncmp(lang, "zh_TW", 5)) // XXX zh_TW:zh_HK
-	ret = gFontset_zhTW;
+    orig_locale = strdup(setlocale(LC_CTYPE, NULL));
+
+    if (strcmp(locale, orig_locale))
+	setlocale(LC_CTYPE, locale);
+
+    if (!strcmp(im_lang, "ja"))
+	name = fontset_ja;
+    else if (!strcmp(im_lang, "ko"))
+	name = fontset_ko;
+    else if (!strcmp(im_lang, "zh_CN"))
+	name = fontset_zhCN;
+    else if (!strcmp(im_lang, "zh_TW:zh_HK"))
+	name = fontset_zhTW;
     else
-	ret = gFontset_ja; // XXX fallback font
+	name = fontset_ja; // XXX fallback fontset
+
+    ret = get_font_set(name, locale);
+
+    if (strcmp(locale, orig_locale))
+	setlocale(LC_CTYPE, orig_locale);
+    free(orig_locale);
 
     return ret;
 }
 
+static XFontSet
+choose_default_fontset(const char *im_lang, const char *locale) {
+    return create_default_fontset(im_lang, locale);
+}
+
 struct char_ent {
     uchar c;
     int stat;
@@ -150,7 +183,7 @@ struct char_ent {
 // Preedit Window
 class PeWin : public WindowIf {
 public:
-    PeWin(Window pw, const char *lang, const char *encoding);
+    PeWin(Window pw, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd);
     virtual ~PeWin();
 
     virtual void release();
@@ -161,6 +194,9 @@ public:
     void draw_char(int x, int y, uchar ch, int stat);
     void set_back(unsigned long p);
     void set_fore(unsigned long p);
+#if HAVE_XFT_UTF8_STRING
+    void set_xftfont(const char *xfld);
+#endif
     void set_fontset(XFontSet f);
     
     virtual void set_size(int w, int h);
@@ -171,23 +207,37 @@ public:
     void clear();
     void draw();
     void unmap();
+
+#if HAVE_XFT_UTF8_STRING
+    XftFont *mXftFont;
+    int mXftFontSize;
+#endif
 protected:
+#if HAVE_XFT_UTF8_STRING
+    int get_fontsize(const char *xfld);
+#endif
     Window mParentWin;
     Window mWin;
     Pixmap mPixmap;
     GC mGC, mClearGC;
-    GC mHilitGC;
     unsigned int mFore, mBack;
+#if HAVE_XFT_UTF8_STRING
+    XftDraw *mXftDraw;
+    XftColor mXftColorFg;
+    XftColor mXftColorFgRev;
+    int mGlyphWidth;
+#endif
     XFontSet mFontset;
     const char *mEncoding;
     int mWidth, mHeight;
     bool mIsMapped;
+    Convdisp *mConvdisp;
 };
 
 // one line preedit window for RootWindowStyle
 class PeLineWin : public PeWin {
 public:
-    PeLineWin(Window w, const char *lang, const char *encoding);
+    PeLineWin(Window w, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd);
     virtual ~PeLineWin();
 
     void draw_pe(pe_stat *p);
@@ -203,7 +253,7 @@ private:
 // window for over the spot style
 class PeOvWin : public PeWin {
 public:
-    PeOvWin(Window w, const char *lang, const char *encoding);
+    PeOvWin(Window w, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd);
     
     void draw_ce(char_ent *ce, int len);
     virtual void set_size(int w, int h);
@@ -222,7 +272,8 @@ public:
     virtual void clear_preedit();
     virtual void update_icxatr();
     virtual void move_candwin();
-    virtual void set_lang(const char *lang);
+    virtual void set_im_lang(const char *im_lang);
+    virtual bool use_xft();
 private:
     bool check_win();
     bool check_atr();
@@ -241,9 +292,6 @@ private:
     char_ent *m_ce;
     int m_ce_len;
     PeOvWin *m_ov_win;
-    XFontSet m_initial_fontset;
-    char *m_initial_lang;
-    bool m_lang_changed;
 };
 
 // Preedit window for RootWindowStyle
@@ -256,11 +304,11 @@ public:
     virtual void clear_preedit();
     virtual void update_icxatr();
     virtual void move_candwin();
+    virtual bool use_xft();
 private:
     PeLineWin *mPeWin;
 };
 
-class Connection;
 class ConvdispOs : public Convdisp {
 public:
     ConvdispOs(InputContext *, icxatr *, Connection *);
@@ -269,6 +317,7 @@ public:
     virtual void clear_preedit();
     virtual void update_icxatr();
     virtual void move_candwin();
+    virtual bool use_xft();
 
 private:
     void compose_preedit_array(TxPacket *);
@@ -298,9 +347,10 @@ Convdisp *create_convdisp(int style, InputContext *k,
 //
 // PeWin(PreEdit Window)
 //
-PeWin::PeWin(Window pw, const char *lang, const char *encoding)
+PeWin::PeWin(Window pw, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd)
 {
     mParentWin = pw;
+    mConvdisp = cd;
     int scr_num = DefaultScreen(XimServer::gDpy);
     // tentative
     mWidth = 1;
@@ -319,7 +369,6 @@ PeWin::PeWin(Window pw, const char *lang, const char *encoding)
     
     mGC = XCreateGC(XimServer::gDpy, mPixmap, 0, 0);
     mClearGC = XCreateGC(XimServer::gDpy, mPixmap, 0, 0);
-    mHilitGC = XCreateGC(XimServer::gDpy, mPixmap, 0, 0);
     
     XSetBackground(XimServer::gDpy, mGC, WhitePixel(XimServer::gDpy, scr_num));
     XSetForeground(XimServer::gDpy, mGC, BlackPixel(XimServer::gDpy, scr_num));
@@ -329,7 +378,40 @@ PeWin::PeWin(Window pw, const char *lang, const char *encoding)
     add_window_watch(mWin, this, EXPOSE_MASK|STRUCTURE_NOTIFY_MASK);
     mIsMapped = false; //not mapped now
     
-    mFontset = choose_default_fontset(lang);
+    if (mConvdisp->use_xft() == true) {
+#if HAVE_XFT_UTF8_STRING
+	mXftFontSize = DEFAULT_FONT_SIZE;
+	if (!strcmp(gXftFontLocale, locale)) {
+	    mXftFont = gXftFont;
+	} else {
+	    mXftFont = XftFontOpen(XimServer::gDpy,
+			    DefaultScreen(XimServer::gDpy),
+			    XFT_FAMILY, XftTypeString, gXftFontName,
+			    XFT_PIXEL_SIZE, XftTypeDouble, (double)mXftFontSize,
+			    NULL);
+	}
+	mXftDraw = XftDrawCreate(XimServer::gDpy, mPixmap,
+			DefaultVisual(XimServer::gDpy, scr_num),
+			DefaultColormap(XimServer::gDpy, scr_num));
+	XColor dummyc, fg;
+	XAllocNamedColor(XimServer::gDpy, DefaultColormap(XimServer::gDpy, scr_num),"black", &fg, &dummyc);
+	mXftColorFg.color.red = dummyc.red;
+	mXftColorFg.color.green = dummyc.green;
+	mXftColorFg.color.blue = dummyc.blue;
+	mXftColorFg.color.alpha = 0xffff;
+	mXftColorFg.pixel = fg.pixel;
+
+	XAllocNamedColor(XimServer::gDpy, DefaultColormap(XimServer::gDpy, scr_num),"white", &fg, &dummyc);
+	mXftColorFgRev.color.red = dummyc.red;
+	mXftColorFgRev.color.green = dummyc.green;
+	mXftColorFgRev.color.blue = dummyc.blue;
+	mXftColorFgRev.color.alpha = 0xffff;
+	mXftColorFgRev.pixel = fg.pixel;
+#endif
+    } else {
+	mFontset = choose_default_fontset(im_lang, locale);
+    }
+
     mEncoding = encoding;
     
     XFlush(XimServer::gDpy);
@@ -344,7 +426,13 @@ PeWin::~PeWin()
     
     XFreeGC(XimServer::gDpy, mGC);
     XFreeGC(XimServer::gDpy, mClearGC);
-    XFreeGC(XimServer::gDpy, mHilitGC);
+#if HAVE_XFT_UTF8_STRING 
+    if (mConvdisp->use_xft() == true) {
+	XftDrawDestroy(mXftDraw);
+	if (mXftFont != gXftFont)
+	    XftFontClose(XimServer::gDpy, mXftFont);
+    }
+#endif
 
     XFlush(XimServer::gDpy);
 }
@@ -364,7 +452,7 @@ void PeWin::destroy(Window w)
 
 void PeWin::expose(Window w)
 {
-    XCopyArea(XimServer::gDpy, mPixmap, mWin, mGC,
+    XCopyArea(XimServer::gDpy, mPixmap, w, mGC,
 	      0, 0, mWidth, mHeight, 0, 0);
 
     XFlush(XimServer::gDpy);
@@ -376,27 +464,38 @@ void PeWin::draw_char(int x, int y, uchar ch, int stat)
     if (stat & PE_REVERSE)
 	gc = mClearGC;
 
-#if 0
-    if (stat & PE_HILIGHT)
-	gc = mHilitGC;
-#endif
     char utf8[6];
     int len = utf8_wctomb((unsigned char *)utf8, ch);
     utf8[len] = '\0';
 
-    if (!strcmp(mEncoding, "UTF-8"))
-    	XwcDrawImageString(XimServer::gDpy, mPixmap, mFontset,
+    if (mConvdisp->use_xft() == true) {
+#ifdef HAVE_XFT_UTF8_STRING
+	XGlyphInfo ginfo;
+	XftTextExtentsUtf8(XimServer::gDpy, mXftFont, (unsigned char *)utf8, len, &ginfo);
+	mGlyphWidth = ginfo.xOff;
+	if (stat & PE_REVERSE) {
+	    XftDrawRect(mXftDraw, &mXftColorFg, x, y - (mXftFontSize - 2), ginfo.xOff, mXftFontSize);
+	    XftDrawStringUtf8(mXftDraw, &mXftColorFgRev, mXftFont, x, y, (unsigned char *)utf8, len);
+	} else {
+	    XftDrawStringUtf8(mXftDraw, &mXftColorFg, mXftFont, x, y, (unsigned char *)utf8, len);
+	}
+#endif
+    } else {
+	if (!strcmp(mEncoding, "UTF-8")) {
+    	    XwcDrawImageString(XimServer::gDpy, mPixmap, mFontset,
 			gc, x, y, &ch, 1);
-    else {
-	char *native_str;
+	} else {
+	    char *native_str;
+	    XimIM *im = get_im_by_id(mConvdisp->get_context()->get_ic()->get_imid());
 	
-	native_str = utf8_to_native_str(utf8, mEncoding);
-	if (!native_str)
-	    return;
-	int len = strlen(native_str);
-	XmbDrawImageString(XimServer::gDpy, mPixmap, mFontset,
+	    native_str = im->utf8_to_native_str(utf8);
+	    if (!native_str)
+		return;
+	    len = strlen(native_str);
+	    XmbDrawImageString(XimServer::gDpy, mPixmap, mFontset,
 			   gc, x, y, native_str, len);
-	free(native_str);
+	    free(native_str);
+	}
     }
 }
 
@@ -405,6 +504,18 @@ void PeWin::set_back(unsigned long p)
     mBack = p;
     XSetBackground(XimServer::gDpy, mGC, p);
     XSetForeground(XimServer::gDpy, mClearGC, p);
+#if HAVE_XFT_UTF8_STRING
+    if (mConvdisp->use_xft() == true) {
+	XColor xcolor;
+	xcolor.pixel = p;
+	XQueryColor(XimServer::gDpy, DefaultColormap(XimServer::gDpy, DefaultScreen(XimServer::gDpy)), &xcolor);
+	mXftColorFgRev.pixel = p;
+	mXftColorFgRev.color.red = xcolor.red;
+	mXftColorFgRev.color.green = xcolor.green;
+	mXftColorFgRev.color.blue = xcolor.blue;
+	mXftColorFgRev.color.alpha = 0xffff;
+    }
+#endif
 }
 
 void PeWin::set_fore(unsigned long p)
@@ -412,13 +523,75 @@ void PeWin::set_fore(unsigned long p)
     mFore = p;
     XSetForeground(XimServer::gDpy, mGC, p);
     XSetBackground(XimServer::gDpy, mClearGC, p);
+#if HAVE_XFT_UTF8_STRING
+    if (mConvdisp->use_xft() == true) {
+	XColor xcolor;
+	xcolor.pixel = p;
+	XQueryColor(XimServer::gDpy, DefaultColormap(XimServer::gDpy, DefaultScreen(XimServer::gDpy)), &xcolor);
+	mXftColorFg.pixel = p;
+	mXftColorFg.color.red = xcolor.red;
+	mXftColorFg.color.green = xcolor.green;
+	mXftColorFg.color.blue = xcolor.blue;
+	mXftColorFg.color.alpha = 0xffff;
+    }
+#endif
 }
 
 void PeWin::set_fontset(XFontSet f)
 {
-    mFontset = f;
+    if (f)
+	mFontset = f;
 }
 
+#if HAVE_XFT_UTF8_STRING
+void PeWin::set_xftfont(const char *xfld)
+{
+	int size = get_fontsize(xfld);
+	const char *locale = mConvdisp->get_locale_name();
+	if (size != -1 && (mXftFontSize != size || strcmp(locale, gXftFontLocale))) {
+	    if (mXftFont != gXftFont)
+		XftFontClose(XimServer::gDpy, mXftFont);
+
+	    mXftFont = XftFontOpen(XimServer::gDpy,
+			    DefaultScreen(XimServer::gDpy),
+			    XFT_FAMILY, XftTypeString, gXftFontName,
+			    XFT_PIXEL_SIZE, XftTypeDouble, (double)size,
+			    NULL);
+	    mXftFontSize = size;
+	}
+}
+
+int PeWin::get_fontsize(const char *xfld)
+{
+#define MAX_DIGIT_OF_PIXEL_SIZE	3
+    int size;
+    char str[MAX_DIGIT_OF_PIXEL_SIZE + 1];
+    const char *p = xfld;
+    int count = 0;
+    int i, j = 0;
+
+    for (i = 0; i < (int)strlen(xfld); i++) {
+	if (p[i] == '-')
+	    count++;
+	if (count == 7) {
+	    i++;
+	    while (p[i] != '-' && p[i] != '\0') {
+		str[j] = p[i];
+		i++;
+		j++;
+		if (j > MAX_DIGIT_OF_PIXEL_SIZE)
+		    return -1;
+	    }
+	    str[j] = '\0';
+	    break;
+	}
+    }
+    if (!sscanf(str, "%d", &size))
+	return -1;
+    return size;
+}
+#endif
+
 void PeWin::set_size(int w, int h)
 {
     if (w == mWidth && h == mHeight)
@@ -428,6 +601,10 @@ void PeWin::set_size(int w, int h)
     XFreePixmap(XimServer::gDpy, mPixmap);
     mPixmap = XCreatePixmap(XimServer::gDpy, DefaultRootWindow(XimServer::gDpy), w, h,
 			    DefaultDepth(XimServer::gDpy, DefaultScreen(XimServer::gDpy)));
+#if HAVE_XFT_UTF8_STRING 
+    if (mConvdisp->use_xft() == true)
+	XftDrawChange(mXftDraw, mPixmap);
+#endif
     mWidth = w;
     mHeight = h;
     clear();
@@ -476,9 +653,9 @@ void PeWin::unmap()
 //
 // PeLineWin
 //
-PeLineWin::PeLineWin(Window w, const char *lang, const char *encoding) : PeWin(w, lang, encoding)
+PeLineWin::PeLineWin(Window w, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd) : PeWin(w, im_lang, encoding, locale, cd)
 {
-    set_size(400, 28); // set window height as twice as the font height
+    set_size(400, 28); // set window height wider than its font height 16
     clear();
 }
 
@@ -503,13 +680,22 @@ void PeLineWin::draw_segment(pe_ustring *s)
     for (i = s->s.begin(); i != s->s.end(); i++) {
 	uchar ch = *i;
 	draw_char(m_x, 20, ch, s->stat);
+
+	int width;
+
+#if HAVE_XFT_UTF8_STRING
+	if (mConvdisp->use_xft() == true)
+	    width = mGlyphWidth;
+	else
+#endif
+	    width = 16;
+
 	if (s->stat & PE_UNDERLINE) {
 	    XDrawLine(XimServer::gDpy, mPixmap, mGC,
 			    m_x, 20 + UNDERLINE_HEIGHT,
-			    m_x + 16, 20 + UNDERLINE_HEIGHT);
-
+			    m_x + width, 20 + UNDERLINE_HEIGHT);
 	}
-	m_x += 16; // XXX font width
+	m_x += width;
     }
 }
 
@@ -524,7 +710,7 @@ void PeLineWin::calc_extent(pe_stat *p)
 //
 // PeOvWin
 //
-PeOvWin::PeOvWin(Window w, const char *lang, const char *encoding) : PeWin(w, lang, encoding)
+PeOvWin::PeOvWin(Window w, const char *im_lang, const char *encoding, const char *locale, Convdisp *cd) : PeWin(w, im_lang, encoding, locale, cd)
 {
     m_mask_pix = 0;
     m_mask_pix_gc = 0;
@@ -597,17 +783,28 @@ Convdisp::Convdisp(InputContext *k, icxatr *a)
 {
     mKkContext = k;
     m_atr = a;
-    mLang = get_im_lang_from_engine(k->get_engine_name());
+    mIMLang = get_im_lang_from_engine(k->get_engine_name());
     mEncoding = k->get_ic()->get_encoding();
+    mLocaleName = k->get_locale_name();
 }
 
 Convdisp::~Convdisp()
 {
 }
 
-void Convdisp::set_lang(const char *lang)
+void Convdisp::set_im_lang(const char *im_lang)
 {
-    mLang = lang;
+    mIMLang = im_lang;
+}
+
+void Convdisp::set_locale_name(const char *locale)
+{
+    mLocaleName = locale;
+}
+
+const char *Convdisp::get_locale_name()
+{
+    return mLocaleName;
 }
 
 void Convdisp::set_pe(pe_stat *p)
@@ -637,6 +834,10 @@ void Convdisp::unset_focus()
     disp->hide();
 }
 
+InputContext *Convdisp::get_context()
+{
+    return mKkContext;
+}
 
 // Root window style
 ConvdispRw::ConvdispRw(InputContext *k, icxatr *a) : Convdisp(k, a)
@@ -662,7 +863,7 @@ void ConvdispRw::update_preedit()
 
     // preedit string exists
     if (!mPeWin)
-        mPeWin = new PeLineWin(DefaultRootWindow(XimServer::gDpy), mLang, mEncoding);
+	mPeWin = new PeLineWin(DefaultRootWindow(XimServer::gDpy), mIMLang, mEncoding, mLocaleName, this);
     
     if (m_atr->has_atr(ICA_ClientWindow)) {
     	int x, y;
@@ -709,61 +910,41 @@ void ConvdispRw::move_candwin()
     }
 }
 
-//Over the spot style
+bool ConvdispRw::use_xft()
+{
+    return m_atr->use_xft();
+}
+
+// Over the spot style
 ConvdispOv::ConvdispOv(InputContext *k, icxatr *a) : Convdisp(k, a)
 {
     m_ov_win = 0;
-    m_initial_lang = strdup(mLang);
-    m_initial_fontset = NULL;
-    m_lang_changed = false;
+#ifdef FLASHPLAYER_WORKAROUND
+    revised_spot_y = -1;
+#endif
 }
 
 ConvdispOv::~ConvdispOv()
 {
     if (m_ov_win)
 	delete m_ov_win;
-
-    free(m_initial_lang);
 }
 
-void ConvdispOv::set_lang(const char *lang)
+void ConvdispOv::set_im_lang(const char *im_lang)
 {
-    mLang = lang;
-
-    if (!strcmp(m_initial_lang, lang)) {
-#if 0
-	if (m_initial_fontset)
-	    m_atr->font_set = m_initial_fontset;
-	else
-	    m_atr->font_set = choose_default_fontset(mLang);
-#endif
-    } else {
-#if 0
-	m_atr->font_set = choose_default_fontset(mLang);
-#endif
-	m_lang_changed = true;
-    }
+    mIMLang = im_lang;
 }
 
 
 void ConvdispOv::update_preedit()
 {
-    if (!m_pe)
-	return;
-
-    if (!m_pe->get_char_count()) {
-	clear_preedit();
-	return;
-    }
-
     draw_preedit();
     move_candwin();
 }
 
 void ConvdispOv::move_candwin()
 {
-    if (m_atr->has_atr(ICA_SpotLocation) &&
-		mKkContext == InputContext::focusedContext()) {
+    if (m_atr->has_atr(ICA_SpotLocation) ) {
 	int x = -1, y = -1;
 	Window win;
 
@@ -797,7 +978,7 @@ void ConvdispOv::move_candwin()
 	    Canddisp *disp = canddisp_singleton();
 	    disp->move(x, y + UNDERLINE_HEIGHT + 1);
 	}
-#ifdef FLASHPLAYER_WORKAROUND
+#if 0
     } else if (m_atr->has_atr(ICA_SpotLocation)) {
 	int x = -1, y = -1;
 	Window win;
@@ -864,7 +1045,13 @@ void ConvdispOv::update_icxatr()
 	m_atr->unset_change_mask(ICA_Background);
     }
     if (m_atr->is_changed(ICA_FontSet)) {
-	m_ov_win->set_fontset(m_atr->font_set);
+	if (use_xft() == true) {
+#if HAVE_XFT_UTF8_STRING
+	    m_ov_win->set_xftfont(m_atr->font_set_name);
+#endif
+	} else {
+	    m_ov_win->set_fontset(m_atr->font_set);
+	}
 	m_atr->unset_change_mask(ICA_FontSet);
     }
   
@@ -873,7 +1060,15 @@ void ConvdispOv::update_icxatr()
 
 void ConvdispOv::draw_preedit()
 {
+    if (!m_pe)
+	return;
+
     m_ce_len = m_pe->get_char_count();
+    if (!m_ce_len) {
+	clear_preedit();
+	return;
+    }
+
     if (!check_win())
 	return;
 
@@ -933,14 +1128,14 @@ void ConvdispOv::do_draw_preedit()
 		if (y > (topattr.height + topattr.y)) {
 		    // preedit area goes beyond the top window's geometry
 		    if ((y - m_atr->area.y) < topattr.y) {
-			// Set preedit at the upper side of the top
-			// window.  But it may doesn't work because it
+			// Set preedit at upper side of the top
+			// window.  But it may not work because it
 			// lacks height for window manager's title bar
 			// and some toolbars of browser...
 			m_ov_win->set_pos(m_atr->area.x, topattr.y - (y - m_atr->area.y));
 			revised_spot_y = font_height + topattr.y - (y - m_atr->area.y);
 		    } else {
-			// Set preedit at the upper side of the flash
+			// Set preedit at upper side of the flash
 			// player's window.
 			m_ov_win->set_pos(m_atr->area.x, 0);
 			revised_spot_y = font_height;
@@ -989,11 +1184,17 @@ bool ConvdispOv::check_win()
     else
 	w = m_atr->client_window;
 
-    m_ov_win = new PeOvWin(w, mLang, mEncoding);
+    m_ov_win = new PeOvWin(w, mIMLang, mEncoding, mLocaleName, this);
     m_ov_win->set_size(m_atr->area.width, m_atr->area.height);
     m_ov_win->set_fore(m_atr->foreground_pixel);
     m_ov_win->set_back(m_atr->background_pixel);
-    m_ov_win->set_fontset(m_atr->font_set);
+    if (use_xft() == true) {
+#if HAVE_XFT_UTF8_STRING
+	m_ov_win->set_xftfont(m_atr->font_set_name);
+#endif
+    } else {
+	m_ov_win->set_fontset(m_atr->font_set);
+    }
   
     return true;
 }
@@ -1015,14 +1216,13 @@ bool ConvdispOv::check_atr()
 	m_atr->area.x = 0;
 	m_atr->area.y = 0;
     }
-    if (!m_atr->has_atr(ICA_FontSet))
-	m_atr->font_set = choose_default_fontset(mLang);
-    else {
-	if (!m_initial_fontset && !m_lang_changed)
-	    m_initial_fontset = m_atr->font_set;
+    if (!m_atr->has_atr(ICA_FontSet)) {
+	if (use_xft() == false)
+	    m_atr->font_set = choose_default_fontset(mIMLang, mLocaleName);
+    }
+    if (!m_atr->has_atr(ICA_LineSpace)) {
+	m_atr->line_space = 16;
     }
-    if (!m_atr->has_atr(ICA_LineSpace))
-	m_atr->line_space = 14;
 
     if (!m_atr->has_atr(ICA_Foreground))
 	m_atr->foreground_pixel = BlackPixel(XimServer::gDpy,
@@ -1062,27 +1262,44 @@ void ConvdispOv::layoutCharEnt()
 
     for (i = 0; i < m_ce_len; i++) {
 	uchar ch = m_ce[i].c;
-	XRectangle ink, logical;
 
-	if (!strcmp(mEncoding, "UTF-8"))
-	    XwcTextExtents(m_atr->font_set, &ch, 1, &ink, &logical);
-	else {
-	    char utf8[6];
-	    int len = utf8_wctomb((unsigned char *)utf8, ch);
+	char utf8[6];
+	int len;
+	if (use_xft() == true) {
+#if HAVE_XFT_UTF8_STRING
+	    len = utf8_wctomb((unsigned char *)utf8, ch);
 	    utf8[len] = '\0';
-	    char *str;
-	    str = utf8_to_native_str(utf8, mEncoding);
-	    if (!str) {
-		logical.width = 0;
-		logical.height = (i > 0) ? m_ce[i - 1].height : 0;
+
+	    XGlyphInfo ginfo;
+	    XftTextExtentsUtf8(XimServer::gDpy, m_ov_win->mXftFont, (unsigned char *)utf8, len, &ginfo);
+	    m_ce[i].width = ginfo.xOff;
+	    m_ce[i].height = m_ov_win->mXftFontSize;
+#endif
+	} else {
+	    XRectangle ink, logical;
+
+	    if (!strcmp(mEncoding, "UTF-8")) {
+		XwcTextExtents(m_atr->font_set, &ch, 1, &ink, &logical);
+		m_ce[i].width = logical.width;
+		m_ce[i].height = logical.height;
 	    } else {
-		len = strlen(str);
-		XmbTextExtents(m_atr->font_set, str, len, &ink, &logical);
+		len = utf8_wctomb((unsigned char *)utf8, ch);
+		utf8[len] = '\0';
+		XimIM *im = get_im_by_id(mKkContext->get_ic()->get_imid());
+		char *str = im->utf8_to_native_str(utf8);
+		if (!str) {
+		    logical.width = 0;
+		    logical.height = (i > 0) ? m_ce[i - 1].height : 0;
+		} else {
+		    len = strlen(str);
+		    XmbTextExtents(m_atr->font_set, str, len, &ink, &logical);
+		    free(str);
+		}
+		m_ce[i].width = logical.width;
+		m_ce[i].height = logical.height;
 	    }
 	}
 
-	m_ce[i].width = logical.width;
-	m_ce[i].height = logical.height;
 	int right_limit = m_atr->area.width;
 	if (m_atr->has_atr(ICA_Area))
 	    right_limit += m_atr->area.x;
@@ -1131,6 +1348,11 @@ int ConvdispOv::get_ce_font_height(char_ent *ce, int len)
 }
 #endif
 
+bool ConvdispOv::use_xft()
+{
+    return m_atr->use_xft();
+}
+
 // On the spot style
 ConvdispOs::ConvdispOs(InputContext *k, icxatr *a, Connection *c)
     : Convdisp(k, a)
@@ -1148,6 +1370,7 @@ ConvdispOs::~ConvdispOs()
 
 void ConvdispOs::update_preedit()
 {
+    move_candwin();
     if (!m_pe)
 	return;
 
@@ -1190,16 +1413,16 @@ void ConvdispOs::update_preedit()
 	mConn->push_passive_packet(t);
     }
     mPrevLen = len;
-    /*
-      t = createTxPacket(XIM_PREEDIT_CARET,0);
-      t->pushC16(mImid);
-      t->pushC16(mIcid);
-      t->pushC32(pos);
-      t->pushC32(dir);
-      t->pushC32(style);
-      mConn->push_packet(t);
-      */
-    move_candwin();
+
+    if (len) {
+	t = createTxPacket(XIM_PREEDIT_CARET, 0);
+	t->pushC16(mImid);
+	t->pushC16(mIcid);
+	t->pushC32(caret_pos);
+	t->pushC32(XIMAbsolutePosition);
+	t->pushC32(XIMIsPrimary);
+	mConn->push_passive_packet(t);
+    }
 }
       
 void ConvdispOs::move_candwin()
@@ -1223,8 +1446,9 @@ void ConvdispOs::move_candwin()
 	Canddisp *disp = canddisp_singleton();
 
 	if (m_atr->has_atr(ICA_SpotLocation))
-	    disp->move(x, y + 36); /* 36 pixel seems too long.
-				      Is there any way to get current preedit height? */
+	    disp->move(x, y + 36); // 36 pixel seems too long.
+				   // Is there any way to get current preedit
+				   // height?
 	else {
 	    XGetWindowAttributes(XimServer::gDpy, m_atr->client_window,
 			      &xattr);
@@ -1252,11 +1476,12 @@ void ConvdispOs::compose_preedit_array(TxPacket *t)
 	append_ustring(&s, &(*it).s);
     }
 
-    char *c = mKkContext->getServer()->uStringToCtext(&s, mEncoding);
+    XimIM *im = get_im_by_id(mKkContext->get_ic()->get_imid());
+    char *c = im->uStringToCtext(&s);
     int i, len = 0;
     if (c)
 	len = strlen(c);
-    t->pushC16(len);// LENGTH
+    t->pushC16(len); // LENGTH
     for (i = 0; i < len; i++) {
 	t->pushC8(c[i]); // CTEXT
     }
@@ -1294,6 +1519,11 @@ void ConvdispOs::compose_feedback_array(TxPacket *t)
     }
 }
 
+bool ConvdispOs::use_xft()
+{
+    return true;
+}
+
 #ifdef FLASHPLAYER_WORKAROUND
 static Window getTopWindow(Display *d, Window w)
 {
