@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -51,7 +51,11 @@ public:
     virtual void set_focus();
     virtual void unset_focus();
     virtual void move_candwin() = 0;
-    virtual void set_lang(const char *lang);
+    virtual void set_im_lang(const char *im_lang);
+    virtual void set_locale_name(const char *locale);
+    virtual const char *get_locale_name();
+    virtual InputContext *get_context();
+    virtual bool use_xft() = 0;
 
 protected:
     // Owner of mKkContext is XimIC. This is set at the time of
@@ -60,11 +64,13 @@ protected:
     InputContext *mKkContext;
     icxatr *m_atr;
     pe_stat *m_pe; // Initially NULL.
-    const char *mLang;
+    const char *mIMLang;
     const char *mEncoding;
+    const char *mLocaleName;
 };
 
 Convdisp *create_convdisp(int style, InputContext *, icxatr *, Connection *);
+XFontSet get_font_set(const char *name, const char *locale);
 
 #endif
 /*
