@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -37,21 +37,27 @@
 # include "config.h"
 #endif
 
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
 #include <stdio.h>
 #include <string.h>
 #include <locale.h>
-#include <ctype.h>
 #include <stdlib.h>
-#include <pwd.h>
-#include <unistd.h>
 #include <signal.h>
-#include <sys/types.h>
+#include <sys/select.h>
+
 #include "xim.h"
 #include "xdispatch.h"
 #include "ximserver.h"
 #include "canddisp.h"
+#include "connection.h"
+#include "util.h"
+#include "helper.h"
+
 #include "uim/uim-util.h"
 #include "uim/uim-im-switcher.h"
+#include "uim/uim-compat-scm.h"
 
 Display *XimServer::gDpy;
 std::map<Window, XimServer *> XimServer::gServerMap;
@@ -61,14 +67,14 @@ int g_option_mask;
 int scr_width, scr_height;
 int host_byte_order;
 
-int lib_uim_fd;
-
 #define VERSION_NAME "uim-xim under the way! Version "PACKAGE_VERSION"\n"
 const char *version_name=VERSION_NAME;
 const char *usage=
 "--help , --version :Show usage or version\n"
 "--list             :Show available backend conversion engines\n"
 "--engine=ENGINE    :Use ENGINE as a backend conversion engine at startup\n"
+"--async            :Use on-demand-synchronous method of XIM event flow\n"
+"                    (using this option is not safe for Tcl/Tk GUI toolkit)\n"
 "--trace            :trace-connection\n"
 "--trace-xim        :trace-xim-message\n";
 const char *default_engine;
@@ -78,10 +84,6 @@ static Atom atom_locales;
 static Atom atom_transport;
 Atom xim_servers;
 
-// fd dispatch
-#define READ_OK 1
-#define WRITE_OK 2
-
 struct fd_watch_struct {
     int mask;
     void (*fn)(int, int);
@@ -91,6 +93,7 @@ static std::map<unsigned int, WindowIf *> window_watch_stat;
 
 static char *supported_locales;
 std::list<UIMInfo> uim_info;
+static void check_pending_xevent(void);
 
 bool
 pretrans_register()
@@ -135,12 +138,17 @@ void add_fd_watch(int fd, int mask, void (*fn)(int, int))
     fd_watch_stat.insert(p);
 }
 
-void main_loop()
+static void main_loop()
 {
     fd_set rfds, wfds;
+    struct timeval tv;
+    
     while (1) {
 	FD_ZERO(&rfds);
 	FD_ZERO(&wfds);
+	tv.tv_sec = 2;
+	tv.tv_usec = 0;
+
 	std::map<int, fd_watch_struct>::iterator it;
 	int  fd_max = 0;
 	for (it = fd_watch_stat.begin(); it != fd_watch_stat.end(); it++) {
@@ -152,7 +160,11 @@ void main_loop()
 	    if (fd_max < fd)
 		fd_max = fd;
 	}
-	select(fd_max + 1, &rfds, &wfds, NULL, NULL);
+	if ((select(fd_max + 1, &rfds, &wfds, NULL, &tv)) == 0) {
+	    check_pending_xevent();
+	    continue;
+	}
+
 	for (it = fd_watch_stat.begin(); it != fd_watch_stat.end(); it++) {
 	    int fd = it->first;
 	    if (FD_ISSET(fd, &rfds))
@@ -169,7 +181,7 @@ add_window_watch(Window id, WindowIf *w, int mask)
     std::pair<unsigned int, WindowIf *> p(id, w);
     window_watch_stat.insert(p);
 
-    // Event mash has same value defined in X,
+    // Event mask is the same value defined in X,
     // but do not depend on.
     int emask = 0;
     if (mask & EXPOSE_MASK)
@@ -203,8 +215,13 @@ findWindowIf(Window w)
 static int
 X_ErrorHandler(Display *d, XErrorEvent *e)
 {
-    if (g_option_mask & OPT_TRACE)
-	printf("X error occured.\n");
+    if (g_option_mask & OPT_TRACE) {
+	if (e->error_code) {
+	    char buf[64];
+	    XGetErrorText(d, e->error_code, buf, 63);
+	    printf("X error occured. %s\n", buf);
+	}
+    }
 
     return 0;
 }
@@ -212,12 +229,13 @@ X_ErrorHandler(Display *d, XErrorEvent *e)
 static int
 X_IOErrorHandler(Display *d)
 {
+    fprintf(stderr, "%s: X IO error.\n", DisplayString(d));
     return 0;
 }
 
 
 static void
-sendSelectionNotify(XEvent *ev, char *buf, int len)
+sendSelectionNotify(XEvent *ev, const char *buf, int len)
 {
     XEvent e;
     e.type = SelectionNotify;
@@ -294,10 +312,8 @@ ProcXEvent(XEvent *e)
 }
 
 static void
-xEventRead(int fd, int ev)
+check_pending_xevent(void)
 {
-    XFlush(XimServer::gDpy);
-
     XEvent e;
     while (XPending(XimServer::gDpy)) {
 	XNextEvent(XimServer::gDpy, &e);
@@ -305,11 +321,23 @@ xEventRead(int fd, int ev)
     }
 }
 
-static int
-pretrans_setup()
+static void
+xEventRead(int fd, int ev)
+{
+    XFlush(XimServer::gDpy);
+    check_pending_xevent();
+}
+
+static void
+error_handler_setup()
 {
     XSetErrorHandler(X_ErrorHandler);
     XSetIOErrorHandler(X_IOErrorHandler);
+}
+
+static int
+pretrans_setup()
+{
     int fd = XConnectionNumber(XimServer::gDpy);
 
     add_fd_watch(fd, READ_OK, xEventRead);
@@ -319,14 +347,14 @@ pretrans_setup()
 static void
 print_version()
 {
-    printf(version_name);
+    printf("%s", version_name);
 }
 
 static void
 print_usage()
 {
     print_version();
-    printf(usage);
+    printf("%s", usage);
     exit(0);
 }
 
@@ -349,10 +377,7 @@ get_uim_info()
     for (int i = 0; i < nr; i++) {
 	ui.name = strdup(uim_get_im_name(uc, i));
 	ui.lang = strdup(uim_get_im_language(uc, i));
-	ui.locale = createLocale(ui.lang);
 	ui.desc = strdup(uim_get_im_short_desc(uc, i));
-	if (!ui.locale)
-	    continue;
 	uim_info.push_back(ui);
     }
     uim_release_context(uc);
@@ -372,49 +397,48 @@ print_uim_info()
     
 }
 
+static void
+clear_uim_info()
+{
+    std::list<UIMInfo>::iterator it;
+    for (it = uim_info.begin(); it != uim_info.end(); it++) {
+	free(it->name);
+	free(it->lang);
+	free(it->desc);
+    }
+    uim_info.clear();
+}
+
 static void
 init_supported_locales()
 {
     std::list<char *> locale_list;
-    char *name;
-    char locale[20];
-    unsigned int len;
+    char *locales;
+    int len;
 
     asprintf(&supported_locales, "@locale=");
+    len = strlen(supported_locales);
 
-    std::list<UIMInfo>::iterator i;
-    std::list<char *>::iterator j;
-    for (i = uim_info.begin(); i != uim_info.end(); i++) {
-	name = strdup(i->locale->compose_localename_from_lang(i->lang));
-	// Hack for IM with several locales.  Should rewite these
-	// locale handling later.
-	if (!strcmp(name, "zh_TW")) {
-		free(name);
-		name = strdup("zh_TW,zh_HK");
-	}
-
-	bool is_exist = false;
-	for (j = locale_list.begin(); j != locale_list.end(); j++) {
-	    if (!strcmp(*j, name)) {
-		is_exist = true;
-		break;
-	    }
-	}
-	if (is_exist == true)
-		continue;
-
-	locale_list.push_back(name);
-	snprintf(locale, sizeof(locale), "%s,", name); // Trailing comma seems to be OK...
-	len = strlen(supported_locales) + strlen(locale);
-	supported_locales = (char *)realloc(supported_locales,
-			     sizeof(char) * len + 1);
-	if (!supported_locales)
-	    break;
-	strcat(supported_locales, locale);
+    // get all locales
+    locales = strdup(compose_localenames_from_im_lang("*"));
+    // replace ':' with ','
+    char *sep;
+    char *tmp = locales;
+    while ((sep = strchr(tmp, ':')) != NULL) {
+	*sep = ',';
+	tmp = sep;
     }
-    for (j = locale_list.begin(); j != locale_list.end(); j++) {
-	free(*j);
+
+    len += strlen(locales);
+    supported_locales = (char *)realloc(supported_locales,
+		    sizeof(char) * len + 1);
+    if (!supported_locales) {
+	fprintf(stderr, "Error: failed to register supported_locales. Aborting....");
+	exit(1);
     }
+
+    strcat(supported_locales, locales);
+    free(locales);
 }
 
 static void
@@ -440,6 +464,8 @@ parse_args(int argc, char **argv)
 		g_option_mask |= OPT_TRACE_XIM;
 	    } else if (!strncmp(opt, "engine=", 7)) {
 		default_engine = strdup(&argv[i][9]);
+	    } else if (!strcmp(opt, "async")) {
+		g_option_mask |= OPT_ON_DEMAND_SYNC;
 	    }
 	}
     }
@@ -474,178 +500,49 @@ get_runtime_env()
 }
 
 static void
-parse_helper_str_im_change(const char *level, const char *engine) {
+terminate_x_connection()
+{
+    int fd = XConnectionNumber(XimServer::gDpy);
+
+    remove_current_fd_watch(fd);
+}
+
+static void
+reload_uim(int x)
+{
+    fprintf(stderr, "\nReloading uim...\n\n");
+
+    terminate_canddisp_connection();
+    helper_disconnect_cb();
+    terminate_x_connection();
+
+    std::map<Window, XimServer *>::iterator it;
+    std::list<InputContext *>::iterator it_c;
+
+    for (it = XimServer::gServerMap.begin(); it != XimServer::gServerMap.end(); it++) {
+	XimServer *xs = it->second;
+	for (it_c = xs->ic_list.begin(); it_c != xs->ic_list.end(); it_c++)
+	    (*it_c)->clear();
+    }
+
+    uim_quit();
+    clear_uim_info();
+    get_uim_info();
+    print_uim_info();
+
+    for (it = XimServer::gServerMap.begin(); it != XimServer::gServerMap.end(); it++) {
+	XimServer *xs = it->second;
+	for (it_c = xs->ic_list.begin(); it_c != xs->ic_list.end(); it_c++) {
+	    const char *engine = (*it_c)->get_engine_name();
+	    (*it_c)->createUimContext(engine);
+	}
+    }
+
     InputContext *focusedContext = InputContext::focusedContext();
+    if (focusedContext)
+	focusedContext->focusIn();
 
-    if (!strcmp(level, "im_change_whole_desktop")) {
-	std::map<Window, XimServer *>::iterator it;
-	for (it = XimServer::gServerMap.begin(); it != XimServer::gServerMap.end(); it++) {
-	    (*it).second->changeContext(engine);
-	}
-    }
-
-    if (focusedContext) {
-	if (!strcmp(level, "im_change_this_text_area_only"))
-	    focusedContext->changeContext(engine);
-	else if (!strcmp(level, "im_change_this_application_only"))
-	    get_im_by_id(focusedContext->get_ic()->get_imid())->changeContext(engine);
-    }
-}
-
-static void
-send_im_list(void)
-{
-    char *buf = NULL, *tmp = NULL;
-    int len;
-    InputContext *focusedContext = InputContext::focusedContext();
-    const char *current_im_name =
-	    uim_get_current_im_name(focusedContext->getUC());
-    const char *encoding = focusedContext->get_ic()->get_encoding();
-    char *supported_lang = NULL;
-    bool need_check = false;
-    
-    if (strcmp(encoding, "UTF-8")) {
-	need_check = true;
-	supported_lang = strdup(find_localename_from_encoding(encoding));
-	tmp = strstr(supported_lang, ".");
-	*tmp = '\0';
-    }
-
-    asprintf(&buf, "im_list\ncharset=UTF-8\n");
-    if (!buf)
-	return;
-
-    std::list<UIMInfo>::iterator it;
-    for (it = uim_info.begin(); it != uim_info.end(); it++) {
-	if (need_check == true) {
-	    // XXX zh_CN or zh_TW?
-	    if (!strncmp(supported_lang, "zh", 2)) {
-		if (strncmp(it->lang, supported_lang, 5) &&
-				strcmp(it->lang, "*"))
-		    continue;
-	    } else {
-		if (strncmp(it->lang, supported_lang, 2) &&
-				strcmp(it->lang, "*"))
-		    continue;
-	    }
-	}
-
-	const char *language;
-	language = uim_get_language_name_from_locale(it->lang);
-	asprintf(&tmp, "%s\t%s\t%s\t", it->name,
-				       language ? language : "",
-				       it->desc ? it->desc : "");
-	if (!tmp)
-	    return;
-	len = strlen(buf) + strlen(tmp);
-	buf = (char *)realloc(buf, sizeof(char) * len + 1);
-	if (!buf)
-	    return;
-	strcat(buf, tmp);
-	free(tmp);
-
-	if (!strcmp(it->name, current_im_name)) {
-	    asprintf(&tmp, "selected\n");
-	    if (!tmp)
-		return;
-	    len = strlen(buf) + strlen(tmp);
-	    buf = (char *)realloc(buf, sizeof(char) * len + 1);
-	    if (!buf)
-		return;
-	    strcat(buf, tmp);
-	    free(tmp);
-	} else {
-	    asprintf(&tmp, "\n");
-	    if (!tmp)
-		return;
-	    len = strlen(buf) + strlen(tmp);
-	    buf = (char *)realloc(buf, sizeof(char) * len + 1);
-	    if (!buf)
-		return;
-	    strcat(buf, tmp);
-	    free(tmp);
-	}
-    }
-    uim_helper_send_message(lib_uim_fd, buf);
-    free(buf);
-    if (supported_lang)
-	free(supported_lang);
-}
-
-static void
-helper_str_parse(char *str)
-{
-    InputContext *focusedContext = InputContext::focusedContext();
-    
-    if (focusedContext) {
-	char *line = str;	
-	char *eol = strchr(line, '\n');
-	if (eol != NULL)
-	    *eol = '\0';
-
-	if (strcmp("prop_list_get", line) == 0) {
-	    uim_prop_list_update(focusedContext->getUC());
-	} else if (strcmp("prop_label_get", line) == 0) {
-	    uim_prop_label_update(focusedContext->getUC());
-	} else if (strcmp("prop_activate", line) == 0) {
-	    line = eol + 1;
-	    eol = strchr(line, '\n');
-	    if (eol != NULL) {
-		*eol = '\0';
-	    }
-	    uim_prop_activate(focusedContext->getUC(), line);
-	} else if (strncmp("focus_in", line, 8) == 0) {
-	    InputContext::deletefocusedContext();
-	    Canddisp *disp = canddisp_singleton();
-	    disp->hide();
- 	} else if (strcmp("im_list_get", line) == 0) {
-	    send_im_list();
-	} else if (strncmp("im_change_", line, 10) == 0) {
-	    char *engine;
-	    engine = eol + 1;
-	    eol = strchr(engine, '\n');
-	    if (eol != NULL) {
-		*eol = '\0';
-	    }
-	    parse_helper_str_im_change(line, engine);
-	}
-    } else {
-	char *line = str;	
-	char *eol = strchr(line, '\n');
-	if (eol != NULL) {
-	    *eol = '\0';
-	}
-
-	if (strncmp("im_change_", line, 10) == 0) {
-	    char *engine;
-	    engine = eol + 1;
-	    eol = strchr(engine, '\n');
-	    if (eol != NULL) {
-		*eol = '\0';
-	    }
-	    parse_helper_str_im_change(line, engine);
-	} else if (strcmp("prop_update_custom", line) == 0) {
-	    fprintf(stderr, "prop_update_custom is not implemented yet\n");
-	}
-    }
-}
-
-static void
-helper_read_cb(int fd, int ev)
-{
-    uim_helper_read_proc(fd);
-    char *tmp;
-    while ((tmp = uim_helper_get_message())) {
-	helper_str_parse(tmp);
-	free(tmp);
-    }
-}
-
-static void
-helper_disconnect_cb(void)
-{
-    remove_current_fd_watch(lib_uim_fd);
-    lib_uim_fd = -1;
+    pretrans_setup();
 }
 
 int
@@ -653,16 +550,21 @@ main(int argc, char **argv)
 {
     const char *locale;
 
-    printf("UIM-XIM bridge. Now testing to support multiple locales.\n");
+    printf("UIM-XIM bridge. Now supporting multiple locales.\n");
 
     get_runtime_env();
 
     parse_args(argc, argv);
-    signal(SIGPIPE, SIG_IGN);
 
-    lib_uim_fd = uim_helper_init_client_fd(helper_disconnect_cb);
-    if (lib_uim_fd >= 0)
-	add_fd_watch(lib_uim_fd, READ_OK, helper_read_cb);
+    if (g_option_mask & OPT_ON_DEMAND_SYNC)
+	printf("Using on-demand-synchronous XIM event flow (not safe for Tcl/TK)\n");
+    else
+	printf("Using full-synchronous XIM event flow\n");
+
+    signal(SIGPIPE, SIG_IGN);
+    signal(SIGUSR1, reload_uim);
+
+    check_helper_connection();
 
     XimServer::gDpy = XOpenDisplay(NULL);
     if (!XimServer::gDpy) {
@@ -684,7 +586,6 @@ main(int argc, char **argv)
     check_default_engine(locale);
     init_supported_locales();
     init_modifier_keys();
-    init_convdisp();
 
     std::list<UIMInfo>::iterator it;
     bool res = false;
@@ -693,7 +594,7 @@ main(int argc, char **argv)
     // "default-im-name" on ~/.uim.
     for (it = uim_info.begin(); it != uim_info.end(); it++) {
 	if (strcmp(it->name, default_engine) == 0) {
-	    XimServer *xs = new XimServer(it->locale, it->name, it->lang);
+	    XimServer *xs = new XimServer(it->name, it->lang);
 	    res = xs->setupConnection(true);
 	    if (res)
 		printf("XMODIFIERS=@im=uim registered, selecting %s (%s) as default conversion engine\n", it->name, it->lang);
@@ -703,68 +604,28 @@ main(int argc, char **argv)
 	}
     }
 
-    // XXX Setup each backend as a separated XimServer, aimed for
-    // backward compatibility, and should be removed in the future.
-    for (it = uim_info.begin(); it != uim_info.end(); it++) {
-	XimServer *xs = new XimServer(it->locale, it->name, it->lang);
-	res = xs->setupConnection(false);
-	if (res)
-	    printf("XMODIFIERS=@im=uim-%s (%s) registered\n",
-			    it->name, it->lang);
-	else
-	    delete xs;
-    }
-
     if (!res) {
 	printf("aborting...\n");
 	return 1;
     }
 
-    // Please don't insert time consuming procedure between
-    // connection_setup() and main_loop().
     connection_setup();
+    error_handler_setup();
     if (pretrans_setup() == -1)
 	return 0;
 
+#if HAVE_XFT_UTF8_STRING
+    if (uim_scm_symbol_value_bool("uim-xim-use-xft-font?"))
+	init_default_xftfont(); // setup Xft fonts for Ov/Rw preedit
+#endif
+
+    // Handle pending events to prevent hang just after startup
+    check_pending_xevent();
+
     main_loop();
     return 0;
 }
 
-int
-pad4(int x)
-{
-    return (4 - (x % 4)) % 4;
-}
-
-void
-hex_dump(unsigned char *buf, int len)
-{
-    int i, j;
-    unsigned char c[16];
-    for (i = 0; i < len; i += 16) {
-	printf("%8.8x ", i);
-	for (j = 0; j < 16; j++) {
-	    if (i + j < len) {
-		c[j] = buf[i + j];
-		if (j == 7)
-		    printf("%2.2x-", c[j]);
-		else
-		    printf("%2.2x ", c[j]);
-	    } else {
-		c[j] = 0;
-		printf("-- ");
-	    }
-	}
-	printf(" ");
-	for (j = 0; j < 16; j++) {
-	    if (isprint(c[j]))
-		printf("%c", c[j]);
-	    else
-		printf(".");
-	}
-	printf("\n");
-    }
-}
 /*
  * Local variables:
  *  c-indent-level: 4
