@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -37,9 +37,11 @@
 #include <X11/X.h>
 #include <X11/Xlib.h>
 #include <X11/keysym.h>
+#include <X11/Xos.h>
 #include <list>
 #include "ximserver.h"
 #include "ximpn.h"
+#include "compose.h"
 
 // The header file contains about connection of XIM protocol, and the
 // definition of IM and IC.
@@ -68,7 +70,7 @@ public:
     virtual int pushC16(unsigned int) = 0;
     virtual int pushC32(unsigned int) = 0;
     virtual int pushSTRING(char *) = 0;
-    virtual int pushBytes(char *, int) = 0;
+    virtual int pushBytes(const char *, int) = 0;
 
     virtual int pop_back() = 0;
 };
@@ -106,11 +108,10 @@ public:
     void OnRecv();
     void OnSend();
     void OnClose();
-    virtual void OnPushPacket() {}; // Called when packet is pushed into sending queue.  But doesn't do anything for now.
     void push_packet(TxPacket *); // for normal packet for reply
     void push_passive_packet(TxPacket *); // for preceding packet for reply
     int byte_order() {return mByteorder;};
-    void push_error_packet(int imid, int icid, int er, char *str);
+    void push_error_packet(int imid, int icid, int er, const char *str);
 
     unsigned short to_hs(unsigned short s);
     unsigned int to_hl(unsigned int l);
@@ -123,17 +124,19 @@ protected:
     virtual void setPreeditStartSyncFlag();
     virtual void unsetPreeditStartSyncFlag();
     virtual bool hasPreeditStartSyncFlag();
+    virtual void setPreeditCaretSyncFlag();
+    virtual void unsetPreeditCaretSyncFlag();
+    virtual bool hasPreeditCaretSyncFlag();
 
     std::list<RxPacket *> mRxQ;
-    std::list<RxPacket *> mPendingRxQ; // pending queue for XIM_FORWARD Rx event
     std::list<TxPacket *> mTxQ;
     std::list<TxPacket *> mPTxQ;
-    std::list<TxPacket *> mPendingTxQ; // pending queue for mTxQ
+    std::list<TxPacket *> mPendingTxQ; // pending queue for mTxQ and mPTxQ
     int mByteorder;
     bool mIsCloseWait; // true when the last packet has handled
 private:
     void xim_connect(RxPacket *);
-    void xim_disconnect(RxPacket *);
+    void xim_disconnect();
     void xim_open(RxPacket *);
     void xim_query_extension(RxPacket *);
     void xim_encoding_negotiation(RxPacket *);
@@ -150,18 +153,20 @@ private:
     void xim_reset_ic(RxPacket *);
 
     void xim_forward_event(RxPacket *);
-    void xim_sync_reply(RxPacket *);
-    void xim_preedit_start_reply(RxPacket *);
+    void xim_sync_reply();
+    void xim_preedit_start_reply();
+    void xim_preedit_caret_reply();
     void xim_error(RxPacket *);
 
     bool is_xim_sync_reply_timeout();
-    void clear_pending_rx();
+    void clear_pending_queue();
 private:
     XimIC *get_ic(RxPacket *);
     std::list<int> mCreatedIm;
     XimServer *mServer;
     bool mSyncFlag;
     bool mPreeditStartSyncFlag;
+    bool mPreeditCaretSyncFlag;
     struct timeval mSyncStartTime;
 };
 
@@ -169,7 +174,7 @@ private:
 class XimIM {
 public:
     XimIM(Connection *, int id);
-    virtual ~XimIM() {};
+    virtual ~XimIM();
     
     virtual void create_ic(RxPacket *) = 0;
     virtual void destroy_ic(int) = 0;
@@ -184,11 +189,30 @@ public:
     virtual void changeContext(const char *engine) = 0;
     void set_encoding(const char *encoding);
     const char *get_encoding();
+    void set_lang_region(const char *name);
+    const char *get_lang_region();
+    char *uStringToCtext(uString *us);
+    char *utf8_to_native_str(char *str);
+    struct input_style *getInputStyles();
+    // for Compose
+    void create_compose_tree();
+    DefTree *get_compose_tree();
 
 protected:
     Connection *mConn;
+    Locale *mLocale;
     int mID;
     char *mEncoding;
+    char *mLangRegion;
+
+    // for Compose
+    char *get_compose_filename();
+    char *TransFileName(char *name);
+    void ParseComposeStringFile(FILE *fp);
+    void FreeComposeTree(DefTree *top);
+    int parse_compose_line(FILE *fp, char *tokenbuf);
+    int get_mb_string(char *buf, KeySym ks);
+    DefTree *mTreeTop;
 };
 
 int unused_im_id();
@@ -213,12 +237,14 @@ class icxatr {
 public:
     icxatr();
     ~icxatr();
-    void set_atr(int id, C8 *v, int len, int byte_order);
+    void set_atr(int id, C8 *v, int byte_order);
     bool has_atr(int id);
     bool is_changed(int id);
     void unset_change_mask(int id);
     void print();
     int getSize(int id);
+    void set_locale_name(const char *locale);
+    bool use_xft();
 
     unsigned long input_style;
     Window client_window;
@@ -230,13 +256,15 @@ public:
     XPoint spot_location;
     XRectangle area;
 
-    char *font_set_name;
     XFontSet font_set;
+    char *font_set_name;
     C16 line_space;
 
 private:
+    char *m_locale;
     int atr_mask;
     int change_mask;
+    bool m_use_xft;
 };
 
 // definition of IC
@@ -261,6 +289,8 @@ public:
     void force_send_packet();
     void changeContext(const char *engine);
     const char *get_encoding();
+    const char *get_lang_region();
+    void move_candwin();
 
 public:
     static XimIC *get_current_ic();
@@ -284,6 +314,7 @@ private:
     int mIMid;
     uString mPending;
     bool mIsActive;
+    keyState *m_keyState;
 private:
     static XimIC *current_ic;
     static int nrActiveIC;
