@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,23 +31,27 @@
 */
 
 // Handle IC (Input Context) part of XIM protocol.
-// Intermidiate key handling between backend IM engine.
+// Intermediate key handling between backend IM engine.
  
 #ifdef HAVE_CONFIG_H
 # include "config.h"
 #endif
  
 #include <stdlib.h>
+#include <string.h>
+#include <stdio.h>
 #include <ctype.h>
+
 #include "xim.h"
 #include "convdisp.h"
 #include "ximserver.h"
 #include "connection.h"
+#include "util.h"
+
+#include "uim/uim-compat-scm.h"
  
-#ifndef __GNUC__
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
-# endif
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
 #endif
 
 char invalid_style_msg[]=
@@ -56,20 +60,21 @@ char invalid_style_msg[]=
 XimIC *XimIC::current_ic;
 int XimIC::nrActiveIC;
 
-struct FSCache{
+struct FSCache {
     int refc;
     XFontSet fs;
     char *name;
+    char *locale;
 };
 
 static std::list<FSCache> fs_cache;
 
-static XFontSet
-get_font_set(char *name)
+XFontSet
+get_font_set(const char *name, const char *locale)
 {
     std::list<FSCache>::iterator it;
     for (it = fs_cache.begin(); it != fs_cache.end(); it++) {
-	if (!strcmp(it->name, name)) {
+	if (!strcmp(it->name, name) && !strcmp(it->locale, locale)) {
 	    it->refc++;
 	    return it->fs;
 	}
@@ -82,22 +87,29 @@ get_font_set(char *name)
     if (missing)
 	XFreeStringList(missing);
 
+    if (fc.fs == NULL) {
+	fprintf(stderr, "Critical: XCreateFontSet failed!\n");
+	return NULL;
+    }
+
     fc.refc = 1;
     fc.name = strdup(name);
+    fc.locale = strdup(locale);
     fs_cache.push_back(fc);
     return fc.fs;
 }
 
 static void
-release_font_set(char *name)
+release_font_set(const char *name, const char *locale)
 {
     std::list<FSCache>::iterator it;
     for (it = fs_cache.begin(); it != fs_cache.end(); it++) {
-	if (!strcmp(it->name, name)) {
+	if (!strcmp(it->name, name) && !strcmp(it->locale, locale)) {
 	    it->refc--;
 	    if (!it->refc) {
 		XFreeFontSet(XimServer::gDpy, it->fs);
 		free(it->name);
+		free(it->locale);
 		fs_cache.erase(it);
 	    }
 	    return;
@@ -149,16 +161,25 @@ icxatr::icxatr()
     change_mask = 0;
     font_set_name = NULL;
     font_set = NULL;
+    m_locale = NULL;
     foreground_pixel = BlackPixel(XimServer::gDpy, DefaultScreen(XimServer::gDpy));
     background_pixel = WhitePixel(XimServer::gDpy, DefaultScreen(XimServer::gDpy));
+    line_space = 0;
+#if HAVE_XFT_UTF8_STRING
+    m_use_xft = uim_scm_symbol_value_bool("uim-xim-use-xft-font?");
+#else
+    m_use_xft = false;
+#endif
 }
 
 icxatr::~icxatr()
 {
     if (font_set_name) {
-	release_font_set(font_set_name);
+	if (use_xft() == false)
+	    release_font_set(font_set_name, m_locale);
 	free((void *)font_set_name);
     }
+    free(m_locale);
 }
 
 bool icxatr::has_atr(int id)
@@ -166,7 +187,7 @@ bool icxatr::has_atr(int id)
     return atr_mask & (1 << id);
 }
 
-void icxatr::set_atr(int id, C8 *val, int len, int o)
+void icxatr::set_atr(int id, C8 *val, int o)
 {
     switch (id) {
     case ICA_InputStyle:
@@ -192,16 +213,15 @@ void icxatr::set_atr(int id, C8 *val, int len, int o)
     {
 	int len = readC16(val, o);
 	char *new_fsn = (char *)alloca(len + 1);
-	new_fsn[len] = 0;
+	new_fsn[len] = '\0';
 	memcpy(new_fsn, &val[2], len);
 	if (font_set_name && !strcmp(font_set_name, new_fsn))
 	    break;
 
-	if (font_set_name)
-	    free(font_set_name);
-
+	free(font_set_name);
 	font_set_name = strdup(new_fsn);
-	font_set = get_font_set(font_set_name);
+	if (use_xft() == false)
+	    font_set = get_font_set(font_set_name, m_locale);
     }
     break;
     case ICA_Area:
@@ -213,7 +233,7 @@ void icxatr::set_atr(int id, C8 *val, int len, int o)
     }
     break;
     case ICA_LineSpace:
-	line_space=readC16(val, o);
+	line_space = readC16(val, o);
 	break;
     default:
 	// unknown attribute
@@ -301,6 +321,17 @@ int icxatr::getSize(int id)
     return 0;
 }
 
+void icxatr::set_locale_name(const char *locale)
+{
+    if (m_locale)
+	free(m_locale);
+    m_locale = strdup(locale);
+}
+
+bool icxatr::use_xft() {
+    return m_use_xft;
+}
+
 XimIC::XimIC(Connection *c, int imid, int icid, const char *engine)
 {
     mConn = c;
@@ -310,7 +341,12 @@ XimIC::XimIC(Connection *c, int imid, int icid, const char *engine)
 
     XimServer *svr = mConn->getXimServer();
     m_kkContext = svr->createContext(this, engine);
+
+    const char *locale = m_kkContext->get_locale_name();
+    m_xatr.set_locale_name(locale);
+
     mConvdisp = 0;
+    m_keyState = new keyState(this);
     if (g_option_mask & OPT_TRACE)
 	printf("imid=%d, icid=%d ic created.\n", mIMid, mICid);
 }
@@ -328,6 +364,7 @@ XimIC::~XimIC()
     delete m_kkContext;
     if (mConvdisp)
 	delete mConvdisp;
+    delete m_keyState;
 }
 
 bool XimIC::isActive()
@@ -345,6 +382,12 @@ int XimIC::get_imid()
     return mIMid;
 }
 
+void XimIC::move_candwin()
+{
+    if (mConvdisp)
+	mConvdisp->move_candwin();
+}
+
 void XimIC::setFocus()
 {
     if (!mIsActive)
@@ -352,18 +395,22 @@ void XimIC::setFocus()
 
     current_ic = this;
     mIsActive = true;
+
+    bool is_candwin_active = m_kkContext->hasActiveCandwin();
+
+    if (mConvdisp && is_candwin_active == false) {
+	// unset_focus before move_candwin in m_kkContext->focusIn()
+	mConvdisp->unset_focus();
+    }
+
+    mConvdisp->move_candwin();
     m_kkContext->focusIn();
 
-    if (mConvdisp) {
+    if (mConvdisp && is_candwin_active == true) {
 	// Updating preedit here causes string mismatch if the context
 	// receives XIM_RESET_IC after XIM_SET_IC_FOCUS.  Should only
 	// update candidate window.
-	if (m_kkContext->hasActiveCandwin()) {
-	    mConvdisp->move_candwin();
-	    m_kkContext->candidate_update();
-	} else {
-	    mConvdisp->unset_focus();
-	}
+	m_kkContext->candidate_update();
     }
 }
 
@@ -375,24 +422,22 @@ void XimIC::unsetFocus()
     mIsActive = false;
     nrActiveIC--;
     m_kkContext->focusOut();
-     /*
-      * Since the sequence of XIM_SET_IC_FOCUS and XIM_UNSET_FOCUS
-      * events is not consistent, unsetting focus of candidate window
-      * is now handled in XimIC::setFocus() and focus_in message from
-      * helper application.
-      *
-      * if (mConvdisp && m_kkContext->hasActiveCandwin()) {
-      *     mConvdisp->unset_focus();
-      * }
-      */
+    // Since the sequence of XIM_SET_IC_FOCUS and XIM_UNSET_FOCUS
+    // events is not consistent, unsetting focus of candidate window
+    // is now handled in XimIC::setFocus() and focus_in message from
+    // helper application.
+    //
+    // if (mConvdisp && m_kkContext->hasActiveCandwin()) {
+    //     mConvdisp->unset_focus();
+    // }
 }
 
 void XimIC::OnKeyEvent(keyEventX e)
 {
     int s;
-    keyState k(&e);
+    m_keyState->check_key(&e);
 
-    s = m_kkContext->pushKey(&k);
+    s = m_kkContext->pushKey(m_keyState);
     if (s & COMMIT_RAW)
 	send_key_event(&e.ev.xkey);
 
@@ -411,7 +456,7 @@ void XimIC::send_key_event(XKeyEvent *e)
     t = createTxPacket(XIM_FORWARD_EVENT, 0);
     t->pushC16(mIMid);
     t->pushC16(mICid);
-    t->pushC16(1); //flag , synchronous
+    t->pushC16(1); // flag, synchronous
     t->pushC16((e->serial >> 16) & 0xffff);
 
     t->pushC8(e->type);
@@ -440,14 +485,16 @@ void XimIC::commit_string(const char *str)
 
 void XimIC::extra_input(char *s)
 {
-    if (!m_kkContext->extra_input(s))
-	commit_string(s);
-    if (m_xatr.has_atr(ICA_FocusWindow))
-	force_event(m_xatr.focus_window);
+    if (s == NULL)
+	return;
+
+    commit_string(s);
+    onSendPacket();
+    force_send_packet();
 }
 
 void XimIC::force_send_packet(void) {
-    ((XConnection *)mConn)->writeProc();
+    (dynamic_cast<XConnection *>(mConn))->writeProc();
 }
 
 void XimIC::setICAttrs(void *val, int len)
@@ -486,7 +533,12 @@ int XimIC::get_ic_atr(int id, TxPacket *t)
 	t->pushC32(m_xatr.focus_window);
 	break;
     case ICA_FilterEvents:
-	t->pushC32(KeyPressMask);
+	if (g_option_mask & OPT_ON_DEMAND_SYNC)
+	    t->pushC32(KeyPressMask|KeyReleaseMask);
+	else // Filtering KeyRelease event with full-synchronous method
+	     //	causes problem with mozilla (1.7.3) gtk2 on navigation
+	     //	toolbar's auto text completion...
+	    t->pushC32(KeyPressMask); 
 	break;
     case ICA_InputStyle:
 	t->pushC32(m_xatr.input_style);
@@ -501,7 +553,7 @@ int XimIC::get_ic_atr(int id, TxPacket *t)
 int XimIC::lookup_style(unsigned long s)
 {
     int i;
-    struct input_style *is = mConn->getXimServer()->getInputStyles();
+    struct input_style *is = get_im_by_id(mIMid)->getInputStyles();
     for (i = 0; is[i].x_style; i++) {
 	if (is[i].x_style == (int)s)
 	    return is[i].style;
@@ -514,7 +566,7 @@ void XimIC::set_ic_attr(int id, C8 *val, int len)
     if (id == ICA_PreeditAttribute || id == ICA_StatusAttributes)
 	setICAttrs(val, len); // list of attribute
     else
-	m_xatr.set_atr(id, val, len, mConn->byte_order());
+	m_xatr.set_atr(id, val, mConn->byte_order());
 
     if (mConvdisp)
 	mConvdisp->update_icxatr();
@@ -538,7 +590,6 @@ void XimIC::reset_ic()
     t->pushC16(mICid);
 
     uString s;
-    const char *encoding = get_encoding();
 
     // m_kkContext->get_preedit_string() returns uncommitted preedit
     // strings, which will be committed in client applications.
@@ -546,7 +597,7 @@ void XimIC::reset_ic()
     if (s.size()) {
 	char *p;
 	int i, len = 0;
-	p = mConn->getXimServer()->uStringToCtext(&s, encoding);
+	p = get_im_by_id(mIMid)->uStringToCtext(&s);
 	if (p) {
 	    len = strlen(p);
 	    t->pushC16(len); // length of committed strings
@@ -572,6 +623,9 @@ void XimIC::reset_ic()
     // After sending RESET_IC_REPLY, uim-xim needs to clear and reset
     // the input context.
     m_kkContext->clear();
+
+    // Also reset key state
+    m_keyState->reset();
 }
 
 Convdisp *XimIC::get_convdisp()
@@ -585,8 +639,7 @@ void XimIC::onSendPacket()
 	return;
 
     char *p;
-    const char *encoding = get_encoding();
-    p = mConn->getXimServer()->uStringToCtext(&mPending, encoding);
+    p = get_im_by_id(mIMid)->uStringToCtext(&mPending);
 
     if (!p) {
 	erase_ustring(&mPending);
@@ -598,7 +651,7 @@ void XimIC::onSendPacket()
     t->pushC16(mIMid);
     t->pushC16(mICid);
 
-    t->pushC16(3); //XLookupChars|synchronous
+    t->pushC16(3); // XLookupChars|synchronous
 
     int i, len;
     len = strlen(p);
@@ -637,12 +690,18 @@ const char *XimIC::get_encoding()
     return im->get_encoding();
 }
 
+const char *XimIC::get_lang_region()
+{
+    XimIM *im = get_im_by_id(mIMid);
+    return im->get_lang_region();
+}
+
 XimIC *create_ic(Connection *c, RxPacket *p, int imid, int icid, const char *engine)
 {
     XimIC *ic;
     ic = new XimIC(c, imid, icid, engine);
     p->rewind();
-    p->getC16();//discard
+    p->getC16(); // discard
     int atr_len = p->getC16();
     unsigned char *v;
     v = (unsigned char *)alloca(atr_len);
