@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -36,20 +36,46 @@
 # include "config.h"
 #endif
 
+#include <string.h>
+#include <stdlib.h>
 #include <map>
+
 #include "xim.h"
 #include <X11/Xutil.h>
 #define NEED_EVENTS	// for declaration of xEvent
 #include <X11/Xproto.h>
 
-#ifndef __GNUC__
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
-# endif
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
 #endif
 
 static std::map<int, XimIM *> g_ims;
 
+// tables
+static input_style input_style_tab_with_over_the_spot[] = {
+    {XIMPreeditNothing|XIMStatusNothing, IS_ROOT_WINDOW},
+    //{XIMPreeditPosition|XIMStatusArea, IS_OVER_THE_SPOT},// emacs
+    {XIMPreeditPosition|XIMStatusNothing, IS_OVER_THE_SPOT},
+    //{XIMPreeditCallbacks|XIMStatusCallbacks, IS_ON_THE_SPOT},// OOo
+    //{XIMPreeditArea|XIMStatusArea, IS_ROOT_WINDOW},
+    {XIMPreeditCallbacks|XIMStatusNothing, IS_ON_THE_SPOT},
+    {0, 0},
+};
+static input_style input_style_tab_without_over_the_spot[] = {
+    {XIMPreeditNothing|XIMStatusNothing, IS_ROOT_WINDOW},
+    //{XIMPreeditPosition|XIMStatusArea, IS_OVER_THE_SPOT},// emacs
+    //{XIMPreeditPosition|XIMStatusNothing, IS_OVER_THE_SPOT},
+    //{XIMPreeditCallbacks|XIMStatusCallbacks, IS_ON_THE_SPOT},// OOo
+    //{XIMPreeditArea|XIMStatusArea, IS_ROOT_WINDOW},
+    {XIMPreeditCallbacks|XIMStatusNothing, IS_ON_THE_SPOT},
+    {0, 0},
+};
+// XIMPreeditArea,XIMPreeditCallbacks,XIMPreeditPosition
+// XIMPreeditNothing,XIMPreeditNone
+// XIMStatusArea,XIMStatusCallbacks
+// XIMStatusNothing,XIMStatusNone
+
+
 class XimIM_impl : public XimIM {
 public:
     XimIM_impl(Connection *c, int id);
@@ -89,6 +115,11 @@ void XimIM_impl::create_ic(RxPacket *p)
 {
     XimIC *ic;
     int icid= unused_ic_id();
+
+    // create compose table with the first ic
+    if (icid == 1)
+	create_compose_tree();
+
     ic = ::create_ic(mConn, p, mID, icid, mEngineName);
     if (!ic) {
 	mConn->push_error_packet(mID, icid,
@@ -263,7 +294,7 @@ void XimIM_impl::forward_event(RxPacket *p)
 
     switch (k.ev.type) {
     case KeyPress:
-	//case KeyRelease:
+    case KeyRelease:
 	k.ev.xkey.window =
 	    mConn->to_hl(ev_raw.u.keyButtonPointer.event);
 	k.ev.xkey.root =
@@ -292,6 +323,8 @@ void XimIM_impl::forward_event(RxPacket *p)
 
 	if (ic)
 	    ic->OnKeyEvent(k);
+	if (!(g_option_mask & OPT_ON_DEMAND_SYNC))
+	    send_sync_reply(icid);
 	break;
     default:
 	printf("unknown type of forwarded event.(%d)\n", k.ev.type);
@@ -343,6 +376,31 @@ XimIM::XimIM(Connection *c, int id)
     mConn = c;
     mID = id;
     mEncoding = NULL;
+    mLangRegion = NULL;
+    mTreeTop = NULL;
+    mLocale = NULL;
+}
+
+XimIM::~XimIM()
+{
+    free(mEncoding);
+    free(mLangRegion);
+    FreeComposeTree(mTreeTop);
+    delete mLocale;
+}
+
+void XimIM::FreeComposeTree(DefTree *top)
+{
+   if (!top)
+	return;
+
+   if (top->succession)
+	FreeComposeTree(top->succession);
+   if (top->next)
+	FreeComposeTree(top->next);
+   free(top->mb);
+   free(top->utf8);
+   free(top);
 }
 
 void XimIM::set_encoding(const char *encoding)
@@ -350,6 +408,11 @@ void XimIM::set_encoding(const char *encoding)
     if (mEncoding)
 	free(mEncoding);
     mEncoding = strdup(encoding);
+
+    // set iconv environment
+    if (mLocale)
+      delete mLocale;
+    mLocale = createLocale(mEncoding);
 }
 
 const char *XimIM::get_encoding()
@@ -357,6 +420,44 @@ const char *XimIM::get_encoding()
     return mEncoding;
 }
 
+void XimIM::set_lang_region(const char *lang_and_region)
+{
+    if (mLangRegion)
+	free(mLangRegion);
+    mLangRegion = strdup(lang_and_region);
+}
+
+const char *XimIM::get_lang_region()
+{
+    return mLangRegion;
+}
+
+struct input_style *XimIM::getInputStyles()
+{
+    if (mLocale && mLocale->supportOverTheSpot())
+	return input_style_tab_with_over_the_spot;
+
+    return input_style_tab_without_over_the_spot;
+}
+
+char *XimIM::uStringToCtext(uString *us)
+{
+    char *ret = NULL;
+    if (mLocale)
+	ret = mLocale->uStringToCtext(us);
+
+    return ret;
+}
+
+char *XimIM::utf8_to_native_str(char *str)
+{
+    char *ret = NULL;
+    if (mLocale)
+	ret = mLocale->utf8_to_native_str(str);
+
+    return ret;
+}
+
 int unused_im_id()
 {
     int max_id;
