@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -35,7 +35,11 @@
 #endif
 
 #include <list>
+#include <string.h>
+#include <stdlib.h>
+
 #include "xim.h"
+#include "util.h"
 
 //
 // Routines for byte manipulating
@@ -196,7 +200,7 @@ private:
 
 class TxBytes : public TxElement {
 public:
-    TxBytes(char *s, int len) {
+    TxBytes(const char *s, int len) {
 	m_str = (char *)malloc(len);
 	m_len = len;
 	memcpy(m_str, s, len);
@@ -231,7 +235,7 @@ public:
     virtual int pushC16(unsigned int);
     virtual int pushC32(unsigned int);
     virtual int pushSTRING(char *);
-    virtual int pushBytes(char *, int);
+    virtual int pushBytes(const char *, int);
 
     virtual int pop_back();
 private:
@@ -315,7 +319,7 @@ int TxPacket_impl::pushSTRING(char *s)
     return e->get_size();
 }
 
-int TxPacket_impl::pushBytes(char *b, int len)
+int TxPacket_impl::pushBytes(const char *b, int len)
 {
     TxElement *e;
     e = new TxBytes(b, len);
@@ -406,7 +410,7 @@ RxPacket_impl::RxPacket_impl(unsigned char *b, int byte_order)
     rewind();
 }
 
-RxPacket_impl::RxPacket_impl(const RxPacket_impl& rhs)
+RxPacket_impl::RxPacket_impl(const RxPacket_impl& rhs) : RxPacket(rhs)
 {
     mLen = rhs.mLen;
     mBuf = (unsigned char *)malloc(mLen);
