@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -30,14 +30,31 @@
   SUCH DAMAGE.
 */
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE 
+#endif
+#include <stdio.h>
 #include <ctype.h>
 #include <locale.h>
+#include <X11/Xlib.h>
 #include <X11/Xatom.h>
+#include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#include <X11/keysymdef.h>
+
 #include "xim.h"
 #include "convdisp.h"
 #include "canddisp.h"
 #include "ximserver.h"
+#include "util.h"
+#include "helper.h"
+
+#include "uim/uim-helper.h"
+#include "uim/uim-compat-scm.h"
 
 #ifndef __GNUC__
 # ifdef HAVE_ALLOCA_H
@@ -48,41 +65,17 @@
 extern int lib_uim_fd;
 extern Atom xim_servers;
 InputContext *InputContext::mFocusedContext = NULL;
-extern std::list<UIMInfo> uim_info;
 
 static int check_modifier(std::list<KeySym> list);
 static int gShiftMask, gLockMask, gControlMask, gMod1Mask,
 	   gMod2Mask, gMod3Mask, gMod4Mask, gMod5Mask;
+static int gXNumLockMask;
 
 
-// tables
-static input_style input_style_tab_with_over_the_spot[] = {
-    {XIMPreeditNothing|XIMStatusNothing, IS_ROOT_WINDOW},
-    //{XIMPreeditPosition|XIMStatusArea, IS_OVER_THE_SPOT},// emacs
-    {XIMPreeditPosition|XIMStatusNothing, IS_OVER_THE_SPOT},
-    //{XIMPreeditCallbacks|XIMStatusCallbacks, IS_ON_THE_SPOT},// OOo
-    //{XIMPreeditArea|XIMStatusArea, IS_ROOT_WINDOW},
-    {XIMPreeditCallbacks|XIMStatusNothing, IS_ON_THE_SPOT},
-    {0, 0},
-};
-static input_style input_style_tab_without_over_the_spot[] = {
-    {XIMPreeditNothing|XIMStatusNothing, IS_ROOT_WINDOW},
-    //{XIMPreeditPosition|XIMStatusArea, IS_OVER_THE_SPOT},// emacs
-    //{XIMPreeditPosition|XIMStatusNothing, IS_OVER_THE_SPOT},
-    //{XIMPreeditCallbacks|XIMStatusCallbacks, IS_ON_THE_SPOT},// OOo
-    //{XIMPreeditArea|XIMStatusArea, IS_ROOT_WINDOW},
-    {XIMPreeditCallbacks|XIMStatusNothing, IS_ON_THE_SPOT},
-    {0, 0},
-};
-// XIMPreeditArea,XIMPreeditCallbacks,XIMPreeditPosition
-// XIMPreeditNothing,XIMPreeditNone
-// XIMStatusArea,XIMStatusCallbacks
-// XIMStatusNothing,XIMStatusNone
-
 void print_ustring(uString *s)
 {
     uString::iterator i;
-    printf("length=%d : ", s->size());
+    printf("length=%d : ", (int)s->size());
     uchar ch;
     char utf8[6];
     int nbyte;
@@ -90,7 +83,7 @@ void print_ustring(uString *s)
 	ch = *i;
 	nbyte = utf8_wctomb((unsigned char *)utf8, ch);
 	utf8[nbyte] = '\0';
-	printf(utf8);
+	printf("%s", utf8);
     }
     printf("\n");
 }
@@ -108,11 +101,11 @@ void append_ustring(uString *d, uString *s)
     }
 }
 
-XimServer::XimServer(Locale *lc, const char *name, const char *lang)
+XimServer::XimServer(const char *name, const char *lang)
 {
-    mName = strdup(name);
-    mLang = lang;
-    mLocale = lc;
+    mIMName = strdup(name);
+    mIMLang = lang;
+    mUsePreservedDefaultIM = false;
 }
 
 InputContext *XimServer::createContext(XimIC *xic, const char *engine)
@@ -139,24 +132,47 @@ void XimServer::changeContext(const char *engine) {
     for (it = ic_list.begin(); it != ic_list.end(); it++) {
 	(*it)->changeContext(engine);
     }
+    // make sure to update locale of focused context
+    InputContext *focusedContext = InputContext::focusedContext();
+    if (focusedContext)
+	focusedContext->focusIn();
 }
 
-struct input_style *
-XimServer::getInputStyles()
-{
-    if (mLocale->supportOverTheSpot())
-	return input_style_tab_with_over_the_spot;
+void XimServer::customContext(const char *custom, const char *val) {
+#if 1
+    // Updated ximserver's global im with customized one.
+    // This is temporal hack.
+    //fprintf(stderr, "custom %s, val %s\n", custom, val);
+    if (!strcmp(custom, "custom-activate-default-im-name?")) {
+	if (!strcmp(val, "#t"))
+	    mUsePreservedDefaultIM = true;
+	else
+	    mUsePreservedDefaultIM = false;
+    }
+    if (!strcmp(custom, "custom-preserved-default-im-name") &&
+		    mUsePreservedDefaultIM == true) {
+	val++;
+	set_im(val);
+	val--;
+    }
+#if HAVE_XFT_UTF8_STRING
+    if (!strcmp(custom, "uim-xim-xft-font-name"))
+	update_default_xftfont(val);
+#endif
+#endif
 
-    return input_style_tab_without_over_the_spot;
+    std::list<InputContext *>::iterator it;
+    for (it = ic_list.begin(); it != ic_list.end(); it++) {
+	(*it)->customContext(custom, val);
+    }
 }
 
 bool
 XimServer::setupConnection(bool useDefaultIM)
 {
-    char *buf;
+    const char *buf;
     if (!useDefaultIM) {
-	buf = (char *)alloca(15 + strlen(mName));
-	sprintf(buf, "@server=uim-%s", mName);
+	return false;
     } else {
 	buf = "@server=uim";
     }
@@ -164,7 +180,7 @@ XimServer::setupConnection(bool useDefaultIM)
     Window owner = XGetSelectionOwner(XimServer::gDpy, mServerAtom);
     if (owner != None) {
 	if (!useDefaultIM)
-	    printf("Another instance exists (uim-%s).\n", mName);
+	    printf("Another instance exists (uim-%s).\n", mIMName);
 	else
 	    printf("Another instance exists (uim).\n");
 	return false;
@@ -213,12 +229,6 @@ XimServer::setupConnection(bool useDefaultIM)
     return true;
 }
 
-char *
-XimServer::uStringToCtext(uString *us, const char *encoding)
-{
-    return mLocale->uStringToCtext(us, encoding);
-}
-
 void
 XimServer::strToUstring(uString *d, const char *s)
 {
@@ -247,34 +257,28 @@ XimServer *XimServer::findServer(Window w)
 
 const char *XimServer::getIMName()
 {
-    return mName;
+    return mIMName;
 }
 
 const char *XimServer::getIMLang()
 {
-    return mLang;
-}
-
-Locale *XimServer::getLocale()
-{
-    return mLocale;
+    return mIMLang;
 }
 
 void XimServer::set_im(const char *engine)
 {
-    if (mName)
-	free(mName);
+    if (mIMName)
+	free(mIMName);
 
-    mName = strdup(engine);
-    mLang = get_im_lang_from_engine(engine);
-    mLocale->set_localename_from_lang(mLang);
+    mIMName = strdup(engine);
+    mIMLang = get_im_lang_from_engine(engine);
 }
 
 const char *get_im_lang_from_engine(const char *engine)
 {
     std::list<UIMInfo>::iterator it;
     for (it = uim_info.begin(); it != uim_info.end(); it++) {
-        if (!strcmp(it->name, engine))
+	if (!strcmp(it->name, engine))
 	    return it->lang;
     }
     return "en"; // For safety...
@@ -288,12 +292,13 @@ InputContext::InputContext(XimServer *svr, XimIC *ic, const char *engine)
     m_pe = new pe_stat(this);
     mConvdisp = 0;
     mServer = svr;
-    mUc = createUimContext(engine);
-    mCandwinActive = false;
+    mEngineName = NULL;
+    mLocaleName = NULL;
     mFocusedContext = this;
+    createUimContext(engine);
+    mCandwinActive = false;
     mNumPage = 1;
     mDisplayLimit = 0;
-    mEngineName = strdup(engine);
 }
 
 InputContext::~InputContext()
@@ -308,39 +313,44 @@ InputContext::~InputContext()
     uim_release_context(mUc);
     mServer->deleteContext(this);
     free(mEngineName);
+    free(mLocaleName);
 }
 
-uim_context
+void
 InputContext::createUimContext(const char *engine)
 {
-    char *locale, *engine_locale = NULL, *supported_lang = NULL;
+    char *locale;
+    const char *client_locale, *engine_locales;
     const char *encoding;
     const char *real_im;
 
-    Locale *lc = this->getServer()->getLocale();
     encoding = mXic->get_encoding();
-    engine_locale  = strdup(lc->compose_localename_from_lang(get_im_lang_from_engine(engine)));
+    client_locale = mXic->get_lang_region();
+    engine_locales = compose_localenames_from_im_lang(get_im_lang_from_engine(engine));
 
     if (!strcmp(encoding, "UTF-8")) {
-	locale = engine_locale;
 	real_im = engine;
-    } else {
-	supported_lang = strdup(find_localename_from_encoding(encoding));
-	char *tmp = strstr(supported_lang, ".");
-	*tmp = '\0';
-	// Use default engine for corresponding encoding unless
-	// encoding matches with selected engine.
-	if (strcmp(supported_lang, engine_locale) &&
-			strcmp(engine_locale, "en_US")) {
-	    // FIXME This is not the best way to select im in this
-	    // case if user set default-im-name which is not
-	    // compatible with client encoding.
-	    real_im = uim_get_default_im_name(supported_lang);
-	    locale = supported_lang;
-	} else {
-	    real_im = engine;
-	    locale = engine_locale;
+	if (is_locale_included(engine_locales, client_locale))
+	    locale = strdup(client_locale);
+	else {
+	    locale = get_prefered_locale(engine_locales);
 	}
+    } else {
+	// Use default engine for corresponding encoding of the client
+	// unless encoding matches with selected engine.
+	if (!is_locale_included(engine_locales, client_locale)) {
+	    const char *test_im = uim_get_default_im_name(client_locale);
+	    const char *test_im_lang = get_im_lang_from_engine(test_im);
+	    const char *test_im_locales = compose_localenames_from_im_lang(test_im_lang);
+	    if (is_locale_included(test_im_locales, client_locale))
+		real_im = test_im;
+	    else
+		real_im = uim_get_im_name_for_locale(client_locale);
+
+	} else
+	    real_im = engine;
+
+	locale = strdup(client_locale);
     }
 
     locale = (char *)realloc(locale, strlen(locale) + strlen(encoding) + 2);
@@ -348,7 +358,14 @@ InputContext::createUimContext(const char *engine)
     strcat(locale, encoding);
 
     setlocale(LC_CTYPE, locale);
-    free(locale);
+
+    free(mLocaleName);
+    mLocaleName = locale;
+
+    if (mEngineName != real_im) {
+      free(mEngineName);
+      mEngineName = strdup(real_im);
+    }
 
     uim_context uc = uim_create_context((void *) this, "UTF-8",
 					NULL, real_im, uim_iconv,
@@ -368,9 +385,10 @@ InputContext::createUimContext(const char *engine)
     uim_set_prop_label_update_cb(uc,
 				 InputContext::update_prop_label_cb);
 
-    uim_prop_list_update(uc);
+    if (mFocusedContext == this)
+	uim_prop_list_update(uc);
 
-    return uc;
+    mUc = uc;
 }
 
 void
@@ -379,31 +397,28 @@ InputContext::changeContext(const char *engine)
     const char *encoding = mXic->get_encoding();
     const char *im_lang = get_im_lang_from_engine(engine);
 
-    // Don't change im-engine unless encoding matches.
+    // Don't change im unless encoding matches for clients with legacy locales.
     if (strcmp(encoding, "UTF-8")) {
-	char *supported_lang = strdup(find_localename_from_encoding(encoding));
-	char *tmp = strstr(supported_lang, ".");
-	*tmp = '\0';
-	if (!strncmp(supported_lang, "zh", 2)) {
-	    if (strncmp(im_lang, supported_lang, 5) &&
-			    strncmp(im_lang, "*", 2))
-		return;
-	} else {
-	    if (strncmp(im_lang, supported_lang, 2) &&
-			    strncmp(im_lang, "*", 2))
-		return;
-	}
-	free(supported_lang);
+	const char *client_locale = mXic->get_lang_region();
+	const char *engine_locales = compose_localenames_from_im_lang(im_lang);
+
+	if (!is_locale_included(engine_locales, client_locale))
+	    return;
     }
 
     clear();
     uim_release_context(mUc);
-    mUc = createUimContext(engine);
-    if (mEngineName)
-	free(mEngineName);
-    mEngineName = strdup(engine);
-    if (mConvdisp)
-	mConvdisp->set_lang(get_im_lang_from_engine(mEngineName));
+    createUimContext(engine); // mUc, mEngineName, and locale will be set here.
+    if (mConvdisp) {
+	mConvdisp->set_im_lang(get_im_lang_from_engine(mEngineName));
+	mConvdisp->set_locale_name(mLocaleName);
+    }
+}
+
+void
+InputContext::customContext(const char *custom, const char *val)
+{
+    uim_prop_update_custom(mUc, custom, val);
 }
 
 InputContext *
@@ -421,24 +436,13 @@ InputContext::deletefocusedContext()
 void
 InputContext::focusIn()
 {
-    char *locale;
-    const char *encoding;
+    setlocale(LC_CTYPE, mLocaleName);
 
-    Locale *lc = this->getServer()->getLocale();
-    locale  = strdup(lc->compose_localename_from_lang(get_im_lang_from_engine(mEngineName)));
-    encoding = mXic->get_encoding();
-
-    locale = (char *)realloc(locale, strlen(locale) + strlen(encoding) + 2);
-    strcat(locale, ".");
-    strcat(locale, encoding);
-
-    setlocale(LC_CTYPE, locale);
-    free(locale);
-
-    uim_prop_list_update(mUc);	
-    uim_prop_label_update(mUc);	
+    check_helper_connection();
     uim_helper_client_focus_in(mUc);
     mFocusedContext = this;
+    uim_prop_list_update(mUc);	
+    uim_prop_label_update(mUc);	
 }
 
 void
@@ -515,13 +519,17 @@ void InputContext::candidate_deactivate_cb(void *ptr)
 void InputContext::update_prop_list_cb(void *ptr, const char *str)
 {
     InputContext *ic = (InputContext *)ptr;
-    ic->update_prop_list(str);
+    InputContext *focusedContext = InputContext::focusedContext();
+    if (ic == focusedContext)
+      ic->update_prop_list(str);
 }
 
 void InputContext::update_prop_label_cb(void *ptr, const char *str)
 {
     InputContext *ic = (InputContext *)ptr;
-    ic->update_prop_label(str);
+    InputContext *focusedContext = InputContext::focusedContext();
+    if (ic == focusedContext)
+      ic->update_prop_label(str);
 }
 
 void InputContext::clear_preedit()
@@ -540,6 +548,10 @@ uString InputContext::get_preedit_string()
 
 void InputContext::pushback_preedit_string(int attr, const char *str)
 {
+    if (str == NULL) {
+	fprintf(stderr, "Warning: str is NULL in pushback_cb\n");
+	return;
+    }
     // Need to check caret pos at first.
     if (attr & UPreeditAttr_Cursor)
 	m_pe->caret_pos = m_pe->get_char_count();
@@ -570,15 +582,29 @@ void InputContext::update_preedit()
 int InputContext::pushKey(keyState *k)
 {
     int key = k->key();
+    int rv = 1;
+
     if (key != UKey_Other) {
-	int rv;
-	rv = uim_press_key(mUc, key, k->modifier());
-	if (rv)
-	    return COMMIT_RAW;
+	if (k->is_push()) {
+	    rv = uim_press_key(mUc, key, k->modifier());
+	    if (!(g_option_mask & OPT_ON_DEMAND_SYNC)) {
+		// Call uim_release_key here since we don't filter key
+		// release event with full-synchronous-method for now.
+		uim_release_key(mUc, key, k->modifier());
+	    }
+	}
 	else
+	    rv = uim_release_key(mUc, key, k->modifier());
+    }
+
+    if (rv) {
+	if (k->check_compose())
 	    return UPDATE_MODE;
+	else
+	    return COMMIT_RAW;
+    } else {
+	return UPDATE_MODE;
     }
-    return COMMIT_RAW;
 }
 
 bool InputContext::hasActiveCandwin()
@@ -607,9 +633,9 @@ void InputContext::commit_string(char *s)
     mXic->commit_string(s);
 }
 
-bool InputContext::extra_input(char *s)
+void InputContext::extra_input(char *s)
 {
-    return false;
+    mXic->extra_input(s);
 }
 
 XimIC *InputContext::get_ic()
@@ -620,14 +646,29 @@ XimIC *InputContext::get_ic()
 void InputContext::candidate_activate(int nr, int display_limit)
 {
     int i;
+    const char *cand_str;
+    const char *heading_label;
     uim_candidate cand[nr];
     std::vector<const char *> candidates;
+    char *str;
 
     Canddisp *disp = canddisp_singleton();
 
     for (i = 0; i < nr; i++) {
-	cand[i] = uim_get_candidate(mUc, i, i);
-	candidates.push_back((const char *)strdup(uim_candidate_get_cand_str(cand[i])));
+	cand[i] = uim_get_candidate(mUc, i,
+			display_limit ? i % display_limit : i);
+	cand_str = uim_candidate_get_cand_str(cand[i]);
+	heading_label = uim_candidate_get_heading_label(cand[i]);
+	//annotation_str = uim_candidate_get_annotation(cand[i]);
+	if (cand_str && heading_label) {
+	    str = (char *)malloc(strlen(cand_str) + strlen(heading_label) + 2);
+	    sprintf(str, "%s\t%s", heading_label, cand_str);
+	    candidates.push_back((const char *)str);
+	}
+	else {
+	    fprintf(stderr, "Warning: cand_str at %d is NULL\n", i);
+	    candidates.push_back((const char *)strdup("\t"));
+	}
     }
     disp->activate(candidates, display_limit);
     for (i = 0; i < nr; i++) {
@@ -637,7 +678,7 @@ void InputContext::candidate_activate(int nr, int display_limit)
     active_candidates = candidates;
     mDisplayLimit = display_limit;
     if (display_limit)
-	mNumPage = nr / display_limit + 1;
+	mNumPage = (nr - 1) / display_limit + 1;
 }
 
 void InputContext::candidate_update()
@@ -715,12 +756,18 @@ void InputContext::update_prop_list(const char *str)
 void InputContext::update_prop_label(const char *str)
 {
     char *buf;
+    uim_bool show_caret_state = uim_scm_symbol_value_bool("bridge-show-input-state?");
 
     asprintf(&buf, "prop_label_update\ncharset=UTF-8\n%s", str);
     if (!buf)
 	return;
     uim_helper_send_message(lib_uim_fd, buf);
     free(buf);
+    
+    if (show_caret_state == UIM_TRUE) {
+	Canddisp *disp = canddisp_singleton();
+	disp->show_caret_state(str);
+    }
 }
 
 const char *InputContext::get_engine_name()
@@ -728,9 +775,94 @@ const char *InputContext::get_engine_name()
     return mEngineName;
 }
 
-keyState::keyState(keyEventX *x)
+const char *InputContext::get_locale_name()
+{
+    return mLocaleName;
+}
+
+keyState::keyState(XimIC *ic)
+{
+    XimIM *im;
+    DefTree *top;
+
+    mAltOn = false;
+    mMetaOn = false;
+    mSuperOn = false;
+    mHyperOn = false;
+    mIc = ic;
+
+    im = get_im_by_id(mIc->get_imid());
+    top = im->get_compose_tree();
+
+    mCompose = new Compose(top, mIc);
+}
+
+keyState::~keyState()
+{
+    delete mCompose;
+}
+
+void keyState::check_key(keyEventX *x)
 {
     mModifier = 0;
+    mXKeySym = x->key_sym;
+    mXKeyState = x->state;
+
+    if (x->press) {
+	m_bPush = true;
+
+	if (!(g_option_mask & OPT_ON_DEMAND_SYNC)) {
+	    // Only KeyPress is forwarded with full-synchronous
+	    // method.  So reset modifiers here.
+	    if (!(x->state) || x->state == LockMask || x->state == gXNumLockMask)
+		mAltOn = mMetaOn = mSuperOn = mHyperOn = false;
+	}
+
+	switch (x->key_sym) {
+	case XK_Alt_L:
+	case XK_Alt_R:
+	    mAltOn = true;
+	    break;
+	case XK_Meta_L:
+	case XK_Meta_R:
+	    mMetaOn = true;
+	    break;
+	case XK_Super_L:
+	case XK_Super_R:
+	    mSuperOn = true;
+	    break;
+	case XK_Hyper_L:
+	case XK_Hyper_R:
+	    mHyperOn = true;
+	    break;
+	default:
+	    break;
+	}
+    } else {
+	m_bPush = false;
+
+	switch (x->key_sym) {
+	case XK_Alt_L:
+	case XK_Alt_R:
+	    mAltOn = false;
+	    break;
+	case XK_Meta_L:
+	case XK_Meta_R:
+	    mMetaOn = false;
+	    break;
+	case XK_Super_L:
+	case XK_Super_R:
+	    mSuperOn = false;
+	    break;
+	case XK_Hyper_L:
+	case XK_Hyper_R:
+	    mHyperOn = false;
+	    break;
+	default:
+	    break;
+	}
+    }
+
     if (x->state & ShiftMask)
 	mModifier |= gShiftMask;
     if (x->state & LockMask)
@@ -738,19 +870,22 @@ keyState::keyState(keyEventX *x)
     if (x->state & ControlMask)
 	mModifier |= gControlMask;
     if (x->state & Mod1Mask)
-	mModifier |= gMod1Mask;
+	mModifier |= revise_mod(gMod1Mask);
     if (x->state & Mod2Mask)
-	mModifier |= gMod2Mask;
+	mModifier |= revise_mod(gMod2Mask);
     if (x->state & Mod3Mask)
-	mModifier |= gMod3Mask;
+	mModifier |= revise_mod(gMod3Mask);
     if (x->state & Mod4Mask)
-	mModifier |= gMod4Mask;
+	mModifier |= revise_mod(gMod4Mask);
     if (x->state & Mod5Mask)
-	mModifier |= gMod5Mask;
+	mModifier |= revise_mod(gMod5Mask);
+
     if (x->key_sym < 128 && x->key_sym >= 32)
 	mKey = x->key_sym;
     else if (x->key_sym >= XK_F1 && x->key_sym <= XK_F35)
 	mKey = x->key_sym - XK_F1 + UKey_F1;
+    else if (x->key_sym >= 0xfe50 && x->key_sym <= 0xfe62)
+	mKey = UKey_Other; // Don't forward deadkeys to libuim.
     else {
 	switch (x->key_sym) {
 	case XK_BackSpace: mKey = UKey_Backspace; break;
@@ -768,7 +903,9 @@ keyState::keyState(keyEventX *x)
 	case XK_End: mKey = UKey_End; break;
 	case XK_Kanji:
 	case XK_Zenkaku_Hankaku: mKey = UKey_Zenkaku_Hankaku; break;
-	case XK_Multi_key: mKey = UKey_Multi_key; break;
+	// Don't forward Multi_key to libuim.
+	//case XK_Multi_key: mKey = UKey_Multi_key; break;
+	case XK_Multi_key: mKey = UKey_Other; break;
 	case XK_Mode_switch: mKey = UKey_Mode_switch; break;
 	case XK_Henkan_Mode: mKey = UKey_Henkan_Mode; break;
 	case XK_Muhenkan: mKey = UKey_Muhenkan; break;
@@ -778,21 +915,28 @@ keyState::keyState(keyEventX *x)
 	case XK_Control_R: mKey = UKey_Control_key; break;
 	case XK_Alt_L: mKey = UKey_Alt_key; break;
 	case XK_Alt_R: mKey = UKey_Alt_key; break;
+	case XK_Meta_L: mKey = UKey_Meta_key; break;
+	case XK_Meta_R: mKey = UKey_Meta_key; break;
 	case XK_Super_L: mKey = UKey_Super_key; break;
 	case XK_Super_R: mKey = UKey_Super_key; break;
 	case XK_Hyper_L: mKey = UKey_Hyper_key; break;
 	case XK_Hyper_R: mKey = UKey_Hyper_key; break;
-	default: mKey = UKey_Other;
+	default:
+	    mKey = UKey_Other;
 	}
     }
 }
 
+bool keyState::check_compose()
+{
+    return mCompose->handleKey(mXKeySym, mXKeyState, m_bPush);
+}
+
 int keyState::key()
 {
     return mKey;
 }
 
-
 int keyState::modifier()
 {
     return mModifier;
@@ -803,6 +947,36 @@ bool keyState::is_push()
     return m_bPush;
 }
 
+int keyState::revise_mod(int uim_mod)
+{
+    if ((uim_mod & UMod_Alt) && (mAltOn == false))
+	uim_mod &= ~UMod_Alt;
+    if ((uim_mod & UMod_Meta) && (mMetaOn == false))
+	uim_mod &= ~UMod_Meta;
+    if ((uim_mod & UMod_Super) && (mSuperOn == false))
+	uim_mod &= ~UMod_Super;
+    if ((uim_mod & UMod_Hyper) && (mHyperOn == false))
+	uim_mod &= ~UMod_Hyper;
+
+    return uim_mod;
+}
+
+KeySym keyState::xkeysym()
+{
+    return mXKeySym;
+}
+
+int keyState::xkeystate()
+{
+    return mXKeyState;
+}
+
+void keyState::reset()
+{
+    mAltOn = mMetaOn = mHyperOn = mSuperOn = false;
+    mCompose->reset();
+}
+
 void keyState::print()
 {
     printf("key code=%x,modifier=%d.\n",
@@ -865,9 +1039,9 @@ void init_modifier_keys() {
 		KeySym ks;
 		int index = 0;
 		do {
-			ks = XKeycodeToKeysym(XimServer::gDpy,
-					map->modifiermap[k], index);
-			index++;
+		    ks = XKeycodeToKeysym(XimServer::gDpy,
+				    map->modifiermap[k], index);
+		    index++;
 		} while (!ks && index < keysyms_per_keycode);
 
 		switch (i) {
@@ -881,7 +1055,11 @@ void init_modifier_keys() {
 		case Mod5MapIndex: Mod5MaskSyms.push_back(ks); break;
 		default: break;
 		}
-	    }k++;
+		// Check NumLock key
+		if (ks == XK_Num_Lock)
+		    gXNumLockMask |= (1 << i);
+	    }
+	    k++;
 	}
     }
     XFreeModifiermap(map);
