@@ -1,6 +1,6 @@
 /*
 
-  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+  Copyright (c) 2003-2005 uim Project http://uim.freedesktop.org/
 
   All rights reserved.
 
@@ -17,10 +17,10 @@
      may be used to endorse or promote products derived from this software
      without specific prior written permission.
 
-  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -37,16 +37,16 @@
 #endif
 
 #include <stdio.h>
+#include <string.h>
 #include <X11/Xos.h>
 #include "xim.h"
+#include "util.h"
 
-#ifndef __GNUC__
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
-# endif
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
 #endif
 
-char *xim_packet_name[] = {
+const char *xim_packet_name[] = {
     // 0
     0, "XIM_CONNECT", "XIM_CONNECT_REPLY",
     "XIM_DISCONNECT", "XIM_DISCONNECT_REPLY",
@@ -88,16 +88,16 @@ char *xim_packet_name[] = {
 };
 
 static struct XIMATTRIBUTE{
-    XIMATTRIBUTE(char *n, int t);
+    XIMATTRIBUTE(const char *n, int t);
     static void write_imattr_to_packet(TxPacket *p);
 
-    char *name;
+    const char *name;
     int type;
 }xim_attributes[]={
     XIMATTRIBUTE(XNQueryInputStyle, TYPE_XIMSTYLE),
 };
 
-XIMATTRIBUTE::XIMATTRIBUTE(char *n, int t)
+XIMATTRIBUTE::XIMATTRIBUTE(const char *n, int t)
 {
     name = n;
     type = t;
@@ -128,9 +128,9 @@ void XIMATTRIBUTE::write_imattr_to_packet(TxPacket *p)
 }
 
 static struct XICATTRIBUTE{
-    XICATTRIBUTE(char *n, int t);
+    XICATTRIBUTE(const char *n, int t);
     static void write_icattr_to_packet(TxPacket *p);
-    char *name;
+    const char *name;
     int type;
 }xic_attributes[]={
     // the sequence is required to be same as the order of
@@ -156,7 +156,7 @@ static struct XICATTRIBUTE{
     XICATTRIBUTE(XNSeparatorofNestedList, TYPE_SEPARATOR),
 };
 
-XICATTRIBUTE::XICATTRIBUTE(char *n, int t)
+XICATTRIBUTE::XICATTRIBUTE(const char *n, int t)
 {
     name = n;
     type = t;
@@ -203,6 +203,7 @@ Connection::Connection(XimServer *svr)
     mServer = svr;
     mSyncFlag = false;
     mPreeditStartSyncFlag = false;
+    mPreeditCaretSyncFlag = false;
 }
 
 Connection::~Connection()
@@ -217,9 +218,6 @@ Connection::~Connection()
     for (ir = mRxQ.begin(); ir != mRxQ.end(); ir++) {
 	delete *ir;
     }
-    for (ir = mPendingRxQ.begin(); ir != mPendingRxQ.end(); ir++) {
-	delete *ir;
-    }
     std::list<TxPacket *>::iterator it;
     for (it = mTxQ.begin(); it != mTxQ.end(); it++) {
 	delete *it;
@@ -267,7 +265,7 @@ void Connection::OnRecv()
 	    xim_close(p);
 	    break;
 	case XIM_DISCONNECT:
-	    xim_disconnect(p);
+	    xim_disconnect();
 	    break;
 	case XIM_GET_IM_VALUES:
 	    xim_get_im_values(p);
@@ -297,16 +295,16 @@ void Connection::OnRecv()
 	    xim_forward_event(p);
 	    break;
 	case XIM_SYNC_REPLY:
-	    xim_sync_reply(p);
+	    xim_sync_reply();
 	    break;
 	case XIM_RESET_IC:
 	    xim_reset_ic(p);
 	    break;
 	case XIM_PREEDIT_START_REPLY:
-	    xim_preedit_start_reply(p);
+	    xim_preedit_start_reply();
 	    break;
 	case XIM_PREEDIT_CARET_REPLY:
-	    // do nothing
+	    xim_preedit_caret_reply();
 	    break;
 	case XIM_ERROR:
 	    xim_error(p);
@@ -357,7 +355,7 @@ void Connection::push_passive_packet(TxPacket *p)
     mPTxQ.push_back(p);
 }
 
-void Connection::push_error_packet(int imid, int icid, int er, char *str)
+void Connection::push_error_packet(int imid, int icid, int er, const char *str)
 {
     TxPacket *t;
     t = createTxPacket(XIM_ERROR, 0);
@@ -439,6 +437,21 @@ bool Connection::hasPreeditStartSyncFlag()
     return mPreeditStartSyncFlag;
 }
 
+void Connection::setPreeditCaretSyncFlag()
+{
+    mPreeditCaretSyncFlag = true;
+}
+
+void Connection::unsetPreeditCaretSyncFlag()
+{
+    mPreeditCaretSyncFlag = false;
+}
+
+bool Connection::hasPreeditCaretSyncFlag()
+{
+    return mPreeditCaretSyncFlag;
+}
+
 //
 // Packet handlers
 //
@@ -463,7 +476,7 @@ void Connection::xim_connect(RxPacket *p)
 	printf("accept xim connection.\n");
 }
 
-void Connection::xim_disconnect(RxPacket *p)
+void Connection::xim_disconnect()
 {
     TxPacket *t;
     t = createTxPacket(XIM_DISCONNECT_REPLY, 0);
@@ -490,9 +503,11 @@ void Connection::xim_open(RxPacket *p)
 
     TxPacket *t;
     int imid;
+    XimIM *im;
     imid = unused_im_id();
-    mCreatedIm.push_back(imid); // had to be delete by the creater Connetcion
-    create_im(this, imid);
+    mCreatedIm.push_back(imid); // had to be deleted by the creator Connection
+    im = create_im(this, imid);
+    im->set_lang_region(buf);
     t = createTxPacket(XIM_OPEN_REPLY, 0);
     t->pushC16(imid);
     XIMATTRIBUTE::write_imattr_to_packet(t);
@@ -505,8 +520,16 @@ void Connection::xim_open(RxPacket *p)
     t = createTxPacket(XIM_SET_EVENT_MASK, 0);
     t->pushC16(imid);
     t->pushC16(0);
-    t->pushC32(KeyPressMask);
-    t->pushC32(~KeyPressMask); // no need to send XIM_SYNC_REPLY from XIM server
+    if (g_option_mask & OPT_ON_DEMAND_SYNC) {
+	t->pushC32(KeyPressMask|KeyReleaseMask);
+	t->pushC32(~(KeyPressMask|KeyReleaseMask)); // no need to send
+						    // XIM_SYNC_REPLY from
+						    // XIM server
+    } else {
+	t->pushC32(KeyPressMask);
+	t->pushC32(KeyPressMask); // need to send XIM_SYNC_REPLY from XIM server
+    }
+
     push_packet(t);
 }
 
@@ -565,7 +588,7 @@ void Connection::xim_encoding_negotiation(RxPacket *p)
 	}
 	p->getStr8(buf);
 
-	/* use COMPOUND_TEXT (eucJP is Buggy) */
+	// use COMPOUND_TEXT
 	if (!strcmp("COMPOUND_TEXT", buf))
 	    index = idx;
 	else {
@@ -603,7 +626,7 @@ void Connection::xim_get_im_values(RxPacket *p)
 
     // XIMATTRIBUTE
     int nr_style;
-    struct input_style *is = mServer->getInputStyles();
+    struct input_style *is = get_im_by_id(imid)->getInputStyles();
     for (nr_style = 0; is[nr_style].style; nr_style++);
 
     t->pushC16(0); // attribute id
@@ -691,49 +714,16 @@ void Connection::xim_forward_event(RxPacket *p)
 	if (is_xim_sync_reply_timeout()) {
 	    // XIM protocol error?
 	    push_error_packet(imid, icid, ERR_BadProtocol, "Bad Protocol");
-	    clear_pending_rx();
+	    clear_pending_queue();
 	    unsetSyncFlag();
-	    im->forward_event(p);
-	} else {
-	    // Copy this packet since it will soon be deleted, then
-	    // add to pending queue.
-	    RxPacket *packet = copyRxPacket(p);
-	    mPendingRxQ.push_back(packet);
 	}
     }
-    else
-	im->forward_event(p);
+    im->forward_event(p);
 }
 
-void Connection::xim_sync_reply(RxPacket *p)
+void Connection::xim_sync_reply()
 {
-    if (hasSyncFlag()) {
-	unsetSyncFlag();
-
-	// send consecutive XIM_FORWARD event just after XIM_COMMIT
-	std::list<TxPacket *>::iterator i;
-	while (mPendingTxQ.size()) {
-	    i = mPendingTxQ.begin();
-	    mTxQ.push_back(*i);
-	    mPendingTxQ.pop_front();
-	}
-
-	// process pending Rx queue
-	int imid, icid;
-	XimIM *im;
-	std::list<RxPacket *>::iterator j;
-	while (mPendingRxQ.size()) {
-	    j = mPendingRxQ.begin();
-	    imid = (*j)->getC16();
-	    icid = (*j)->getC16();
-	    (*j)->rewind();
-	    im = get_im_by_id(imid);
-	    im->forward_event(*j);
-	    delete *j;
-	    mPendingRxQ.pop_front();
-	    break; // handle only one pending event is safe...
-	}
-    }
+    unsetSyncFlag();
 }
 
 void Connection::xim_reset_ic(RxPacket *p)
@@ -743,19 +733,14 @@ void Connection::xim_reset_ic(RxPacket *p)
 	ic->reset_ic();
 }
 
-void Connection::xim_preedit_start_reply(RxPacket *p)
+void Connection::xim_preedit_start_reply()
 {
-    std::list<TxPacket *>::iterator i;
-
     unsetPreeditStartSyncFlag();
-#if 0
-    // not used for now
-    while (mPPendingTxQ.size()) {
-	i = mPPendingTxQ.begin();
-	mPTxQ.push_back(*i);
-	mPPendingTxQ.pop_front();
-    }
-#endif
+}
+
+void Connection::xim_preedit_caret_reply()
+{
+    unsetPreeditCaretSyncFlag();
 }
 
 void Connection::xim_error(RxPacket *p)
@@ -795,10 +780,10 @@ XimIC *Connection::get_ic(RxPacket *p)
 
 #define XIM_SYNC_REPLY_TIMEOUT_SEC	5
 #define TIMEDELTA(dest, src1, src2) { \
-	if(((dest).tv_usec = (src1).tv_usec - (src2).tv_usec) < 0) {\
+	if (((dest).tv_usec = (src1).tv_usec - (src2).tv_usec) < 0) {\
 		(dest).tv_usec += 1000000;\
 		(dest).tv_sec = (src1).tv_sec - (src2).tv_sec - 1;\
-	} else  (dest).tv_sec = (src1).tv_sec - (src2).tv_sec;  }
+	} else (dest).tv_sec = (src1).tv_sec - (src2).tv_sec;  }
 
 bool Connection::is_xim_sync_reply_timeout(void)
 {
@@ -812,11 +797,11 @@ bool Connection::is_xim_sync_reply_timeout(void)
 	return false;
 }
 
-void Connection::clear_pending_rx() {
-    std::list<RxPacket *>::iterator i;
-    while (mPendingRxQ.size()) {
-	i = mPendingRxQ.begin();
-	mPendingRxQ.pop_front();
+void Connection::clear_pending_queue() {
+    std::list<TxPacket *>::iterator i;
+    while (mPendingTxQ.size()) {
+	i = mPendingTxQ.begin();
+	mPendingTxQ.pop_front();
 	delete *i;
     }
 }
